
# # # # # # # # # # # # # # # # # # # # # # # # # #
# Routines to combine multiple datafiles from MPI #
#                                                 #
# Jonas Sukys                                     #
# SAM, ETH Zurich, Switzerland                    #
# sukys.jonas@gmail.com                           #
# # # # # # # # # # # # # # # # # # # # # # # # # #

# === global imports

import numpy
import sys, os

sys.path.append( os.path.dirname(__file__) +  "/scripts" )

# === local imports

from data_io import Config
from variables import *

# === stripped and extended version of the 'Run' class from 'plot.py'

class Run (object):
    def __init__(self, name=None):
        if name and name != 'None':
          self.outputf = 'output_'+name+'.dat'
          self.timesf = 'times_'+name+'.py'
          self.configf = 'config_'+name+'.py'
          self.runtimef = 'runtime_'+name+'.py'
          self.name = name
        else:
          self.outputf = 'output.dat'
          self.timesf = 'times.py'
          self.configf = 'config.py'
          self.runtimef = 'runtime.py'
          self.name = 'run'
        config = {}
        times = {}
        execfile(self.configf, globals(), config)
        execfile(self.timesf, globals(), times)
        self.times = times['time']
        self.opts = config['opts']
        self.vars = config['used']
        self.i = Config(self, int)
        self.f = Config(self, float)
        
        # detecting data path
        self.data_path = self.vars['DATA_PATH'][1:-1]
        
        # setting STATS_SIZE variable for legacy runs
        if 'STATS_SIZE' not in self.vars:
          self.vars['STATS_SIZE'] = '1'
          self.i.STATS_SIZE = 1
        
        # set flag for variance
        self.VARIANCE = self.i.STATS_COUNT == 2 and (self.i.L != 0 or self.i.ML != 1) and not self.i.USE_ONE_SAMPLE_PER_LEVEL
        
        # detecting number of variables from equation type
        if 'equation' in self.opts:
          self.NVARS = get_NVARS (self.opts['equation'])
        else:
          self.NVARS = 16
        
        # set correct MULTI*
        if 'MULTI_CORES' in self.vars and self.i.MULTI_CORES > 0:
          runtime = {}
          execfile(self.runtimef, globals(), runtime)
          self.i.MULTIX = runtime['default_parallelization'][0][0]
          self.i.MULTIY = runtime['default_parallelization'][0][1]
          self.i.MULTIZ = runtime['default_parallelization'][0][2]
    
    def read_variable (self, ts, vs, i, sub_shape, slice_shape, prefix, level_suffix):
      
      # allocate array
      variable = numpy.zeros (slice_shape * self.i.STATS_SIZE, dtype=numpy.float64)
      
      # iterate through all sub domains in Y and Z directions
      for j in range(self.i.MULTIY):
        for k in range(self.i.MULTIZ):
          f = open ( "%s.%d.%d.%d" % (self.data_path + prefix + self.outputf + level_suffix, i, j, k), 'rb')
          f.seek ( (ts*self.NVARS + vs) * 8 * sub_shape[0] * sub_shape[1] * sub_shape[2] * self.i.STATS_SIZE )
          sub_variable = numpy.fromfile(f, dtype = numpy.float64, count = sub_shape[0] * sub_shape[1] * sub_shape[2] * self.i.STATS_SIZE)
          sub_variable.shape = sub_shape
          x = 0
          y = j*sub_shape[1]
          z = k*sub_shape[2]
          variable [ x : x + sub_shape[0], y : y + sub_shape[1], z : z + sub_shape[2] ] = sub_variable
          f.close()
      
      return variable
    
    def combine (self, split, begin, end, level=None, type=None, prefix=''):
      
      # define shape of the subdomain
      if level == None:
        sub_shape = (self.i.NX / self.i.MULTIX, self.i.NY / self.i.MULTIY, self.i.NZ / self.i.MULTIZ)
      else:
        sub_shape = (max(1,self.i.NX / self.i.MULTIX / 2**(self.i.L-level)), max(1,self.i.NY / self.i.MULTIY / 2**(self.i.L-level)), max(1,self.i.NZ / self.i.MULTIZ / 2**(self.i.L-level)))
      
      # define slice shape - entire domain in sliced within x-axis
      # to prevent heavy usage of working memory
      if level == None:
        slice_shape = (self.i.NX / self.i.MULTIX, self.i.NY, self.i.NZ)
      else:
        slice_shape = (max(1,self.i.NX / self.i.MULTIX / 2**(self.i.L-level)), max(1,self.i.NY / 2**(self.i.L-level)), max(1,self.i.NZ / 2**(self.i.L-level)))
      
      # set first and last snapshot to be combined
      if begin == None: begin = 0
      if end   == None: end = len(self.times)
      
      # define suffix for level datafiles
      level_suffix = '' if level == None else '_MESH_LEVEL=' + str(level) + '_TYPE=' + str(type)
      
      # initialize progress indicator (only if not in batch mode)
      if not split:
        from helpers import Progress
        progress = Progress(" :: Combining " + prefix + self.outputf + level_suffix + ": ", self.NVARS*(end-begin))
        progress.update(0)
      
      # if not split, copy entire data to ONE output file
      if not split:
        cf = open(prefix + self.outputf + level_suffix, 'wb')
      
      # iterate through all snapshots
      for ts in range(begin, end):
        
        # if split, copy entire data to MULTIPLE output files
        # each output file corresponding to one time step
        if split:
          cf = open(prefix + self.outputf + level_suffix + "-" + str(ts), 'wb')
        
        # iterate through all variables
        for vs in range(self.NVARS):
          
          # iterate through all sub domains in X direction
          for i in range(self.i.MULTIX):
            
            # get combined slice of the domain
            variable = self.read_variable (ts, vs, i, sub_shape, slice_shape, prefix, level_suffix)
            
            # write combined slice to the output file
            variable.tofile(cf)
          
          # update progress indicator
          if split:
            print "Progress: %d" % round( 100 * float( (ts - begin) * self.NVARS + vs + 1 ) / ( self.NVARS*(end-begin) ) ) + '%'
          else:
            progress.update ( (ts - begin) * self.NVARS + vs + 1 )
        
        # close output files
        if split: cf.close()
      
      # close ouput file
      if not split: cf.close()
      
      # print new line
      print

if __name__ == '__main__':
    
    import sys
    
    name  =     sys.argv[1]  if len(sys.argv) >= 2 else None
    split = int(sys.argv[2]) if len(sys.argv) >= 3 else 0
    begin = int(sys.argv[3]) if len(sys.argv) >= 4 else None
    end   = int(sys.argv[4]) if len(sys.argv) >= 5 else None
    
    # load run
    r = Run (name)
    
    # check if there is anything to combine
    if r.i.MULTIX == 1 and r.i.MULTIY == 1 and r.i.MULTIZ == 1:
      Exception (" :: ERROR: nothing to combine.")
    
    print
    print " :: === COMBINING STATISTICAL MLMC ESTIMATES ==="
    print
    
    # combine datafiles
    r.combine (split, begin, end)
    
    # combine variance datafiles
    if r.VARIANCE:
      r.combine (split, begin, end, prefix='var_')
    
    # combine level output files
    
    if (r.i.L > 0 or r.i.ML > 1) and 'WRITE_LEVEL_DATAFILES' in r.vars and r.i.WRITE_LEVEL_DATAFILES:
      
      print
      print " :: === COMBINING STATISTICAL MC ESTIMATES FOR EACH MESH LEVEL AND TYPE ==="
      print
      
      # combine datafiles for each MESH_LEVEL and TYPE
      
      for level in xrange(r.i.L+1):
        r.combine (split, begin, end, level=level, type=0)
        if level != r.i.L:
          r.combine (split, begin, end, level=level, type=1)
      
      # combine variance datafiles for each MESH_LEVEL and TYPE
      
      if r.VARIANCE:
        for level in xrange(r.i.L+1):
          r.combine (split, begin, end, level=level, type=0, prefix='var_',)
          if level != r.i.L:
            r.combine (split, begin, end, level=level, type=1, prefix='var_')
  
