
# # # # # # # # # # # # # # # # # # # # # # # # # #
# Make file for ALSVID-UQ                         #
#                                                 #
# Jonas Sukys                                     #
# SAM, ETH Zurich, Switzerland                    #
# sukys.jonas@gmail.com                           #
# # # # # # # # # # # # # # # # # # # # # # # # # #

# === global imports

import re
import os
import sys

sys.path.append( os.path.dirname(__file__) +  "/scripts" )

# === local imports

# import routines for generation of 'bc.cpp' file (boundary conditions)
from make_bc import *

# === configuration

# specify default values for alternative
OPTIONS = [('equation', 'mhd'),
           ('flux','det'),
           ('model', ''),
           ('topography','flat'),
           ('material','const'),
           ('solver', 'hll'),
           ('space', 'o1'),
           ('multi', 'single'),
           ('balancer', 'static'),
           ('rng', 'well512a'),
           ('stats','mean-var')
          ]

# specify default models for each equation
DEFAULT_MODELS = {'burgers': 'sine',
                  'euler'  : 'cloudshock',
                  'linadv' : 'sine',
                  'mhd'    : 'orszagtang',
                  'bl'     : 'riemann',
                  'sw'     : 'dambreak',
                  'wave'   : 'gaussian'
                 }

# specify which alternatives are equation dependent
# i.e. the source file names have additional appropriate suffix '-equation'
EQUATION_DEPENDENT = ['flux', 'model', 'solver']

# specify which alternatives are equation specific and hence should be ignored otherwise
EQUATION_SPECIFIC  = {'material'   : 'wave',
                      'topography' : 'sw'
                     }

# === helpers

# load contents of 'local.py'
local = {}
lfn = os.path.join(os.path.split(__file__)[0], 'local.py')
if os.path.exists(lfn):
    execfile(lfn, local, local)

# define regular expression
META_RE = re.compile(r'///::(\w+)::\s+(.+)$')

# set default values for parallel job submission
if local.get('WALLTIME'):
    DEFAULT_WALLTIME = local.get('WALLTIME')
else:
    DEFAULT_WALLTIME = '1'

if local.get('MEMORY'):
    DEFAULT_MEMORY = local.get('MEMORY')
else:
    DEFAULT_MEMORY = '1024'

if local.get('PTILE'):
    DEFAULT_PTILE = local.get('PTILE')
else:
    DEFAULT_PTILE = '32'

class FileMeta (dict):
    def space_list(self, k):
        r = self.get(k)
        if r is None: return []
        return re.split('\s+', r)
    
    def line_list(self, k):
        r = self.get(k)
        if r is None: return []
        return r.split('\n')

# === Make object

class MakeMLMC (object):
    def __init__(self, dir='.'):
        self.dir  = dir
        self.opts = dict(OPTIONS)
        self.vars = {}
        self.meta_store = {}
        
        self.quiet   = False
        self.debug   = False
        self.verbose = False
        self.log     = False
        
        self.configf = 'config.py'
        self.name    = ''
        self.logf    = 'log.txt'
        
        self.norun    = False
        self.parallel = False
        self.walltime = float(DEFAULT_WALLTIME)
        self.memory   = int(DEFAULT_MEMORY)
        self.ptile    = int(DEFAULT_PTILE)
        self.nodes    = 1
        self.profile  = False
        self.xopts    = ''
        
        self.directives = ''
    
    def read_meta(self, fn):
        fname = os.path.join(self.dir, fn)
        meta = self.meta_store.get(fname)
        if meta is not None: return meta
        meta = FileMeta()
        f = open(fname)
        for l in f:
            m = META_RE.match(l.strip())
            if m:
                o = meta.get(m.group(1))
                if o:
                    meta[m.group(1)] = o + '\n' + m.group(2)
                else:
                    meta[m.group(1)] = m.group(2)
        f.close()
        self.meta_store[fname] = meta
        return meta
    
    def get_alternative (self, c):
        if c in EQUATION_DEPENDENT:
          return c + '-' + self.opts['equation']
        else:
          return c
    
    def get_vars(self, c, v, consts=False):
        if consts: name='consts'
        else: name='vars'
        try:
          vars = self.read_meta(self.get_alternative(c) + '_' + v + '.cpp').space_list(name)
        except:
          print (' :: ERROR: The requested "' + c + '" is not available.\n' + '  : -> Try directive -a in order to see available options.')
          raise SystemExit
        return dict([x.split('=') for x in vars])
    
    def get_macros(self, c, v):
        vals = {}
        incs = {}
        for m in self.read_meta(self.get_alternative(c) + '_' + v + '.cpp').line_list('macro'):
            (n,i,v) = re.split('\s+', m, 2)
            if i == '@': i = ''
            else: i = '#include "'+i+'"\n'
            if v == '@': v = ''
            vals[n] = v
            incs[n] = i
        return (vals,incs)
    
    def filter_opts(self):
        for (o,req) in EQUATION_SPECIFIC.items():
            if self.opts['equation'] != req and o in self.opts:
                del self.opts[o]
    
    def get_settings(self):
        vars = self.get_vars('mlmc','main')
        consts = self.get_vars('mlmc','main', consts=True)
        (m,incs) = self.get_macros('mlmc','main')
        consts.update(m)
        if self.opts['model'] == '':
            self.opts['model'] = DEFAULT_MODELS[self.opts['equation']]
        self.filter_opts()
        for o,v in self.opts.items():
            vars.update(self.get_vars(o,v))
            consts.update(self.get_vars(o,v, consts=True))
            (m,i) = self.get_macros(o,v)
            consts.update(m)
            incs.update(i)
        vars.update(self.vars)
        
        # replace '@' by spaces in COMPILER_DIRECTIVES
        if 'COMPILER_DIRECTIVES' in vars:
          self.directives = vars['COMPILER_DIRECTIVES'].replace('@',' ')[1:-1]
        else:
          self.directives = ''
        
        # replace '@' by spaces in LINKER_LIBS
        if 'LINKER_LIBS' in vars:
          self.libs = vars['LINKER_LIBS'].replace('@',' ')[1:-1]
        else:
          self.libs = ''
        
        self.settings = (vars, consts, incs)
        self.parallel = (self.opts['multi'] != 'single')
        self.compiler = CompilerLocal(self.opts['multi'])
    
    def check_compatibility (self):
        
        '''
        # TODO: MULTI* are available only in multi:mpi, need to do some checking.. (based on MIN_CELLS_PER_CORE?)
        # check if MIN_CELLS_PER_CORE can be guaranteed
        MIN_CELLS_PER_CORE = int(self.settings[0]['MIN_CELLS_PER_CORE'])
        for d in ['X','Y','Z']:
          N = int(self.settings[0]['N'+d])
          MULTI = int(self.settings[0]['MULTI' + d])
          if N > 1 and MULTI > 1 and N / MULTI < MIN_CELLS_PER_CORE:
            print (" :: ERROR: Enforcement of MIN_CELLS_PER_CORE = " + str(MIN_CELLS_PER_CORE) + " is NOT possible.")
            raise SystemExit
        '''
        '''
        # check if ENABLE_SLICED_PARALLELIZATION = 1 can be guaranteed
        ENABLE_SLICED_PARALLELIZATION = int(self.settings[1]['ENABLE_SLICED_PARALLELIZATION'])
        if ENABLE_SLICED_PARALLELIZATION:
          if int(self.settings[0]['NX']) < int(self.settings[0]['MULTIX']) * int(self.settings[0]['MULTIY']) * int(self.settings[0]['MULTIZ']):
            print (" :: ERROR: Enforcement of ENABLE_SLICED_PARALLELIZATION = 1 is NOT possible.")
            raise SystemExit
        '''
    
    def depend(self, f):
        return self.read_meta(f).space_list('depends')
    
    def get_files(self):
        files = ['mlmc_main']+[self.get_alternative(c) + '_' + v for (c,v) in self.opts.items()]
        self.files = []
        while files:
            f = files[0]
            del files[0]
            self.files.append(f)
            files += self.depend(f+'.cpp')
    
    def read_config(self, fname):
        cfg = {}
        execfile(fname, globals(), cfg)
        self.opts.update(cfg['opts'])
        self.vars.update(cfg['vars'])
    
    def set_name(self, name):
        self.configf = 'config_'+name+'.py'
        self.vars['RUNTIME_FILE'] = '"runtime_'+name+'.py"'
        self.vars['DATA_FILE'] = '"output_'+name+'.dat"'
        self.vars['TIME_FILE'] = '"times_'+name+'.py"'
        self.vars['LOG_FILE'] = '"log_'+name+'.txt"'
        self.vars['BALANCER_REPORT_FILE'] = '"balancer_'+name+'.txt"'
        self.vars['SAMLPES_FILE'] = '"samples_'+name+'.py"'
        self.vars['LOADS_FILE'] = '"loads_'+name+'.dat"'
        self.vars['BALANCER_FILE'] = '"balancer_'+name+'.dat"'
        self.name = '_'+name
    
    def write_config_py(self, fname):
        f = open(fname, 'w')
        f.write('opts={ '+', '.join(["'"+k+"':"+repr(v)
                                     for k,v in sorted(self.opts.items())]))
        f.write(' }\nvars={ ')
        f.write(', '.join(["'"+k+"':"+repr(v)
                           for k,v in sorted(self.vars.items())])+' }\nused={ ')
        used = {}
        used.update(self.settings[0])
        used.update(self.settings[1])
        
        # setting STATS_SIZE variable to its numeric value
        if 'STATS_SIZE' in used:
          if not used['STATS_SIZE'].isdigit():
            used['STATS_SIZE'] = used[used['STATS_SIZE']]
        else:
          used['STATS_SIZE'] = '1'
        
        f.write(', '.join(["'"+k+"':"+repr(v)
                          for k,v in sorted(used.items())])+' }\n')
        cluster = local.get('CLUSTER')
        if cluster:
          f.write('info={ ' + "'CLUSTER':'" + cluster + "' }\n")
        f.close()
    
    def show_config(self):
        
        if self.quiet: return
        
        # print alternatives
        print
        for c,v in sorted(self.opts.items()):
            MAX_LENGTH_C = 10
            MAX_LENGTH_V = 16
            space_c = ' '*(MAX_LENGTH_C+1-len(c))
            space_v = ' '*(MAX_LENGTH_V+1-len(v))
            # get title
            title = self.read_meta(self.get_alternative(c) + '_' + v + '.cpp').get('title')
            # truncate title, if too long
            max_length = 52;
            if len(title) > max_length:
              title = title[0:max_length-1-5] + '[...]'
            print c + space_c + v + space_v + title
        
        # print main parameters related to mesh, MLMC and parallelization
        print
        print "CONFIG:",
        if 'NX' in self.settings[0]: print 'NX=' + self.settings[0]['NX'],
        if 'NY' in self.settings[0]: print 'NY=' + self.settings[0]['NY'],
        if 'NZ' in self.settings[0]: print 'NZ=' + self.settings[0]['NZ'],
        if int(self.settings[0]['L']) != 0 or int(self.settings[0]['ML']) != 1:
          print "L=%s ML=%s" % (self.settings[0]['L'], self.settings[0]['ML']),
        if 'MULTI_CORES' not in self.settings[0] or int(self.settings[0]['MULTI_CORES']) == 0:
          if 'MULTIX' in self.settings[0]: print 'MULTIX=' + self.settings[0]['MULTIX'],
          if 'MULTIY' in self.settings[0]: print 'MULTIY=' + self.settings[0]['MULTIY'],
          if 'MULTIZ' in self.settings[0]: print 'MULTIZ=' + self.settings[0]['MULTIZ'],
          if 'MULTIM' in self.settings[0] and int(self.settings[0]['MULTIM']) != 1:
            print "MULTIM=" + self.settings[0]['MULTIM'],
        else:
          print "MULTI_CORES=" + self.settings[0]['MULTI_CORES'],
        print
        
        # print all parameters, including the default ones
        if self.verbose:
          for n,v in sorted(self.settings[0].items()):
            print n,'=',v
        
        # print explicitely specified parameters, excluding the already printed ones and paths
        if len(self.vars) > 0:
          
          # exclude unwanted entries
          vars_to_print = self.vars.copy()
          for key in list(vars_to_print):
            if key in ['NX','NY','NZ','L','ML','MULTI_CORES','MULTIX','MULTIY','MULTIZ','MULTIM'] or key .endswith ( ('_FILE', '_PATH') ):
              del vars_to_print [ key ]
          
          # print vars
          print
          print "VARS:",
          for p,v in sorted(vars_to_print.items()):
            print p + "=" + v,
          print #endl
          
          # print paths
          print
          print "PATHS+FILES:",
          for p,v in sorted(self.vars.items()):
            if p .endswith ( ('_FILE', '_PATH') ):
              print p + "=" + v,
          print #endl
        
        print
        print " :: Use -v to display all parameters, compiler output and job submission query."
        print
    
    def write_config_h(self):
        f = open(os.path.join(self.dir, 'config.h'), 'w')
        if self.quiet: f.write('#define QUIET 1\n')
        if self.debug: f.write('#define DEBUG %d\n'%self.debug)
        if self.log: f.write('#define LOG 1\n')
        used = {}
        used.update(self.settings[0])
        used.update(self.settings[1])
        f.write(''.join(['#define %s %s\n'%x for x in sorted(used.items())]))
        f.close()
        f = open(os.path.join(self.dir, 'include.h'), 'w')
        f.write(''.join([x for x in self.settings[2].values()]))
        f.write('#include "structures_' + self.opts['equation'] + '.h"\n')
        f.close()
    
    def make_bc(self):
        spec = self.read_meta('model-' + self.opts['equation'] + '_' + self.opts['model'] + '.cpp').space_list('bc')
        if len(spec) == 1:
          spec = spec * 6
        elif len(spec) == 3:
          spec = spec * 2
        elif len(spec) != 6:
          raise Exception(" :: ERROR: unsupported boundary conditions specified in '" + self.opts['model'] + ".cpp" + "'.")
        f = open(os.path.join(self.dir, 'bc.cpp'), 'w')
        bc (f, spec, self.settings)
    
    def default_compiler(self):
        return DEFAULT_COMPILER()
    
    def compile(self, f):
        cc = self.compiler.ccpp(f+'.cpp', f+'.o')
        if self.verbose:
          print cc
        if not self.simulate:
          return os.system("cd "+self.dir+"; "+cc) == 0
        else:
          return 1
    
    def link(self):
        lc = self.compiler.link([f+'.o' for f in self.files], 'alsvid-uq', self.libs)
        if self.verbose: print lc
        return os.system("cd "+self.dir+"; "+lc) == 0
    
    def build(self):
        self.get_settings()
        self.check_compatibility()
        self.compiler.profile = self.profile
        self.compiler.debug = self.debug
        self.make_bc()
        self.write_config_h()
        self.show_config()
        self.get_files()
        self.write_config_py(self.configf)
        
        # append additional compiler directives
        if self.profile: self.compiler.COMPILER['profile'] += " " + self.directives;
        elif self.debug: self.compiler.COMPILER['debug']   += " " + self.directives;
        else:            self.compiler.COMPILER['compile'] += " " + self.directives;
        
        if self.log:
            f = open(self.logf, 'w')
            f.close()
            self.compiler.log_file = self.logf
        
        if not self.verbose:
          print " :: COMPILING with directives: "
          if self.profile: print "  : -> " + self.compiler.COMPILER['profile']
          elif self.debug: print "  : -> " + self.compiler.COMPILER['debug']
          else:            print "  : -> " + self.compiler.COMPILER['compile']
        
        if not self.verbose:
          from helpers import Progress
          progress = Progress("  : -> ", len(self.files))
          progress.update(0)
        
        for (f_step,f) in enumerate(self.files):
            if not self.compile(f):
                raise SystemExit
                raise Exception("Compile failed for "+f)
            if not self.verbose: progress.update(f_step+1)
        print
        print " :: LINKING...",
        if not self.simulate and not self.link():
            raise SystemExit
            raise Exception("Link failed (see make.log)")
        print "FINISHED."
        print
    
    # see multi_initialize_default_parallelization_setup() in the source code for explanations
    def get_multi_total (self, d):
        
        if self.parallel == False:
          
          return 1
        
        else:
          
          MULTI_CORES = int(d.get('MULTI_CORES',1))
          if MULTI_CORES > 0:
            return MULTI_CORES
          
          MULTIX = int(d.get('MULTIX',1))
          MULTIY = int(d.get('MULTIY',1))
          MULTIZ = int(d.get('MULTIZ',1))
          MULTIM = int(d.get('MULTIM',1))
          L      = int(d.get('L',1))
          NX     = int(d.get('NX',1))
          NY     = int(d.get('NY',1))
          NZ     = int(d.get('NZ',1))
          ORDER  = int(d.get('ORDER',1))
          SMOOTH = int(d.get('SMOOTH_SOLUTION',1)) * int(d.get('SMOOTH_FLUX',1))
          VISC   = int(d.get('HAVE_VISCOSITY',1))
          
          ACTIVE_DIRS = (NX > 1) + (NY > 1) + (NZ > 1)
          if SMOOTH or VISC:
            EFFECTIVE_DOUBLED_ORDER = 2*ORDER
          else:
            EFFECTIVE_DOUBLED_ORDER = min(2,ORDER)
          
          WORK_FACTOR_EXPONENT = ACTIVE_DIRS+1 + VISC - EFFECTIVE_DOUBLED_ORDER
          
          MULTI_TOTAL = MULTIM * MULTIX * MULTIY * MULTIZ
          
          WORK_FACTOR = 2 ** WORK_FACTOR_EXPONENT
          MULTI_ADD = MULTI_TOTAL
          for multi_level in range (0, L):
            MULTI_ADD_OLD = MULTI_ADD
            if MULTI_ADD >= WORK_FACTOR:
              MULTI_ADD /= WORK_FACTOR
            else:
              MULTI_ADD = 1
            MULTI_TOTAL += MULTI_ADD
            if MULTI_ADD == 1 and MULTI_ADD_OLD == 1:
              break
          
          return MULTI_TOTAL
    
    def run(self):
        if self.norun: return
        d = {}
        d.update(self.settings[0])
        d.update(self.settings[1])
        
        d['nprocs']   = int(self.get_multi_total(d))
        d['walltime'] = int(self.walltime)
        d['walltime-mins'] = min (59, int ( 100 * (self.walltime - d['walltime']) ) )
        d['memory']   = self.memory
        d['ptile']    = self.ptile
        d['xopts']    = self.xopts
        d['suffix']   = ''
        
        from math import ceil
        d['nodes']    = ceil(float(d['nprocs']) / d['ptile'])
        
        if d['nprocs'] < d['ptile']:
          d['ptile'] = d['nprocs']
        
        d['name']     = 'alsvid-uq'+self.name
        
        jsc = local.get('JOB_SUBMIT')
        
        # inline run
        if jsc is None:
            run_cmd = os.path.join(self.dir, 'alsvid-uq')
            d['cmd'] = run_cmd
            if self.parallel:
                run_cmd = local.get('%s_RUN'%(self.opts['multi'].upper(),),
                                    '%(cmd)s') % d
            if self.verbose: print run_cmd % d
            if not self.simulate and os.system(run_cmd) != 0:
                raise SystemExit
            return
        
        # submission using queues
        if 'DATA_PATH' in self.vars:
          run_cmd = self.vars['DATA_PATH'][1:-1]+d['name']
        else:
          run_cmd = d['name']
        if not self.simulate:
          os.system( '/bin/mv %s/alsvid-uq %s' % (self.dir,run_cmd) )
        d['cmd'] = run_cmd
        
        # profiling prepocessing and executable suffix
        pp = local.get('PROFILE_PREPROC')
        if self.profile and pp:
            print pp % d
            if not self.simulate and os.system(pp % d) != 0:
                raise SystemExit
            d['suffix'] = local.get('PROFILE_SUFFIX')
        
        if self.parallel:
            run_cmd = local.get('%s_RUN'%(self.opts['multi'].upper(),), '%(cmd)s') % d
            d['cmd'] = run_cmd
        if self.verbose: print jsc % d
        else: print ' :: SUBMITTING on ' + str(d['nprocs']) + ' core(s) in ' + str(int(d['nodes'])) + ' nodes for ' + str(d['walltime']) + ' hour(s) and ' +  str(d['walltime-mins']) + ' minute(s)...'
        if not self.simulate and os.system(jsc % d) != 0:
            raise SystemExit

# === Options object

class Options (object):
    
    def __init__(self):
        self.make = MakeMLMC(os.path.join(os.path.split(__file__)[0], 'src'))
    
    # process command line arguments
    def process_args(self):
        from optparse import OptionParser
        usage = "usage: %prog [options] [config:choice] [var=value]"
        p = OptionParser(usage)
        p.add_option('-g', '--debug', action="count",
                     help="compile debug version")
        p.add_option('-q', '--quiet', action="store_true", default=False,
                     help="run quietly (no output)")
        p.add_option('-v', '--verbose', action="store_true", default=False,
                     help="display compiler output, job parameters and job submission query")
        p.add_option('-l', '--log', action="store_true", default=False,
                     help="log output to file")
        p.add_option('-r', '--rerun', action="store_true", dest="config",
                     help="use config from config.py")
        p.add_option('-c', '--config', action="store", default=False,
                     help="specify config file")
        p.add_option('-o', '--xopts', action="store", default='',
                     help="specify extra options for job submission")
        p.add_option('-n', '--name', action="store", default='',
                     help="specify run name for output files")
        p.add_option('-x', '--norun', action="store_true", default=False,
                     help="compile only, do not run")
        p.add_option('-a', '--ask', action="count",
                     help="ask user about config options (twice: vars too)")
        p.add_option('-w', '--walltime', action="store", default=DEFAULT_WALLTIME,
                     help="specify estimated runtime limit (in hours)")
        p.add_option('-m', '--memory', action="store", default=DEFAULT_MEMORY,
                     help="specify the required memory (in MB, per core)")
        p.add_option('-s', '--simulate', action="store_true", default=False,
                     help="simulate 'make' (no actual compilation or running)")
        p.add_option('-t', '--ptile', action="store", default=DEFAULT_PTILE,
                     help="specify the number cores per node")
        p.add_option('-p', '--profile', action="store_true", default=False,
                     help="enable profiling")
        (opts, args) = p.parse_args()
        if opts.config == True:
            self.make.read_config('config.py')
        elif opts.config:
            self.make.read_config(opts.config)
        for a in args:
            if ':' in a:
                (c,v) = a.split(':')
                self.make.opts[c] = v
            elif '=' in a:
                (n,v) = a.split('=')
                self.make.vars[n] = v
        if opts.name:
            self.make.set_name(opts.name)
        if opts.ask: self.ask(opts.ask)
        self.make.debug = opts.debug
        self.make.quiet = opts.quiet
        self.make.verbose = opts.verbose
        self.make.log = opts.log
        self.make.norun = opts.norun
        self.make.walltime = float(opts.walltime)
        self.make.memory = int(opts.memory)
        self.make.ptile = int(opts.ptile)
        self.make.profile = int(opts.profile)
        self.make.xopts = str(opts.xopts)
        self.make.simulate = int(opts.simulate)
    
    def get_options(self, o):
        if o in EQUATION_DEPENDENT:
            req_eqn = '-' + self.make.opts['equation']
        else:
            req_eqn = ''
        for fn in os.listdir(self.make.dir):
            m = re.match(o + req_eqn + r'_([A-Za-z0-9_-]+)' + r'.cpp$', fn)
            if m is None: continue
            yield m.group(1), self.make.read_meta(fn).get('title')
    
    def ask(self, askn):
        print "Major Options:"
        for o,dv in OPTIONS:
            if o in EQUATION_SPECIFIC and EQUATION_SPECIFIC[o] != self.make.opts['equation']:
              continue;
            if o == 'model' and self.make.opts['model'] == '':
              self.make.opts['model'] = DEFAULT_MODELS[self.make.opts['equation']]
            opts = dict(self.get_options(o))
            print 'Select '+o+':'
            for c,t in sorted(opts.items()):
                tabs = '\t'
                if len(c) < 15: tabs += '\t'
                if len(c) < 7:  tabs += '\t'
                max_length = 52;
                if len(t) > max_length:
                  t = t[0:max_length-1-5] + '[...]'
                print c, tabs + '"' + t + '"'
            while True:
                r = raw_input(o+': ['+self.make.opts[o]+'] ')
                if r == '':
                    break
                elif r in opts:
                    self.make.opts[o] = r
                    break
                else:
                    print "Invalid value for",o
            print
        if askn == 1: return
        print "Variables:"
        self.make.get_settings()
        for vn, vv in sorted(self.make.settings[0].items()):
            if vv[0] == '"' and vv[-1] == '"':
                s = 1
                vv = vv[1:-1]
            else:
                s = 0
            while True:
                nv = raw_input(vn+' ['+vv+'] = ')
                if nv == '':
                    break
                elif s:
                    self.make.vars[vn] = '"'+nv+'"'
                    break
                else:
                    try:
                        n = float(nv)
                    except ValueError:
                        print "Invalid number"
                    else:
                        self.make.vars[vn] = nv
                        break
        print

# Compiler object

class Compiler (object):
    def __init__(self, debug=False, profile=False):
        self.log_file=None
        self.debug = debug
        self.profile = profile

class CompilerLocal (Compiler):

    DEFAULTS = {
        'compile': 'g++ -O3',
        'debug'  : 'g++ -g',
        'profile': 'g++ -g -pg',
        'libs'   : ''
        }
    
    def __init__(self, type):
        super(CompilerLocal, self).__init__()
        self.COMPILER = local.get(type.upper()+'_COMPILER')
        if self.COMPILER is None:
            self.COMPILER = local.get('COMPILER', self.DEFAULTS)
    
    def ccpp(self, s, o):
        if self.log_file: lf = ' >> '+self.log_file+' 2>&1'
        else: lf = ''
        if self.profile: opts = self.COMPILER['profile']
        elif self.debug: opts = self.COMPILER['debug']
        else: opts = self.COMPILER['compile']
        return '%s -c -o %s %s%s'%(opts, o, s, lf)
    
    def link(self, olist, x, libs):
        if self.log_file: lf = ' >> '+self.log_file+' 2>&1'
        else: lf = ''
        if self.profile: opts = self.COMPILER['profile']
        elif self.debug: opts = self.COMPILER['debug']
        else: opts = self.COMPILER['compile']
        return '%s -o %s %s %s%s%s' % (opts, x, ' '.join(olist), self.COMPILER['libs'], libs, lf)

DEFAULT_COMPILER = CompilerLocal

if __name__=='__main__':
    o = Options()
    o.process_args()  
    o.make.build()
    o.make.run()
