
# # # # # # # # # # # # # # # # # # # # # # # # # #
# FRONT-END for plotting and I/O routines         #
# of multi-dimensional simulation data            #
#                                                 #
# Jonas Sukys                                     #
# SAM, ETH Zurich, Switzerland                    #
# sukys.jonas@gmail.com                           #
#                                                 #
# Documentation:                                  #
# http://www.sam.math.ethz.ch/alsvid-uq           #
# # # # # # # # # # # # # # # # # # # # # # # # # #

# === global imports

import sys, os
sys.path.append( os.path.dirname(__file__) +  "/scripts" )

# === local imports

import data_io

# === simulation ('Run') class

class Run (object):
  
  # === constructor
  # 3d backend: mayavi, visit, paraview
  def __init__(self, name=None, path='', load_times=True):
    data_io.init_run (self, name, path, load_times)
  
  # returns numpy array with simulation data at time 'ts' of variable 'v'
  # if 'cut_slice' is specified, only data required for that slice is loaded
  # if 'level' and 'type' are specified, only MC estimate for that mesh_level of that type is loaded (requires WRITE_LEVEL_DATAFILES=1)
  def var (self, v, ts, stat="mean", cut_slice=None, level=None, type=None):
    return data_io.var (self, v, ts, stat, cut_slice, level, type)
  
  # === plotting routines
  
  def plot (self, v, frame=False, ts=None, extent=None, infolines=True, save=None, topography=None, cut_slice=None, loc='best', extent_mean=None, extent_var=None, log_mean=False, log_var=False, contour=False, poster=False, view=None, num_contours=9, contour_levels_mean=None, contour_levels_var=None, clip=None, highlight=None, batch=False, level=None, type=None, engine=None):
    
    if level != None: data_io.check_level_datafiles (self, level)
    
    if ts is None: ts = len(self.times) - 1
    
    if self.ny == 1:
      import plot1d
      plot1d.plot_1d (self, v, ts, extent, frame, infolines, save, topography, cut_slice, loc, level, type)
    
    elif self.nz == 1:
      if not cut_slice:
        import plot2d
        plot2d.subplot_2d (self, v, ts, extent_mean, extent_var, infolines, save, topography, cut_slice, log_mean, log_var, contour, poster, num_contours, level, type)
      else:
        import plot1d
        plot1d.plot_1d (self, v, ts, extent, frame, infolines, save, topography, cut_slice, loc, level, type)
    
    else:
      if not cut_slice:
        plot3d = __import__('plot3d_' + self.backend_3d)
        plot3d.plot_3d(self, v, ts, extent_mean, extent_var, infolines, view, save, log_mean, log_var, contour, num_contours, contour_levels_mean, contour_levels_var, clip, highlight, batch, level, type, engine)
      
      else:
        if len(cut_slice) == 2:
          import plot2d
          plot2d.subplot_2d (self, v, ts, extent_mean, extent_var, infolines, save, topography, cut_slice, log_mean, log_var, contour, poster, num_contours, level, type)
        else:
          import plot1d
          plot1d.plot_1d (self, v, ts, extent, frame, infolines, save, topography, cut_slice, loc, level, type)
  
  def plot0 (self, v, frame=False, ts=None, extent=None, infolines=True, save=None, topography=None, cut_slice=None, loc='best', extent_mean=None, extent_var=None, log_mean=False, log_var=False, contour=False, poster=False, view=None, num_contours=9, contour_levels_mean=None, contour_levels_var=None, clip=None, highlight=None, batch=False, level=None, type=None, engine=None):
    
    if not self.i.SAVE0:
      raise Exception (" :: ERROR: Snapshot at t=0 is not available! Use SAVE0=1 when compiling.")
    
    if level != None: data_io.check_level_datafiles (self, level)
    
    if ts is None: ts = len(self.times) - 1
    
    if self.ny == 1:
      import plot1d
      plot1d.plot0_1d (self, v, ts, extent, frame, infolines, save, topography, cut_slice, loc, level, type)
    
    elif self.nz == 1:
      if not cut_slice:
        import plot2d
        plot2d.subplot_2d (self, v, 0,  extent_mean, extent_var, infolines, save, topography, cut_slice, log_mean, log_var, contour, poster, num_contours, level, type)
        plot2d.subplot_2d (self, v, ts, extent_mean, extent_var, infolines, save, topography, cut_slice, log_mean, log_var, contour, poster, num_contours, level, type)
      else:
        import plot1d
        plot1d.plot0_1d (self, v, ts, extent, frame, infolines, save, topography, cut_slice, loc, level, type)
    
    else:
      if not cut_slice:
        plot3d = __import__('plot3d_' + self.backend_3d)
        plot3d.plot_3d (self, v, 0,  extent_mean, extent_var, infolines, view, save, log_mean, log_var, contour, num_contours, contour_levels_mean, contour_levels_var, clip, highlight, batch, level, type, engine)
        plot3d.plot_3d (self, v, ts, extent_mean, extent_var, infolines, view, save, log_mean, log_var, contour, num_contours, contour_levels_mean, contour_levels_var, clip, highlight, batch, level, type, engine)
      
      else:
        if len(cut_slice) == 2:
          import plot2d
          plot2d.subplot_2d (self, v, 0,  extent_mean, extent_var, infolines, save, topography, cut_slice, log_mean, log_var, contour, poster, num_contours, level, type)
          plot2d.subplot_2d (self, v, ts, extent_mean, extent_var, infolines, save, topography, cut_slice, log_mean, log_var, contour, poster, num_contours, level, type)
        else:
          import plot1d
          plot1d.plot0_1d (self, v, ts, extent, frame, infolines, save, topography, cut_slice, loc, level, type)
  
  def subplot (self, v, frame=False, ts=None, extent_mean=None, extent_var=None, infolines=True, save=None, topography=None, cut_slice=None, loc='best', log_mean=False, log_var=False, contour=False, poster=False, num_contours=9, movie=False, level=None, type=None):
    
    if level != None: data_io.check_level_datafiles (self, level)
    
    if self.ny == 1:
      import plot1d
      plot1d.subplot_1d (self, v, ts, extent_mean, extent_var, frame, infolines, save, topography, cut_slice, log_var, loc, level, type, movie)
    
    elif self.nz == 1:
      if not cut_slice:
        import plot2d
        plot2d.subplot_2d (self, v, ts, extent_mean, extent_var, infolines, save, topography, cut_slice, log_mean, log_var, contour, poster, num_contours, level, type, movie)
      else:
        import plot1d
        plot1d.subplot_1d (self, v, ts, extent_mean, extent_var, infolines, save, topography, cut_slice, log_mean, log_var, loc, level, type, movie)
    
    else:
      if not cut_slice:
        raise Exception (" :: ERROR: 'subplot()' is NOT available for 3d simulation data! Use 'plot()', 'plot0()' or 'cut_slice' option.")
      else:
        if len(cut_slice) == 2:
          import plot2d
          plot2d.subplot_2d (self, v, ts, extent_mean, extent_var, infolines, save, topography, cut_slice, log_mean, log_var, contour, poster, num_contours, level, type, movie)
        else:
          import plot1d
          plot1d.subplot_1d (self, v, ts, extent_mean, extent_var, frame, infolines, save, topography, cut_slice, log_var, loc, level, type, movie)
  
  # === movie generation routines
  
  def movie (self, v, duration=None, infolines=False, last_extent=True, output_formats=['pdf','mp4'], still_frames=25, cut_slice=None, topography=None, log_mean=False, log_var=False, view=None, contour=False, num_contours=9, contour_levels_mean=None, contour_levels_var=None, clip=None, highlight=None, save=None, level=None, type=None, engine=None):
    
    if level != None: data_io.check_level_datafiles (self, level)
    
    # for movie generation, 'cut_slice' option is relevant only for 1d and 2d,
    # hence 'self.subplot()' routine from this file is used in 'movie_1d()' and 'movie_2d()'
    # for 3d case, 'cut_slice' option is irrelevant, hence internal plotting routine 'plot_3d()' can be used
    
    if self.ny == 1:
      import plot1d
      plot1d.movie_1d (self, v, duration, infolines, last_extent, output_formats, still_frames, cut_slice, topography, log_var, level, type, save)
    
    elif self.nz == 1:
      import plot2d
      plot2d.movie_2d (self, v, duration, infolines, last_extent, output_formats, still_frames, cut_slice, topography, log_mean, log_var, contour, num_contours, level, type, save)
    
    else:
      plot3d = __import__('plot3d_' + self.backend_3d)
      plot3d.movie_3d (self, v, duration, infolines, last_extent, output_formats, still_frames, view, log_mean, log_var, contour, num_contours, contour_levels_mean, contour_levels_var, clip, highlight, level, type, save, engine)
  
  def movie_rotate (self, v, ts=None, duration=None, frames=100, infolines=False, output_formats=['pdf','mp4'], still_frames=0, log_mean=False, log_var=False, view=None, contour=False, num_contours=9, contour_levels_mean=None, contour_levels_var=None, clip=None, highlight=None, save=None, level=None, type=None, engine=None):
    
    if level != None: data_io.check_level_datafiles (self, level)
    
    if ts is None: ts = len(self.times) - 1
    
    # only for 3d
    
    if self.nz == 1 or cut_slice != None:
      raise Exception (" :: ERROR: 'movie_rotate()' is available only for 3d simulation data! Use 'movie()' routine.")
    
    else:
      plot3d = __import__('plot3d_' + self.backend_3d)
      plot3d.movie_rotate_3d (self, v, ts, duration, frames, infolines, output_formats, still_frames, view, log_mean, log_var, contour, num_contours, contour_levels_mean, contour_levels_var, clip, highlight, level, type, save, engine)
  
  # === specific plotting and movie generation routines
  
  # plotting routines for 2d visualization of 1d and sliced 2d simulation data with histograms
  
  def plot_histograms (self, v, ts=None, infolines=True, save=None, cut_slice=None, pick=0, smoothen=False, support=True, movie=False, level=None, type=None):
    
    if level != None: data_io.check_level_datafiles (self, level)
    
    if self.ny == 1:
      import plot2d_histograms
      plot2d_histograms.plot_histograms (self, v, ts, infolines, save, cut_slice, pick, smoothen, support, level, type, movie)
    
    elif self.nz == 1:
      if cut_slice:
        import plot2d_histograms
        plot2d_histograms.plot_histograms (self, v, ts, infolines, save, cut_slice, pick, smoothen, support, level, type, movie)
      else:
        raise Exception (" :: ERROR: 'plot_histograms()' is NOT available for 2d simulation data! Use 'cut_slice' option.")
    
    else:
      raise Exception (" :: ERROR: 'plot_histograms()' is NOT available for 3d simulation data!")
  
  # movie generation routines for 2d visualization of 1d and sliced 2d simulation data with histograms
  
  def movie_histograms (self, v, ts=None, duration=None, frames=500, infolines=False, output_formats=['pdf','mp4'], cut_slice=None, smoothen=False, support=True, save=None, level=None, type=None):
    
    if level != None: data_io.check_level_datafiles (self, level)
    
    if ts is None: ts = len(self.times) - 1
    
    if self.ny == 1:
      import plot2d_histograms
      plot2d_histograms.movie_histograms (self, v, ts, duration, frames, infolines, output_formats, cut_slice, smoothen, support, level, type, save)
    
    elif self.nz == 1:
      if cut_slice:
        import plot2d_histograms
        plot2d_histograms.movie_histograms (self, v, ts, duration, frames, infolines, output_formats, cut_slice, smoothen, support, level, type, save)
      else:
        raise Exception (" :: ERROR: 'movie_histograms()' is NOT available for 2d simulation data! Use 'cut_slice' option.")
    
    else:
      raise Exception (" :: ERROR: 'movie_histograms()' is NOT available for 3d simulation data!")
  
  # plot wave speeds
  
  def plot_speeds (self, infolines=True, bins=20, save=None):
    
    import plot1d
    plot1d.plot_loads (self, infolines, bins, save, compute_speeds=True)
  
  # plot loads
  
  def plot_loads (self, infolines=True, bins=20, save=None):
    
    import plot1d
    plot1d.plot_loads (self, infolines, bins, save)
  
  # plot imbalance
  
  def plot_imbalance (self, infolines=True, save=None):
    
    import plot1d
    plot1d.plot_imbalance (self, infolines, save)
  
  # for 3d visualization of 2d simulation data from shallow water equations
  
  def plot_sw_3d (self, ts=None, infolines=False, view={'azimuth':270, 'elevation':75, 'distance':-0.4, 'yaw':4}, save=None):
    
    if backend_3d == 'mayavi':
      import plot3d_mayavi as plot3d
    else:
      Exception (" :: ERROR: 'plot_sw_3d()' is available only for 'backend_3d' == 'mayavi'.")
    
    plot3d.plot_sw_3d (self, ts, infolines, view, save)
  
  def plot_sw_3d_topography (self, infolines=False, view={'azimuth':270, 'elevation':75, 'distance':-0.8, 'yaw':4}, title=True, save=None):
    
    if backend_3d == 'mayavi':
      import plot3d_mayavi as plot3d
    else:
      Exception (" :: ERROR: 'plot_sw_3d_topography()' is available only for 'backend_3d' == 'mayavi'.")
    
    plot3d.plot_sw_3d_topography (self, infolines, view, title, save)
  
  def movie_sw_3d (self, duration=4, infolines=False, output_formats=['pdf','mp4'], view={'azimuth':270, 'elevation':75, 'distance':-0.4, 'yaw':4}):
    
    if backend_3d == 'mayavi':
      import plot3d_mayavi as plot3d
    else:
      Exception (" :: ERROR: 'movie_sw_3d()' is available only for 'backend_3d' == 'mayavi'.")
    
    plot3d.movie_sw_3d(self, duration, infolines, output_formats, view)
  
  # === exporters
  
  def matlab (self, v, ts=None, path=''):
    data_io.matlab(self, v, ts, path)
  
  # === helpers
  
  def generateInfo (self, prefix='', subplots=True, level=None, type=None):
    return data_io.generateInfo (self, prefix, subplots, level, type)
  
  def generateTexTable (self, save=None):
    data_io.generateTexTable (self, save)
  
  def generateTexTableVertical (self, save=None):
    data_io.generateTexTableVertical (self, save)

print ("")

try:
  r = Run();
  print (" :: Default run loaded as \"r\".")
except:
  print (" :: Default run not found!")
  # for debugging only
  #raise

print ("")

# import useful functions
from variables import *
from helpers import *
