
# # # # # # # # # # # # # # # # # # # # # # # # # #
# I/O routines of multi-dimensional data          #
#                                                 #
# Jonas Sukys                                     #
# SAM, ETH Zurich, Switzerland                    #
# sukys.jonas@gmail.com                           #
# # # # # # # # # # # # # # # # # # # # # # # # # #

# === global imports

import os

# === helpers

class Config (object):
  def __init__(self, run, conv):
    self.run = run
    self.conv = conv
  
  def __getattr__(self, n):
    return self.conv(self.run.vars[n])

# === configuration

import sys
sys.path.append( os.path.dirname(__file__) +  "/.." )
from local_visualization import backend_3d

# === routines

# initializes "Run" object
def init_run (self, name, path, load_times):
  
  if name:
    self.outputf   = path + 'output_'            + name + '.dat'
    self.varf      = path + 'var_output_'        + name + '.dat'
    self.histf     = path + 'histograms_output_' + name + '.dat'
    self.debugf    = path + 'debug_output_'      + name + '.dat'
    self.timesf    = path + 'times_'             + name + '.py'
    self.configf   = path + 'config_'            + name + '.py'
    self.runtimef  = path + 'runtime_'           + name + '.py'
    self.loadsf    = path + 'loads_'             + name + '.dat'
    self.balancerf = path + 'balancer_'          + name + '.dat'
    self.samplesf  = path + 'samples_'           + name + '.py'
    self.movief    = path + 'movie_'             + name
    self.rotatef   = path + 'rotate_'            + name
    self.name      = name
  else:
    self.outputf   = path + 'output.dat'
    self.varf      = path + 'var_output.dat'
    self.histf     = path + 'histograms_output.dat'
    self.debugf    = path + 'debug_output.dat'
    self.timesf    = path + 'times.py'
    self.configf   = path + 'config.py'
    self.runtimef  = path + 'runtime.py'
    self.loadsf    = path + 'loads.dat'
    self.balancerf = path + 'balancer.dat'
    self.samplesf  = path + 'samples.py'
    self.movief    = path + 'movie'
    self.rotatef   = path + 'rotate'
    self.name      = path + 'run'
  
  self.load_times = load_times
  if load_times:
    times = {}
    execfile(self.timesf, globals(), times)
    self.times = times['time']
  
  config = {}
  execfile(self.configf, globals(), config)
  self.opts = config['opts']
  self.vars = config['used']
  if 'info' in config:
    self.info = config['info']
  else:
    self.info = None
  
  self.runtime = {}
  execfile(self.runtimef, globals(), self.runtime)
  
  if os.path.exists (self.samplesf):
    samples = {}
    execfile(self.samplesf, globals(), samples)
    self.samples = samples['samples']
  else:
    self.samples = None
  
  data = {}
  execfile(self.runtimef, globals(), data)
  self.default_parallelization = data['default_parallelization'] if 'default_parallelization' in data else None
  
  # setting STATS_SIZE variable for legacy runs
  if 'STATS_SIZE' not in self.vars:
    self.vars['STATS_SIZE'] = '1'
  if not self.vars['STATS_SIZE'].isdigit():
    self.vars['STATS_SIZE'] = self.vars[self.vars['STATS_SIZE']]
  
  # setting STATS_COUNT variable for legacy runs
  if 'STATS_COUNT' not in self.vars:
    if int(self.vars['COMPUTE_VARIANCE']):
      self.vars['STATS_COUNT'] = '2'
    else:
      self.vars['STATS_COUNT'] = '1'
  
  self.i = Config(self, int)
  self.f = Config(self, float)
  self.nx = self.i.NX
  self.ny = self.i.NY
  self.nz = self.i.NZ
  self.dxyz = (self.f.MAXX/self.nx, self.f.MAXY/self.ny, self.f.MAXZ/self.nz)
  
  # extracting the data (infolines) to be displayed on the plots
  self.info_opts = self.opts
  self.info_vars = config['vars']
  for key in list(self.info_vars):
    if key .endswith ( ('_FILE', '_PATH') ):
      del self.info_vars [ key ]
  
  # detecting number of variables from equation type
  if 'equation' in self.opts:
    from variables import get_NVARS
    self.NVARS = get_NVARS (self.opts['equation'])
  else:
    self.NVARS = 16
  
  # set 3d plotting backend
  self.backend_3d = backend_3d
  
  # set flag for variance
  self.VARIANCE = self.i.STATS_COUNT == 2 and (self.i.L != 0 or self.i.ML != 1) and ('USE_ONE_SAMPLE_PER_LEVEL' not in self.vars or not self.i.USE_ONE_SAMPLE_PER_LEVEL)
  
  # set correct MULTI*
  if 'MULTI_CORES' in self.vars and self.i.MULTI_CORES > 0:
    self.i.MULTIX = self.runtime['default_parallelization'][0][0]
    self.i.MULTIY = self.runtime['default_parallelization'][0][1]
    self.i.MULTIZ = self.runtime['default_parallelization'][0][2]
    self.i.MULTIM = self.runtime['default_parallelization'][0][3]

# returns numpy array with simulation data at time 'ts' of variable 'v'
# if 'cut_slice' is specified, only data required for that slice is loaded
# if 'level' and 'type' are specified, only MC estimate for that mesh_level of that type is loaded (requires WRITE_LEVEL_DATAFILES=1)
def var (self, v, ts, stat, cut_slice, level, type):
  
  if self.load_times and ts > len(self.times):
    raise Exception ("Timestep out of bounds")
  if stat == "histogram":
    filename = self.histf
  elif stat == "mean":
    filename = self.outputf
  elif stat == "variance":
    filename = self.varf
  else:
    raise Exception ("Data file for '" + stat + "' does not exist")
  
  if level != None and type != None:
    filename += '_MESH_LEVEL=' + str(level) + '_TYPE=' + str(type)
  
  f = open(filename, 'rb')
  
  # reads segments of floats from binary file:
  # seeks to float 'start + offset', then reads 'count' segments of size 'read_size'
  # 'stride' indicates that there are gaps of size 'stride - read_size' between segments
  def load (f, start, offset, read_size, stride, count):
    
    # adjust arguments for STATS_SIZE > 1
    if self.i.STATS_SIZE > 1:
      start     *= self.i.STATS_SIZE
      offset    *= self.i.STATS_SIZE
      read_size *= self.i.STATS_SIZE
    
    import numpy
    
    d = numpy.array([], dtype=numpy.float64);
    for i in range(count):
      f.seek(8*(start+offset+i*stride))
      d = numpy.append(d, numpy.fromfile(f, dtype=numpy.float64, count=read_size))
    
    f.close()
    return d
  
  # compute nx, ny, nz according to the required mesh level
  nx = self.nx
  ny = self.ny
  nz = self.nz
  if level != None:
    if (self.nx > 1): nx = self.nx / (2**(self.i.L - level))
    if (self.ny > 1): ny = self.ny / (2**(self.i.L - level))
    if (self.nz > 1): nz = self.nz / (2**(self.i.L - level))
  
  # special treatment of functional data files
  if 'STATS_FUNCTIONAL' in self.vars and self.i.STATS_FUNCTIONAL:
    
    start     = ts
    offset    = 0
    read_size = 1
    stride    = 0
    count     = 1
    
    return load (f, start, offset, read_size, stride, count) [0]
  
  # file needs to be read in segments
  if cut_slice:
    
    # 1d case
    if self.ny == 1:
      raise Exeption ("No more slices to cut for (d=1)")
    
    # 2d case
    elif self.nz == 1:
      
      # cut at some x
      if cut_slice['dir'].upper() == 'X':
        slice_at = round(float(cut_slice['pos']) / self.f.MAXX * nx)
        
        start     = nx * ny * (v.i + self.NVARS*ts)
        offset    = slice_at * ny
        read_size = ny
        stride    = 0
        count     = 1
        
        d = load (f, start, offset, read_size, stride, count)
        
        if self.i.STATS_SIZE > 1:
          d.shape = (ny, self.i.STATS_SIZE)
        else:
          d.shape = (ny)
        
        return d
      
      # cut at some y
      else:
        slice_at = round(float(cut_slice['pos']) / self.f.MAXY * ny)
        
        start     = nx * ny * (v.i + self.NVARS*ts)
        offset    = slice_at
        read_size = 1
        stride    = ny
        count     = nx
        
        d = load (f, start, offset, read_size, stride, count)
        
        if self.i.STATS_SIZE > 1:
          d.shape = (nx, self.i.STATS_SIZE)
        else:
          d.shape = (nx)
        
        return d
    
    # 3d case
    else:
      
      # cut at some x
      if cut_slice['dir'].upper() == 'X':
        slice_at  = round(float(cut_slice['pos']) / self.f.MAXX * nx)
        start     = nx * ny * nz * (v.i + self.NVARS*ts)
        offset    = slice_at * ny * nz
        read_size = ny * nz
        stride    = 0
        count     = 1
        
        if self.i.STATS_SIZE > 1:
          d.shape = (ny, nz, self.i.STATS_SIZE)
        else:
          d.shape = (ny, nz)
        
        if len(cut_slice) == 2:
          return d
        
        else:
          
          # cut additionally at y
          if cut_slice['dir2'].upper() == 'Y':
            slice_at  = round(float(cut_slice['pos2']) / self.f.MAXY * ny)
            if self.i.STATS_SIZE > 1:
              return d[slice_at,:,:]
            else:
              return d[slice_at,:]
          
          # cut additionally at z
          else:
            slice_at  = round(float(cut_slice['pos2']) / self.f.MAXZ * nz)
            if self.i.STATS_SIZE > 1:
              return d[:,slice_at,:]
            else:
              return d[:,slice_at]
      
      # cut at some y
      elif cut_slice['dir'].upper() == 'Y':
        slice_at  = round(float(cut_slice['pos']) / self.f.MAXY * ny)
        start     = nx * ny * nz * (v.i + self.NVARS*ts)
        offset    = slice_at * nz
        read_size = nz
        stride    = ny * nz
        count     = nx
        
        d = load (f, start, offset, read_size, stride, count)
        
        if self.i.STATS_SIZE > 1:
          d.shape = (nx, nz, self.i.STATS_SIZE)
        else:
          d.shape = (nx, nz)
        
        if len(cut_slice) == 2:
          return d
        
        else:
          
          # cut additionally at x
          if cut_slice['dir2'].upper() == 'X':
            slice_at  = round(float(cut_slice['pos2']) / self.f.MAXX * nx)
            if self.i.STATS_SIZE > 1:
              return d[slice_at,:,:]
            else:
              return d[slice_at,:]
          
          # cut additionally at z
          else:
            slice_at  = round(float(cut_slice['pos2']) / self.f.MAXZ * nz)
            if self.i.STATS_SIZE > 1:
              return d[:,slice_at,:]
            else:
              return d[:,slice_at]
      
      # cut at some z
      else:
        slice_at  = round(float(cut_slice['pos']) / self.f.MAXZ * nz)
        start     = nx * ny * nz * (v.i + self.NVARS*ts)
        offset    = slice_at
        read_size = 1
        stride    = nz
        count     = nx * ny
        
        d = load (f, start, offset, read_size, stride, count)
        
        if self.i.STATS_SIZE > 1:
          d.shape = (nx, ny, self.i.STATS_SIZE)
        else:
          d.shape = (nx, ny)
        
        if len(cut_slice) == 2:
          return d
        
        else:
          
          # cut additionally at x
          if cut_slice['dir2'].upper() == 'X':
            slice_at  = round(float(cut_slice['pos2']) / self.f.MAXX * nx)
            if self.i.STATS_SIZE > 1:
              return d[slice_at,:,:]
            else:
              return d[slice_at,:]
          
          # cut additionally at y
          else:
            slice_at  = round(float(cut_slice['pos2']) / self.f.MAXY * ny)
            if self.i.STATS_SIZE > 1:
              return d[:,slice_at,:]
            else:
              return d[:,slice_at]
  
  # file needs to be read contiguously
  else:
    
    n = nx * ny * nz
    d = load (f, n * (v.i + self.NVARS*ts), 0, n, 0, 1)
    
    if self.i.STATS_SIZE > 1:
      
      if self.ny == 1:
        d.shape = (nx, self.i.STATS_SIZE)
      elif self.nz == 1:
        d.shape = (nx, ny, self.i.STATS_SIZE)
      else:
        d.shape = (nx, ny, nz, self.i.STATS_SIZE)
      
      return d
    
    else:
      
      if self.ny == 1:
        d.shape = (nx)
      elif self.nz == 1:
        d.shape = (nx, ny)
      else:
        d.shape = (nx, ny, nz)
      
      return d

# export data to Matlab's MAT files
def matlab (self, v, ts, path):
  
  import numpy
  
  if ts is None: ts = len(self.times) - 1
  data = {}
  
  data['x'] = numpy.arange(self.nx)*(self.f.MAXX/self.nx)+0.5*self.f.MAXX/self.nx
  
  if not self.ny == 1:
    data['y'] = numpy.arange(self.ny)*(self.f.MAXY/self.ny)+0.5*self.f.MAXY/self.ny
  
  if not self.nz == 1:
    data['z'] = numpy.arange(self.nz)*(self.f.MAXZ/self.nz)+0.5*self.f.MAXZ/self.nz
  
  data['mean'] = self.var(v, ts, 'mean')
  
  if self.VARIANCE:
    data['variance'] = self.var(v, ts, 'variance')
  
  if 'wallclock_runtime_global' in self.runtime:
    data['time'] = self.runtime['wallclock_runtime_global']
  elif 'wall_runtime_total' in self.runtime:
    data['time'] = self.runtime['wall_runtime_total'] # legacy
  else:
    data['time'] = self.runtime['clock_runtime']
  
  from scipy.io import savemat
  savemat(path + self.name + '_' + v.name + '_' + str(ts) + '.mat', data)

# generate infolines
def generateInfo (self, prefix, subplots, level, type):
  
  # filter out default alternative choices to reduce output size
  info_opts = self.info_opts
  
  # if MULTI_CORES > 0, other MULTI* variables are ignored
  if 'MULTI_CORES' in self.vars and self.i.MULTI_CORES:
    if 'MULTIX' in info_opts: del info_opts['MULTIX']
    if 'MULTIY' in info_opts: del info_opts['MULTIY']
    if 'MULTIZ' in info_opts: del info_opts['MULTIZ']
    if 'MULTIM' in info_opts: del info_opts['MULTIM']
  
  # create strings from dictionaries
  string_opts = ''.join([' %s:%s,' % x for x in sorted(info_opts.items())])
  string_vars = ''.join([' %s=%s' % x for x in sorted(self.info_vars.items())])
  
  # text formatter
  def format_text (text, subplots):
    if not subplots and len(text) > 65:
      cut_pos = len(text)/2
      cut_pos = text.find(' ', cut_pos)
      text = text[:cut_pos] + '\n' + ' ' * 9 + text[cut_pos:]
    return text
  
  # OPTS
  text  = prefix + format_text('OPTS:' + string_opts[0:-1], subplots) + '\n'
  
  # VARS
  text += prefix + format_text('VARS:' + string_vars, subplots) + '\n'
  
  if 'wallclock_runtime' in self.runtime:
    runtime = self.runtime['wallclock_runtime']
  elif 'wall_runtime' in self.runtime:
    runtime = self.runtime['wall_runtime'] # legacy
  else:
    runtime = self.runtime['clock_runtime']
  
  import datetime
  runtime = str(datetime.timedelta(seconds=round(runtime)))
  
  string_info = ' '
  
  if self.info:
    string_info += str(self.info['CLUSTER']) + ', '
  
  if 'cores' in self.runtime:
    string_info += 'cores: ' + str(self.runtime['cores']) + ', '
  else: # legacy
    string_info += 'cores: ' + str(self.runtime['nodes']) + ', '
  
  string_info += 'runtime: ' + runtime
  
  if 'wallclock_runtime_global_balancer' in self.runtime:
    string_info += ' (Bal: ' + str(int(round(100 * (self.runtime['wallclock_runtime_global_balancer'] + self.runtime['wallclock_runtime_global_loads']) / self.runtime['wallclock_runtime_global']))) + r'%)'
  
  if 'mpi_efficiency' in self.runtime:
    string_info += ', ' + 'efficiency: ' + str(int(round(100 * self.runtime['mpi_efficiency']))) + r'%'
  
  # legacy
  if 'mpi_scaling' in self.runtime:
    string_info += ', ' + 'efficiency: ' + str(int(round(100 * self.runtime['mpi_scaling']))) + r'%'
  
  if 'wallclock_runtime_global_mpi_ddm' in self.runtime:
    runtime_global = self.runtime['wallclock_runtime_global']
    string_info += ', ' + 'overhead: '
    string_info   += 'DDM '  + str(int(round(100 * self.runtime['wallclock_runtime_global_mpi_ddm']  / runtime_global)))      + r'%, '
    string_info   += 'MC '   + str(int(round(100 * self.runtime['wallclock_runtime_global_mpi_mc']   / runtime_global)))      + r'%, '
    string_info   += 'MLMC ' + str(int(round(100 * self.runtime['wallclock_runtime_global_mpi_mlmc'] / runtime_global)))      + r'%, '
    string_info   += 'idle ' + str(int(round(100 * self.runtime['wallclock_runtime_global_mpi_idle'] / runtime_global)))      + r'%'
  
  # INFO
  text += prefix + format_text('INFO:' + string_info, subplots)
  
  # SAMPLES
  samples_available = self.samples and (self.i.L != 0 or self.i.ML != 1 or self.i.READ_NUMBER_OF_SAMPLES_FROM_FILE)
  if samples_available:
    text += '\n' + prefix + 'SAMPLES: ' + str(self.samples[::-1])[1:-1]
  
  # MESH_LEVEL and TYPE
  if level != None:
    text += ' (only the %d-sample MC estimate from MESH_LEVEL=%d and TYPE=%d is displayed)' % (self.samples[level+type], level, type)
  
  # default parallelization
  if self.default_parallelization:
    if samples_available:
      text += ' | '
    else:
      text += '\n'
    text += 'PARALLELIZATION:'
    for level in range(len(self.default_parallelization)):
      text += ' ' + str(self.default_parallelization[level][3]) + 'x(' + str(self.default_parallelization[level][0])
      if self.i.NY > 1: text += 'x' + str(self.default_parallelization[level][1])
      if self.i.NZ > 1: text += 'x' + str(self.default_parallelization[level][2])
      text += ')'
  
  return text

# compute parameters needed for the TexTable
def getTexTableConfig (self, vertical=False):
  
  # config
  
  keys     =  ['L',    'ML',     'grid_size', 'CFL', 'cores', 'runtime']
  captions =  [r'$L$', r'$M_L$', 'grid size', 'CFL', 'cores', 'runtime']
  
  if 'mpi_efficiency' in self.runtime or 'mpi_scaling' in self.runtime:
    keys     += ['efficiency']
    captions += ['efficiency']
  
  # aggregation of information
  
  opts = {}
  opts.update(self.opts)
  opts.update(self.vars)
  
  grid_size = str(opts['NX'])
  if vertical:
    grid_size = r'$%s^%s$' % ( grid_size, str( (self.nx != 1) + (self.ny != 1) + (self.nz != 1) ) )
  else:
    if self.ny != 1:
      grid_size += 'x' + str(opts['NY'])
    if self.nz != 1:
      grid_size += 'x' + str(opts['NZ'])
  opts.update( { 'grid_size' : grid_size } )
  
  if 'cores' in self.runtime:
    opts.update( { 'cores' : str(self.runtime['cores']) } )
  else: # legacy
    opts.update( { 'cores' : str(self.runtime['nodes']) } )
  
  if 'wallclock_runtime' in self.runtime:
    runtime = self.runtime['wallclock_runtime']
  elif 'wall_runtime' in self.runtime:
    runtime = self.runtime['wall_runtime'] # legacy
  else:
    runtime = self.runtime['clock_runtime']
  
  import datetime
  runtime = str(datetime.timedelta(seconds=round(runtime)))
  
  opts.update( { 'runtime' : str(runtime) } )
  
  if 'mpi_efficiency' in self.runtime:
    opts.update( { 'efficiency' : str(round(1000 * self.runtime['mpi_efficiency']) / 10.0) + r'\%' } )
  
  # legacy
  if 'mpi_scaling' in self.runtime:
    opts.update( { 'efficiency' : str(round(1000 * self.runtime['mpi_scaling']) / 10.0) + r'\%' } )
  
  return [keys, captions, opts]

# generate LaTeX table (horizontal version)
def generateTexTable (self, save):
  
  # get the config
  
  [keys, captions, opts] = getTexTableConfig (self)
  
  # TeX code generation
  
  columns =  '|' + 'c|' * len(keys)
  
  text =  '\n'
  text += r'\begin{tabular}{%s}' % columns + '\n'
  
  text += r'\hline' + '\n'
  text += (r'%s & ' * len(captions))[0:-2] % tuple(captions) + r'\\' + '\n'
  text += r'\hline' + '\n'
  text += (r'%s & ' * len(keys))[0:-2] % tuple([opts[key] for key in keys]) + r'\\' + '\n'
  text += r'\hline' + '\n'
  
  text += r'\end{tabular}' + '\n'
  
  # saving
  
  f = open(save[0:-4] + '.tex', 'w')
  f.write(text)
  f.close()

def generateTexTableVertical (self, save):
  
  # get the config (specialized for vertical setting)
  
  [keys, captions, opts] = getTexTableConfig (self, vertical=True)
  
  # TeX code generation
  
  #columns =  '|' + 'c|' * len(keys)
  
  text =  '\n'
  text += r'\begin{tabular}{|c|c|}' + '\n'
  
  text += r'\hline' + '\n'
  
  for row, key in enumerate(keys):
    text += r'%s & %s \\' % (captions[row], opts[key]) + '\n'
    text += r'\hline' + '\n'
  
  text += r'\end{tabular}' + '\n'
  
  # saving
  
  f = open(save[0:-4] + '-vertical.tex', 'w')
  f.write(text)
  f.close()

# check if level-datafiles are available
def check_level_datafiles (self, level):
  
  if not self.i.WRITE_LEVEL_DATAFILES:
    raise Exception (" :: ERROR: Level-datafiles not available. Use WRITE_LEVEL_DATAFILES=1.")
  if self.i.L == 0 and self.i.ML == 1:
    raise Exception (" :: ERROR: Level-datafiles are not available for deterministic runs.")
