
# # # # # # # # # # # # # # # # # # # # # # # # # #
# Generation of 'bc.cpp' file (boundary cond's)   #
#                                                 #
# Jonas Sukys                                     #
# SAM, ETH Zurich, Switzerland                    #
# sukys.jonas@gmail.com                           #
# # # # # # # # # # # # # # # # # # # # # # # # # #

# === configuration

SUPPORTED_BC = ['NONE', '*', 'REPEAT', 'NEUMANN', 'WAVE_REFLECT', 'ZERO']

BC_HEADER  = "\n// This file is automatically generated by 'make_bc.py' script called from 'make.py'\n\n"
BC_HEADER += '#include "equation.h"\n'
BC_HEADER += '#include "multi.h"\n'

INDICES = ['i', 'j', 'k']
PARTS   = ["HEAD", "TAIL"]
HEAD    = 0
TAIL    = 1

BC_STD_LHS = "v [%s][%s][%s]"
BC_STD_RHS = "v [%s][%s][%s]"

# === template

BC_FUNC = """
template <class T>
void set_bc_%s_%s (volume<T> &v, const int &i, const int &j, const int &k) {
  %s
}
"""

# === generator

def bc (f, spec, settings):
    
    # write the header of the bc.cpp
    f.write (BC_HEADER)
    
    # reshape spec for easier indexing
    spec = [[spec[0], spec[3]], [spec[1], spec[4]], [spec[2], spec[5]]]
    
    # process given boundary conditions
    for d in range(3):
        
        for p, ps in enumerate(PARTS):
          
          # === set the required flags for the source code
          
          if spec[d][p] not in SUPPORTED_BC:
            Exception (" :: ERROR: unsupported boundary conditions specified.")
        
          if spec[d][p] == 'REPEAT':
              settings[1]['BC_REPEAT_' + 'XYZ'[d]] = 1
          
          if spec[d][p] == 'ZERO_FLUX':
              settings[1]['BC_ZERO_FLUX_' + 'XYZ'[d] + '_' + ps ] = 1
          
          # === specify the indices of the GHOST cells
          
          # copy the lists of indices
          vl = INDICES + []
          
          # specify the HEAD ghost cells in direction 'd'
          if p == HEAD:
            vl[d] = '-1-' + vl[d]
          
          # specify the TAIL ghost cells in direction 'd'
          if p == TAIL:
            vl[d] += '+N' + 'XYZ'[d] + 'l'
          
          # === specify the indices of the DOMAIN cells
          
          # copy the lists of indices
          # HEAD domain cells in direction 'd' are already correctly specified
          vr = INDICES + []
          
          # specify the correct TAIL domain cells in direction 'd'
          if p == TAIL and spec[d][p] != 'REPEAT':
            vr[d] = 'N' + 'XYZ'[d] + 'l' + '-1-' + vr[d]
          
          # === special cases
          
          # for periodic b.c., HEAD ghost cells should copy values from the TAIL domain cells
          if p == HEAD and spec[d][p] == 'REPEAT':
            vr[d] = 'N' + 'XYZ'[d] + 'l' + '-1-' + vr[d]
          
          # === form the boundary condition
          
          fv = ['xyz'[d]] + [ps]
          
          # special cases
          
          if spec[d][p] == 'ZERO_FLUX':
            fv += [ '' ]
          
          elif spec[d][p] == 'ZERO':
            fv += [ 'Zero ( ' + BC_STD_LHS % tuple(vl) + ')' + ';' ]
          
          elif spec[d][p] == 'WAVE_REFLECT':
            bc_P = BC_STD_LHS % tuple(vl) + ' .p ' + ' = ' + BC_STD_RHS % tuple(vr) + ' .p '
            bc_U = BC_STD_LHS % tuple(vl) + ' .u ' + ' = (-1) * ' + BC_STD_RHS % tuple(vr) + ' .u '
            fv += [ bc_P + ';\n  ' + bc_U + ';' ]
          
          # generic case
          else:
            fv += [ BC_STD_LHS % tuple(vl) + ' = ' + BC_STD_RHS % tuple(vr) + ';' ]
          
          f.write (BC_FUNC % tuple(fv))
          
    f.close()
