
# # # # # # # # # # # # # # # # # # # # # # # # # #
# Plotting routines for 1d simulation data        #
#                                                 #
# Jonas Sukys                                     #
# SAM, ETH Zurich, Switzerland                    #
# sukys.jonas@gmail.com                           #
# # # # # # # # # # # # # # # # # # # # # # # # # #

# === global imports

import pylab

import numpy

# === local imports

from variables import *
from helpers import *

# === configuration

line_format_single    = 'b'
line_format_upper_std = 'g--'
line_format_lower_std = 'r--'

# for B&W plots only
#line_format_single    = 'k'
#line_format_upper_std = 'k:'
#line_format_lower_std = 'k--'

import matplotlib as mpl
mpl.rcParams['font.size'] = 18
mpl.rcParams['legend.fontsize'] = 16

# === plotting routines

def plot_topography_1d(self, xr, ts, cut_slice, level, type):
    
    b_mean = self.var(sw.B, ts, "mean", cut_slice, level, type)
    
    if not self.VARIANCE:
      pylab.plot(xr, b_mean, "k-", label="topography")
    else:
      pylab.plot(xr, b_mean, "k-")
    
    if self.VARIANCE:
      b_var = self.var(sw.B, ts, "variance", cut_slice, level, type)
      b_std = numpy.sqrt(b_var)
      pylab.plot(xr, b_mean + b_std, "k--")
    
    ymin, ymax = pylab.ylim()
    pylab.fill_between(xr, b_mean, ymin, alpha=0.5, color='black')

def plot_1d(self, v, ts, extent, frame, infolines, save, topography, cut_slice, loc, level, type, adjust_subplots=True):
  
  if ts is None: ts = len(self.times) - 1
  
  (xr, title_suffix) = process_slice_data (self, cut_slice, ts)
  
  mean = self.var(v, ts, "mean", cut_slice, level, type)
  
  if int(self.vars['L']) != 0 or int(self.vars['ML']) != 1:
    mean_label = "mean"
  elif self.name[-3:] != "run":
    mean_label = self.name
  else:
    mean_label = "solution"
  
  pylab.plot(xr, mean, line_format_single, label=mean_label)
  
  if self.VARIANCE:
    var = self.var(v, ts, "variance", cut_slice, level, type)
    std = numpy.sqrt(var);
    upper_std = mean + std;
    lower_std = mean - std;
    pylab.plot(xr, upper_std, line_format_upper_std, label="mean + std")
    pylab.plot(xr, lower_std, line_format_lower_std, label="mean - std")
  
  if topography is None: topography = True
  
  if self.opts['equation'] == 'sw' and topography:
    plot_topography_1d(self, xr, ts, cut_slice, level, type)
  
  if extent:
      pylab.ylim(*extent)
  
  pylab.title(v.name + title_suffix)
  
  pylab.legend(loc=loc)
  
  if adjust_subplots:
    pylab.subplots_adjust(right=0.96)
    pylab.subplots_adjust(left=0.10)
    # bugfix in pylab for topography plotting
    if not self.opts['equation'] == 'sw' or not topography:
      pylab.subplots_adjust(top=0.94)
  
  if infolines:
    show_info(self, False, level, type)
  
  if save:
    pylab.savefig(save)
    pylab.savefig(save[:-3] + 'eps')
    self.generateTexTable(save)
  if not frame:
    pylab.show()

def plot0_1d(self, v, ts, extent, frame, infolines, save, topography, cut_slice, loc, level, type):
  
  if not frame:
    if infolines:
      pylab.figure(figsize=(2*8,6))
    else:
      pylab.figure(figsize=(2*8,5))
  
  pylab.subplot(121)
  plot_1d (self, v, 0, extent, True, False, None, topography, cut_slice, loc, level, type, adjust_subplots=False)
  
  pylab.subplot(122)
  plot_1d (self, v, ts, extent, True, False, None, topography, cut_slice, loc, level, type, adjust_subplots=False)
  
  pylab.subplots_adjust(right=0.96)
  pylab.subplots_adjust(left=0.04)
  # bugfix in pylab for topography plotting
  if not self.opts['equation'] == 'sw' or not topography:
    pylab.subplots_adjust(top=0.94)
  
  if infolines:
    show_info(self, self.VARIANCE, level, type)
  
  if save:
    pylab.savefig(save)
    pylab.savefig(save[:-3] + 'eps')
    self.generateTexTable(save)
  
  if not frame: pylab.show()

def subplot_1d_stat (self, v, ts, stat, extent, frame, infolines, save, topography, cut_slice, log, loc, level, type, movie):
  
  (xr, title_suffix) = process_slice_data (self, cut_slice, ts)
  
  d = self.var(v, ts, stat, cut_slice, level, type)
  #if log: (extent,d) = apply_log (extent,d)
  
  title_prefix = get_title_prefix (self, stat, False)
  
  if log:
    pylab.semilogy(xr, d, line_format_single, label=self.name)
  else:
    pylab.plot(xr, d, line_format_single, label=self.name)
  
  if self.opts['stats'] == 'treshold-sw':
    pylab.title('probability of E' + title_suffix)
  else:
    pylab.title(title_prefix + v.name + title_suffix)
  
  if self.opts['equation'] == 'sw' and topography:
    plot_topography_1d(self, xr, ts, cut_slice)
  
  pylab.legend(loc=loc)
  
  if extent:
    pylab.ylim(*extent)

def subplot_1d(self, v, ts, extent_mean, extent_var, frame, infolines, save, topography, cut_slice, log_var, loc, level, type, movie):
  
  if ts is None: ts = len(self.times) - 1
  
  if movie:
    if self.VARIANCE:
      pylab.figure(figsize=(2*8,6))
    else:
      pylab.figure(figsize=(8,6))
  
  # mean
  
  if self.VARIANCE: pylab.subplot(121)
  
  if topography is None: topography = True
  
  subplot_1d_stat (self, v, ts, "mean", extent_mean, frame, infolines, save, topography, cut_slice, False, loc, level, type, movie)
  
  # variance
  
  if self.VARIANCE:
    
    pylab.subplot(122)
    
    topography = False
    
    subplot_1d_stat (self, v, ts, "variance", extent_var, frame, infolines, save, topography, cut_slice, log_var, loc, level, type, movie)
  
  # finalization
  
  pylab.subplots_adjust(right=0.98)
  pylab.subplots_adjust(left=0.03)
  pylab.subplots_adjust(top=0.94)
  
  if infolines: show_info (self, self.VARIANCE, level, type)
  
  if save:
    pylab.savefig(save)
    pylab.savefig(save[:-3] + 'eps')
    if not movie: self.generateTexTable(save)
  if not frame and not movie:
    pylab.show()

def plot_loads (self, infolines, bins, save, compute_speeds=False):
  
  if 'ESTIMATE_LOADS' in self.vars and self.i.ESTIMATE_LOADS:
    loads = {}
    execfile(self.loadsf, globals(), loads)
  else:
    raise Exception ("ERROR: Loads are not available! Use ESTIMATE_LOADS=1 when compiling.")
  
  # if needed, compute wave speeds from loads
  if compute_speeds:
    ACTIVE_DIRS = (self.i.NX > 1) + (self.i.NY > 1) + (self.i.NZ > 1)
    for level in xrange(0,self.i.L+1):
      factor = 2 ** (self.i.L - level)
      loads['loads'][level] = numpy.array(loads['loads'][level])
      loads['loads'][level] *= factor ** ACTIVE_DIRS
      loads['loads'][level] *= self.f.CFL
      loads['loads'][level] /= ACTIVE_DIRS
      loads['loads'][level] *= self.dxyz[0] * factor
  
  # computing positioning of the subplots
  if self.i.L == 0:
      num_plots_x = 1
      num_plots_y = 1
  else:
      num_plots_x = 3
      num_plots_y = numpy.floor ( self.i.L/3.0 + 1 )
  
  if infolines:
    pylab.figure(figsize=(7*num_plots_x,3*num_plots_y+1))
  else:
    pylab.figure(figsize=(7*num_plots_x,3*num_plots_y))
  
  for level in xrange(0,self.i.L+1):
    pylab.subplot(num_plots_y, num_plots_x, level+1)
    log = len(loads['loads'][level]) >= 100
    pylab.hist(loads['loads'][level], label='level %d' % level, bins=bins, log=log)
    pylab.ylabel('count')
    pylab.legend(loc='best')
  
  pylab.subplots_adjust(top=0.96)
  pylab.subplots_adjust(right=0.96)
  pylab.subplots_adjust(left=0.06)
  
  if infolines:
    show_info(self)
    pylab.subplots_adjust(bottom=0.15)
  else:
    pylab.subplots_adjust(bottom=0.05)
  
  if save:
      pylab.savefig(save)
      pylab.savefig(save[:-3] + 'eps')
      self.generateTexTable(save)
  
  pylab.show()

def plot_imbalance (self, infolines, save):
  
  # === load all required data
  
  try:
    data = {}
    execfile(self.balancerf, globals(), data)
  except:
    raise Exception ("ERROR: Balancer datafile " + self.balancerf + " is not available!")
  
  total_loads_estimated = data['total_loads']
  
  try:
    data = {}
    execfile(self.runtimef, globals(), data)
  except:
    raise Exception ("ERROR: Runtime information datafile " + self.runtimef + " is not available!")
  
  total_loads_measured = numpy.array(data['wallclock_runtime_gathered_imbalance'])
  default_parallelization = data['default_parallelization']
  
  # === compute estimated total loads
  
  # 'inflate' total_loads_estimated by copying samplers as many times as there are cores in them
  total_loads_estimated_full = []
  level_indices = [0]
  for level in range(len(total_loads_estimated)):
    inflated_total_loads = []
    for sampler in range(len(total_loads_estimated[level])):
      inflated_total_loads += numpy.prod(default_parallelization[level][:3]) * [ total_loads_estimated [level] [sampler] ]
    total_loads_estimated [level] = inflated_total_loads
    total_loads_estimated_full += inflated_total_loads
    level_indices += [level_indices[-1]+ len(inflated_total_loads)]
  
  total_loads_estimated = total_loads_estimated_full
  
  # compute the average of total_loads_estimated
  
  avg_estimated = numpy.mean(total_loads_estimated)
  
  # normalize total_loads_estimated
  total_loads_estimated /= avg_estimated
  
  # === compute measured total loads
  
  # compute the average of total_loads_estimated
  avg_measured = numpy.mean(total_loads_measured)
  
  # normalize total_loads_estimated
  total_loads_measured /= avg_measured
  
  # === assuming that the shortest runtime was estimated correctly
  # compute differences and scale such that total_loads_measured >= total_loads_estimated
  
  diff = total_loads_measured - total_loads_estimated
  min = numpy.amin(diff)
  ind = numpy.where(diff==min) [0] [0]
  total_loads_measured /= (1-abs(diff[ind]))
  
  # then update the resulting differences
  diff = total_loads_measured - total_loads_estimated
  
  # === plot
  
  pylab.figure(figsize=(2*8,7))
  indices = numpy.arange(0, len(total_loads_estimated))
  
  # plot total_loads_estimated
  pe = pylab.bar(indices, total_loads_estimated, width=1, linewidth=0)
  
  # plot total_loads_measured (on top of total_loads_estimated)
  pm = pylab.bar(indices, diff, width=1, linewidth=0, color='r', bottom=total_loads_estimated)
  #pm = pylab.plot(indices + 0.5, total_loads_measured, color='r', linestyle='-', linewidth=2)
  
  # add red line between bars from different levels
  for index in level_indices[1:-1]:
    pl = pylab.axvline(x=index, color='k', linestyle='--', linewidth=2)
  
  pylab.legend([pe,pm,pl], ['estimated', 'measured', 'level separator'], loc='lower center')
  pylab.ylabel('normalized total load')
  pylab.xlabel('core index')
  pylab.xlim([0,len(indices)])
  
  pylab.subplots_adjust(top=0.96)
  pylab.subplots_adjust(right=0.96)
  pylab.subplots_adjust(left=0.06)
  
  if infolines:
    show_info(self)
    pylab.subplots_adjust(bottom=0.25)
  else:
    pylab.subplots_adjust(bottom=0.1)
  
  if save:
    pylab.savefig(save)
    pylab.savefig(save[:-3] + 'eps')
    self.generateTexTable(save)
  
  pylab.show()

# === movie generation routines

def movie_1d(self, v, duration, infolines, last_extent, output_formats, still_frames, cut_slice, topography, log_var, level, type, save):
  
  margin = 0.1
  
  if topography is None: topography = True
  
  [ext_mean, ext_var] = getExtents (self, v, last_extent, topography, margin, level, type)
  
  from helpers import Progress
  progress = Progress(" :: Generating frames: ", len(self.times))
  progress.update(0)
  
  if 'pdf' in output_formats:
    from matplotlib.backends.backend_pdf import PdfPages
    if save:
      pdf = PdfPages ( save + '.pdf')
    else:
      pdf = PdfPages ( getFileName (self, self.movief + '-' + v.name, file_suffix='pdf') )
  
  for ts in range(len(self.times)):
    self.subplot(v, False, ts, ext_mean, ext_var, infolines, topography=topography, movie=True, cut_slice=cut_slice, log_var=log_var, level=level, type=type)
    if 'pdf' in output_formats:
      pdf.savefig()
    if 'mp4' in output_formats:
      pylab.savefig('frame_%04d.png' % ts)
    if ts != len(self.times) - 1:
      pylab.close()
    progress.update(ts+1)
  
  # output PDF file with multiple pages
  if 'pdf' in output_formats:
    generate_PDF_movie (self, pdf, pylab, still_frames)
  
  # generate MP4 movies from PNG files; afterwards, delete PNG files
  if 'mp4' in output_formats:
    generate_movies (self, len(self.times), save, None, self.movief, ['mp4'], duration, still_frames)

# === helpers

def process_slice_data (self, cut_slice, ts):
  
  title_suffix = ' at t=%.2f' % self.times[ts]
  
  if cut_slice:
    
    title_suffix += ', %s=%.2f' % (cut_slice['dir'].lower(), float(cut_slice['pos']))
    
    # 1d slice from 2d data
    if len(cut_slice) == 2:
      xr = get_range (self, cut_slice['dir'].upper())
    
    # 1d slice from 3d data
    else:
      title_suffix += ', %s=%.2f' % (cut_slice['dir2'].lower(), float(cut_slice['pos2']))
      if 'X' not in cut_slice.values():
        xr = get_range (self, 'X')
      elif 'Y' not in cut_slice.values():
        xr = get_range (self, 'Y')
      else:
        xr = get_range (self, 'Z')
  
  else:
    # no cutting
    xr = get_range (self, 'X')
  
  return (xr, title_suffix)

def show_info(self, subplots=True, level=None, type=None):
  
  text = self.generateInfo(prefix='', subplots=subplots, level=level, type=type)
  
  if subplots:
    pylab.figtext(0.01, 0.015, text, fontsize='small')
    pylab.subplots_adjust(bottom=0.20)
  else:
    pylab.figtext(0.01, 0.015, text, fontsize=14)
    pylab.subplots_adjust(bottom=0.32)
