
# # # # # # # # # # # # # # # # # # # # # # # # # #
# Plotting routines for 2d simulation data        #
#                                                 #
# Jonas Sukys                                     #
# SAM, ETH Zurich, Switzerland                    #
# sukys.jonas@gmail.com                           #
# # # # # # # # # # # # # # # # # # # # # # # # # #

# === global imports

import pylab
import numpy

# === local imports

from variables import *
from helpers import *

# === configuration

import matplotlib as mpl
mpl.rcParams['font.size'] = 18
mpl.rcParams['legend.fontsize'] = 16

# === plotting routines

def plot_topography_2d (self, extent, ts, stat, cut_slice, num_contours, level, type):
  
  b = self.var(sw.B, ts, stat, cut_slice, level, type)
  pylab.contour(numpy.transpose(b), num_contours, extent=extent, colors='black')

def subplot_2d_stat (self, v, ts, stat, extent, infolines, save, topography, cut_slice, log, contour, poster, num_contours, level, type, movie, subplots_adjust):
  
  d = self.var(v, ts, stat, cut_slice, level, type)
  
  if log:
    [values_extent,d] = apply_log (extent,d)
  else:
    values_extent = extent
  
  title_prefix = get_title_prefix (self, stat, log)
  
  s = {}
  if extent:
    s['vmin'] = values_extent[0]
    s['vmax'] = values_extent[1]
  
  title_suffix = ' at t=%.2f' % self.times[ts]
  if cut_slice:
    title_suffix += ', %s=%.2f' % (cut_slice['dir'].lower(), float(cut_slice['pos']))
  
  if cut_slice:
    if cut_slice['dir'].upper() == 'X':
      axes_extent = (0,self.f.MAXY,0,self.f.MAXZ)
    elif cut_slice['dir'].upper() == 'Y':
      axes_extent = (0,self.f.MAXX,0,self.f.MAXZ)
    else:
      axes_extent = (0,self.f.MAXX,0,self.f.MAXY)
  else:
    axes_extent = (0,self.f.MAXX,0,self.f.MAXY)
  
  if contour:
    pylab.contour(numpy.transpose(d), num_contours, origin='lower', extent=axes_extent, **s)
  else:
    pylab.imshow(numpy.transpose(d), origin='lower', extent=axes_extent, **s)
  
  if 'stats' in self.opts and self.opts['stats'] == 'treshold-sw':
    pylab.title('probability of E' + title_suffix)
  else:
    pylab.title(title_prefix + v.name + title_suffix)
  
  if poster: pylab.axis('off')
  
  # MUST be before all 'pylab.colorbar()' or otherwise plot is distorted (any idea why?)
  if subplots_adjust:
    pylab.subplots_adjust(right=0.96)
    pylab.subplots_adjust(left=0.00)
    pylab.subplots_adjust(top=0.92)
    pylab.subplots_adjust(wspace=0.05)
  
  pylab.colorbar()
  if self.opts['equation'] == 'sw':
    pylab.winter()
  
  if self.opts['equation'] == 'sw' and topography and not contour:
    plot_topography_2d(self, axes_extent, ts, "mean", cut_slice, level, type)

def subplot_2d (self, v, ts, extent_mean, extent_var, infolines, save, topography, cut_slice, log_mean, log_var, contour, poster, num_contours, level, type, movie=False):
  
  if ts is None: ts = len(self.times) - 1
  
  if infolines:
    if self.VARIANCE:
      pylab.figure(figsize=(14,6))
    else:
      pylab.figure(figsize=(7,6))
  elif poster:
    pylab.figure(figsize=(56,40))
  else:
    if self.VARIANCE:
      pylab.figure(figsize=(14,5))
    else:
      pylab.figure(figsize=(7,5))
  
  # mean
  
  if self.VARIANCE: pylab.subplot(121)
  
  subplots_adjust = True
  
  subplot_2d_stat (self, v, ts, "mean", extent_mean, infolines, save, topography, cut_slice, log_mean, contour, poster, num_contours, level, type, movie, subplots_adjust)
  
  # variance
  
  if self.VARIANCE:
    
    pylab.subplot(122)
    
    subplots_adjust = False
    
    subplot_2d_stat (self, v, ts, "variance", extent_var, infolines, save, topography, cut_slice, log_var, contour, poster, num_contours, level, type, movie, subplots_adjust)
  
  # infolines
  if infolines:
    show_info(self, self.VARIANCE, level, type)
  
  # finalization
  
  if save:
    pylab.savefig(save)
    pylab.savefig(save[:-3] + 'eps')
    if not movie: self.generateTexTable(save)
  if not movie:
    pylab.show()

# === movie generation routines

def movie_2d (self, v, duration, infolines, last_extent, output_formats, still_frames, cut_slice, topography, log_mean, log_var, contour, num_contours, level, type, save):
  
  [ext_mean, ext_var] = getExtents (self, v, last_extent, topography, level, type)
  
  from helpers import Progress
  progress = Progress(" :: Generating frames: ", len(self.times))
  progress.update(0)
  
  if 'pdf' in output_formats:
    from matplotlib.backends.backend_pdf import PdfPages
    if save:
      pdf = PdfPages ( save + '.pdf' )
    else:
      pdf = PdfPages ( getFileName (self, self.movief + '-' + v.name, file_suffix='pdf') )
  
  for ts in range(len(self.times)):
    self.subplot(v, None, ts, ext_mean, ext_var, infolines, topography=topography, movie=True, cut_slice=cut_slice, log_mean=log_mean, log_var=log_var, contour=contour, num_contours=num_contours, level=level, type=type)
    if 'pdf' in output_formats:
      pdf.savefig()
    if 'mp4' in output_formats:
      pylab.savefig('frame_%04d.png' % ts)
    if ts != len(self.times) - 1:
      pylab.close()
    progress.update(ts+1)
  
  # output PDF file with multiple pages
  if 'pdf' in output_formats:
    generate_PDF_movie (self, pdf, pylab, still_frames)
  
  # generate MP4 movies from PNG files; afterwards, delete PNG files
  if 'mp4' in output_formats:
    generate_movies (self, len(self.times), save, None, self.movief, ['mp4'], duration, still_frames)

# === helpers

def show_info (self, subplots, level, type):
      
  text = self.generateInfo(prefix='', subplots=subplots, level=level, type=type)
  
  if subplots:
    pylab.figtext(0.01, 0.015, text, fontsize='small')
    pylab.subplots_adjust(bottom=0.23)
  else:
    pylab.figtext(0.01, 0.015, text, fontsize=14)
    pylab.subplots_adjust(bottom=0.32)
