
# # # # # # # # # # # # # # # # # # # # # # # # # #
# Plotting routines for 1d histograms data        #
# from 1d or sliced 2d numerical simulations      #
#                                                 #
# Jonas Sukys                                     #
# SAM, ETH Zurich, Switzerland                    #
# sukys.jonas@gmail.com                           #
# # # # # # # # # # # # # # # # # # # # # # # # # #

# === global imports

import pylab
import numpy

# === local imports

from variables import *
from helpers import *

# === configuration

import matplotlib as mpl
mpl.rcParams['font.size'] = 18
mpl.rcParams['legend.fontsize'] = 16

# === plotting routines

def plot_histograms (self, v, ts, infolines, save, cut_slice, pick, smoothen, support, level, type, movie):
  
  if ts is None: ts = len(self.times) - 1
  
  desired_ratio = 0.75
  
  if infolines:
    pylab.figure(figsize=(14,6.5))
  else:
    pylab.figure(figsize=(14,5.5))
  
  d = self.var (v, ts, "histogram", cut_slice, level, type)
  
  max_value = numpy.max(d)
  
  dx = ( self.f.HISTOGRAM_MAX - self.f.HISTOGRAM_MIN ) / (self.f.NBINS - 2)
  
  title_prefix = get_title_prefix (self, "PDF")
  
  # 2d plot of color-coded histogram values for each cell
  
  ax = pylab.subplot(121)
  
  s = {'vmin':0,'vmax':numpy.max(d)*0.9}
  
  title_suffix = ' at t=%.2f' % self.times[ts]
  if cut_slice:
    title_suffix += ', %s=%.2f' % (cut_slice['dir'].lower(), float(cut_slice['pos']))
  
  axes_extent = (0, self.f.MAXX, self.f.HISTOGRAM_MIN - dx, self.f.HISTOGRAM_MAX + dx)
  
  # invalidate zero elements of d
  # such that any colormap can be used on the remaining values
  # keeping white color for d == 0
  d_masked = numpy.where(d == 0, 0, d + numpy.max(d)/50)
  pylab.get_cmap().set_bad(color = 'w', alpha = 1.)
  
  # plot
  pylab.imshow(numpy.transpose(d_masked), cmap='binary', origin='lower', extent=axes_extent, interpolation='bicubic', **s)
  
  # set aspect ratio
  inversed_ratio = float(axes_extent[1] - axes_extent[0]) / (axes_extent[3] - axes_extent[2])
  ax.set_aspect(desired_ratio * inversed_ratio)
  
  pylab.title(title_prefix + v.name + title_suffix)
  
  pylab.xlabel('%s-domain' % pick['dir'])
  pylab.ylabel('values of ' + v.name)
  
  # add red line at pick position
  pylab.axvline(x=float(pick['pos']), color='r', linestyle='--', linewidth=2)
  
  # 1d plot of the histogram at a given domain point 'pick'
  
  ax = pylab.subplot(122)
  
  title_suffix += ', %s=%.2f' % (pick['dir'], float(pick['pos']))
  
  xr = self.f.HISTOGRAM_MIN - dx + dx * numpy.arange(self.i.NBINS)
  
  if pick['dir'].upper() == 'X':
    pick_at = min ( self.nx-1, round(float(pick['pos']) / self.f.MAXX * self.nx) )
  if pick['dir'].upper() == 'Y':
    pick_at = min ( self.ny-1, round(float(pick['pos']) / self.f.MAXY * self.ny) )
  
  d = d[pick_at,:]
  
  if smoothen:
    # plot interpolated and smoothened PDF using cubic splines
    from scipy.interpolate import UnivariateSpline
    s = UnivariateSpline(xr, d, bbox=[self.f.HISTOGRAM_MIN-dx, self.f.HISTOGRAM_MAX+dx])
    xs = numpy.linspace(self.f.HISTOGRAM_MIN-dx, self.f.HISTOGRAM_MAX+dx, 1000)
    pylab.plot(s(xs), xs, 'r')
  
  else:
    # plot raw PDF
    width = max ( (self.f.HISTOGRAM_MAX - self.f.HISTOGRAM_MIN + 2*dx) / 100, dx)
    pylab.barh(xr, d, height=dx, color='r', edgecolor='r')
  
  # annotate the atomic part (dirac measures) of the PDF
  dirac_indices = numpy.where(d > 0.3 / dx)[0]
  for dirac_index in dirac_indices:
    annotation_loc = ( d[dirac_index], xr[dirac_index]-0.02*(self.f.HISTOGRAM_MAX - self.f.HISTOGRAM_MIN + 2*dx) )
    pylab.annotate(' %.2f' % (d[dirac_index] * dx), xy=annotation_loc,  xycoords='data')
  
  pylab.xlim([0, 1.1*max_value])
  pylab.ylim([self.f.HISTOGRAM_MIN - dx, self.f.HISTOGRAM_MAX + dx])
  
  # set aspect ratio
  inversed_ratio = max_value / (self.f.HISTOGRAM_MAX - self.f.HISTOGRAM_MIN + 2*dx)
  ax.set_aspect(desired_ratio * inversed_ratio)
  
  pylab.xlabel('density')
  pylab.ylabel('values of %s' % v.name)
  
  pylab.title(title_prefix + v.name + title_suffix)
  
  # adjust
  if infolines:
    pylab.subplots_adjust(right=0.96)
    pylab.subplots_adjust(left=0.05)
    pylab.subplots_adjust(top=0.94)
    pylab.subplots_adjust(wspace=0.2)
  else:
    pylab.subplots_adjust(right=0.96)
    pylab.subplots_adjust(left=0.05)
    pylab.subplots_adjust(top=0.92)
    pylab.subplots_adjust(wspace=0.2)
  
  # infolines
  if infolines:
    show_info(self, level=level, type=type)
  
  if save:
    pylab.savefig(save)
    pylab.savefig(save[:-3] + 'eps')
    if not movie: self.generateTexTable(save)
  if not movie:
    pylab.show()

# === movie generation routines

def movie_histograms (self, v, ts, duration, frames, infolines, output_formats, cut_slice, smoothen, support, type, movie, save):
  
  from helpers import Progress
  progress = Progress(" :: Generating frames: ", frames)
  progress.update(0)
  
  if 'pdf' in output_formats:
    from matplotlib.backends.backend_pdf import PdfPages
    if save:
      pdf = PdfPages ( save + '.pdf' )
    else:
      pdf = PdfPages ( getFileName (self, self.movief + '-histograms' + '-' + v.name, file_suffix='pdf') )
  
  # set 'pick' direction and compute positions
  pick = {}
  if cut_slice:
    if cut_slice['dir'].upper() == 'X':
      pick['dir'] = 'y'
      pick_positions = numpy.arange(frames, dtype=float) / frames * self.f.MAXY
    if cut_slice['dir'].upper() == 'Y':
      pick['dir'] = 'x'
      pick_positions = numpy.arange(frames, dtype=float) / frames * self.f.MAXX
  else:
    pick['dir'] = 'x'
    pick_positions = numpy.arange(frames, dtype=float) / frames * self.f.MAXX
  
  for i, pick_position in enumerate(pick_positions):
    pick['pos'] = str(pick_position)
    self.plot_histograms(v, ts, infolines, cut_slice=cut_slice, pick=pick, smoothen=smoothen, level=level, type=type, movie=True)
    if 'pdf' in output_formats:
      pdf.savefig()
    if 'mp4' in output_formats:
      pylab.savefig('frame_%04d.png' % i)
    if i != frames - 1:
      pylab.close()
    progress.update(i+1)
  
  # output PDF file with multiple pages
  if 'pdf' in output_formats:
    generate_PDF_movie (self, pdf, pylab, 0)
  
  # generate MP4 movies from PNG files; afterwards, delete PNG files
  if 'mp4' in output_formats:
    generate_movies (self, frames, save, None, self.movief, ['mp4'], duration, 0)

# === helpers

def show_info (self, prefix='', adjust=True, level=None, type=None):
      
  text = self.generateInfo(prefix=prefix, level=level, type=type)
  
  pylab.figtext(0.01, 0.015, text, fontsize='small')
  
  if adjust:
    pylab.subplots_adjust(bottom=0.20)
