
# # # # # # # # # # # # # # # # # # # # # # # # # #
# Plotting routines for 3d simulation data        #
# Mayavi backend                                  #
#                                                 #
# Jonas Sukys                                     #
# SAM, ETH Zurich, Switzerland                    #
# sukys.jonas@gmail.com                           #
# # # # # # # # # # # # # # # # # # # # # # # # # #

# === global imports

import numpy

from enthought.mayavi.scripts import mayavi2
mayavi2.standalone(globals())

from enthought.mayavi import mlab

# === local imports

from variables import *
from helpers import *

# === configuration

#mlab.options.backend = 'envisage'

blue  = (0.0,0.0,1.0)
brown = (0.4,0.2,0.0)
black = (0.0,0.0,0.0)
red   = (1.0,0.0,0.0)
white = (1.0,1.0,1.0)

import warnings
warnings.filterwarnings("ignore", message="Unable to import the wx backend for pyface due to traceback: Traceback (most recent call last):")

# === plotting routines

def get_view (view=None):
  
  current_view = {}
  
  current_view['azimuth']   = mlab.view()[0]
  current_view['elevation'] = mlab.view()[1]
  current_view['distance']  = mlab.view()[2]
  current_view['roll']      = mlab.roll()
  
  if view:
    if 'azimuth'   in view and view['azimuth']   != None : current_view['azimuth']   = view['azimuth']
    if 'elevation' in view and view['elevation'] != None : current_view['elevation'] = view['elevation']
    if 'distance'  in view and view['distance']  != None : current_view['distance']  = view['distance']
    if 'roll'      in view and view['roll']      != None : current_view['roll']      = view['roll']
  
  return current_view

def set_view (view):
  
  mlab.roll (view['roll'])
  mlab.view ( azimuth=view['azimuth'], elevation=view['elevation'], distance=view['distance'] )

def plot_3d_stat (self, v, ts, stat, extent, infolines, view, save, log, contour, num_contours, contour_levels, clip, highlight, level, type, movie=False, engine=None):
  
  d = self.var(v, ts, stat, False, level, type)
  if log:
    [values_extent,d] = apply_log (extent,d)
  else:
    values_extent = extent
  
  if values_extent:
    vmin = values_extent[0]
    vmax = values_extent[1]
  else:
    vmin = None
    vmax = None
  
  title_prefix = get_title_prefix (self, stat, log)
  
  [xr, yr, zr] = numpy.mgrid[0:self.nx, 0:self.ny, 0:self.nz]
  xr = xr * self.f.MAXX / self.nx
  yr = yr * self.f.MAXY / self.ny
  zr = zr * self.f.MAXZ / self.nz
  
  if not movie: mlab.figure(size=(800, 800))
  
  if contour:
    mlab.contour3d(xr, yr, zr, d, contours=num_contours, opacity=0.5, vmin=vmin, vmax=vmax, name=title_prefix+v.name)
  else:
    mlab.pipeline.volume(mlab.pipeline.scalar_field(d), vmin=vmin, vmax=vmax, name=title_prefix+v.name)
  
  mlab.title(title_prefix + v.name + ' at t=%.2f' % self.times[ts], height=0.95, size=0.3, color=black)
  
  mlab.axes ( x_axis_visibility=False, y_axis_visibility=False, z_axis_visibility=False )
  
  ''' BUGGY - WHY?
  if not movie:
    mlab.axes ( color=black, extent=(0, self.f.MAXX, 0, self.f.MAXY, 0, self.f.MAXZ ) )
    mlab.xlabel('x')
    mlab.ylabel('y')
    mlab.zlabel('z')
  else:
    mlab.axes ( x_axis_visibility=False, y_axis_visibility=False, z_axis_visibility=False )
  '''
  
  mlab.colorbar()
  
  if view: set_view ( get_view (view) )
  
  if infolines: show_info(self, level, type)
  
  if save: mlab.savefig (save[0:-4] + '-' + stat + save[len(save)-4:len(save)], size=(800,800))
  
  if not movie: mlab.show()

def plot_3d (self, v, ts, extent_mean, extent_var, infolines, view, save, log_mean, log_var, contour, num_contours, contour_levels_mean, contour_levels_var, clip, highlight, level, type, batch, engine):
  
  if ts is None: ts = len(self.times) - 1
  
  # plot mean
  plot_3d_stat (self, v, ts, "mean", extent_mean, infolines, view, save, log_mean, contour, num_contours, contour_levels_mean, clip, highlight, level, type, engine)
  
  # plot variance
  if self.VARIANCE:
    plot_3d_stat (self, v, ts, "variance", extent_var, infolines, view, save, log_var, contour, num_contours, contour_levels_var, clip, highlight, level, type, engine)
  
  # finalization
  if save:
    self.generateTexTable(save)
    self.generateTexTableVertical(save)

# === movie generation

def movie_3d_stat (self, v, stat, duration, infolines, last_extent, output_formats, still_frames, view, log, contour, num_contours, contour_levels, clip, highlight, level, type, save):
  
  extent = getExtent(self, v, stat, last_extent, level, type)
  
  from helpers import Progress
  progress = Progress(" :: Generating frames: ", len(self.times))
  progress.update(0)
  
  # generate snapshots
  for ts in range(len(self.times)):
    
    # initial view parameters are given as input
    if ts == 0: fixed_view = view
    
    plot_3d_stat (self, v, ts, stat, extent, infolines=infolines, view=fixed_view, log=log, save=None, contour=contour, num_contours=num_contours, contour_levels=contour_levels, clip=clip, highlight=highlight, level=level, type=type, movie=True, engine=engine)
    
    mlab.savefig('frame_%04d.png' % ts, size=(800,800))
    
    # save initial view
    if ts == 0: fixed_view = get_view()
    
    mlab.close()
    
    progress.update (ts+1)
  
  # generate MP4 movies from PNG files; afterwards, delete PNG files
  generate_movies (self, len(self.times), save, stat, self.movief, output_formats, duration, still_frames)

def movie_3d (self, v, duration, infolines, last_extent, output_formats, still_frames, view, log_mean, log_var, contour, num_contours, contour_levels, clip, highlight, level, type, save, engine):
  
  movie_3d_stat (self, v, "mean", duration, infolines, last_extent, output_formats, still_frames, view, log_mean, contour, num_contours, contour_levels, clip, highlight, level, type, save, engine)
  
  if self.VARIANCE:
    movie_3d_stat (self, v, "variance", duration, infolines, last_extent, output_formats, still_frames, view, log_var, contour, num_contours, contour_levels, clip, highlight, level, type, save, engine)

def movie_rotate_3d_stat (self, v, ts, stat, duration, frames, infolines, output_formats, still_frames, view, log, contour, num_contours, contour_levels, clip, highlight, level, type, save, engine):
  
  from helpers import Progress
  progress = Progress(" :: Generating frames: ", frames)
  progress.update(0)
  
  plot_3d_stat (self, v, ts, stat, extent=None, infolines=infolines, view=view, log=log, save=None, contour=contour, num_contours=num_contours, contour_levels=contour_levels, clip=clip, highlight=highlight, level=level, type=type, movie=True, engine=engine)
  fixed_view = get_view()
  
  for frame in range(frames):
    fixed_view['azimuth'] += 360.0/frames
    set_view(fixed_view)
    mlab.savefig('frame_%04d.png' % frame, size=(800,800))
    progress.update (frame+1)
  
  mlab.close()
  
  # generate movies from PNG files; afterwards, delete PNG files
  generate_movies (self, frames, save, stat, self.rotatef, output_formats, duration, still_frames)

def movie_rotate_3d (self, v, ts, duration, frames, infolines, output_formats, still_frames, view, log_mean, log_var, contour, num_contours, contour_levels, clip, highlight, level, type, save, engine):
  
  movie_rotate_3d_stat (self, v, ts, "mean", duration, frames, infolines, output_formats, still_frames, view, log_mean, contour, num_contours, contour_levels, clip, highlight, level, type, save, engine)
  
  if self.VARIANCE:
    movie_rotate_3d_stat (self, v, ts, "variance", duration, frames, infolines, output_formats, still_frames, view, log_var, contour, num_contours, contour_levels, clip, highlight, level, type, save, engine)

# === specific plotting routines for 3d visualization of 2d simulation data from shallow water equations

def plot_sw_3d (self, ts, infolines, view, save, movie=False):
  
  if save:
    self.generateTexTable(save)
    self.generateTexTableVertical(save)
  
  if ts is None: ts = len(self.times)-1
  
  if not movie: mlab.figure(size=(1024, 768))
  
  xr = numpy.arange(self.nx)*(self.f.MAXX/self.nx)
  yr = numpy.arange(self.ny)*(self.f.MAXY/self.ny)
  
  S_mean = self.var(sw.S, ts, 'mean')
  b_mean = self.var(sw.B, ts, 'mean')
  
  mlab.surf(xr, yr, S_mean[:,:,0], color=blue, opacity=0.5, name='mean of S')
  mlab.surf(xr, yr, b_mean[:,:,0],                          name='mean of b')
  
  if self.i.COMPUTE_VARIANCE:
    S_var = self.var(sw.S, ts, 'variance')
    b_var = self.var(sw.B, ts, 'variance')
    S_std = numpy.sqrt(S_var)
    b_std = numpy.sqrt(b_var)
    
    mlab.surf(xr, yr, (S_mean+S_std)[:,:,0], colormap='cool', opacity=0.2, name='mean+std of S')
    mlab.surf(xr, yr, (b_mean+b_std)[:,:,0],                  opacity=0.2, name='mean+std of b')
  
  mlab.title('bottom topography and free surface at t=%.2f' % self.times[ts], height=0.95, size=0.3, color=black)
  
  mlab.xlabel('x')
  mlab.ylabel('y')
  mlab.zlabel('')
  
  if self.i.COMPUTE_VARIANCE:
    mlab.axes(color=black, extent=(0, self.f.MAXX, 0, self.f.MAXY, numpy.min(b_mean), numpy.max( numpy.max(S_mean+S_std), numpy.max(b_mean+b_std)) ))
  else:
    mlab.axes(color=black, extent=(0, self.f.MAXX, 0, self.f.MAXY, numpy.min(b_mean), numpy.max( numpy.max(S_mean), numpy.max(b_mean)) ))
  
  if 'azimuth' in view and view['azimuth']: mlab.view (azimuth = view['azimuth'])
  if 'elevation' in view and view['elevation']: mlab.view (elevation = view['elevation'])
  if 'distance' in view and view['distance']:
    dist = mlab.view()[2]
    mlab.view (distance = dist + view['distance'])
  if 'yaw' in view and view['yaw']: mlab.yaw(view['yaw'])
  
  if infolines: show_info(self)
  
  if save: mlab.savefig (save, size=(1024,768))
  
  if not movie: mlab.show()

def plot_sw_3d_topography (self, infolines, view, title, save):
  
  ts = len(self.times)-1
  
  mlab.figure(size=(1024,768))
  
  xr = numpy.arange(self.nx)*(self.f.MAXX/self.nx)
  yr = numpy.arange(self.ny)*(self.f.MAXY/self.ny)
  
  b_mean = self.var(sw.B, ts, 'mean')
  
  mlab.surf(xr, yr, b_mean[:,:,0], name='mean of b')
  
  if self.i.COMPUTE_VARIANCE:
    b_var = self.var(sw.B, ts, 'variance')
    b_std = numpy.sqrt(b_var)
    mlab.surf(xr, yr, (b_mean+b_std)[:,:,0], opacity=0.2, name='mean+std of b')
  
  if title: mlab.title('mean and mean+std of bottom topography', height=0.95, size=0.3, color=black)
  
  mlab.xlabel('x')
  mlab.ylabel('y')
  mlab.zlabel('z')
  
  if self.i.COMPUTE_VARIANCE:
    mlab.axes(color=black, extent=(0, self.f.MAXX, 0, self.f.MAXY, numpy.min(b_mean), numpy.max(b_mean+b_std)))
  else:
    mlab.axes(color=black, extent=(0, self.f.MAXX, 0, self.f.MAXY, numpy.min(b_mean), numpy.max(b_mean)))
  
  if 'azimuth' in view and view['azimuth']: mlab.view (azimuth = view['azimuth'])
  if 'elevation' in view and view['elevation']: mlab.view (elevation = view['elevation'])
  if 'distance' in view and view['distance']:
    dist = mlab.view()[2]
    mlab.view (distance = dist + view['distance'])
  if 'yaw' in view and view['yaw']: mlab.yaw(view['yaw'])
  
  if infolines: show_info(self)
  
  if save:
    mlab.savefig (save, size=(1024,768))
    self.generateTexTable (save, include_vertical=True)
  
  mlab.show()

def movie_sw_3d (self, duration, infolines, output_formats, view):
  
  from helpers import Progress
  progress = Progress(" :: Generating frames: ", len(self.times))
  progress.update(0)
  
  for ts in range(len(self.times)):
    self.plot_sw_3d(ts, infolines, view=view, movie=True)
    mlab.savefig('movie_%04d.png' % ts, size=(1024,768))
    mlab.close()
    progress.update(ts+1)
  
  print
  
  import os
  
  # output PDF file with multiple pages
  if 'pdf' in output_formats:
    
    if save:
      file_name = save + '-3d'
    else:
      file_name = getFileName (self, self.movief + '-' + v.name + '-3d', file_suffix=None)
    
    print ' :: Converting into PDF...',
    os.system('convert movie_*.png %s.pdf' % file_name )
    print ' done.'
  
  # encode mp4 movie from PNG snapshots
  if 'mp4' in output_formats:
    
    fps = self.i.NSAVES / duration
    
    print " :: Frame rate: " + str(fps)
    if fps < 25:
      print "  : -> consider increasing parameter 'NSAVES' to increase the frame rate!"
    print
    
    if save:
      file_name = save + '-3d'
    else:
      file_name = getFileName (self, self.movief + '-' + v.name + '-3d', file_suffix=None)
    
    print ' :: Converting into MP4...',
    os.system('mencoder "mf://*.png" -mf fps=%d -o %s.mp4 -ovc lavc -lavcopts vcodec=mpeg4:vbitrate=800' % (fps, file_name))
    print ' done.'
  
  os.system('rm movie_*.png')

# === helpers

def show_info(self, level=None, type=None):
      
  text = self.generateInfo (prefix='', level=level, type=type)
  
  mlab.text(0.01, 0.01, text, color=black, line_width=2.0, width=1.0, name='infolines')
