
# # # # # # # # # # # # # # # # # # # # # # # # # #
# Plotting routines for 3d simulation data        #
# Visit backend                                   #
#                                                 #
# Jonas Sukys                                     #
# SAM, ETH Zurich, Switzerland                    #
# sukys.jonas@gmail.com                           #
# # # # # # # # # # # # # # # # # # # # # # # # # #

# === global imports

import sys
sys.path.append("/scratch/data/visit2_6_0.linux-x86_64/2.6.0/linux-x86_64/lib/site-packages")
from visit import *

import math

# === local imports

from variables import *
from helpers import *
from wrappers import *

# === configuration

log10_cutoff = 1e-3

sys.path.append( os.path.dirname(__file__) +  "/.." )
from local_visualization import *

# === plotting routines

def get_view (view=None):
  
  ResetView()
  current_view = GetView3D()
  
  # set default zoom
  current_view.imageZoom = 1.0/0.85
  
  if view:
    if 'azimuth'   in view and view['azimuth']:
      current_view.RotateAxis ( 1, view['azimuth'] )
    if 'elevation' in view and view['elevation']:
      current_view.RotateAxis ( 0, view['elevation'] )
    if 'distance' in view and view['distance']:
      current_view.imageZoom = 1.0/view['distance']
  
  return current_view

def set_view (view):
  
  SetView3D(view)

def print_view ():
  
  current_view = GetView3D()
  print "VIEW distance: " + str(1.0/current_view.imageZoom)

# apply parameters (extent, log, num_contours, contour_levels) for the main plot
def set_parameters (atts, extent, log, num_contours, contour_levels):
  
  # set number of contours
  if contour_levels:
    atts.contourNLevels = len(contour_levels)
  else:
    atts.contourNLevels = num_contours
  
  # set countour levels
  if contour_levels:
    atts.contourMethod = 1
    atts.contourValue = contour_levels
  
  # set extent
  if extent and not contour_levels:
    if extent[0]:
      atts.minFlag = 1
      atts.min = extent[0]
      if log and extent[0] < log10_cutoff:
        atts.min = log10_cutoff
    if extent[1]:
      atts.maxFlag = 1
      atts.max = extent[1]
  
  # set log scaling if needed
  if log and not contour_levels:
    atts.scaling = 1
    atts.minFlag = 1
    atts.min = log10_cutoff

# re-apply main plot parameters (extent, log, num_contours, contour_levels) for the additional plots
def reapply_parameters (atts, extent, log, num_contours, contour_levels):
  
  # set correct extent
  if contour_levels:
    atts.minFlag = 1
    atts.min     = contour_levels[1]
    atts.maxFlag = 1
    atts.max     = contour_levels[-1]
  
  # set extent
  if extent and not contour_levels:
    if extent[0]:
      atts.minFlag = 1
      atts.min = extent[0]
      if log and extent[0] < log10_cutoff:
        atts.min = log10_cutoff
    if extent[1]:
      atts.maxFlag = 1
      atts.max = extent[1]
  
  # set correct extent for log scaling
  if log and not contour_levels:
    atts.minFlag = 1
    atts.min     = log10_cutoff
  
  # set log scale
  if log:
    atts.scaling = 1

# parse highlight parameters (for mean and variance)
def parse_highlight (highlight):
  
  # parse highlight argument
  if isinstance(highlight, (list, tuple)):
    highlight_mean = highlight[0]
    highlight_var  = highlight[1]
  else:
    highlight_mean = highlight
    highlight_var  = highlight
  
  return highlight_mean, highlight_var

# open the specified compute engine
def open_compute_engine (engine):
  
  # initialize Visit
  '''
  if batch:
    LaunchNowin()
  else:
    Launch()
    #Launch(vdir="path/to/visit/bin")
  '''
  
  # open the specified compute engine
  if engine:
    OpenComputeEngine("localhost", engine)
  
  # open the default engine
  else:
    OpenComputeEngine("localhost", get_engine())

def plot_3d_stat (self, v, ts, stat, extent, infolines, view, save, log, contour, num_contours, contour_levels, clip, highlight, level, type, batch, movie=False):
  
  # write BOV file, get file name
  bov_file_name, variable = write_BOV (self, v, ts, stat, level, type)
  
  # open database using BOV header (2nd input indicates time)
  OpenDatabase (bov_file_name, 0, "BOV_1.0")
  
  # contour plot
  if contour:
    
    # add contour plot of data
    AddPlot ("Contour", variable)
    
    # get plot attributes
    atts = ContourAttributes()
    
    # set plot parameters
    set_parameters (atts, extent, log, num_contours, contour_levels)
    
    # set color table and opacity
    half    = float(num_contours / 2)
    quarter = float(num_contours / 4)
      
    def evaluate (slope, zero, x):
      return slope * x / quarter - slope * zero / quarter;
    
    for index in range(num_contours):
      blue     = round ( 255 * max ( 0, min( 1, evaluate(-1, half, index) ) ) )
      red      = round ( 255 * max ( 0, min( 1, evaluate(1,  half, index) ) ) )
      greenval = min ( evaluate(1, 0, index), evaluate(-1, num_contours-1, index) )
      green    = round ( 255 * max ( 0, min( 1, greenval                 ) ) )
      opacity  = round((float(index + 1) / num_contours)**2 * 255)
      if opacity > 200: opacity = 200
      if opacity < 30: opacity = 30
      atts.SetMultiColor (index, ( red, green, blue, opacity ) )
    
    # set plot attributes
    SetPlotOptions(atts)
    
    # clip a part of the domain
    if clip:
      
      # compute inverted normals
      clip['inverted_normals'] = [None, None, None]
      for i in xrange(3):
        if clip['normals'][i]: clip['inverted_normals'][i] = tuple( [-a for a in clip['normals'][i] ] )
      clip['inverted_normals'] = tuple(clip['inverted_normals'])
      
      # === CLIP OPERATOR
      
      # add clip operator
      AddOperator("Clip")
      
      # get clip attributes
      atts = ClipAttributes()
      
      # activate clip plane 1
      if clip['normals'][0]:
        atts.plane1Status = 1
        atts.plane1Origin = clip['origin']
        atts.plane1Normal = clip['normals'][0]
      
      # activate clip plane 2
      if clip['normals'][1]:
        atts.plane2Status = 1
        atts.plane2Origin = clip['origin']
        atts.plane2Normal = clip['normals'][1]
      
      # activate clip plane 3
      if clip['normals'][2]:
        atts.plane3Status = 1
        atts.plane3Origin = clip['origin']
        atts.plane3Normal = clip['normals'][2]
      
      # set attributes of all clip planes
      SetOperatorOptions(atts)
      
      # === FIRST CUSTOM SLICE OPERATOR
      
      # === PSEUDOCOLOR PLOT
      
      # add pseudo plot of data
      AddPlot ("Pseudocolor", variable)
      
      # get pseudocolor attributes
      atts = PseudocolorAttributes()
      
      # remove lighting
      atts.lightingFlag = 0
      
      # set plot parameters
      reapply_parameters (atts, extent, log, num_contours, contour_levels)
      
      # remove legend
      atts.legendFlag = 0
      
      # apply plot attributes
      SetPlotOptions(atts)
      
      # === CLIP OPERATOR
      if clip['normals'][1]:
        
        # add clip operator
        AddOperator("Clip")
        
        # get clip attributes
        atts = ClipAttributes()
        
        # activate clip plane 1
        atts.plane1Status = 1
        atts.plane1Origin = clip['origin']
        atts.plane1Normal = clip['inverted_normals'][1]
        
        # set attributes of all clip planes
        SetOperatorOptions(atts)
      
      # === ISOVOLUME OPERATOR
      if 'slice_cutoff' in clip and clip['slice_cutoff'] != None:
        
        # add isovolume operator
        AddOperator("Isovolume")
        
        # get isovolume attributes
        atts = IsovolumeAttributes()
        
        # set isovolume extents
        atts.lbound = clip['slice_cutoff']
        
        # set attributes of isovolume
        SetOperatorOptions(atts)
      
      # === SLICE OPERATOR
      
      # add slice operator
      AddOperator("Slice")
      
      # get slice attributes
      atts = SliceAttributes()
      
      # set slice normal
      atts.normal = clip['normals'][0]
      
      # set slice origin at a given origin
      atts.originType = 0
      atts.originPoint = clip['origin']
      
      # no 2D projection is needed
      atts.project2d = 0
      
      # set attributes of slice
      SetOperatorOptions(atts)
      
      # === SECOND CUSTOM SLICE OPERATOR
      if clip['normals'][1]:
        
        # === PSEUDOCOLOR PLOT
        
        # add pseudo plot of data
        AddPlot ("Pseudocolor", variable)
        
        # get pseudocolor attributes
        atts = PseudocolorAttributes()
        
        # remove lighting
        atts.lightingFlag = 0
        
        # set plot parameters
        reapply_parameters (atts, extent, log, num_contours, contour_levels)
        
        # remove legend
        atts.legendFlag = 0
        
        # apply plot attributes
        SetPlotOptions(atts)
        
        # === CLIP OPERATOR
        
        # add clip operator
        AddOperator("Clip")
        
        # get clip attributes
        atts = ClipAttributes()
        
        # activate clip plane 1
        atts.plane1Status = 1
        atts.plane1Origin = clip['origin']
        atts.plane1Normal = clip['inverted_normals'][0]
        
        # set attributes of all clip planes
        SetOperatorOptions(atts)
        
        # === ISOVOLUME OPERATOR
        if 'slice_cutoff' in clip and clip['slice_cutoff'] != None:
          
          # add isovolume operator
          AddOperator("Isovolume")
          
          # get isovolume attributes
          atts = IsovolumeAttributes()
          
          # set isovolume extents
          atts.lbound = clip['slice_cutoff']
          
          # set attributes of isovolume
          SetOperatorOptions(atts)
        
        # === SLICE OPERATOR
        
        # add slice operator
        AddOperator("Slice")
        
        # get slice attributes
        atts = SliceAttributes()
        
        # set slice normal
        atts.normal = clip['normals'][1]
        
        # set slice origin
        atts.originType = 0
        atts.originPoint = clip['origin']
        
        # no 2D projection is needed
        atts.project2d = 0
        
        # set attributes of slice
        SetOperatorOptions(atts)
    
    # highlight contours within the clipped region
    if highlight:
    
      # === CREATE EXPRESSION FOR VORTICITY
      
      # create expression
      #AddExpression("vorticity",variable)
      
      # === PSEUDOCOLOR PLOT
      
      # add pseudo plot of data
      AddPlot ("Pseudocolor", variable)
      
      # get pseudocolor attributes
      atts = PseudocolorAttributes()
      
      # set plot parameters
      reapply_parameters (atts, extent, log, num_contours, contour_levels)
      
      # set color
      #atts.colorTableName = "Purples"
      
      # remove legend
      atts.legendFlag = 0
      
      # apply plot attributes
      SetPlotOptions(atts)
      
      # === ISOVOLUME OPERATOR
      
      # add isovolume operator
      AddOperator("Isovolume")
      
      # get isovolume attributes
      atts = IsovolumeAttributes()
      
      # set required isovolumes
      atts.lbound = highlight
      
      # set attributes of isovolume
      SetOperatorOptions(atts)
      
      # TODO: try to add expression 'vorticity = curl(U)' to the plot, visualize as rings in the center, use colormap 'Purples'
      # TODO: volume render for vorticity (with isovolume operator)
  
  # volume plot
  else:
    
    # add volume plot of data
    AddPlot("Volume", variable, 1, 0)
    
    # get plot attributes
    atts = VolumeAttributes()
    
    # set extent
    if extent:
      if extent[0]:
        atts.useColorVarMin = 1
        atts.colorVarMin = extent[0]
        if log and extent[0] < log10_cutoff:
          atts.colorVarMin = log10_variance_cutoff
      if extent[1]:
        atts.useColorVarMax = 1
        atts.colorVarMax = extent[1]
    
    # set log scaling if needed
    if log:
      atts.scaling = 1
      atts.useColorVarMin = 1
      atts.colorVarMin = log10_cutoff
    
    # set opacity
    atts.freeformOpacity = range(256)
    
    # set plot attributes
    SetPlotOptions(atts)
  
  # customize annotations
  annt = AnnotationAttributes()
  annt.axes3D.visible = 0       # axes labels, grid and values
  annt.axes3D.triadFlag = 0     # small x-y-z arrows
  annt.axes3D.bboxFlag = 0      # bounding box
  annt.userInfoFlag = 0         # user information, date and time
  annt.databaseInfoFlag = 1     # database info
  annt.timeInfoFlag = 0         # time info
  SetAnnotationAttributes(annt)
  
  # renders the plot
  DrawPlots()
  
  # set the view
  if view: set_view ( get_view (view) )
  
  if infolines: show_info(self, level, type)
  
  # save plot
  if save:
    s = SaveWindowAttributes()
    s.format = s.PNG
    s.fileName = save[0:-4] + '-' + stat
    s.width, s.height = 1200, 1200
    s.quality = 90
    s.screenCapture = 0
    s.family = 0
    SetSaveWindowAttributes(s)
    SaveWindow()
  
  if not movie:
    
    # print view parameters
    print_view ()
    
    # display plot
    if not batch: raw_input("Press Enter to continue...")
    
    # clear plots
    DeleteAllPlots()
    ResetView()
    
    # close database
    CloseDatabase(bov_file_name)
    
    # delete BOV file
    delete_BOV (bov_file_name)
  
  # return database name
  return bov_file_name

def plot_3d (self, v, ts, extent_mean, extent_var, infolines, view, save, log_mean, log_var, contour, num_contours, contour_levels_mean, contour_levels_var, clip, highlight, level, type, batch, engine):
  
  # set default time snapshot index
  if ts is None: ts = len(self.times) - 1
  
  # parse highlight parameters
  highlight_mean, highlight_var = parse_highlight (highlight)
  
  # initialize Visit
  open_compute_engine (engine)
  
  # supress messages (display only warning and errors)
  SuppressMessages(3)
  
  # plot mean
  plot_3d_stat (self, v, ts, "mean", extent_mean, infolines, view, save, log_mean, contour, num_contours, contour_levels_mean, clip, highlight_mean, level, type, batch)
  
  # plot variance
  if self.VARIANCE:
    plot_3d_stat (self, v, ts, "variance", extent_var, infolines, view, save, log_var, contour, num_contours, contour_levels_var, clip, highlight_var, level, type, batch)
  
  # save TeX table
  if save:
    self.generateTexTable(save)
    self.generateTexTableVertical(save)
  
  # finalize Visit
  CloseComputeEngine()

# === movie generation

def movie_3d_stat (self, v, stat, duration, infolines, last_extent, output_formats, still_frames, view, log, contour, num_contours, contour_levels, clip, highlight, level, type, save, engine):
  
  # initialize Visit
  open_compute_engine (engine)
  
  # supress messages (display only warning and errors)
  SuppressMessages(3)
  
  # determine extent
  if contour_levels:
    
    # not needed
    extent = None
  
  else:
    
    # plot last snapshot and determine extent
    print " :: Computing extent...",
    ts = len(self.times) - 1
    database = plot_3d_stat (self, v, ts, stat, extent=None, infolines=infolines, view=view, log=log, save=None, contour=contour, num_contours=num_contours, contour_levels=contour_levels, clip=clip, highlight=highlight, level=level, type=type, batch=True, movie=True)
    SuppressQueryOutputOn()
    Query("MinMax")
    extent = GetQueryOutputValue()
    print " done."
    
    # clear plots
    DeleteAllPlots()
    ResetView()
    
    # close database
    CloseDatabase(database)
    
    # delete BOV file
    delete_BOV (database)
  
  from helpers import Progress
  progress = Progress(" :: Generating frames: ", len(self.times))
  progress.update(0)
  
  # generate snapshots
  for ts in range(len(self.times)):
    
    database = plot_3d_stat (self, v, ts, stat, extent, infolines=infolines, view=view, log=log, save=None, contour=contour, num_contours=num_contours, contour_levels=contour_levels, clip=clip, highlight=highlight, level=level, type=type, batch=True, movie=True)
    
    # save plot
    s = SaveWindowAttributes()
    s.format = s.PNG
    s.fileName = 'frame_%04d.png' % ts
    s.width, s.height = 1200, 1200
    s.quality = 90
    s.screenCapture = 0
    s.family = 0
    SetSaveWindowAttributes(s)
    SaveWindow()
    
    # clear plot
    DeleteAllPlots()
    ResetView()
    
    # close database
    CloseDatabase(database)
    
    # delete BOV file
    delete_BOV (database)
    
    # update progress
    progress.update (ts+1)
  
  # finalize Visit
  CloseComputeEngine()
  
  # generate movies from PNG files; afterwards, delete PNG files
  generate_movies (self, len(self.times), save, stat, self.movief, output_formats, duration, still_frames)

def movie_3d (self, v, duration, infolines, last_extent, output_formats, still_frames, view, log_mean, log_var, contour, num_contours, contour_levels_mean, contour_levels_var, clip, highlight, level, type, save, engine):
  
  if last_extent == False:
    Exception(" :: ERROR: last_extent == False is not supported for 3d movie with Visit backend.")
  
  # parse highlight parameters
  highlight_mean, highlight_var = parse_highlight (highlight)
  
  # generate movie for the mean
  movie_3d_stat (self, v, "mean", duration, infolines, last_extent, output_formats, still_frames, view, log_mean, contour, num_contours, contour_levels_mean, clip, highlight_mean, level, type, save, engine)
  
  # generate movie for the variance
  if self.VARIANCE:
    movie_3d_stat (self, v, "variance", duration, infolines, last_extent, output_formats, still_frames, view, log_var, contour, num_contours, contour_levels_var, clip, highlight_var, level, type, save, engine)

def movie_rotate_3d_stat (self, v, ts, stat, duration, frames, infolines, output_formats, still_frames, view, log, contour, num_contours, contour_levels, clip, highlight, level, type, save, engine):
  
  # initialize Visit
  open_compute_engine (engine)
  
  # supress messages (display only warning and errors)
  SuppressMessages(3)
  
  database = plot_3d_stat (self, v, ts, stat, extent=None, infolines=infolines, view=view, log=log, save=None, contour=contour, num_contours=num_contours, contour_levels=contour_levels, clip=clip, highlight=highlight, level=level, type=type, batch=True, movie=True)
  
  s = SaveWindowAttributes()
  s.format = s.PNG
  s.width, s.height = 1200, 1200
  s.quality = 90
  s.screenCapture = 0
  s.family = 0
  
  if view == None: view = {}
  if 'azimuth' not in view: view['azimuth'] = 0
  
  from helpers import Progress
  progress = Progress(" :: Generating frames: ", frames)
  progress.update(0)
  
  # generate snapshots
  for frame in range(frames):
    view['azimuth'] += 360.0/frames
    set_view(get_view(view))
    s.fileName = 'frame_%04d.png' % frame
    SetSaveWindowAttributes(s)
    SaveWindow()
    progress.update (frame+1)
  
  # clear plot
  DeleteAllPlots()
  ResetView()
  
  # close database
  CloseDatabase(database)
  
  # delete BOV file
  delete_BOV (database)
  
  # finalize Visit
  CloseComputeEngine()
  
  # generate movies from PNG files; afterwards, delete PNG files
  generate_movies (self, frames, save, stat, self.rotatef, output_formats, duration, still_frames)

def movie_rotate_3d (self, v, ts, duration, frames, infolines, output_formats, still_frames, view, log_mean, log_var, contour, num_contours, contour_levels_mean, contour_levels_var, clip, highlight, level, type, save, engine):
  
  # parse highlight parameters
  highlight_mean, highlight_var = parse_highlight (highlight)
  
  # generate movie for the mean
  movie_rotate_3d_stat (self, v, ts, "mean", duration, frames, infolines, output_formats, still_frames, view, log_mean, contour, num_contours, contour_levels_mean, clip, highlight_mean, level, type, save, engine)
  
  # generate movie for the variance
  if self.VARIANCE:
    movie_rotate_3d_stat (self, v, ts, "variance", duration, frames, infolines, output_formats, still_frames, view, log_var, contour, num_contours, contour_levels_var, clip, highlight_var, level, type, save, engine)

# === helpers

def show_info(self, level, type):
  
  info = CreateAnnotationObject("Text2D")
  info.text = self.generateInfo(prefix='', level=level, type=type)
  info.position = (0.01, 0.01)
  info.height = 0.012
