
# # # # # # # # # # # # # # # # # # # # # # # # # #
# Classes for variables of each equation type     #
#                                                 #
# Jonas Sukys                                     #
# SAM, ETH Zurich, Switzerland                    #
# sukys.jonas@gmail.com                           #
# # # # # # # # # # # # # # # # # # # # # # # # # #

# === helpers

class Variable (object):
    def __init__(self, i, name):
        self.i = i
        self.name = name
    
    def load(self, r, ts, name):
        d = r.load(self.i, ts)
        return mlab.pipeline.scalar_field(d, name=name)

class VectorVariable (object):
    def __init__(self, i, name):
        self.i = i
        self.name = name

    def load(self, r, ts, name):
        x = r.load(self.i, ts)
        y = r.load(self.i+1, ts)
        z = r.load(self.i+2, ts)
        return mlab.pipeline.vector_field(x, y, z, name=name)

# === classes

# MHD
class MHD (object):
    RHO = Variable(0, 'rho')
    
    M  = VectorVariable(1, 'm')
    MX = Variable      (1, 'mx')
    MY = Variable      (2, 'my')
    MZ = Variable      (3, 'mz')
    
    E = Variable(4, 'E')
    
    B  = VectorVariable(5, 'B')
    BX = Variable      (5, 'Bx')
    BY = Variable      (6, 'By')
    BZ = Variable      (7, 'Bz')
    
    P = Variable(8, 'p')
    
    U  = VectorVariable(9,  'u')
    UX = Variable      (9,  'ux')
    UY = Variable      (10, 'uy')
    UZ = Variable      (11, 'uz')
    
    A = Variable(12, 'a')
    
    CF  = VectorVariable(13, 'cf')
    CFX = Variable      (13, 'cfx')
    CFY = Variable      (14, 'cfy')
    CFZ = Variable      (15, 'cfz')
    
    NVARS=16

mhd = MHD()

# Burgers'
class Burgers (object):
    U = Variable(0, 'u')
    E = Variable(1, 'E')
    NVARS = 2

br = Burgers()

# Euler
class Euler (object):
    RHO = Variable(0, 'rho')
    
    M  = VectorVariable(1, 'm')
    MX = Variable      (1, 'mx')
    MY = Variable      (2, 'my')
    MZ = Variable      (3, 'mz')
    
    E = Variable(4, 'E')
    P = Variable(5, 'p')
    
    U  = VectorVariable(6,  'u')
    UX = Variable      (6, 'ux')
    UY = Variable      (7, 'uy')
    UZ = Variable      (8, 'uz')
    
    #A  = Variable(9, 'A')
    
    #NVARS = 10
    NVARS = 9

eu = Euler()

# Shallow Water
class ShallowWater (object):
    H = Variable(0, 'h')
    
    HU  = VectorVariable(1, 'hu')
    HUX = Variable      (1, 'hux')
    HUY = Variable      (2, 'huy')
    #HUZ should not exist
    
    U  = VectorVariable(4,  'u')
    UX = Variable      (4, 'ux')
    UY = Variable      (5, 'uy')
    #UZ should not exist
    
    E = Variable(7, 'E')
    B = Variable(8, 'b')
    S = Variable(9, 'h+b')
    
    NVARS = 10

sw = ShallowWater()

class ShallowWaterLEGACY (object):
    H = Variable(0, 'h+b')
    
    HU  = VectorVariable(1, 'hu')
    HUX = Variable      (1, 'hux')
    HUY = Variable      (2, 'huy')
    #HUZ should not exist
    
    U  = VectorVariable(4,  'u')
    UX = Variable      (4, 'ux')
    UY = Variable      (5, 'uy')
    #UZ should not exist
    
    E = Variable(7, 'E')
    B = Variable(8, 'b')
    
    NVARS = 9

swl = ShallowWaterLEGACY()

# Wave (isotropic)
class Wave (object):
    P  = Variable(0, 'p')
    
    U  = VectorVariable(1, 'u')
    UX = Variable      (1, 'ux')
    UY = Variable      (2, 'uy')
    UZ = Variable      (3, 'uz')
    
    E = Variable(4, 'E')
    C = Variable(5, 'c')
    
    NVARS = 6

wa = Wave()

# Wave (anisotropic)
class WaveAnisotropic (object):
    P  = Variable(0, 'p')
    
    U  = VectorVariable(1, 'u')
    UX = Variable      (1, 'ux')
    UY = Variable      (2, 'uy')
    UZ = Variable      (3, 'uz')
    
    E = Variable(4, 'E')
    CXX = Variable(5,  'cxx')
    CXY = Variable(6,  'cxy')
    CXZ = Variable(7,  'cxz')
    CYX = Variable(8,  'cyx')
    CYY = Variable(9,  'cyy')
    CYZ = Variable(10, 'cyz')
    CZX = Variable(11, 'czx')
    CZY = Variable(12, 'czy')
    CZZ = Variable(13, 'czz')
    
    NVARS = 14

wa_ani = WaveAnisotropic()

# Buckley-Leverett
class BuckleyLeverett (object):
    So = Variable(0, 'So')
    
    Sw = Variable(1, 'Sw')
    
    NVARS = 2

bl = BuckleyLeverett()

# returns number of variables for a specific equation
def get_NVARS (equation):
    if equation == 'bl':
      return bl.NVARS
    elif equation == 'burgers':
      return br.NVARS
    elif equation == 'sw':
      return sw.NVARS
    elif equation == 'swl':
      return swl.NVARS
    elif equation == 'wave':
      return wa.NVARS
    elif equation == 'euler':
      return eu.NVARS
    elif equation == 'mhd':
      return mhd.NVARS
