
/*
 * General BALANCER routines:
 */

#include "balancer.h"

#include "multi.h"
#include "chaos.h"

#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <cstdio>
#include <cstdlib>

// needed for std::sort()
#include <algorithm>

std::vector<long_int> SAMPLES [L+1][MAX_SAMPLERS][L+1][2];

long_int SAMPLES_COUNT [L+1][L+1][2];

std::vector<Load> LOADS;

long_int LOADS_COUNT [MAX_SAMPLERS];

real TOTAL_LOAD [L+1][MAX_SAMPLERS];
real AVG_LOAD, MAX_LOAD;

Timer timer_balancer_distribute, timer_balancer_estimate;

void balancer () {
  
  #ifndef QUIET
  if (multi_main && MULTI_TOTAL > 1) {
    std::cout << std::endl << " :: BALANCER started... " << std::endl;
    #if ESTIMATE_LOADS
    std::cout << "  : -> Assuming stochastic CFL" << std::endl;
    #else
    std::cout << "  : -> Assuming deterministic CFL" << std::endl;
    #endif
  }
  #endif
  
  // synchronize before starting balancer
  multi_sync();
  
  #if ESTIMATE_LOADS
  
  timer_balancer_estimate.start ();
  
  // compute loads
  balancer_compute_loads ();
  
  // gather loads
  multi_balancer_gather_loads ();
  
  // compute AVG_LOAD
  balancer_compute_average_load ();
  
  // synchronize (to includes imbalance of parallelization of load estimation)
  multi_sync ();
  
  timer_balancer_estimate.stop ();
  
  // compute the parallelization overhead for load computation
  multi_compute_mpi_ddm_loads ();
  
  #endif
  
  timer_balancer_distribute.start();
  
  #ifndef QUIET
  if (multi_main && MULTI_TOTAL > 1) {
    std::cout << "  : -> Distributing loads... ";
    std::cout.flush();
  }
  #endif
  
  // distribute samples
  balancer_distribute_samples();
  
  #ifndef QUIET
  if (multi_main && MULTI_TOTAL > 1)
    std::cout << "done." << std::endl;
  #endif
  
  // compute MAX_LOAD
  balancer_compute_max_load();
  
  #if ENSURE_PARALLEL_REPRODUCIBILITY && SORT_SAMPLES
  
  // sort SAMPLES if required:
  // sorted SAMPLES improve performance of 'chaos_draw()', if ENSURE_PARALLEL_REPRODUCIBILITY == 1
  balancer_sort_samples ();
  
  #endif
  
  balancer_compute_samples_count();
  
  #ifndef QUIET
  if (multi_main && MULTI_TOTAL > 1) {
    #if ESTIMATE_LOADS
    std::cout << "  : -> Expected estimated imbalance: ";
    #else
    std::cout << "  : -> Expected theoretical imbalance: ";
    #endif
    std::cout << (round(1000 * (1.0 - AVG_LOAD / MAX_LOAD)) / 10) << "%";
    std::cout << std::endl;
  }
  #endif
  
  // write balancer information to disk
  balancer_write_datafiles();
  
  #ifndef QUIET
  if (multi_main && MULTI_TOTAL > 1)
    std::cout << " :: BALANCER done." << std::endl << std::endl;
  #endif
  
  // sync and stop the timer
  multi_sync();
  timer_balancer_distribute.stop();
}

#if ESTIMATE_LOADS

// computes loads 'LOADS[*]' for a fixed mesh level, etc..
// used in 'compute_loads()'
void balancer_compute_level_loads () {
  
  // setup RNG and chaos
  mlmc_setup_chaos ();
  
  // compute sizes of sample batches and remaining surplus
  long_int batch_size   = NM [SAMPLES_LEVEL()] / MULTIMl;
  long_int surplus_size = NM [SAMPLES_LEVEL()] % MULTIMl;
  
  // compute total_size and shift_size
  long_int total_size   = batch_size + (MULTI_SAMPLER < surplus_size);
  long_int shift_size   = MULTI_SAMPLER * batch_size + min(surplus_size, MULTI_SAMPLER);
  
  // compute LOADS_COUNT for each multi_sampler
  // and accumulate them in the corresponding multi_sampler
  // for the parallelization configuration of MESH_LEVEL = 0
  for ( int multi_sampler = 0; multi_sampler < MULTIMl; multi_sampler++ ) {
    int corr_multi_sampler = multi_sampler * MULTI[M] [L] [0] /  MULTI[M] [L] [MESH_LEVEL];
    LOADS_COUNT [corr_multi_sampler] += batch_size + (multi_sampler < surplus_size);
  }
  
  // loop over MC samples belonging to current sampler
  for (long_int i = shift_size + 1; i <= shift_size + total_size; i++) {
    
    // draw a sample
    mlmc_get_chaos (i);
    
    // initialize initial data, sources, fluxes and bc
    init_fvm ();
    
    // compute local time step
    real time_step = calc_time_step();
    
    // sync all cores within the same domain in order to measure communication overhead
    multi_sync_domain();
    
    // compute global time step
    time_step = multi_dt ( time_step );
    
    // loads are computed only on domain_root cores
    // this is required for easy gathering of loads
    if ( multi_domain_root ) {
      
      // compute relative load (scale down to avoid overflow in floating point arithmetic)
      real load = ( (real) N[X][MESH_LEVEL] / N[X][L] ) * ( (real) N[Y][MESH_LEVEL] / N[Y][L] ) * ( (real) N[Z][MESH_LEVEL] / N[Z][L] ) / time_step;
      
      // check if load is not infinite
      if ( 1/load == 0 ) {
        std::cerr << std::endl << " :: ERROR: infinite load encountered on MULTI_LEVEL = " << MULTI_LEVEL << " and MESH_LEVEL = " << MESH_LEVEL << std::endl;
        std::exit(1);
      }
      
      // append load to 'LOADS'
      LOADS .push_back ( Load( MESH_LEVEL, TYPE, i, load ) );
    }
  }
}

void balancer_compute_loads () {
  
  #ifndef QUIET
  if (multi_main && MULTI_TOTAL > 1) {
    std::cout << "  : -> Estimating loads";
    std::cout << " (using " << round ( 100.0 * MULTI[M][L][L] * MULTIN[L][L] / MULTI_TOTAL ) << "% of all cores)... ";
    std::cout.flush();
  }
  #endif
  
  // initialize LOADS_COUNT array
  for ( int multi_sampler = 0; multi_sampler < MAX_SAMPLERS; multi_sampler++ )
    LOADS_COUNT [multi_sampler] = 0;
  
  // cores from MULTI_LEVEL = L compute relative runtimes for all samples
  if ( MULTI_LEVEL == L ) {
    
    // compute loads for all MLMC levels
    for (int mlmc_level=L; mlmc_level>=0; mlmc_level--) {
      
      // === initialize TYPE = 0 samples for MC estimate E_{M_l}[U_l]
      
      // initialize to MESH_LEVEL = mlmc_level
      MESH_LEVEL = mlmc_level;
      
      // set type = 0
      TYPE = 0;
      
      // initialize parallelization setup for current MESH_LEVEL and TYPE
      multi_level_init ();
      
      // initialize to MESH_LEVEL
      mesh_level_init ();
      
      // compute loads
      balancer_compute_level_loads ();
      
      mesh_level_exit();
      
      #if ! DISTRIBUTE_IN_DIFFERENCES
      
      // === initialize TYPE = 1 samples for MC estimate E_{M_l}[U_{l-1}]
      
      if ( mlmc_level != 0 ) {
        
        // initialize to MESH_LEVEL = mlmc_level-1
        MESH_LEVEL = mlmc_level - 1;
        
        // set type = 1
        TYPE = 1;
        
        // initialize parallelization setup for current MESH_LEVEL and TYPE
        multi_level_init ();
        
        // initialize to MESH_LEVEL
        mesh_level_init ();
        
        // compute loads
        balancer_compute_level_loads ();
        
        mesh_level_exit();
      }
      
      #endif
    }
  }
  
  #ifndef QUIET
  if (multi_main && MULTI_TOTAL > 1)
    std::cout << "done." << std::endl;
  #endif
}

void balancer_compute_average_load () {
  
  // backward summation to ensure numerical stability
  AVG_LOAD = 0;
  for (int i=LOADS.size()-1; i>=0; i--)
    AVG_LOAD += LOADS[i].size;
  AVG_LOAD /= MULTI_TOTAL;
}

#endif

void balancer_compute_max_load () {
  
  MAX_LOAD = 0;
  
  for (int multi_level=L; multi_level>=LOWEST_MULTI_LEVEL; multi_level--)
    
    for (int multi_sampler=0; multi_sampler<MULTI[M][multi_level][L]; multi_sampler++)
      
      if ( TOTAL_LOAD[multi_level][multi_sampler] > MAX_LOAD )
        
        MAX_LOAD = TOTAL_LOAD[multi_level][multi_sampler];
}

#ifdef SORT_SAMPLES

void balancer_sort_samples () {
  
  // sort samples to improve performance of 'chaos_draw()'
  
  for (int multi_level=L; multi_level>=LOWEST_MULTI_LEVEL; multi_level--)
    
    for (int mesh_level=0; mesh_level<=L; mesh_level++)
      
      for (int multi_sampler=0; multi_sampler<MULTI[M][multi_level][mesh_level]; multi_sampler++)
        
        for (int type=0; type<2; type++)
          
          std::sort ( SAMPLES[multi_level][multi_sampler][mesh_level][type].begin(), SAMPLES[multi_level][multi_sampler][mesh_level][type].end() );
}

#endif

void balancer_compute_samples_count () {
  
  for ( int multi_level=L; multi_level>=LOWEST_MULTI_LEVEL; multi_level-- )
    
    for ( int mesh_level = 0; mesh_level <= L; mesh_level++ )
      
      for ( int type = 0; type < 2; type++ ) {
        
        SAMPLES_COUNT [multi_level][mesh_level][type] = 0;
        
        for ( int multi_sampler=0; multi_sampler<MULTI[M][multi_level][mesh_level]; multi_sampler++ )
          
          SAMPLES_COUNT [multi_level][mesh_level][type] += SAMPLES[multi_level][multi_sampler][mesh_level][type].size();
      }
}

// adds additional space characters for table formatting
template <class T>
std::string format ( T input, const int width ) {
  std::ostringstream oss;
  oss << input;
  if ( width - oss.str().size() >= 0 )
    oss << std::string ( width - oss.str().size(), ' ');
  return oss.str();
}


void balancer_write_report () {
  
  std::ostringstream oss;
  
  const int width  = 5;
  const int offset = 9;
  
  // header
  oss << " :: BALANCER REPORT: (rows: multi level - sampler, columns: mesh level)" << std::endl << std::endl;
  
  // column labels
  oss << format ( ' ', offset );
  oss << "|" << format ( ' ', width/2 ) << format ( L, width - width/2  ) << " ";
  for (int mesh_level=L-1; mesh_level>=0; mesh_level--)
    oss << "| " << format ( ' ', width - 1 ) << format ( mesh_level, width + 1 );
  oss << "|" << std::endl;
  
  // top border
  oss << std::string( offset + (width + 2) + L * 2 * (width + 1), '-') << '+' << std::endl;
  
  // the body of the table
  for (int multi_level=L; multi_level>=LOWEST_MULTI_LEVEL; multi_level--) {
    
    for (int multi_sampler=0; multi_sampler<MULTI[M][multi_level][L]; multi_sampler++) {
      
      // multi_level-multi_sampler
      std::ostringstream oss_label;
      oss_label << multi_level << "-" << multi_sampler;
      if ( MULTIN[multi_level][L] != 1 )
        oss_label << 'x' << MULTIN[multi_level][L];
      oss << format ( oss_label.str(), offset );
      
      for (MESH_LEVEL=L; MESH_LEVEL>=0; MESH_LEVEL--) {
        
        oss << "| ";
        
        for (TYPE=0; TYPE<2; TYPE++)
          
          if ( TYPE == 0 || MESH_LEVEL != L ) {
            
            // compute the sum all samples from all 'sub-samplers'
            int samples_count = 0;
            int samplers_count = MULTI[M][multi_level][MESH_LEVEL] / MULTI[M][multi_level][L];
            int samplers_start = samplers_count * multi_sampler;
            for (int sub_sampler = samplers_start; sub_sampler < samplers_start + samplers_count; sub_sampler ++)
              samples_count += SAMPLES[multi_level][sub_sampler][MESH_LEVEL][TYPE].size();
            
            if ( samples_count != 0 )
              oss << format ( round( 100.0 * samples_count / NM[SAMPLES_LEVEL()] ) / 100, width );
            else
              oss << format ( ' ', width );
          }
      }
      
      // end of loads
      oss << "|";
      
      // output estimated load of the sampler,
      // i.e. percentage of fraction TOTAL_LOAD[multi_level][multi_sampler] / AVG_LOAD - 1;
      real load = round( 1000 * (TOTAL_LOAD[multi_level][multi_sampler] / AVG_LOAD - 1) ) / 10;
      if ( load != 0 )
        oss << " " << (( load > 0 ) ? "+" : "") << load << "%";
      oss << std::endl;
      
      // bottom border
      oss << std::string( offset + (width + 2) + L * 2 * (width + 1), '-') << '+' << std::endl;
    }
  }
  
  // save generated table to BALANCER_FILE
  std::ostringstream name;
  name << BALANCER_REPORT_FILE;
  std::ofstream *file = new std::ofstream(name.str().c_str());
  *file << oss.str() << std::endl;
  file -> flush();
  file -> close();
  delete file;
}

void balancer_save_total_loads () {
  std::ostringstream hfn;
  hfn << BALANCER_FILE;
  std::ofstream *file = new std::ofstream(hfn.str().c_str());
  *file << "total_loads = [ ";
  int multi_level, sampler;
  for (multi_level=L; multi_level>=LOWEST_MULTI_LEVEL; multi_level--) {
    *file << "[ ";
    for (sampler=0; sampler<MULTI [M] [multi_level] [L]; sampler++)
      *file << TOTAL_LOAD [multi_level] [sampler] << ", ";
    *file << " ], ";
  }
  *file << " ]" << std::endl;
  file -> flush();
  file -> close();
  delete file;
}

#if ESTIMATE_LOADS
void balancer_save_loads () {
  std::ostringstream hfn;
  hfn << LOADS_FILE;
  std::ofstream *file = new std::ofstream(hfn.str().c_str());
  *file << "loads = [ ";
  int mesh_level, i;
  for (mesh_level=0; mesh_level<=L; mesh_level++) {
    *file << "[ ";
    for (i=0; i<LOADS.size(); i++)
      if ( LOADS[i].mesh_level == mesh_level )
        *file << LOADS[i].size << ", ";
    *file << " ], ";
  }
  *file << " ]" << std::endl;
  file -> flush();
  file -> close();
  delete file;
}
#endif

void balancer_write_datafiles () {
  
  // generate balancer report
  
  if (multi_main && MULTI_TOTAL > 1)  {
    
    #ifndef QUIET
    std::cout << " :: Writing balancer reports (TXT and DAT)..." ;
    std::cout.flush();
    #endif
    
    balancer_write_report();
    balancer_save_total_loads();
    
    #ifndef QUIET
    std::cout << " done." << std::endl;
    #endif
  }
  
  // save balancer LOADS
  
  #if ESTIMATE_LOADS
  
  if (multi_main && MULTI_TOTAL > 1)  {
    
    #ifndef QUIET
    std::cout << " :: Writing all estimated loads..." ;
    std::cout.flush();
    #endif
    
    balancer_save_loads();
    
    #ifndef QUIET
    std::cout << " done." << std::endl;
    #endif
  }
  
  #endif
}
