
#ifndef BALANCER_H
#define BALANCER_H

#include "mlmc.h"

#include <vector>

/*
 * Declarations of BALANCER routines
 */

const int MAX_SAMPLERS = MULTI_CORES > 0 ? MULTI_CORES : MULTIX * MULTIY * MULTIZ * MULTIM;

// SAMPLES[multi_level][multi_sampler][mesh_level][*] - vector of indices of samples
// to be run on sampler 'multi_sampler' of 'multi_level' at the mesh level 'mesh_level'
// SAMPLES[*][*][*][0] contains samples which should be added      (type 0)
// SAMPLES[*][*][*][1] contains samples which should be subtracted (type 1)
extern std::vector<long_int> SAMPLES [L+1][MAX_SAMPLERS][L+1][2];

// SAMPLES_COUNT[multi_level][mesh_level][type] - sum of the number of samples
// to be run on ALL 'multi_samplers' of 'multi_level'
// at a given mesh level 'mesh_level' and of type 'type'
extern long_int SAMPLES_COUNT [L+1][L+1][2];

// structure for a relative load (runtime) of a sample
struct Load
{
  int      mesh_level; // mesh level of the load
  int      type;       // type=0 - 1st term; type=1 - 2nd term (in the MC differences of the MLMC estimator)
  long_int index;      // index of the sample
  real     size;       // relative size (runtime) of the sample
  
  // constructors
  Load (int l, int t, long_int i, real s) : mesh_level(l), type(t), index(i), size(s) {}
  Load () : mesh_level(0), type(0), index(0), size(0) {}
  
  bool operator < (const Load &load) const {
    return (size < load.size);
  }
};

// vector to store loads loads
extern std::vector<Load> LOADS;

// number of loads (samples) that are estimated on each sampler
// depends on the parallelization configuration for a given MESH_LEVEL
extern long_int LOADS_COUNT [MAX_SAMPLERS];

// TOTAL_LOAD[multi_level][multi_sampler] - total load for a sampler
extern real TOTAL_LOAD [L+1][MAX_SAMPLERS];

// AVG_LOAD - average load per 1 core
extern real AVG_LOAD;

// MAX_LOAD - maximal load among all cores
extern real MAX_LOAD;

// timers
extern Timer timer_balancer_estimate, timer_balancer_distribute;

// === routines defined in balancer.cpp file

// check compatibility
void balancer_compatibility_checks();

// manage balancer routines and I/O feedback
void balancer ();

// computes 'loads[*][*]' if 'ESTIMATE_LOADS=1' is specified
void balancer_compute_loads ();

// using 'LOADS', compute 'AVG_LOAD'
void balancer_compute_average_load ();

// using 'TOTAL_LOAD', compute 'MAX_LOAD'
void balancer_compute_max_load ();

// compute SAMPLES_COUNT
void balancer_compute_samples_count ();

// sort resulting SAMPLES array
void balancer_sort_samples ();

// report the results of the balancer (generates table)
void balancer_write_report ();

// save TOTAL_LOADS
void balancer_save_total_loads ();

// save loads
#if ESTIMATE_LOADS
void balancer_save_loads ();
#endif

// write balancer information to disk
// executes balancer_write_report(), balancer_save_total_loads() and balancer_save_loads()
void balancer_write_datafiles ();

// === routines defined in corresponding balancer_*.cpp file

// distribute samples to all cores (i.e. compute SAMPLES[*][*][*])
void balancer_distribute_samples ();

#endif
