
///::title:: Adaptive (for sample-dependent wave speeds (CFL))
///::consts:: ESTIMATE_LOADS=1 SORT_SAMPLES=1 ENFORCE_DDM_SETUP=0

/*
 * Loads are distributed by assigning samples to multi levels
 * using parallel adaptive (based on precomputation of CFL condition) load balancing.
 * Load balancer is parallel, i.e. loads of samples are computed in parallel.
 * 
 * Step 1:
 * Loads of each sample are computed based on (random) CFL condition,
 * i.e. time step is determined to be CFL / (max wave speed).
 * 
 * Step 2:
 * Loads are distributed by assigning samples to multi levels
 * using extension of the "greedy" algorithm to variable size computing units:
 *   
 *   - always assigning the next largest load to the 'best' sampler.
 *     where 'best' indicates sample for which the TOTAL_LOAD is MINIMIZED
 * 
 * Step 3:
 * Finally, all samples are sorted by index,
 * i.e. all values in SAMPLES[*][*][*][*] are in increasing order.
 */

#include "balancer.h"

#include "multi.h"

// needed for std::sort()
#include <algorithm>

#include <iostream>

// TOTAL_SAMPLER_LOAD [multi_level] [multi_sampler]
std::vector<real> TOTAL_SAMPLER_LOAD [L+1];

int best_multi_level, best_multi_sampler;
real total_load, best_total_level_load, best_total_sampler_load;

void balancer_compatibility_checks() {
  
  // for MULTIM == ML, expect sub-optimal performance of the adaptive balancer
  if ( MULTIM == ML && ! MULTI_TOTAL > 0 )
    if (multi_main) std::cout << " :: WARNING: Since MULTIM >= ML, expect sub-optimal performace of balancer:adaptive." << std::endl << std::endl;
}

// return the maximal total load in 'multi_level' of the specified 'multi_sampler' for the specified parallelization configuration of 'mesh_level' 
real get_total_load (const int &multi_level, const int &multi_sampler, const int &mesh_level ) {
  
  // determine the required range of samplers for the specified 'mesh_level'
  int samplers_count = MULTI[M][multi_level][0] / MULTI[M][multi_level][mesh_level];
  int samplers_start = samplers_count * multi_sampler;
  
  // find the maximal load among the samplers for the specified 'mesh_level'
  std::vector<real>::const_iterator it;
  it = max_element ( TOTAL_SAMPLER_LOAD[multi_level].begin() + samplers_start, TOTAL_SAMPLER_LOAD[multi_level].begin() + samplers_start + samplers_count);
  
  return *it;
}

// return the effective value of the load if the specified 'load' on the specified 'mesh_level' _would_ be computed on the specified 'multi_level'
real get_effective_sample_load (const int &multi_level, const int &mesh_level, const real &load) {
  return load / MULTIN[multi_level][mesh_level];
}

// return the maximal total load in 'multi_level' if the specified 'load' on the specified 'mesh_level' _would_ be assigned to the specified 'multi_sampler'
// the vector TOTAL_SAMPLER_LOAD [multi_level] is _not_ modified
real get_total_load (const int &multi_level, const int &multi_sampler, const int &mesh_level, const real &load) {
  
  // compute and return the effective total load if a sampler with load 'load' would be added
  return get_total_load (multi_level, multi_sampler, mesh_level) + get_effective_sample_load (multi_level, mesh_level, load);
}

// add the specified 'load' to appropriate loads in the vector TOTAL_SAMPLER_LOAD [multi_level]
void update_total_load (const int &multi_level, const int &multi_sampler, const int &mesh_level, const real &load) {
  
  // compute the effective load
  real effective_load = get_effective_sample_load (multi_level, mesh_level, load);
  
  // determine the required range of samplers
  int samplers_count = MULTI[M][multi_level][0] / MULTI[M][multi_level][mesh_level];
  int samplers_start = samplers_count * multi_sampler;
  
  for (int multi_sampler = samplers_start; multi_sampler < samplers_start + samplers_count; multi_sampler++)
    TOTAL_SAMPLER_LOAD [multi_level] [multi_sampler] += effective_load;
}

// find 'best_multi_sampler'
int find_best_multi_sampler_for (const int &multi_level, const int &mesh_level, const real &load) {
  
  // choose initial guess of the 'best_multi_sampler'
  int best_multi_sampler = 0;
  
  // compute 'best_total_sampler_load'
  best_total_sampler_load = get_total_load (multi_level, best_multi_sampler, mesh_level, load);
  
  // iterate over all multi_samplers
  for (int multi_sampler=0; multi_sampler<MULTI[M][multi_level][mesh_level]; multi_sampler++) {
    
    // compute 'total_load' for the given 'multi_level' and 'multi_sampler'
    total_load = get_total_load (multi_level, multi_sampler, mesh_level, load);
    
    // compare 'total_load' with the 'best_total_sampler_load' found so far
    if ( total_load < best_total_sampler_load ) {
      best_multi_sampler      = multi_sampler;
      best_total_sampler_load = total_load;
    }
  }
  
  return best_multi_sampler;
}

// find 'best_multi_level' by iterating over all multi_levels
int find_best_multi_level_for (const int &mesh_level, const real &load) {
  
  // choose initial guess of the 'best_multi_level'
  int best_multi_level = L;
  
  // find the 'best_multi_sampler' for the given initial 'best_multi_level' for 'load' at 'mesh_level'
  int best_multi_sampler = find_best_multi_sampler_for (best_multi_level, mesh_level, load);
  
  // compute the initial 'best_total_load' for the initial 'best_multi_level' and 'best_multi_sampler'
  best_total_level_load = get_total_load (best_multi_level, best_multi_sampler, mesh_level, load);
  
  // iterate over all multi_levels
  for (int multi_level=L; multi_level>=LOWEST_MULTI_LEVEL; multi_level--) {
    
    // find the 'best_multi_sampler' of the given 'multi_level' for 'load' at 'mesh_level'
    best_multi_sampler = find_best_multi_sampler_for (multi_level, mesh_level, load);
    
    // compute 'total_load' for the given 'multi_level' and 'best_multi_sampler'
    total_load = get_total_load (multi_level, best_multi_sampler, mesh_level, load);
    
    // compare 'total_load' with the 'best_total_level_load' found so far
    if ( total_load < best_total_level_load ) {
      best_multi_level      = multi_level;
      best_total_level_load = total_load;
    }
  }
  
  return best_multi_level;
}

void balancer_distribute_samples () {
  
  // sort LOADS (ascending order)
  std::sort ( LOADS.begin(), LOADS.end() );
  
  // reverse LOADS (descending order)
  std::reverse ( LOADS.begin(), LOADS.end() );
  
  // set sizes of TOTAL_SAMPLER_LOAD for each multi_level and initialize loads to 0
  for (int multi_level=L; multi_level>=LOWEST_MULTI_LEVEL; multi_level--)
    TOTAL_SAMPLER_LOAD [multi_level] .resize (MULTI[M][multi_level][0], 0);
  
  // === distribute LOADS to multi levels and samplers (i.e. construct SAMPLES)
  for (long_int i=0; i<LOADS.size(); i++) {
    
    if ( LOADS[i].size > 1e+30 || LOADS[i].size < 0 ) {
      std::cerr << " :: ERROR: infinite (>1e+30) or negative load." << std::endl;
      std::exit(1);
    }
    
    // find best multi_level
    best_multi_level = find_best_multi_level_for (LOADS[i].mesh_level, LOADS[i].size);
    
    // find best_multi_sampler for the best_multi_level
    best_multi_sampler = find_best_multi_sampler_for (best_multi_level, LOADS[i].mesh_level, LOADS[i].size);
    
    // assign the load to SAMPLES [best_multi_level] [best_multi_sampler]
    SAMPLES [best_multi_level] [best_multi_sampler] [LOADS[i].mesh_level] [LOADS[i].type] .push_back (LOADS[i].index);
    
    #if DISTRIBUTE_IN_DIFFERENCES
    
    // additionally assign the corresponding TYPE = 1 load to SAMPLES [best_multi_level] [best_multi_sampler]
    if ( LOADS[i].mesh_level != 0 )
      SAMPLES [best_multi_level] [best_multi_sampler] [LOADS[i].mesh_level - 1] [1] .push_back (LOADS[i].index);
    
    #endif
    
    // update the 'TOTAL_SAMPLER_LOAD [multi_level]'
    update_total_load (best_multi_level, best_multi_sampler, LOADS[i].mesh_level, LOADS[i].size);
  }
  
  // compute TOTAL_LOAD
  for (int multi_level=L; multi_level>=LOWEST_MULTI_LEVEL; multi_level--)
    for (int multi_sampler=0; multi_sampler<MULTI[M][multi_level][L]; multi_sampler++)
      TOTAL_LOAD [multi_level] [multi_sampler] = get_total_load (multi_level, multi_sampler, L);
}
