
///::title:: Static (for sample-independent wave speeds (CFL))
///::consts:: ENFORCE_DDM_SETUP=1

/*
 * Loads are distributed by assigning samples to multi levels
 * using static load balancing.
 * 
 * For detailed description, refer to:
 * 
 * J. Šukys, S. Mishra, and Ch. Schwab
 * Static load balancing for multi-level Monte Carlo finite volume solvers.
 * PPAM 2011, Part I, LNCS 7203, pp. 245--254. Springer, Heidelberg, 2012.
 * DOI:10.1007/978-3-642-31464-3_25
 */

#include "balancer.h"

#include "multi.h"

#include <iostream>
#include <cstdlib>

void balancer_compatibility_checks() {
  
  // MULTI_CORES > 1 is not supported
  if ( MULTI_CORES > 1 ) {
    if (multi_main) std::cerr << " :: ERROR: MULTI_CORES > 1 is NOT supported with balancer:static." << std::endl << std::endl;
    if (multi_main) std::exit(1);
  }
  
  // if L != 0, MULTI[X][L][L] must be a positive divider of NX/2
  if ( L != 0 && NX/2 % MULTI[X][L][L] != 0 ) {
    if (multi_main) std::cerr << " :: ERROR: MULTIX must be a positive divider of NX/2." << std::endl << std::endl;
    if (multi_main) std::exit(1);
  }
  
  // if L != 0, MULTI[Y][L][L] must be a positive divider of NY/2
  if ( L != 0 && NY/2 % MULTI[Y][L][L] != 0 ) {
    if (multi_main) std::cerr << " :: ERROR: MULTIY must be a positive divider of NY/2." << std::endl << std::endl;
    if (multi_main) std::exit(1);
  }
  
  // if L != 0, MULTI[Z][L][L] must be a positive divider of NZ/2
  if ( L != 0 && NZ/2 % MULTI[Z][L][L] != 0 ) {
    if (multi_main) std::cerr << " :: ERROR: MULTIZ must be a positive divider of NZ/2." << std::endl << std::endl;
    if (multi_main) std::exit(1);
  }
  
  // USE_OPTIMAL_NUMBER_OF_SAMPLES == 1 might be poor with balancer:static
  if ( USE_OPTIMAL_NUMBER_OF_SAMPLES ) {
    if (multi_main) std::cout << " :: WARNING: USE_OPTIMAL_NUMBER_OF_SAMPLES = 1 might result in poor balancing with balancer:static." << std::endl << std::endl;
  }
  
  // READ_NUMBER_OF_SAMPLES_FROM_FILE == 1 might be poor with balancer:static
  if ( READ_NUMBER_OF_SAMPLES_FROM_FILE ) {
    if (multi_main) std::cout << " :: WARNING: READ_NUMBER_OF_SAMPLES_FROM_FILE = 1 might result in poor balancing with balancer:static." << std::endl << std::endl;
  }
  
  // USE_ONE_SAMPLE_PER_LEVEL == 1 might be poor with balancer:static
  if ( USE_ONE_SAMPLE_PER_LEVEL ) {
    if (multi_main) std::cout << " :: WARNING: USE_ONE_SAMPLE_PER_LEVEL = 1 might result in poor balancing with balancer:static." << std::endl << std::endl;
  }
}

// creates vector with elements (begin, begin+1, ..., end-1)
std::vector<long_int> get_sequential_vector (long_int begin, long_int end) {
  std::vector<long_int> vec (end - begin);
  for (long_int i=0; i<vec.size(); i++)
    vec[i] = begin + i;
  return vec;
}

// distributes samples from the specified mesh_level and type to the samplers of specified multi_level
void distribute_level_samples (const int &samples_level, const int &type, const int &mesh_level, const int &multi_level) {
  
  long_int batch_size, surplus_size, samples_begin, samples_end, load_per_sample;
  int multi_sampler_L;
  
  // compute sizes of sample batches and remaining surplus
  batch_size   = NM[samples_level] / MULTI[M][multi_level][mesh_level];
  surplus_size = NM[samples_level] % MULTI[M][multi_level][mesh_level];
  
  // compute 'load_per_sample'
  load_per_sample = N[X][mesh_level] * N[Y][mesh_level] * N[Z][mesh_level] * ( 1 << ( (1 + HAVE_VISCOSITY) * mesh_level ) );
  
  // initialize 'samples_end'
  samples_end = 1;
  
  // iterate over samplers
  for (int multi_sampler=0; multi_sampler<MULTI[M][multi_level][mesh_level]; multi_sampler++) {
    
    // compute the part of samples which belongs to sampler 'multi_sampler'
    samples_begin = samples_end;
    samples_end   = samples_begin + batch_size + (multi_sampler < surplus_size);
    
    // set SAMPLES
    SAMPLES [multi_level][multi_sampler][mesh_level][type] = get_sequential_vector (samples_begin, samples_end);
  }
  
  // update AVG_LOAD
  AVG_LOAD += NM[samples_level] * load_per_sample;
  
  // update TOTAL_LOAD
  for (int multi_sampler=0; multi_sampler<MULTI[M][multi_level][L]; multi_sampler++)
    TOTAL_LOAD [multi_level] [multi_sampler] += (batch_size + (multi_sampler < surplus_size)) * load_per_sample / MULTIN[multi_level][L];
}

void balancer_distribute_samples () {
  
  int multi_level, mesh_level, samples_level, type;
  
  // set initial average load to zero
  AVG_LOAD = 0;
  
  // set initial total load to zero
  for (int multi_level=L; multi_level>=LOWEST_MULTI_LEVEL; multi_level--)
    for (int multi_sampler=0; multi_sampler<MULTI[M][multi_level][L]; multi_sampler++)
      TOTAL_LOAD[multi_level][multi_sampler] = 0;
  
  // iterate over MLMC level pairs
  for (int mlmc_level=L; mlmc_level>=0; mlmc_level--) {
    
    // all samples of multi_levels below LOWEST_MULTI_LEVEL
    // must go to multi_level = LOWEST_MULTI_LEVEL
    multi_level = (mlmc_level < LOWEST_MULTI_LEVEL) ? LOWEST_MULTI_LEVEL : mlmc_level;
    
    // set samples level
    samples_level = mlmc_level;
    
    // MC estimator E_{M_l}[U_l]
    // i.e. type = 0 samples with mesh_level = mlmc_level and samples_level = mlmc_level
    mesh_level = mlmc_level;
    type = 0;
    distribute_level_samples (samples_level, type, mesh_level, multi_level);
    
    // MC estimator E_{M_l}[U_{l-1}], only for l > 0
    // i.e. all type = 1 samples with mesh_level = mlmc_level-1 and samples_level = mlmc_level
    if ( mlmc_level == 0 ) continue;
    mesh_level = mlmc_level - 1;
    type = 1;
    distribute_level_samples (samples_level, type, mesh_level, multi_level);
  }
  
  // finalize the computation of AVG_LOAD
  AVG_LOAD /= MULTI_TOTAL;
}
