
/*
 * GENERAL FVM routines
 */

#include "equation.h"
#include "quadrature.h"
#include "multi.h"

#include <iostream>

int NXl, NYl, NZl, NXl_FULL, NYl_FULL, NZl_FULL;
real dt, t, t_halt, halt_times[NSAVES+SAVE0], dx[NDIR];

volume<ConservedVars> U;
volume<ConservedVars> Flux[NDIR];
volume<ConservedVars> Source;
volume<XtraVars>      Xtra;

real get_x (const int &i) { return (i + NXl * multix + 0.5) * dx[X]; }
real get_y (const int &j) { return (j + NYl * multiy + 0.5) * dx[Y]; }
real get_z (const int &k) { return (k + NZl * multiz + 0.5) * dx[Z]; }

real get_x_per (const int &i) { return (i + NXl * multix + 0.5) * dx[X]; }
real get_y_per (const int &j) { return (j + NYl * multiy + 0.5) * dx[Y]; }
real get_z_per (const int &k) { return (k + NZl * multiz + 0.5) * dx[Z]; }

void calc_all_vars(AllVariables &V, const PrimitiveVars &P) {
  P.toConservedVars(V.U);
  calc_xtra (V.XU, V.U);
}

void initial_data (PrimitiveVars &P, int &i, int &j, int &k) {
  
  // use quadrature rule
  
  PrimitiveVars QP;
  real x = get_x(i);
  real y = get_y(j);
  real z = get_z(k);
  
  for (int px=0; px<(HAVE_DIR(X) ? QUAD_NUM : 1); px++)
    for (int py=0; py<(HAVE_DIR(Y) ? QUAD_NUM : 1); py++)
      for (int pz=0; pz<(HAVE_DIR(Z) ? QUAD_NUM : 1); pz++) {
        Zero(QP);
        initial_data (QP, x + QUAD_POINTS[px] * dx[X],
                          y + QUAD_POINTS[py] * dx[Y],
                          z + QUAD_POINTS[pz] * dx[Z]);
        if (HAVE_DIR(X)) QP *= QUAD_WEIGHTS[px];
        if (HAVE_DIR(Y)) QP *= QUAD_WEIGHTS[py];
        if (HAVE_DIR(Z)) QP *= QUAD_WEIGHTS[pz];
        P += QP;
      }
}

void model_init() {
  
  // initialize coefficients
  coefficients_init ();
  
  // initialize initial data
  PrimitiveVars P;
  
  for (int i = 0; i < NXl; i++) {
    for (int j = 0; j < NYl; j++) {
      for (int k = 0; k < NZl; k++) {
        Zero(P);
        initial_data (P, i, j, k);
        if ( INITIAL_DATA_SCALING != 1 ) P *= INITIAL_DATA_SCALING;
        if ( INITIAL_DATA_OFFSET  != 0 ) P += INITIAL_DATA_OFFSET;
        INITIAL_DATA_HOOK
        P.toConservedVars (U[i][j][k]);
      }
    }
  }
}

void init_fvm () {
  t = 0;
  model_init();
  time_step_calc_xtra();
}

void coefficients_init_ghost_cells () {
  
  // Remark: only coefficients should be affected, not all XtraVars
  
  // start parallel communication of multi boundaries for XtraVars
  multi_ddm_start_send_recv_xtra ();
  
  // extend coefficients at the domain boundaries
  time_step_extend_coefficients ();
  
  // confirm receives of multi boundaries for XtraVars
  multi_ddm_confirm_recv ();
  
  // confirm sends of multi boundaries
  multi_ddm_confirm_send ();
}

void run_fvm () {
  
  // initialize ghost cells for coefficients
  if ( t == 0 && t_halt > t ) coefficients_init_ghost_cells ();
  
  // run fvm time-stepping until t_halt is reached
  while (t < t_halt) {
    
    PRE_STEP_HOOK
    
    if (multi_main && std::isnan(t)) {
      std::cerr << " :: ERROR: time is NaN." << std::endl;
      std::exit(1);
    }
    
    // run fvm up to time 't + dt'
    time_step ();
    
    #ifndef QUIET
    if (PRINT_t && multi_main) std::cout << " :: time:  t=" << t << " (dt=" << dt << ")" << std::endl;
    #endif
    
    POST_STEP_HOOK
  }
}

real calc_time_step () {
  
  vector maxspeed;
  maxspeed.zero();
  
  for (int d = 0; d < NDIR; d++)
    if (HAVE_DIR(d))
      for (int i = 0; i < NXl; i++)
        for (int j = 0; j < NYl; j++)
          for (int k = 0; k < NZl; k++)
            maxspeed[d] = max( maxspeed[d], calc_max_speed (U[i][j][k], Xtra[i][j][k], d) );
  
  #if HAVE_VISCOSITY
  for (int d = 0; d < NDIR; d++)
    if (HAVE_DIR(d))
      maxspeed[d] += VISCOSITY / dx[d];
  #endif
  
  return CFL/(maxspeed[X]/dx[X] + maxspeed[Y]/dx[Y] + maxspeed[Z]/dx[Z]);
}

void time_step_calc_xtra () {
  
  for (int i = 0; i < NXl; i++)
    for (int j = 0; j < NYl; j++)
      for (int k = 0; k < NZl; k++)
        calc_xtra (Xtra[i][j][k], U[i][j][k], i, j, k);
}

void time_step_calc_xtra_internal_cells () {
  
  // if boundary layer is sufficiently thick, all cells are 'boundary' cells
  if ( NXl < 2*GX || NYl < 2*GY || NZl < 2*GZ )
    return;
  
  // compute XtraVars for 'internal' cells
  for (int i = GX; i < NXl-GX; i++)
    for (int j = GY; j < NYl-GY; j++)
      for (int k = GZ; k < NZl-GZ; k++)
        calc_xtra (Xtra[i][j][k], U[i][j][k], i, j, k);
}

void time_step_calc_xtra_external_cells () {
  
  // if boundary layer is sufficiently thick, all cells are 'external' cells
  if ( NXl < 2*GX || NYl < 2*GY || NZl < 2*GZ ) {
    time_step_calc_xtra ();
    return;
  }
  
  // compute XtraVars for all external cells except 'corner cells'
  int p, i, j, k;
  if ( ACTIVE_DIRS > 1 && HAVE_DIR(X) )
    for (p = 0; p < 2; p++)
      for (i = (p==0?0:NXl-GX); i < (p==0?GX:NXl); i++)
        for (j = GY; j < NYl-GY; j++)
          for (k = GZ; k < NZl-GZ; k++)
            calc_xtra (Xtra[i][j][k], U[i][j][k], i, j, k);
  if ( ACTIVE_DIRS > 1 && HAVE_DIR(Y) )
    for (p = 0; p < 2; p++)
      for (i = GX; i < NXl-GX; i++)
        for (j = (p==0?0:NYl-GY); j < (p==0?GY:NYl); j++)
          for (k = GZ; k < NZl-GZ; k++)
            calc_xtra (Xtra[i][j][k], U[i][j][k], i, j, k);
  if ( ACTIVE_DIRS > 1 && HAVE_DIR(Z) )
    for (p = 0; p < 2; p++)
      for (i = GX; i < NXl-GX; i++)
        for (j = GY; j < NYl-GY; j++)
          for (k = (p==0?0:NZl-GZ); k < (p==0?GZ:NZl); k++)
            calc_xtra (Xtra[i][j][k], U[i][j][k], i, j, k);
  
  // compute XtraVars for 'corner cells'
  int ci, cj, ck;
  for (ci = 0; ci <= HAVE_DIR(X); ci++)
    for (cj = 0; cj <= HAVE_DIR(Y); cj++)
      for (ck = 0; ck <= HAVE_DIR(Z); ck++)
        for (i = (ci==0?0:max(0,NXl-GX)); i < (ci==0?min(max(GX,1),NXl):NXl); i++)
          for (j = (cj==0?0:max(0,NYl-GY)); j < (cj==0?min(max(GY,1),NYl):NYl); j++)
            for (k = (ck==0?0:max(0,NZl-GZ)); k < (ck==0?min(max(GZ,1),NZl):NZl); k++)
              calc_xtra (Xtra[i][j][k], U[i][j][k], i, j, k);
}

void time_step_calc_xtra_bc () {
  
  #if NX != 1
  
  if (!multi_bc[X][HEAD])
    for (int i = 0; i < GX; i++)
      for (int j = 0; j < NYl; j++)
        for (int k = 0; k < NZl; k++)
          calc_xtra (Xtra[-1-i][j][k], U[-1-i][j][k], -1-i, j, k);
  
  if (!multi_bc[X][TAIL])
    for (int i = 0; i < GX; i++)
      for (int j = 0; j < NYl; j++)
        for (int k = 0; k < NZl; k++)
          calc_xtra (Xtra[i+NXl][j][k], U[i+NXl][j][k], i+NXl, j, k);
  
  #endif
  
  #if NY != 1
  
  if (!multi_bc[Y][HEAD])
    for (int i = 0; i < NXl; i++)
      for (int j = 0; j < GY; j++)
        for (int k = 0; k < NZl; k++)
          calc_xtra (Xtra[i][-1-j][k], U[i][-1-j][k], i, -1-j, k);
  
  if (!multi_bc[Y][TAIL])
    for (int i = 0; i < NXl; i++)
      for (int j = 0; j < GY; j++)
        for (int k = 0; k < NZl; k++)
          calc_xtra (Xtra[i][j+NYl][k], U[i][j+NYl][k], i, j+NYl, k);
  
  #endif
  
  #if NZ != 1
  
  if (!multi_bc[Z][HEAD])
    for (int i = 0; i < NXl; i++)
      for (int j = 0; j < NYl; j++)
        for (int k = 0; k < GZ; k++)
          calc_xtra (Xtra[i][j][-1-k], U[i][j][-1-k], i, j, -1-k);
  
  if (!multi_bc[Z][TAIL])
    for (int i = 0; i < NXl; i++)
      for (int j = 0; j < NYl; j++)
        for (int k = 0; k < GZ; k++)
          calc_xtra (Xtra[i][j][k+NZl], U[i][j][k+NZl], i, j, k+NZl);
  
  #endif
}

void time_step_calc_xtra_ddm () {
  
  #if NX != 1
  
  if (multi_bc[X][HEAD])
    for (int i = 0; i < GX; i++)
      for (int j = 0; j < NYl; j++)
        for (int k = 0; k < NZl; k++)
          calc_xtra (Xtra[-1-i][j][k], U[-1-i][j][k], -1-i, j, k);
  
  if (multi_bc[X][TAIL])
    for (int i = 0; i < GX; i++)
      for (int j = 0; j < NYl; j++)
        for (int k = 0; k < NZl; k++)
          calc_xtra (Xtra[i+NXl][j][k], U[i+NXl][j][k], i+NXl, j, k);
  
  #endif
  
  #if NY != 1
  
  if (multi_bc[Y][HEAD])
    for (int i = 0; i < NXl; i++)
      for (int j = 0; j < GY; j++)
        for (int k = 0; k < NZl; k++)
          calc_xtra (Xtra[i][-1-j][k], U[i][-1-j][k], i, -1-j, k);
  
  if (multi_bc[Y][TAIL])
    for (int i = 0; i < NXl; i++)
      for (int j = 0; j < GY; j++)
        for (int k = 0; k < NZl; k++)
          calc_xtra (Xtra[i][j+NYl][k], U[i][j+NYl][k], i, j+NYl, k);
  
  #endif
  
  #if NZ != 1
  
  if (multi_bc[Z][HEAD])
    for (int i = 0; i < NXl; i++)
      for (int j = 0; j < NYl; j++)
        for (int k = 0; k < GZ; k++)
          calc_xtra (Xtra[i][j][-1-k], U[i][j][-1-k], i, j, -1-k);
  
  if (multi_bc[Z][TAIL])
    for (int i = 0; i < NXl; i++)
      for (int j = 0; j < NYl; j++)
        for (int k = 0; k < GZ; k++)
          calc_xtra (Xtra[i][j][k+NZl], U[i][j][k+NZl], i, j, k+NZl);
  
  #endif
}
