
/*
 * Declarations of two types of FVM routines:
 *  - GENERAL routines for arbitrary system of conservation laws
 *  - EQUATION SPECIFIC (Burgers', advection, MHD, Euler, Shallow Water, Wave, etc.) routines
 */

#ifndef EQUATION_H
#define EQUATION_H

#include "flux.h"

#include "volumes.h"
#include "structures.h"

// === global variables

extern int NXl, NYl, NZl, NXl_FULL, NYl_FULL, NZl_FULL;
extern real dt, t, t_halt, halt_times[NSAVES+SAVE0], dx[NDIR];

// === global volumes

extern volume<ConservedVars> U;
extern volume<ConservedVars> Flux[NDIR];
extern volume<ConservedVars> Source;
extern volume<XtraVars>      Xtra;

// === global functions

static int SIZE (int d) { return d==X?NXl:(d==Y?NYl:NZl); }

static int FLUX_NX (int d) { return d==X?NXl+1:NXl; }
static int FLUX_NY (int d) { return d==Y?NYl+1:NYl; }
static int FLUX_NZ (int d) { return d==Z?NZl+1:NZl; }

// === GENERAL FVM routines (defined in equation.cpp and elsewhere)

// compute x, y, and z coordinates of cell centers from cell indices i, j, k
real get_x (const int &i);
real get_y (const int &j);
real get_z (const int &k);

// compute x, y, and z coordinates of cell centers from cell indices i, j, k (periodic)
real get_x_per (const int &i);
real get_y_per (const int &j);
real get_z_per (const int &k);

// compute AllVariables from PrimitiveVars
void calc_all_vars (AllVariables &V, const PrimitiveVars &P);

// compute needed quantities: mesh sizes, initial data, etc.
void model_init ();

// set CHAOS_SIZE[MODEL] (defined in model_*.cpp files)
void model_set_chaos_size ();

// set CHAOS_SIZE[TOPOGRAPHY] (defined in topography_*.cpp files)
void topography_set_chaos_size ();

// set CHAOS_SIZE[FLUX] (defined in flux_*.cpp files)
void flux_set_chaos_size ();

// use quadrature to compute average of initial data over mesh cell (i,j,k)
void initial_data (PrimitiveVars &P, const int &i, const int &j, const int &k);

// evaluate initial data at domain point (x,y,z)
void initial_data (PrimitiveVars &P, real x, real y, real z);

// initialize ghost cells for coefficients
void coefficients_init_ghost_cells ();

// extend coefficients at the domain boundaries
void time_step_extend_coefficients ();

// initialize FVM
void init_fvm ();

// applies boundary conditions
template <class T>
void time_step_set_bc (volume<T> &v);

template <class T>
void set_bc_x (volume<T> &v, const int &i, const int &j, const int &k);

template <class T>
void set_bc_y (volume<T> &v, const int &i, const int &j, const int &k);

template <class T>
void set_bc_z (volume<T> &v, const int &i, const int &j, const int &k);

// computes XtraVars for boundary conditions ('U' is set with 'set_bc_*()')
void time_step_calc_xtra_bc ();

// computes XtraVars for multi boundaries (only 'U' was communicated)
void time_step_calc_xtra_ddm ();

// calculate timestep using CFL number
// Remark: for HLL-type solvers, this is NOT correctly computed,
// since this does NOT use the reconstructed solution.
// This is used only as an estimate of time step for USE_ESTIMATED_LOADS=1 in balancer:adaptive
real calc_time_step();

// time integration routines
void time_step();

// compute XtraVars for ALL domain cells
void time_step_calc_xtra();

// compute XtraVars for 'internal' cells
void time_step_calc_xtra_internal_cells();

// compute XtraVars for remaining ('external') cells
void time_step_calc_xtra_external_cells();

// run FVM
void run_fvm ();

// === EQUATION SPECIFIC FVM routines (defined in corresponding equation_*.cpp)

// set required CHAOS_SIZE [STREAM] [DISTRIBUTION] each STREAM and DISTRIBUTION
void set_chaos_size ();

// apply constraints to variables; for instance, pressure should be non-negative, etc.
void apply_constraints (ConservedVars &CV, XtraVars &XV);

// calculate indicator for limiters
real indicator (const int &i, const int &j, const int &k, const int &d);

// === DEFINITIONS OF TEMPLATED routines

#include "multi.h"

template <class T>
void time_step_set_bc (volume<T> &v) {
  
  int i, j, k;
  
  #if NX != 1
  
  if (!multi_bc[X][HEAD])
    for (i = 0; i < GX; i++)
      for (j = 0; j < NYl; j++)
        for (k = 0; k < NZl; k++)
          set_bc_x_HEAD (v, i, j, k);
  
  if (!multi_bc[X][TAIL])
    for (i = 0; i < GX; i++)
      for (j = 0; j < NYl; j++)
        for (k = 0; k < NZl; k++)
          set_bc_x_TAIL (v, i, j, k);
  
  #endif
  
  #if NY != 1
  
  if (!multi_bc[Y][HEAD])
    for (i = 0; i < NXl; i++)
      for (j = 0; j < GY; j++)
        for (k = 0; k < NZl; k++)
          set_bc_y_HEAD (v, i, j, k);
  
  if (!multi_bc[Y][TAIL])
    for (i = 0; i < NXl; i++)
      for (j = 0; j < GY; j++)
        for (k = 0; k < NZl; k++)
          set_bc_y_TAIL (v, i, j, k);
  
  #endif
  
  #if NZ != 1
  
  if (!multi_bc[Z][HEAD])
    for (i = 0; i < NXl; i++)
      for (j = 0; j < NYl; j++)
        for (k = 0; k < GZ; k++)
          set_bc_z_HEAD (v, i, j, k);
  
  if (!multi_bc[Z][TAIL])
    for (i = 0; i < NXl; i++)
      for (j = 0; j < NYl; j++)
        for (k = 0; k < GZ; k++)
          set_bc_z_TAIL (v, i, j, k);
  
  #endif
}

// definitions of 'set_bc_*()'
#include "bc.cpp"

#endif
