
/* EQUATION SPECIFIC FVM routines for FLUX
 * Karhunen-Loeve expansion of flux with exponential covariance kernel (Burgers' equation):
 * f(u) = u^2/2 + Karhunen-Loeve expansion with exponential covariance kernel
 */

///::title:: Karhunen-Loeve expansion (exponential covariance)
///::vars:: KL=3 KL_SCALING=0.2 KL_MAX=10

#include "flux.h"
#include "chaos.h"

#include "equation.h"

#include <cstdlib>

void flux_set_chaos_size () {
  CHAOS_SIZE [FLUX] [GLOBAL] [UNIFORM] = 0;
  CHAOS_SIZE [FLUX] [GLOBAL] [NORMAL]  = KL;
  CHAOS_SIZE [FLUX] [LOCAL]  [UNIFORM] = 0;
  CHAOS_SIZE [FLUX] [LOCAL]  [NORMAL]  = 0;
}

const real eta    = 3.0;
const real sigmaY = 1.0;
// computed for eta = 3.0, N = 10, sigmaY = 1, L=MAXX (domain length)
const real w [KL_MAX] = {0.547160757260330, 1.758164169633519, 3.243987484932942, 4.781982505508050, 6.335748362345733, 7.896171137138582, 9.459999472511342, 11.025797211953535, 12.592834514443268, 14.160701917675429};

// returns eigenvalue lambda_m
real calc_lambda (int m) {
  if (m >= KL_MAX) {
    std::cerr << " ERROR: KL < KL_MAX." << std::endl;
    std::exit(1);
  }
  return 2 * eta * sqr(sigmaY) / (sqr(eta) * sqr(w[m]) + 1);
}

// returns eigenfunction Phi_m
real calc_Phi (int m, real u) {
  return ( eta * w[m] * cos(w[m] * u) + sin(w[m] * u) ) / sqrt( (sqr(eta) * sqr(w[m]) + 1) * MAXX/2 + eta );
}

// returns derivative of the eigenfunction Phi_m
real calc_dPhi (int m, real u) {
  return ( - eta * w[m] * w[m] * sin(w[m] * u) + w[m] * cos(w[m] * u) ) / sqrt( (sqr(eta) * sqr(w[m]) + 1) * MAXX/2 + eta );
}

// computes Karhunen-Loeve expansion
real calc_KL_flux (real u) {
  real value = u*u/2.0;
  for (int m=1; m<=KL; m++)
    value += KL_SCALING * sqrt( calc_lambda(m) ) * calc_Phi(m,u) * chaos [FLUX] [GLOBAL] [NORMAL] [m-1];
  return value;
}

void coefficients_init () {}

real calc_max_speed (const ConservedVars &CV, const XtraVars &XV, const int &d)
{
  real value = CV.u;
  for (int m=1; m<=KL; m++)
    value += KL_SCALING * sqrt( calc_lambda(m) ) * calc_dPhi(m,CV.u) * chaos [FLUX] [GLOBAL] [NORMAL] [m-1];
  
  return fabs(value);
}

void calc_point_flux (ConservedVars &F, const AllVariables &V, const int &d)
{
  F.u = calc_KL_flux(V.U.u);
}

void calc_xtra (XtraVars &Xijk, const ConservedVars &Uijk, const int &i, const int &j, const int &k)
{
  Xijk.E = calc_KL_flux(Uijk.u);
}
