
#include "material-fft.h"

#include "material.h"
#include "chaos.h"
#include "mlmc.h"

#include "multi.h"
#include "multi_mpi.h"

#include <fftw3-mpi.h>
#include <iostream>
#include <cstdlib>

// physical sizes of the kernel
ptrdiff_t KN [NDIR];

// sizes of the 'half' of the kernel for real to complex FFT transform
ptrdiff_t HKN [NDIR];

// sizes of the kernel including padding
int VKN [NDIR];

// physical size and the offset of the local kernel in X direction
ptrdiff_t FFT_KNX_MPI, FFT_KNX_MPI_START;

// required memory size of the local kernel
ptrdiff_t alloc_local;

// iterators
ptrdiff_t i, j, k, l, m, n;

// local kernel of the circulant covariance matrix
double* kernel;

// local FFT of the kernel of the circulant covariance matrix
fftw_complex* kernel_FFT;

// local FFT of the Gaussian random vector
fftw_complex* chaos_FFT;

// local sampling buffer
double* buffer_real;

// local material coefficient send buffer
volume <real> send_buffer;

// local material coefficient recv buffer
volume <real> recv_buffer;

// local material coefficient buffer
volume <real> c_buffer;

// MPI intra-communicators and ranks
MPI_Comm   comm_domain_active;
MPI_Comm   comm_alltoall;
int        rank_alltoall;

// indicates if all or only part of all available cores is used for parallel FFT
int FULL_PARALLELIZATION;

// indicates if current core is active in FFTW or not
int ACTIVE;

// parallelization configuration flags

const int INHERIT_RANK_ORDER = 0;

// sizes of local material coefficient
int LOCAL_KN [NDIR];

// === Helpers

// returns location of the multi-index in the linear memory model for REAL arrays
inline int getLocReal (const int &i, const int &j, const int &k) {
  return (i * VKN[Y] + j) * VKN[Z] + k;
}

// returns location of the multi-index in the linear memory model for COMPLEX arrays
inline int getLocComplex (const int &i, const int &j, const int &k) {
  return (i * HKN[Y] + j) * HKN[Z] + k;
}

// returns location of the multi-index in the linear memory model for 'chaos' arrays
inline int getLocChaos (const int &i, const int &j, const int &k) {
  return (i * MATERIAL_N[Y][SAMPLES_LEVEL()] + j) * MATERIAL_N[Z][SAMPLES_LEVEL()] + k;
}

// returns distance vector from cell center at (i,j,k) to cell center at (0,0,0)
inline vector get_distance_vector (const int &i, const int &j, const int &k) {
  return vector (1.0 * i * MAXX / KN[X], 1.0 * j * MAXY / KN[Y], 1.0 * k * MAXZ / KN[Z]);
}

// performs appropriate multi-d real to complex FFT transform
void material_fft_r2c (double *in, fftw_complex *out);

// performs appropriate multi-d complex to real FFT transform
void material_fft_c2r (fftw_complex *in, double *out);

// === main routines

void material_fft_init () {
  
  fftw_mpi_init ();
  
  // FFTW-MPI is not very well supported for one-dimensional transforms
  if ( ACTIVE_DIRS == 1 && multi_main ) {
      std::cerr << " :: ERROR: Parallel FFTW is not very well supported for one-dimensional transforms." << std::endl;
      std::cerr << "  : -> Please use the serial FFTW in 'material_*-fft.cpp'." << std::endl;
      std::exit(1);
  }
  
  // set MATERIAL_N [*] [*]
  for ( int mesh_level = 0; mesh_level <= L; mesh_level++ ) {
    MATERIAL_N [X] [mesh_level] = min ( N[X][mesh_level], MATERIAL_NX );
    MATERIAL_N [Y] [mesh_level] = min ( N[Y][mesh_level], MATERIAL_NY );
    MATERIAL_N [Z] [mesh_level] = min ( N[Z][mesh_level], MATERIAL_NZ );
  }
}

void material_fft_level_init () {
  
  // === set sizes of various arrays to be used
  
  // physical sizes of the kernel
  KN [X] = MATERIAL_N [X] [MESH_LEVEL];
  KN [Y] = MATERIAL_N [Y] [MESH_LEVEL];
  KN [Z] = MATERIAL_N [Z] [MESH_LEVEL];
  
  // set LOCAL_KN [*]
  LOCAL_KN [X] = HAVE_DIR(X) ? max ( KN [X] / MULTIXl, 1 ) : 1;
  LOCAL_KN [Y] = HAVE_DIR(Y) ? max ( KN [Y] / MULTIYl, 1 ) : 1;
  LOCAL_KN [Z] = HAVE_DIR(Z) ? max ( KN [Z] / MULTIZl, 1 ) : 1;
  
  // check if KN[X] x KN[Y] x KN[Z] is equal to (might be smaller) MULTINl x (LOCAL_KN[X] x LOCAL_KN[Y] x LOCAL_KN[Z])
  if ( KN[X] * KN[Y] * KN[Z] != MULTINl * LOCAL_KN[X] * LOCAL_KN[Y] * LOCAL_KN[Z] ) {
    std::cerr << " :: ERROR: KN[X] * KN[Y] * KN[Z] != MULTINl * ( LOCAL_KN[X] * LOCAL_KN[Y] * LOCAL_KN[Z] ):" << std::endl;
    std::cerr << "  : -> " << KN[X] << "*" << KN[Y] << "*" << KN[Z] << " != " << MULTINl << "x(" << LOCAL_KN[X] << "*" << LOCAL_KN[Y] << "*" << LOCAL_KN[Z] << ")" << std::endl;
    std::cerr << "  : -> This can be fixed either by increasing MATERIAL_N* or by decreasing MULTI* (or MULTI_CORES)." << std::endl;
    std::exit(1);
  }
  
  // sizes of the 'half' of the kernel for real to complex FFT transform
  // size of the last dimension needs to be only half of the physical dimension
  HKN [X] = HAVE_DIR(Y) ? KN [X] : KN [X] / 2 + 1;
  HKN [Y] = HAVE_DIR(Z) ? KN [Y] : KN [Y] / 2 + 1;
  HKN [Z] = HAVE_DIR(Z) ? KN [Z] / 2 + 1 : KN [Z];
  
  // sizes of the kernel including padding
  VKN [X] = HAVE_DIR(Y) ? KN [X] : 2 * (KN [X] / 2 + 1);
  VKN [Y] = HAVE_DIR(Z) ? KN [Y] : 2 * (KN [Y] / 2 + 1);
  VKN [Z] = HAVE_DIR(Z) ? 2 * (KN [Z] / 2 + 1) : KN [Z];
  
  // === compute parallelization parameters and create required communicators
  
  // set the parallelization mode (full or partial)
  FULL_PARALLELIZATION = ( MULTINl <= KN [X] );
  
  // if FULL_PARALLELIZATION == 1, then the entire communicator is active
  if ( FULL_PARALLELIZATION ) {
    comm_domain_active = comm_domain;
    ACTIVE = 1;
  }
  
  // otherwise, split comm_domain and use only part of it
  else {
    int color = rank_domain % (MULTINl / KN [X]);
    MPI_Comm_split ( comm_domain, color, INHERIT_RANK_ORDER, &comm_domain_active );
    ACTIVE = ( color == 0 );
  }
  
  // set MULTI_FFT
  MULTI_FFT = FULL_PARALLELIZATION ? MULTINl : KN [X];
  
  // create MPI communicator for data redistribution in 'material_fft_draw_sample()'
  int color = rank_domain / (MULTIYl * MULTIZl);
  MPI_Comm_split ( comm_domain, color, INHERIT_RANK_ORDER, &comm_alltoall );
  MPI_Comm_rank (comm_alltoall, &rank_alltoall);
  
  // === get parallelization parameters from FFTW-MPI
  
  // only active cores participate in parallel FFT
  if ( ! ACTIVE ) return;
  
  // get the required size and the offset of the local kernel
  alloc_local = fftw_mpi_local_size (ACTIVE_DIRS, HKN, comm_domain_active, &FFT_KNX_MPI, &FFT_KNX_MPI_START);
  
  // allocate memory for local arrays
  kernel      = fftw_alloc_real    ( 2 * alloc_local );
  kernel_FFT  = fftw_alloc_complex ( alloc_local );
  buffer_real = fftw_alloc_real    ( 2 * alloc_local );
  chaos_FFT   = fftw_alloc_complex ( alloc_local );
  
  // check if successful
  if (kernel == NULL || kernel_FFT == NULL || buffer_real == NULL || chaos_FFT == NULL) {
      std::cerr << " :: ERROR: out of memory for FFT buffers." << std::endl;
      std::exit(1);
  }
  
  // === initialize local part of the global kernel and compute its FFT
  
  // initialize local kernel at the given resolution
  for (i = 0; i < FFT_KNX_MPI; i++)
    for (j = 0; j < KN[Y]; j++)
      for (k = 0; k < KN[Z]; k++)
        kernel [ getLocReal (i,j,k) ] = material_covariance_kernel ( get_distance_vector (FFT_KNX_MPI_START+i,j,k) );
  
  // compute FFT of kernel
  material_fft_r2c (kernel, kernel_FFT);
  
  // TODO: make this with DCT
  // make sure kernel_FFT is real and non-negative
  for (i = 0; i < FFT_KNX_MPI; i++)
    for (j = 0; j < HKN[Y]; j++)
      for (k = 0; k < HKN[Z]; k++) {
        // check if imaginary parts are zero
        if ( fabs ( kernel_FFT [ getLocComplex (i,j,k) ] [1] ) < 1e-6 ) {
          kernel_FFT [ getLocComplex (i,j,k) ] [1] = 0.0;
        } else {
          std::cerr << " :: ERROR: kernel_FFT has non-zero imaginary part " << kernel_FFT [ getLocComplex (i,j,k) ] [1] << " at index ";
          std::cerr << "(" << i << "," << j << "," << k << ")" << std::endl;
          std::cerr << "  : -> on rank_domain " << rank_domain;
          std::cerr << " at MESH_LEVEL " << MESH_LEVEL << ", TYPE " << TYPE << std::endl;
          std::exit(1);
        }
        // check if real parts are non-negative
        if ( kernel_FFT [ getLocComplex (i,j,k) ] [0] < 0.0 )
          if ( kernel_FFT [ getLocComplex (i,j,k) ] [0] > 1e-6 )
            kernel_FFT [ getLocComplex (i,j,k) ] [0] = 0.0;
          else {
            std::cerr << " :: ERROR: kernel_FFT has negative entry " << kernel_FFT [ getLocComplex (i,j,k) ] [0] << " at index ";
            std::cerr << "(" << i << "," << j << "," << k << ")" << std::endl;
            std::cerr << "  : -> on rank_domain " << rank_domain;
            std::cerr << " at MESH_LEVEL " << MESH_LEVEL << ", TYPE " << TYPE << std::endl;
            std::exit(1);
          }
      }
  
  // take a (complex) square root of the local kernel_FFT (Cholesky factor)
  for (i = 0; i < FFT_KNX_MPI; i++)
    for (j = 0; j < HKN[Y]; j++)
      for (k = 0; k < HKN[Z]; k++)
        kernel_FFT [ getLocComplex (i,j,k) ] [0] = sqrt ( kernel_FFT [ getLocComplex (i,j,k) ] [0] );
}

void material_fft_draw_sample () {
  
  // normalization constant
  real sqrt2 = sqrt (2.0);
  
  if ( ACTIVE ) {
    
    // === copy normal variates from 'chaos' vector to 'buffer_real'
    
    // if TYPE == 0 or the resolution of material coefficient is the same for both TYPEs: copy directly
    if ( TYPE == 0 || ( TYPE == 1 && MATERIAL_N[X][MESH_LEVEL] == MATERIAL_N[X][MESH_LEVEL+1] && MATERIAL_N[Y][MESH_LEVEL] == MATERIAL_N[Y][MESH_LEVEL+1] && MATERIAL_N[Z][MESH_LEVEL] == MATERIAL_N[Z][MESH_LEVEL+1] ) )
      for (i = 0; i < FFT_KNX_MPI; i++)
        for (j = 0; j < KN[Y]; j++)
          for (k = 0; k < KN[Z]; k++)
            #if ENSURE_PARALLEL_REPRODUCIBILITY
            buffer_real [ getLocReal (i,j,k) ] = chaos [FLUX] [GLOBAL] [NORMAL] [ getLocChaos (FFT_KNX_MPI_START+i,j,k) ];
            #else
            buffer_real [ getLocReal (i,j,k) ] = chaos [FLUX] [LOCAL] [NORMAL] [ getLocChaos (i,j,k) ];
            #endif
    
    // if TYPE == 1: upscale 'buffer_real' to the coarser MESH_LEVEL by averaging
    else
      for (i = 0; i < FFT_KNX_MPI; i++)
        for (j = 0; j < KN[Y]; j++)
          for (k = 0; k < KN[Z]; k++) {
            int loc = getLocReal (i,j,k);
            buffer_real [loc] = 0;
            for (l = 0; l < 1 + HAVE_DIR(X); l++)
              for (m = 0; m < 1 + HAVE_DIR(Y); m++)
                for (n = 0; n < 1 + HAVE_DIR(Z); n++)
                  #if ENSURE_PARALLEL_REPRODUCIBILITY
                  buffer_real [loc] += chaos [FLUX] [GLOBAL] [NORMAL] [ getLocChaos (2*(FFT_KNX_MPI_START+i)+l,2*j+m,2*k+n) ];
                  #else
                  buffer_real [loc] += chaos [FLUX] [LOCAL] [NORMAL] [ getLocChaos (2*i+l,2*j+m,2*k+n) ];
                  #endif
            buffer_real [loc] /= pow (sqrt2, ACTIVE_DIRS);
          }
    
    // compute FFT of normal random variates
    material_fft_r2c (buffer_real, chaos_FFT);
    
    // compute kernel_FFT * chaos_FFT
    // which gives the correlated local material coefficient in Fourier domain
    for (i = 0; i < FFT_KNX_MPI; i++)
      for (j = 0; j < HKN[Y]; j++)
        for (k = 0; k < HKN[Z]; k++) {
          chaos_FFT [ getLocComplex (i,j,k) ] [0] *= kernel_FFT [ getLocComplex (i,j,k) ] [0];
          chaos_FFT [ getLocComplex (i,j,k) ] [1] *= kernel_FFT [ getLocComplex (i,j,k) ] [0];
        }
    
    // compute inverse FFT
    material_fft_c2r (chaos_FFT, buffer_real);
    
    // construct 'send_buffer' of the shape LOCAL_KN[X] x LOCAL_KN[Y] x LOCAL_KN[Z]
    if ( FULL_PARALLELIZATION ) {
      
      // prepare 'send_buffer' (must have NO ghost cells!)
      send_buffer.resize (LOCAL_KN[X], LOCAL_KN[Y], LOCAL_KN[Z], 0);
      
      // copy 'buffer_real' to 'send_buffer'
      // reshape the array from FFT_KNX_MPI x KN[Y] x KN[Z] to LOCAL_KN[X] x LOCAL_KN[Y] x LOCAL_KN[Z]
      for (i = 0; i < FFT_KNX_MPI; i++)
        for (j = 0; j < KN[Y]; j++)
          for (k = 0; k < KN[Z]; k++)
            send_buffer [i + FFT_KNX_MPI * (MULTIZl * (j / LOCAL_KN[Y]) + k / LOCAL_KN[Z])] [j % LOCAL_KN[Y]] [k % LOCAL_KN[Z]] = buffer_real [ getLocReal (i,j,k) ];
    }
    
    // construct 'send_buffer' of the shape (MULTIYl * MULTIZl) x LOCAL_KN[Y] x LOCAL_KN[Z]
    else {
      
      // prepare 'send_buffer' (must have NO ghost cells!)
      send_buffer.resize ( MULTIYl * MULTIZl, LOCAL_KN[Y], LOCAL_KN[Z], 0);
      
      // copy 'buffer_real' to 'send_buffer'
      // reshape the array from 1 x KN[Y] x KN[Z] to (MULTIYl * MULTIZl) x LOCAL_KN[Y] x LOCAL_KN[Z]
      for (j = 0; j < KN[Y]; j++)
        for (k = 0; k < KN[Z]; k++)
          send_buffer [MULTIZl * (j / LOCAL_KN[Y]) + k / LOCAL_KN[Z]] [j % LOCAL_KN[Y]] [k % LOCAL_KN[Z]] = buffer_real [ getLocReal (0,j,k) ];
    }
    
  } // END if ( ACTIVE )
  
  // prepare 'recv_buffer' (must have NO ghost cells!)
  recv_buffer.resize (LOCAL_KN[X], LOCAL_KN[Y], LOCAL_KN[Z], 0);
  
  // for FULL_PARALLELIZATION == 1, results need to be globally exchanged (all to all)
  if ( FULL_PARALLELIZATION ) {
    
    // compute the size of the send/recv data blocks
    int count = FFT_KNX_MPI * LOCAL_KN[Y] * LOCAL_KN[Z];
    
    // redistribute 'send_buffer' to 'recv_buffer' according to the layout of the conventional DDM
    // REMARK: MPI_IN_PLACE is not supported for 'MPI_Alltoall()', at least in Open-MPI
    MPI_Alltoall (send_buffer.data, count, mpi_REAL, recv_buffer.data, count, mpi_REAL, comm_alltoall);
  }
  
  // for FULL_PARALLELIZATION == 0, results need to be scattered to the remaining cores
  else {
    
    // compute the size of the send/recv data blocks
    int count = LOCAL_KN[Y] * LOCAL_KN[Z];
    
    // scatter 'send_buffer' to 'recv_buffer' according to the layout of the conventional DDM
    for ( int layer = 0; layer < LOCAL_KN [X]; layer++ ) {
      int rank_scatterer = layer * MULTINl / KN [X];
      int recv_offset = layer * count;
      MPI_Scatter (send_buffer.data, count, mpi_REAL, recv_buffer.data + recv_offset, count, mpi_REAL, rank_scatterer, comm_alltoall);
    }
  }
  
  // complete the inverse FFT (normalization)
  recv_buffer /= KN[X] * KN[Y] * KN[Z];
  
  // compute exponentials of elements in c_buffer
  recv_buffer.exp();
  
  // prepare 'c_buffer'
  c_buffer.reinit (NXl, NYl, NZl, 0);
  
  // copy 'recv_buffer' to 'c_buffer'
  c_buffer += recv_buffer;
}

volume<real> & buffered_material () {
  return c_buffer;
}

void material_fft_level_exit () {
  
  // free memory
  if ( ACTIVE ) {
    if ( kernel      != NULL ) fftw_free (kernel);
    if ( kernel_FFT  != NULL ) fftw_free (kernel_FFT);
    if ( buffer_real != NULL ) fftw_free (buffer_real);
    if ( chaos_FFT   != NULL ) fftw_free (chaos_FFT);
  }
  
  // free communicators
  MPI_Comm_free (&comm_alltoall);
  if ( ! FULL_PARALLELIZATION )
    MPI_Comm_free (&comm_domain_active);
}

void material_fft_exit () {
  fftw_mpi_cleanup ();
}

void material_fft_r2c (double *in, fftw_complex *out) {
  
  fftw_plan p_r2c;
  
  p_r2c = fftw_mpi_plan_dft_r2c (ACTIVE_DIRS, KN, in, out, comm_domain_active, FFTW_ESTIMATE);
  
  fftw_execute(p_r2c);
  fftw_destroy_plan(p_r2c);
}

void material_fft_c2r (fftw_complex *in, double *out) {
  
  fftw_plan p_c2r;
  
  p_c2r = fftw_mpi_plan_dft_c2r (ACTIVE_DIRS, KN, in, out, comm_domain_active, FFTW_ESTIMATE);
  
  fftw_execute(p_c2r);
  fftw_destroy_plan(p_c2r);
}
