
#include "material-fft.h"

#include "material.h"
#include "chaos.h"
#include "mlmc.h"

#include "multi.h"

#include <fftw3.h>
#include <iostream>
#include <cstdlib>

// range of layers that are needed for a current subdomain
int LAYER_BEGIN;
int LAYER_END;

// physical sizes of the kernel
int KN [NDIR];

// sizes of the 'half' of the kernel for real to real FFT transforms
int RKN [NDIR];

// sizes of the 'half' of the kernel for real to complex FFT transform
int HKN [NDIR];

// iterators
int i, j, k, l, m, n;

// kernel of the circulant covariance matrix
double* kernel;

// FFT of the kernel of the circulant covariance matrix
double* kernel_FFT [MATERIAL_LAYERS];

// FFT of the Gaussian random vector
fftw_complex* chaos_FFT;

// sampling buffer
double* buffer_real;

// material coefficient buffer
volume <real> c_buffer;

// === Helpers

// returns location of the multi-index in the linear memory model for REAL (half) arrays
inline int getLocRealHalf (const int &i, const int &j, const int &k) {
  return (i * RKN[Y] + j) * RKN[Z] + k;
}

// returns location of the multi-index in the linear memory model for REAL arrays
inline int getLocReal (const int &i, const int &j, const int &k) {
  return (i * KN[Y] + j) * KN[Z] + k;
}

// returns location of the multi-index in the linear memory model for COMPLEX arrays
inline int getLocComplex (const int &i, const int &j, const int &k) {
  return (i * HKN[Y] + j) * HKN[Z] + k;
}

// returns location of the multi-index in the linear memory model for 'chaos' arrays
inline int getLocChaos (const int &i, const int &j, const int &k) {
  return (i * MATERIAL_N[Y][SAMPLES_LEVEL()] + j) * MATERIAL_N[Z][SAMPLES_LEVEL()] + k;
}

// returns distance vector from cell center at (i,j,k) to cell center at (0,0,0)
inline vector get_distance_vector (const int &i, const int &j, const int &k) {
  return vector (1.0 * i * MAXX / KN[X], 1.0 * j * MAXY / KN[Y], 1.0 * k * MAXZ / KN[Z]);
}

// performs appropriate multi-d real to real FFT transform
void material_fft_r2r (double *in, double *out);

// performs appropriate multi-d real to complex FFT transform
void material_fft_r2c (double *in, fftw_complex *out);

// performs appropriate multi-d complex to real FFT transform
void material_fft_c2r (fftw_complex *in, double *out);

// === main routines

void material_fft_init () {
  
  // set MATERIAL_N [*] [*]
  for ( int mesh_level = 0; mesh_level <= L; mesh_level++ ) {
    MATERIAL_N [X] [mesh_level] = min ( N[X][mesh_level], MATERIAL_NX );
    MATERIAL_N [Y] [mesh_level] = min ( N[Y][mesh_level], MATERIAL_NY );
    MATERIAL_N [Z] [mesh_level] = min ( N[Z][mesh_level], MATERIAL_NZ );
  }
  
  // issue a warning if used in parallel runs
  if ( multi_main && MULTI_TOTAL > 1 ) {
    std::cout << std::endl;
    std::cout << " :: WARNING: this is SERIAL version of FFT only." << std::endl;
    std::cout << "  : -> Consider using parallel 'material-fft-mpi.cpp' based on MPI FFTW." << std::endl;
  }
}

void material_fft_level_init () {
  
  // set LAYERS_PER_CORE, LAYER_BEGIN and LAYER_END
  if ( MATERIAL_LAYERS >= MULTIYl ) {
    LAYER_BEGIN =  multiy      * ( MATERIAL_LAYERS / MULTIYl );
    LAYER_END   = (multiy + 1) * ( MATERIAL_LAYERS / MULTIYl );
  }
  else {
    LAYER_BEGIN = MATERIAL_LAYERS * multiy / MULTIYl;
    LAYER_END   = LAYER_BEGIN + 1;
  }
  
  // === set sizes of various arrays to be used
  
  // physical sizes of the kernel
  KN [X] = MATERIAL_N [X] [MESH_LEVEL];
  KN [Y] = MATERIAL_N [Y] [MESH_LEVEL];
  KN [Z] = MATERIAL_N [Z] [MESH_LEVEL];
  
  // sizes of the 'half' of the kernel for real to real FFT transforms
  RKN [X] = KN [X] / 2 + 1;
  RKN [Y] = KN [Y] / 2 + 1;
  RKN [Z] = KN [Z] / 2 + 1;
  
  // sizes of the 'half' of the kernel for real to complex FFT transform
  // size of the last dimension needs to be only half of the physical dimension
  HKN [X] = HAVE_DIR(Y) ? KN [X] : KN [X] / 2 + 1;
  HKN [Y] = HAVE_DIR(Z) ? KN [Y] : KN [Y] / 2 + 1;
  HKN [Z] = HAVE_DIR(Z) ? KN [Z] / 2 + 1 : KN [Z];
  
  // allocate memory
  kernel      = fftw_alloc_real    ( RKN [X] * RKN [Y] * RKN [Z] );
  buffer_real = fftw_alloc_real    (  KN [X] *  KN [Y] *  KN [Z] );
  chaos_FFT   = fftw_alloc_complex ( HKN [X] * HKN [Y] * HKN [Z] );
  
  // check if successful
  if (kernel == NULL || buffer_real == NULL || chaos_FFT == NULL) {
      std::cerr << " :: ERROR: out of memory for FFT buffers." << std::endl;
      std::exit(1);
  }
  
  // === initialize kernel and compute its FFT
  
  // for each layer
  for (int layer = LAYER_BEGIN; layer < LAYER_END; layer++) {
    
    // allocate layer-dependent memory
    kernel_FFT [layer] = fftw_alloc_real    ( RKN [X] * RKN [Y] * RKN [Z] );
    
    // check if successful
    if (kernel_FFT == NULL) {
        std::cerr << " :: ERROR: out of memory for FFT buffers." << std::endl;
        std::exit(1);
    }
    
    // initialize local kernel at the given resolution and for a specified layer
    for (i = 0; i < RKN[X]; i++)
      for (j = 0; j < RKN[Y]; j++)
        for (k = 0; k < RKN[Z]; k++)
          kernel [ getLocRealHalf (i,j,k) ] = material_covariance_kernel ( get_distance_vector (i,j,k), layer );
    
    // compute FFT of kernel (of the specified layer)
    material_fft_r2r ( kernel, kernel_FFT [layer] );
    
    // make sure kernel_FFT is non-negative
    for (i = 0; i < RKN[X]; i++)
      for (j = 0; j < RKN[Y]; j++)
        for (k = 0; k < RKN[Z]; k++)
          if ( kernel_FFT [layer] [ getLocRealHalf (i,j,k) ] < 0.0 )
            if ( kernel_FFT [layer] [ getLocRealHalf (i,j,k) ] > -1e-6 )
              kernel_FFT [layer] [ getLocRealHalf (i,j,k) ] = 0;
            else {
              std::cerr << " :: ERROR: kernel_FFT has negative entry " << kernel_FFT [layer] [ getLocRealHalf (i,j,k) ];
              std::cerr << " for layer " << layer << " at index ";
              std::cerr << "(" << i << "," << j << "," << k << ")" << std::endl;
              std::cerr << " at MESH_LEVEL " << MESH_LEVEL << ", TYPE " << TYPE << std::endl;
              std::exit(1);
            }
    
    // take a (complex) square root of the local kernel_FFT (Cholesky factor)
    for (i = 0; i < RKN[X]; i++)
      for (j = 0; j < RKN[Y]; j++)
        for (k = 0; k < RKN[Z]; k++)
          kernel_FFT [layer] [ getLocRealHalf (i,j,k) ] = sqrt ( kernel_FFT [layer] [ getLocRealHalf (i,j,k) ] );
  
  } // END FOR (LAYER...
}

void material_fft_draw_sample () {
  
  // normalization constant
  real sqrt2 = sqrt (2.0);
  
  // for each layer
  for (int layer = LAYER_BEGIN; layer < LAYER_END; layer++) {
    
    // copy normal variates from 'chaos' vector to 'buffer_real'
    for (i = 0; i < KN[X]; i++)
      for (j = 0; j < KN[Y]; j++)
        for (k = 0; k < KN[Z]; k++)
          
          // if TYPE == 0 or the resolution of material coefficient is the same for both TYPEs: copy directly
          if ( TYPE == 0 || ( TYPE == 1 && MATERIAL_N[X][MESH_LEVEL] == MATERIAL_N[X][MESH_LEVEL+1] && MATERIAL_N[Y][MESH_LEVEL] == MATERIAL_N[Y][MESH_LEVEL+1] && MATERIAL_N[Z][MESH_LEVEL] == MATERIAL_N[Z][MESH_LEVEL+1] ) )
            buffer_real [ getLocReal (i,j,k) ] = chaos [FLUX] [GLOBAL] [NORMAL] [ getLocChaos (i,j,k) + layer * CHAOS_SIZE [FLUX] [GLOBAL] [NORMAL] / MATERIAL_LAYERS ];
          
          // if TYPE == 1: upscale 'buffer_real' to the coarser MESH_LEVEL by averaging
          else {
            int loc = getLocReal (i,j,k);
            buffer_real [loc] = 0;
            for (l = 0; l < 1 + HAVE_DIR(X); l++)
              for (m = 0; m < 1 + HAVE_DIR(Y); m++)
                for (n = 0; n < 1 + HAVE_DIR(Z); n++)
                  buffer_real [loc] += chaos [FLUX] [GLOBAL] [NORMAL] [ getLocChaos (2*i+l,2*j+m,2*k+n) + layer * CHAOS_SIZE [FLUX] [GLOBAL] [NORMAL] / MATERIAL_LAYERS ];
            buffer_real [loc] /= pow (sqrt2, ACTIVE_DIRS);
          }
    
    // compute FFT of normal random variates
    material_fft_r2c (buffer_real, chaos_FFT);
    
    // compute kernel_FFT * chaos_FFT
    // which gives the correlated material coefficient in Fourier domain
    for (i = 0; i < HKN[X]; i++)
      for (j = 0; j < HKN[Y]; j++)
        for (k = 0; k < HKN[Z]; k++) {
          chaos_FFT [ getLocComplex (i,j,k) ] [0] *= kernel_FFT [layer] [ getLocRealHalf ( min(i,KN[X]-i), min(j,KN[Y]-j), min(k,KN[Z]-k) ) ];
          chaos_FFT [ getLocComplex (i,j,k) ] [1] *= kernel_FFT [layer] [ getLocRealHalf ( min(i,KN[X]-i), min(j,KN[Y]-j), min(k,KN[Z]-k) ) ];
        }
    
    // compute inverse FFT
    material_fft_c2r (chaos_FFT, buffer_real);
    
    // prepare 'c_buffer'
    c_buffer.resize (NXl, NYl, NZl, 0);
    
    // determine the required range of cells in the y-direction
    int NY_BEGIN = MATERIAL_LAYERS <= MULTIYl ? 0   : ( layer - LAYER_BEGIN     ) * NYl / ( LAYER_END - LAYER_BEGIN );
    int NY_END   = MATERIAL_LAYERS <= MULTIYl ? NYl : ( layer - LAYER_BEGIN + 1 ) * NYl / ( LAYER_END - LAYER_BEGIN );
    
    // copy required sub-domain of 'buffer_real' to 'c_buffer'
    for (i = 0; i < NXl; i++)
      for (j = NY_BEGIN; j < NY_END; j++)
        for (k = 0; k < NZl; k++)
          c_buffer [i][j][k] = buffer_real [ getLocReal ( (i + multix*NXl) / (N[X][MESH_LEVEL] / KN[X]), (j + multiy*NYl) / (N[Y][MESH_LEVEL] / KN[Y]), (k + multiz*NZl) / (N[Z][MESH_LEVEL] / KN[Z]) ) ];
  
  } // END FOR (LAYER...
  
  // complete the inverse FFT (normalization)
  c_buffer /= KN[X] * KN[Y] * KN[Z];
  
  // compute exponentials of elements in c_buffer
  c_buffer.exp();
}

volume<real> & buffered_material () {
  return c_buffer;
}

void material_fft_level_exit () {
  if ( kernel      != NULL ) fftw_free (kernel);
  if ( buffer_real != NULL ) fftw_free (buffer_real);
  if ( chaos_FFT   != NULL ) fftw_free (chaos_FFT);
  for (int layer = LAYER_BEGIN; layer < LAYER_END; layer++)
    if ( kernel_FFT [layer] != NULL ) fftw_free ( kernel_FFT [layer] );
}

void material_fft_exit () {}

void material_fft_r2r (double *in, double *out) {
  
  fftw_plan p_r2r;
  
  fftw_r2r_kind fftw_kinds [NDIR] = {FFTW_REDFT00, FFTW_REDFT00, FFTW_REDFT00};
  
  p_r2r = fftw_plan_r2r (ACTIVE_DIRS, RKN, in, out, fftw_kinds, FFTW_ESTIMATE);
  
  fftw_execute(p_r2r);
  fftw_destroy_plan(p_r2r);
}

void material_fft_r2c (double *in, fftw_complex *out) {
  
  fftw_plan p_r2c;
  
  p_r2c = fftw_plan_dft_r2c (ACTIVE_DIRS, KN, in, out, FFTW_ESTIMATE);
  
  fftw_execute(p_r2c);
  fftw_destroy_plan(p_r2c);
}

void material_fft_c2r (fftw_complex *in, double *out) {
  
  fftw_plan p_c2r;
  
  p_c2r = fftw_plan_dft_c2r (ACTIVE_DIRS, KN, in, out, FFTW_ESTIMATE);
  
  fftw_execute(p_c2r);
  fftw_destroy_plan(p_c2r);
}
