
#include "mlmc.h"

#include "rng.h"
#include "chaos.h"
#include "multi.h"
#include "balancer.h"

#if STATS
  #include "stats.h"
#else
  #include "saver.h"
#endif

#if USE_PRIME_SEEDS
  #include "primes.h"
#endif

#include <cstdlib>
#include <iostream>
#include <fstream>

// === MLMC

int  N  [NDIR] [L+1];
real D  [NDIR] [L+1];

long_int NM [L+1], NMT, NMT_TYPE0;
long_int RUN_NUMBER_FROM_FILE;

int MESH_LEVEL, TYPE, SNAPSHOT;

long_int SAMPLE_INDEX;
int SUBDOMAIN_INDEX;

int SAMPLES_LEVEL() {
  return MESH_LEVEL + TYPE;
}

int INCLUDE_LEVEL_SUFFIX;

// === MLMC routines

void compatibility_checks () {
  
  // compatibility checks for multi:* alternative
  multi_compatibility_checks();
  
  // compatibility checks for balancer:* alternative
  balancer_compatibility_checks();
  
  // check if not too many levels were asked for the specified finest resolution
  if ( ( HAVE_DIR(X) && NX % (1 << L) != 0 ) || ( HAVE_DIR(Y) && NY % (1 << L) != 0 ) || ( HAVE_DIR(Z) && NZ % (1 << L) != 0 ) )
    if (multi_main) {
      std::cerr << " :: ERROR: Too many levels OR incompatible finest mesh resolution in mesh hierarchy." << std::endl;
      std::exit(1);
  }
  
  // check for condition 2s <= d+1
  if ( L != 0 && EFFECTIVE_DOUBLED_ORDER > ACTIVE_DIRS + 1 + HAVE_VISCOSITY ) {
    if (multi_main) {
      std::cerr << " :: ERROR: 2 * (FVM convergence order) <= d + 1 + HAVE_VISCOSITY is violated: " << EFFECTIVE_DOUBLED_ORDER << " > " << ACTIVE_DIRS + 1 << "." << std::endl;
      std::cerr << "  : -> Solution is expected to be ";
      if (!SMOOTH_SOLUTION) std::cout << "non-";
      std::cerr << "smooth, i.e. SMOOTH_SOLUTION=" << SMOOTH_SOLUTION << std::endl;
      std::cerr << "  : -> Same MLMC-FVM performance could be achieved with less accurate FVM scheme." << std::endl << std::endl;
      std::exit(1);
    }
  }
  
  // check for compatibility of DISTRIBUTE_IN_DIFFERENCES and COMPUTE_IN_DIFFERENCES
  if ( DISTRIBUTE_IN_DIFFERENCES == 0 && COMPUTE_IN_DIFFERENCES == 1 )
    if (multi_main) {
      std::cerr << " :: ERROR: DISTRIBUTE_IN_DIFFERENCES == 0 and COMPUTE_IN_DIFFERENCES == 1 are NOT compatible." << std::endl;
      std::exit(1);
    }
  
  // ENSURE_PARALLEL_REPRODUCIBILITY == 0 requires DISTRIBUTE_IN_DIFFERENCES == 1
  if ( ENSURE_PARALLEL_REPRODUCIBILITY == 0 && DISTRIBUTE_IN_DIFFERENCES == 0 ) {
    if (multi_main) {
      std::cerr << " :: ERROR: ENSURE_PARALLEL_REPRODUCIBILITY == 0 requires DISTRIBUTE_IN_DIFFERENCES == 1." << std::endl;
      std::exit(1);
    }
  }
  
  // ENSURE_PARALLEL_REPRODUCIBILITY == 0 performs better if RNG_HAS_STREAMS == 1 (in rng_*.cpp)
  if ( ENSURE_PARALLEL_REPRODUCIBILITY == 0 && RNG_HAS_STREAMS == 0 ) {
    if (multi_main) std::cout << " :: WARNING: ENSURE_PARALLEL_REPRODUCIBILITY == 0 performs better if RNG_HAS_STREAMS == 1." << std::endl;
  }
}

void mlmc_init () {
  
  // === init mesh resolutions N[*][*]
  
  // the resolution of the finest level is set by NX, NY, NZ
  N[X][L] = NX;
  N[Y][L] = NY;
  N[Z][L] = NZ;
  
  // remaining levels are coarsened until it is no longer possible to do so
  for (int mesh_level=L-1; mesh_level>=0; mesh_level--)
    for (int d=0; d<NDIR; d++)
      N[d][mesh_level] = HAVE_DIR(d) ? N[d][mesh_level+1] / 2 : N[d][mesh_level+1];
  
  // === init mesh widths D[*][*]
  
  for (int mesh_level=L; mesh_level>=0; mesh_level--)
    for (int d=0; d<NDIR; d++)
      D[d][mesh_level] = 1.0 * MAX[d] / N[d][mesh_level];
  
  // === init number of samples NM[*]
  
  #if READ_NUMBER_OF_SAMPLES_FROM_FILE
  
  // === read the number of samples on every level from file SAMPLES_FILE
  
  std::ifstream is (SAMPLE_NUMBERS_FILE);
  
  if (!is) {
    if (multi_main) std::cerr << " :: ERROR: In the run folder there is no " << READ_NUMBER_OF_SAMPLES_FROM_FILE << " file!" << std::endl;
    if (multi_main) std::exit(1);
  }
  
  if (STATS && multi_main) std::cout << " :: Reading number of samples from file " << name.str().c_str() << std::endl << std::endl;
  
  for (int samples_level=L; samples_level>=0; samples_level--) {
    long_int al;
    is >> al;
    if (al < 1) {
      if (multi_main) std::cerr << " :: ERROR: Number of samples should be positive integer! Encountered: " << al << std::endl;
      if (multi_main) std::exit(1);
    }
    NM[samples_level] = al;
  }
  
  #elif USE_OPTIMAL_NUMBER_OF_SAMPLES
  
  // === optimal distribution of error
  
  if (STATS && multi_main) std::cout << " :: Using optimal number of samples" << std::endl << std::endl;
  
  // the number of samples at finest sample level is set by ML
  NM[L] = ML;
  
  // the number of samples at lower sample levels are set using OPTIMAL_SAMPLES_FACTOR
  for (int samples_level=L-1; samples_level>=0; samples_level--)
    NM[samples_level] = (long_int) round ( NM[samples_level+1] * OPTIMAL_SAMPLES_FACTOR );
  
  #elif VISCOUS
  
  // === equidistribution of error, taking into account viscosity
  
  // number of samples according to Ch. Schwab and A. Barth
  
  if (STATS && multi_main) std::cout << " :: Using equidistributed number of samples, adjusted for viscosity" << std::endl << std::endl;
  
  // fraction of total error that is comming from sampling-error
  const real ETA = 0.0;
  
  // the number of samples at the coarsest sample level
  NM[0] = (long_int) round ( ML * sqr ( min ( D[X][0], (real) VISCOSITY ) / D[X][L] ) );
  
  // the number of samples at higher sample levels
  for (int samples_level=1; samples_level<=L; samples_level++)
    NM[samples_level] = (long_int) round ( ML * sqr ( min ( D[X][samples_level], (real) VISCOSITY ) / D[X][L] ) * pow ( samples_level, 2.0 + 2*ETA ) );
  
  // for the statistical approximation of centered statistical k-th moments,
  // the number of samples should not be smaller than k
  // here we assume that the highest statistical moment is variance
  for (int samples_level=0; samples_level<=L; samples_level++)
    if ( NM[samples_level] < 2 ) NM[samples_level] = 2;
  
  #elif USE_ONE_SAMPLE_PER_LEVEL
  
  if (STATS && multi_main) std::cout << " :: Using one sample per level" << std::endl << std::endl;
  
  // the number of samples all levels are set using SAMPLES_FACTOR
  for (int samples_level=L; samples_level>=0; samples_level--)
    NM[samples_level] = 1;
  
  #else
  
  // === equidistribution of error
  
  if (STATS && multi_main) std::cout << " :: Using equidistributed number of samples" << std::endl << std::endl;
  
  // the number of samples at finest sample level is set by ML
  NM[L] = ML;
  
  // the number of samples at lower sample levels are set using SAMPLES_FACTOR
  for (int samples_level=L-1; samples_level>=0; samples_level--)
    NM[samples_level] = NM[samples_level+1] * SAMPLES_FACTOR;
  
  #endif
  
  // check whether there is no arithmetic overflow
  for (int samples_level=L; samples_level>=0; samples_level--)
    if ( ! (NM[samples_level] <= NM_MAX) ) {
      if (multi_main) std::cerr << " :: ERROR: NM[" << samples_level << "] = " << NM[samples_level] << " exceeded maximum value of " << NM_MAX << "." << std::endl;
      if (multi_main) std::cerr << "  : -> Change typedef of 'long_int' in 'numtypes.h'." << std::endl;
      if (multi_main) std::exit(1);
    }
  
  // check whether there are no levels without samples
  for (int samples_level=L; samples_level>=0; samples_level--)
    if ( NM[samples_level] == 0 ) {
      if (multi_main) std::cerr << " :: ERROR: Encountered level without samples: NM[" << samples_level << "] = " << 0 << "." << std::endl;
      if (multi_main) std::exit(1);
    }
  
  // compute total number of samples (including both types)
  NMT = NM[0];
  for (int samples_level = 1; samples_level <= L; samples_level++)
    NMT += 2*NM[samples_level];
  
  // compute total number of samples (only TYPE == 0)
  NMT_TYPE0 = NM[0];
  for (int samples_level = 1; samples_level <= L; samples_level++)
    NMT_TYPE0 += NM[samples_level];
  
  // output computed number of samples in SAMPLES_FILE
  std::ofstream samples (SAMPLES_FILE);
  samples << "samples = [";
  for (int samples_level = 0; samples_level <= L; samples_level++)
    samples << NM[samples_level] << ", ";
  samples << "]" << std::endl;
  
  // === if READ_RUN_NUMBER_FROM_FILE, then set the RUN_NUMBER_FROM_FILE
  
  #if READ_RUN_NUMBER_FROM_FILE
  
  std::ifstream is (RUN_NUMBER_FILE);
  
  if ( !is ) {
    if (multi_main) std::cerr << " :: ERROR: In the run folder there is no " << RUN_NUMBER_FILE << " file!" << std::endl;
    if (multi_main) std::exit(1);
  }
  
  is >> RUN_NUMBER_FROM_FILE;
  
  #endif
}

void display_info () {
  
  #ifndef QUIET
  
  if (multi_main)
    
    // FVM simulation
    if (L == 0 && ML == 1) {
      std::cout << " :: FVM simulation (deterministic)" << std::endl;
      std::cout << "  : -> FVM convergence rate : " << EFFECTIVE_DOUBLED_ORDER << "/" << 2 << std::endl;
      std::cout << "  : -> Number of cells      : " << NX << "x" << NY << "x" << NZ << std::endl;
    }
    
    // MC-FVM simulation
    else if (L == 0 && ML != 1) {
        std::cout << " :: MC-FVM simulation (stochastic)" << std::endl;
        std::cout << "  : -> FVM convergence rate : " << EFFECTIVE_DOUBLED_ORDER << "/" << 2 << std::endl;
        std::cout << "  : -> Number of cells      : " << NX << "x" << NY << "x" << NZ << std::endl;
        std::cout << "  : -> Number of samples    : " << NM[0] << std::endl;
    }
    
    // MLMC-FVM simulation
    else {
      std::cout << " :: MLMC-FVM simulation (stochastic) with " << L+1 << " level(s)" << std::endl;
      std::cout << "  : -> FVM convergence rate    : " << EFFECTIVE_DOUBLED_ORDER << "/" << 2 << std::endl;
      std::cout << "  : -> Finest mesh             : " << N[X][L] << "x" << N[Y][L] << "x" << N[Z][L] << std::endl;
      std::cout << "  : -> Coarsest mesh           : " << (1 << L0X) << "x" << (1 << L0Y) << "x" << (1 << L0Z) << std::endl;
      std::cout << "  : -> Number of samples       :";
      for (int samples_level = L; samples_level >= 0; samples_level--)
        std::cout << " " << NM[samples_level];
      std::cout << std::endl;
      std::cout << "  : -> Total number of samples : " << NMT << std::endl;
    }
  
  #endif
}

void fvm_volumes_level_init () {
  U      .reinit (N[X][MESH_LEVEL] / MULTIXl, N[Y][MESH_LEVEL] / MULTIYl, N[Z][MESH_LEVEL] / MULTIZl);
  Xtra   .reinit (N[X][MESH_LEVEL] / MULTIXl, N[Y][MESH_LEVEL] / MULTIYl, N[Z][MESH_LEVEL] / MULTIZl);
  Source .reinit (N[X][MESH_LEVEL] / MULTIXl, N[Y][MESH_LEVEL] / MULTIYl, N[Z][MESH_LEVEL] / MULTIZl);
  for (int d=0; d<NDIR; d++)
    Flux[d].reinit (FLUX_NX(d), FLUX_NY(d), FLUX_NZ(d), 0);
}

void mesh_level_init () {
  
  // set local mesh sizes for FVM
  NXl = N[X][MESH_LEVEL] / MULTIXl;
  NYl = N[Y][MESH_LEVEL] / MULTIYl;
  NZl = N[Z][MESH_LEVEL] / MULTIZl;
  
  // set global mesh sizes for FVM
  NXl_FULL = N[X][MESH_LEVEL];
  NYl_FULL = N[Y][MESH_LEVEL];
  NZl_FULL = N[Z][MESH_LEVEL];
  
  // set mesh widths for FVM
  dx[X] = D[X][MESH_LEVEL];
  dx[Y] = D[Y][MESH_LEVEL];
  dx[Z] = D[Z][MESH_LEVEL];
  
  // reinit FVM volumes
  fvm_volumes_level_init();
  
  LEVEL_INIT_HOOK
}

void mesh_level_exit () {
  LEVEL_EXIT_HOOK
}

void display_level_info () {
  #ifndef QUIET
  char pm [2] = {'+', '-'};
  if ( L == LOWEST_MULTI_LEVEL && multi_main && L != 0 ) {
    std::cout << "  : LEVEL " << MESH_LEVEL;
    #if ! COMPUTE_IN_DIFFERENCES
    std::cout << pm[TYPE];
    #endif
    std::cout << ": ";
    std::cout << "resolution: " << N[X][MESH_LEVEL] << "x" << N[Y][MESH_LEVEL] << "x" << N[Z][MESH_LEVEL];
    std::cout << ", samples: " << SAMPLES[MULTI_LEVEL][MULTI_SAMPLER][MESH_LEVEL][TYPE].size();
    std::cout << ", cores: (" << MULTIXl << "x" << MULTIYl << "x" << MULTIZl << ")x" << MULTIMl;
    std::cout << "=" << MULTIXl*MULTIYl*MULTIZl*MULTIMl << "...";
    std::cout.flush();
  }
  #endif
}

void display_level_done () {
  #ifndef QUIET
  if ( multi_main && L == LOWEST_MULTI_LEVEL )
    std::cout << " done." << std::endl;
  #endif
}

int getSeed (const int sample_index) {
  
  // compute the actual RUN_NUMBER
  int ACTUAL_RUN_NUMBER = (READ_RUN_NUMBER_FROM_FILE ? RUN_NUMBER_FROM_FILE : RUN_NUMBER) - 1;
  
  #if ENSURE_PARALLEL_REPRODUCIBILITY
  
  // === seed is UNIQUE for each tuple of SAMPLES_LEVEL() and ACTUAL_RUN_NUMBER
  // seed is IDENTICAL in all samplers (and all their sub-domains) working on the same SAMPLES_LEVEL()
  int seed = pair ( SAMPLES_LEVEL(), ACTUAL_RUN_NUMBER );
  
  #else
  
  #if ! RNG_HAS_STREAMS
  
  // === seed is UNIQUE for each tuple of SAMPLES_LEVEL(), sample_index, MULTI_SUBDOMAIN and ACTUAL_RUN_NUMBER
  // seed is NOWHERE IDENTICAL
  int seed = pair ( SAMPLES_LEVEL(), sample_index - 1, MULTI_SUBDOMAIN, ACTUAL_RUN_NUMBER );
  
  #else
  
  // === seed is UNIQUE for each tuple of MULTI_SUBDOMAIN and ACTUAL_RUN_NUMBER
  // seed is NOWHERE IDENTICAL
  int seed = pair ( MULTI_SUBDOMAIN, ACTUAL_RUN_NUMBER );
  
  #endif
  
  #endif
  
  #if USE_PRIME_SEEDS
  if (seed + PRIME_SHIFT >= NPRIMES) {
    std::cerr << " :: ERROR: Prime number array 'primes' in primes.h is too small!" << std::endl;
    std::cerr << "  : -> Decrease PRIME_SHIFT or L or RUN_NUMBER or increase the 'primes' array." << std::endl;
    std::cerr << "  : -> Alternatively, set USE_PRIME_SEEDS=0." << std::endl;
    std::exit(1);
  }
  
  return primes [seed + PRIME_SHIFT];
  
  #else
  
  return seed;
  
  #endif
}

int get_stream (const int sample_index) {
  
  #if ENSURE_PARALLEL_REPRODUCIBILITY
  
  // === stream is IDENTICAL everywhere
  return 0;
  
  #else
  
  // === stream is UNIQUE for each tuple of SAMPLES_LEVEL() and sample_index
  // stream is IDENTICAL in subdomains working on the same SAMPLES_LEVEL() and sample_index
  return pair ( SAMPLES_LEVEL(), sample_index - 1 );
  
  #endif
}

void rng_and_chaos_init (const int sample_index) {
  
  // initialize RNG
  rng_init ( getSeed (sample_index), get_stream (sample_index) );
  
  // initialize chaos
  chaos_init ( SAMPLES_OFFSET );
}

void mlmc_setup_chaos () {
  
  #if ENSURE_PARALLEL_REPRODUCIBILITY
  
  // === SAME seed/stream will be used for all samples
  
  // set the required amounts of random numbers for each stream and distribution
  set_chaos_size ();
  
  // initialize RNG and chaos
  rng_and_chaos_init ();
  
  #else
  
  // === DIFFERENT seed/stream will be used for each sample
  
  // set the required amounts of random numbers for each stream and distribution
  set_chaos_size ();
  
  #endif
}

void mlmc_get_chaos (const int sample_index) {
  
  #if ENSURE_PARALLEL_REPRODUCIBILITY
  
  // draw a sample
  chaos_draw (sample_index);
  
  #else
  
  // init RNG and chaos
  rng_and_chaos_init (sample_index);
  
  // draw a sample
  chaos_draw ();
  
  #endif
}

void run_sample () {
  
  #if ! STATS && WRITE_DATAFILES
  
  // initialize saver (snapshots will be saved on-the-fly)
  Saver saver ("");
  
  #endif
  
  // initialize initial data, sources, fluxes and bc
  init_fvm ();
  
  // run the simulation while saving the snapshots
  for (SNAPSHOT = 0; SNAPSHOT < NSAVES + SAVE0; SNAPSHOT++) {
    
    // define when simulation should halt
    if (SAVE0)
      t_halt = 1.0 * MAXT / NSAVES * (SNAPSHOT);
    else
      t_halt = 1.0 * MAXT / NSAVES * (SNAPSHOT + 1);
    
    // run simulation until it halts
    run_fvm ();
    
    // record the halt time
    halt_times[SNAPSHOT] = t;
    
    #if STATS
    
    // update statistics
    stats_update ();
    
    #elif WRITE_DATAFILES
    
    // save deterministic snapshot
    saver.snapshot (U, Xtra);
    
    #endif
  }
  
  #if ! STATS && WRITE_DATAFILES
  
  // finalize saving of deterministic snapshots
  saver.finalize();
  
  #endif
}

#if ! COMPUTE_IN_DIFFERENCES

void run_samples () {
  
  // loop over all MC samples
  std::vector<long_int>::iterator it;
  for (it = SAMPLES[MULTI_LEVEL][MULTI_SAMPLER][MESH_LEVEL][TYPE].begin(); it < SAMPLES[MULTI_LEVEL][MULTI_SAMPLER][MESH_LEVEL][TYPE].end(); it++) {
    
    // draw a sample
    mlmc_get_chaos (*it);
    
    // run sample
    run_sample ();
  }
}

void run_level_samples () {
  
  // check if there are any assigned samples for current mesh level and type (among ALL samplers)
  if ( SAMPLES_COUNT[MULTI_LEVEL][MESH_LEVEL][TYPE] == 0 )
    return;
  
  // initialize parallelization setup for current MESH_LEVEL
  // this sets 'MULTI_SAMPLER' which is then required for 'SAMPLES'
  multi_level_init ();
  
  // check if there are any assigned samples for current mesh level and type
  if ( SAMPLES[MULTI_LEVEL][MULTI_SAMPLER][MESH_LEVEL][TYPE].empty() )
    return;
  
  // initialize to MESH_LEVEL
  mesh_level_init ();
  
  // sets up the mesh-size dependent parallelization parameters
  multi_mesh_level_init ();
  
  // ouput information about current level
  display_level_info ();
  
  // setup RNG and chaos
  mlmc_setup_chaos ();
  
  //#pragma acc data create(mean_U_l[0:NSAVES+SAVE0],mean_U_l[0:NSAVES+SAVE0]) create(U,Xtra,Source,Flux[0:NDIR])
  
  #if STATS
  
  // initializing volumes for statistical information
  stats_volumes_level_init ();
  
  // initialize routines for statistics
  stats_init ();
  
  #endif
  
  // get MC estimate
  run_samples ();
  
  #if STATS
  
  // finalizes (intermediate) statistical estimates
  stats_finalize_updates ();
  
  #endif
  
  multi_mesh_level_exit ();
  
  mesh_level_exit ();
  
  display_level_done ();
}

#else

void run_samples_in_differences (const int &samples_level) {
  
  // loop over all MC samples
  std::vector<long_int>::iterator it;
  for (it = SAMPLES[MULTI_LEVEL][MULTI_SAMPLER][samples_level][0].begin(); it < SAMPLES[MULTI_LEVEL][MULTI_SAMPLER][samples_level][0].end(); it++) {
    
    // draw a sample
    mlmc_get_chaos (*it);
    
    // run sample on two different MESH_LEVELs,
    // with TYPE = 0 and TYPE = 1, respectively.
    for ( TYPE=0; TYPE<2; TYPE++ ) {
      
      // no type 1 samples on the coarsest samples_level
      if ( samples_level == 0 && TYPE == 1 )
        continue;
      
      // set MESH_LEVEL according to TYPE
      MESH_LEVEL = samples_level - TYPE;
      
      // initialize to MESH_LEVEL
      mesh_level_init ();
      
      // sets up the mesh-size dependent parallelization parameters
      multi_mesh_level_init ();
      
      // run sample
      run_sample ();
      
      mesh_level_exit ();
      
      multi_mesh_level_exit ();
    }
  }
}

void run_level_samples_in_differences (const int &samples_level) {
  
  MESH_LEVEL = samples_level;
  TYPE       = 0;
  
  // check if there are any assigned samples for current mesh level and type (among ALL samplers)
  if ( SAMPLES_COUNT[MULTI_LEVEL][MESH_LEVEL][TYPE] == 0 )
    return;
  
  // initialize parallelization setup for current MESH_LEVEL
  // this sets 'MULTI_SAMPLER' which is then required for 'SAMPLES'
  multi_level_init ();
  
  // check if there are any assinged samples for current samples_level
  if ( SAMPLES[MULTI_LEVEL][MULTI_SAMPLER][MESH_LEVEL][TYPE].empty() )
    return;
  
  // initialize parallelization setup for current MESH_LEVEL
  multi_level_init ();
  
  // output information about current level
  display_level_info ();
  
  // setup RNG and chaos
  mlmc_setup_chaos ();
  
  #if STATS
  
  for ( TYPE=0; TYPE<2; TYPE++ ) {
    
    // no type 1 samples on the coarsest samples_level
    if ( samples_level == 0 && TYPE == 1 )
      continue;
    
    // set MESH_LEVEL according to TYPE
    MESH_LEVEL = samples_level - TYPE;
    
    // initializing volumes for statistical information
    stats_volumes_level_init ();
    
    // initialize routines for statistics
    stats_init ();
  }
  
  #endif
  
  // get MC estimate
  run_samples_in_differences (samples_level);
  
  #if STATS
  
  for ( TYPE=0; TYPE<2; TYPE++ ) {
    
    // no type 1 samples on the coarsest samples_level
    if ( samples_level == 0 && TYPE == 1 )
      continue;
    
    // set MESH_LEVEL according to TYPE
    MESH_LEVEL = samples_level - TYPE;
    
    // finalizes (intermediate) statistical estimates
    stats_finalize_updates ();
  }
  
  #endif
  
  multi_mesh_level_exit ();
  
  display_level_done ();
}

#endif

void run_mlmc () {
  
  #ifndef QUIET
  if ( multi_main && (L != LOWEST_MULTI_LEVEL || L == 0) ) {
    std::cout << " :: SIMULATION running...";
    std::cout.flush();
  }
  #endif
  
  #if COMPUTE_IN_DIFFERENCES
  
  // must be from L down to 0 in order to have
  // correct parallelization idling times
  for ( int samples_level=L; samples_level>=0; samples_level-- )
    run_level_samples_in_differences (samples_level);
  
  #else
  
  // must be from L down to 0 in order to have
  // correct parallelization idling times
  for ( MESH_LEVEL=L; MESH_LEVEL>=0; MESH_LEVEL-- )
    for ( TYPE=0; TYPE<2; TYPE++ )
      run_level_samples ();
  
  #endif
  
  #ifndef QUIET
  if ( multi_main && L != LOWEST_MULTI_LEVEL )
    std::cout << " done." << std::endl << std::endl;
  #endif
}

void fvm_volumes_cleanup () {
  U      .reinit (1, 1, 1, 0);
  Xtra   .reinit (1, 1, 1, 0);
  Source .reinit (1, 1, 1, 0);
  for (int i=0; i<NDIR; i++)
    Flux[i].reinit (1, 1, 1, 0);
}

#if STATS

void combine_mc() {
  multi_combine_mc();
}

void combine_mlmc() {
  
  multi_combine_mlmc();
  
  // finalizes statistical estimates
  if ( multi_samplers_root && multi_level_root )
    
    for ( MESH_LEVEL = L; MESH_LEVEL >= 0; MESH_LEVEL-- )
      
      for ( TYPE = 0; TYPE < 2; TYPE++ )
        
        if ( TYPE == 0 || MESH_LEVEL != L )
          
          for ( SNAPSHOT=0; SNAPSHOT<NSAVES+SAVE0; SNAPSHOT++ )
            
            stats_finalize_combines();
}

void assemble_mlmc() {
  
  #if L != 0
  
  #ifndef QUIET
  if (multi_main) std::cout << " :: Assembling MLMC estimate...";
  if (multi_main) std::cout.flush();
  #endif
  
  if ( multi_samplers_root && multi_level_root ) {
    
    stats_assemble_mlmc ();
    
    for (SNAPSHOT = 0; SNAPSHOT<NSAVES+SAVE0; SNAPSHOT++)
      stats_apply_constraints ();
  }
  
  #ifndef QUIET
  if (multi_main) std::cout << " done." << std::endl;
  #endif
  
  #endif
}

#endif

void write_mc_datafiles () {
  
  // save MC estimates (for stochastic simulations only)
  
  #if STATS && WRITE_DATAFILES && WRITE_LEVEL_DATAFILES
  
  // save MC estimates for each MESH_LEVEL and TYPE
  if ( multi_samplers_root && multi_level_root ) {
    
    #ifndef QUIET
    if (multi_main) std::cout << " :: Writing level datafiles...";
    if (multi_main) std::cout.flush();
    #endif
    
    INCLUDE_LEVEL_SUFFIX = 1;
    
    for ( MESH_LEVEL = L; MESH_LEVEL >= 0; MESH_LEVEL-- )
      for ( TYPE = 0; TYPE < 2; TYPE++ )
        if ( TYPE == 0 || MESH_LEVEL != L )
          stats_write_datafiles ();
    
    #ifndef QUIET
    if (multi_main) std::cout << " done." << std::endl;
    #endif
  }
  
  #endif
}

void write_mlmc_datafiles () {
  
  // save MLMC estimates (for stochastic simulations only)
  
  #if STATS && WRITE_DATAFILES
  
  if ( multi_samplers_root && multi_level_root ) {
    
    #ifndef QUIET
    if (multi_main) std::cout << " :: Writing datafiles...";
    if (multi_main) std::cout.flush();
    #endif
    
    // MLMC estimates are assembled to MESH_LEVEL = L, TYPE = 0
    MESH_LEVEL = L;
    TYPE = 0;
    INCLUDE_LEVEL_SUFFIX = 0;
    stats_write_datafiles ();
    
    #ifndef QUIET
    if (multi_main) std::cout << " done." << std::endl;
    #endif
  }
  
  #endif
}
