
#ifndef MLMC_H
#define MLMC_H

#include "equation.h"

#include <math.h>
#include <vector>

// === MLMC variables and volumes

// effective doubled order: for solutions with shocks, effective order is halved and bounded by 1
// !!! IF MODIFYING, also CHECK make.py::MakeMLMC::get_multi_total() for possible necessary modifications
const int EFFECTIVE_DOUBLED_ORDER = ( (SMOOTH_FLUX && SMOOTH_SOLUTION) || HAVE_VISCOSITY ) ? 2*ORDER : min(2,ORDER);

// "equilibrated" factor for number of samples in each succesive samples level, i.e. M[samples_level] = FACTOR * M[samples_level+1]
const int SAMPLES_FACTOR = 1 << EFFECTIVE_DOUBLED_ORDER;

// "equilibrated" factor for number of samples in each succesive samples level, i.e. M[samples_level] = FACTOR * M[samples_level+1]
const real OPTIMAL_SAMPLES_FACTOR = pow ( 2, (EFFECTIVE_DOUBLED_ORDER + 2.0*(ACTIVE_DIRS+1)) / 3.0 );

// resolution level at the coarsest mesh
const int L0X = HAVE_DIR(X) ? (int) floor ( log2(NX) ) - L : 0;
const int L0Y = HAVE_DIR(Y) ? (int) floor ( log2(NY) ) - L : 0;
const int L0Z = HAVE_DIR(Z) ? (int) floor ( log2(NZ) ) - L : 0;

// constants indicating first and last snapshots
const int FIRST_SNAPSHOT = 0;
const int  LAST_SNAPSHOT = NSAVES + SAVE0 - 1;

// === global variables

extern int  N [NDIR] [L+1];             // number of cells in the entire domain for each direction and level
extern real D [NDIR] [L+1];             // mesh widths for each direction and level

extern long_int NM [L+1];               // number of samples for each level
extern long_int NMT;                    // total number of samples (from all levels, including BOTH types)
extern long_int NMT_TYPE0;              // total number of samples (from all levels, only TYPE == 0)

extern int MESH_LEVEL;                  // current mesh_level
extern int TYPE;                        // current type of sample
extern int SNAPSHOT;                    // current snapshot index
extern long_int SAMPLE_INDEX;           // current index of the sample
extern int SUBDOMAIN_INDEX;             // current index of the subdomain

int SAMPLES_LEVEL();                    // returns current samples_level

extern int INCLUDE_LEVEL_SUFFIX;        // flag for special suffix in level output datafiles

// === MLMC routines

// run compatibility checks
// must be after 'multi_init()'
void compatibility_checks ();

// computes MLMC paramaters: number of samples NM[*] and mesh resolutions N[*][*]
void mlmc_init ();

// display information such as number of cores, levels, coarsest level, etc.
// should be executed after 'mlmc_init()' and 'multi_mlmc_init()'
void display_info ();

// initializes volumes U and Xtra for currently set MESH_LEVEL
void fvm_volumes_level_init ();

// setup FVM meshes for mesh_level = MESH_LEVEL and call LEVEL_INIT_HOOK
// should be executed after 'mlmc_init()' and 'multi_level_init()'
void mesh_level_init ();

// mainly calles LEVEL_EXIT_HOOK hook
void mesh_level_exit ();

// display information such as resolution, number of samples, number of cores, etc.
// should be executed after 'mesh_level_init()' and 'multi_level_init()'
void display_level_info();

// display that the level has finished
// should be executed after display_level_info()
void display_level_done();

// Szudzik's pairing function
inline int pair (const int a, const int b) { return a >= b ? a * a + a + b : a + b * b; }

// pairing 4-tuple
inline int pair (const int a, const int b, const int c, const int d) { return pair ( pair (a, b), pair (c, d) ); }

// get the pre-set seed (corresponding to the specified 'sample_index', if needed)
int get_seed (const int sample_index = 1);

// get the pre-set stream (corresponding to the specified 'sample_index', if needed)
int get_stream (const int sample_index = 1);

// draw the required samples to set the chaos correctly
void mlmc_get_chaos (const int sample_index);

// setup chaos: set the amount of random numbers; if needed, call 'rng_and_chaos_init()'
void mlmc_setup_chaos ();

// initialize random number generator with the appropriate seed and stream and chaos with the appropriate offset
// the next chaos_draw() should return the chaos corresponding to the specified 'sample_index'
void rng_and_chaos_init (const int sample_index = 1);

// runs one sample (FVM simulation), saves results for each SNAPSHOT
void run_sample ();

#if COMPUTE_IN_DIFFERENCES

// run 'SAMPLERS[MULTI_LEVEL][MULTI_SAMPLER][samples_level][*]'
// use samples from level samples_level
void run_samples_in_differences (const int &samples_level);

// run MC samples given by 'SAMPLERS[MULTI_LEVEL][MULTI_SAMPLER][samples_level][*]'
void run_level_samples_in_differences (const int &samples_level);

#else

// run 'SAMPLERS[MULTI_LEVEL][MULTI_SAMPLER][MESH_LEVEL][TYPE]'
// use samples from level SAMPLES_LEVEL()
void run_samples ();

// run MC samples given by 'SAMPLERS[MULTI_LEVEL][MULTI_SAMPLER][MESH_LEVEL][*]'
void run_level_samples ();

#endif

// run MLMC algorithm: compute samples assigned in 'SAMPLERS[MULTI_LEVEL][*][*][*]'
void run_mlmc();

// cleap up after all samples are computed,
// i.e. deallocates FVM volumes
void fvm_volumes_cleanup ();

#if STATS

// combine parts of MC estimates from 'samplers' to 'samplers roots'
void combine_mc ();

// combine parts MC estimates from 'samplers roots' to 'level roots'
// to obtain all required MC estimates for MLMC estimate
void combine_mlmc ();

// assemble MLMC estimate from combined MC estimates to MESH_LEVEL = L and TYPE = 0
void assemble_mlmc ();

#endif

// data output of MC estimates at each mesh level and type (only if WRITE_LEVEL_DATAFILES = 1)
void write_mc_datafiles ();

// data output of the MLMC estimates (only if WRITE_DATAFILES = 1)
void write_mlmc_datafiles ();

#endif
