
// # # # # # # # # # # # # # # # # # # # # # # # # # #
// # ASLVID-UQ                                       #
// # MLMC-FVM modelling software for nonlinear SHPDE #
// #                                                 #
// # Jonas Sukys                                     #
// # SAM, ETH Zurich, Switzerland                    #
// # sukys.jonas@gmail.com                           #
// #                                                 #
// # For related publications, refer to              #
// # http://pub.sukys.lt                             #
// #                                                 #
// # Documentation:                                  #
// # http://www.sam.math.ethz.ch/alsvid-uq           #
// #                                                 #
// # Based on: ALSVID by Andrew McMurry (CMA, Oslo)  #
// # # # # # # # # # # # # # # # # # # # # # # # # # #

#include "mlmc.h"
#include "multi.h"
#include "balancer.h"

///::title:: MLMC-FVM modelling software for Stochastic systems of Hyperbolic Partial Differential Equations

// Default constants and parameters for models (deterministic and stochastic)
///::consts:: MAXX=1 MAXY=1 MAXZ=1
///::consts:: ORDER=1 SMOOTH_FLUX=1 GHOST_SIZE=1 HAVE_GRAVITY=0
///::consts:: STATS=(L||ML-1)
///::vars:: NX=1 NY=1 NZ=1
///::vars:: ML=1 L=0
///::vars:: SMOOTH_SOLUTION=0 CFL=0.9
///::vars:: HAVE_VISCOSITY=0 VISCOSITY=1e-3
///::vars:: INITIAL_DATA_OFFSET=0.0 INITIAL_DATA_SCALING=1.0

// Default output customization parameters
///::vars:: NSAVES=1 SAVE0=0
///::vars:: WRITE_DATAFILES=1 WRITE_LEVEL_DATAFILES=0
///::vars:: PRINT_t=0

// Default technical parameters
///::vars:: READ_NUMBER_OF_SAMPLES_FROM_FILE=0 USE_OPTIMAL_NUMBER_OF_SAMPLES=0 USE_ONE_SAMPLE_PER_LEVEL=0
///::vars:: USE_PRIME_SEEDS=0 PRIME_SHIFT=10 RUN_NUMBER=1 READ_RUN_NUMBER_FROM_FILE=0 SAMPLES_OFFSET=0
///::vars:: ENSURE_PARALLEL_REPRODUCIBILITY=1
///::vars:: DISTRIBUTE_IN_DIFFERENCES=0 COMPUTE_IN_DIFFERENCES=0

// Default file names and paths
///::vars:: DATA_PATH="" DATA_FILE="output.dat" BALANCER_FILE="balancer.dat" LOADS_FILE="loads.dat"
///::vars:: TIME_FILE="times.py" RUNTIME_FILE="runtime.py" SAMPLES_FILE="samples.py" BALANCER_REPORT_FILE="balancer.txt"
///::vars:: SAMPLE_NUMBERS_FILE="Ml.txt" RUN_NUMBER_FILE="RUN_NUMBER.txt"

// Other parameters
///::consts:: INDICATOR=0
///::vars:: COMPILER_DIRECTIVES="" LINKER_LIBS=""
///::vars:: CHECK_VOLUME_INDICES=0

// Default hooks
///::macro:: INIT_HOOK @ @
///::macro:: LEVEL_INIT_HOOK @ @
///::macro:: PRE_STEP_HOOK @ @
///::macro:: INITIAL_DATA_HOOK @ @
///::macro:: SOURCE_HOOK @ @
///::macro:: VISCOSITY_HOOK @ @
///::macro:: PRE_RECONSTRUCTION_HOOK @ @
///::macro:: POST_RECONSTRUCTION_HOOK @ @
///::macro:: POST_STEP_HOOK @ @
///::macro:: LEVEL_EXIT_HOOK @ @
///::macro:: EXIT_HOOK @ @

// Dependencies
///::depends:: timer structures-base structures time equation multi balancer chaos stats mlmc

/* REMARKS on the implementation of ALSVID-UQ
 * 
 * There are 3 types of 'current level indicators' troughout ALSVID-UQ:
 *   MESH_LEVEL    - indicates current mesh level, i.e. used for accessing N[*][MESH_LEVEL]            (range: 0, ..., L)
 *   SAMPLES_LEVEL - indicates current samples level, i.e. used for accessing NM[SAMPLES_LEVEL]        (range: 0, ..., L)
 *   MULTI_LEVEL   - indicates which group of 'parallelization over levels' current core belongs to    (range: LOWEST_MULTI_LEVEL, ..., L)
 * 
 * All samples have a 'TYPE' assigned to them,
 * indicating to which term of MC difference E_{M_l} [ U_l - U_{l-1} ] they belong to:
 * (here 'l' denotes 'MESH_LEVEL')
 *   TYPE = 0      - sample belongs to the term E_{M_l} [ U_l ]
 *   TYPE = 1      - sample belongs to the term E_{M_l} [ U_{l-1} ]
 */

int main(int argc, char **argv) {
  
  // === INITIALIZATION
  
  // system initialization
  multi_init(argc, argv);
  
  // MLMC initialization
  mlmc_init();
  
  // display information about simulation
  display_info();
  
  // initialization of parallelization setup
  multi_mlmc_init();
  
  // parameter compatibility checks
  compatibility_checks();
  
  // specified initialization routines
  INIT_HOOK
  
  // === BALANCING
  
  // distribute samples to all cores
  balancer();
  
  // === SIMULATION
  
  // synchronize and start timer
  multi_sync();
  timer.start();
  
  // main computations
  run_mlmc();
  
  // compute load imbalance
  multi_compute_imbalance();
  
  // sync and compute idling
  multi_compute_idling();
  
  // cleanup FVM memory
  fvm_volumes_cleanup();
  
  // === DATA ASSEMBLY
  
  #if STATS
  
  // combine parts of MC estimates from samplers
  combine_mc();
  
  // combine parts of MC estimates from samplers roots
  combine_mlmc();
  
  // set the parallelization setup for parallel data output
  multi_level_init_parameters (L);
  
  // data output of MC estimates at each MESH_LEVEL and TYPE
  write_mc_datafiles();
  
  // compute MLMC estimate from combined MC estimates
  assemble_mlmc();
  
  #endif
  
  // === OUTPUT
  
  // stop timer
  timer.stop();
  
  // report runtimes
  multi_report_runtimes();
  
  // data output of the MLMC estimates
  write_mlmc_datafiles();
  
  // == FINALIZATION
  
  // cleanup of parallelization setup
  multi_cleanup();
  
  // specified finalization routines
  EXIT_HOOK
  
  // finalization of parallelization setup
  multi_finalize();
  
  return 0;
}
