
#include "multi-mpi-helper.h"

#include "multi_mpi.h"
#include "multi.h"
#include "mlmc.h"
#include "balancer.h"

#include <iostream>
#include <fstream>
#include <cstdlib>
#include "mpi.h"

int DEFAULT_MULTI [NDIR+1] [L+1];
int DEFAULT_MULTIN [L+1];

int comm_level_root_ranks [L+1];
int comm_level_last_ranks [L+1];

int ddm_ranks_send [MULTIX][MULTIY][MULTIZ];
int ddm_ranks_recv [MULTIX][MULTIY][MULTIZ] [L+1];

int MULTI_MODIFIED [L+1][L+1];

int REMAINING_CORES;

void multi_compatibility_checks () {
  
  // if MULTI_CORES == 0, a certain structure of MULTIX* is assumed
  if ( MULTI_CORES == 0 ) {
    
    // MULTIX can NOT be larger than 1 if not HAVE_DIR(X)
    if ( !HAVE_DIR(X) && MULTIX > 1 && ! MULTI_CORES ) {
      if (multi_main) std::cerr << " :: ERROR: MULTIX can NOT be larger than 1 since x-direction is not present." << std::endl << std::endl;
      if (multi_main) std::exit(1);
    }
    
    // MULTIY can NOT be larger than 1 if not HAVE_DIR(Y)
    if ( !HAVE_DIR(Y) && MULTIY > 1 && ! MULTI_CORES ) {
      if (multi_main) std::cerr << " :: ERROR: MULTIY can NOT be larger than 1 since y-direction is not present." << std::endl << std::endl;
      if (multi_main) std::exit(1);
    }
    
    // MULTIZ can NOT be larger than 1 if not HAVE_DIR(Z)
    if ( !HAVE_DIR(Z) && MULTIZ > 1 && ! MULTI_CORES ) {
      if (multi_main) std::cerr << " :: ERROR: MULTIZ can NOT be larger than 1 since z-direction is not present." << std::endl << std::endl;
      if (multi_main) std::exit(1);
    }
    
    // MULTIX must be a positive divider of NX
    if ( MULTIX < 1 || NX % MULTIX != 0 && ! MULTI_CORES ) {
      if (multi_main) std::cerr << " :: ERROR: MULTIX must be a positive divider of NX." << std::endl << std::endl;
      if (multi_main) std::exit(1);
    }
    
    // MULTIY must be a positive divider of NY
    if ( MULTIY < 1 || NY % MULTIY != 0 && ! MULTI_CORES ) {
      if (multi_main) std::cerr << " :: ERROR: MULTIY must be a positive divider of NY." << std::endl << std::endl;
      if (multi_main) std::exit(1);
    }
    
    // MULTIZ must be a positive divider of NZ
    if ( MULTIZ < 1 || NZ % MULTIZ != 0 && ! MULTI_CORES ) {
      if (multi_main) std::cerr << " :: ERROR: MULTIZ must be a positive divider of NZ." << std::endl << std::endl;
      if (multi_main) std::exit(1);
    }
  }
}

#if ! MULTI_CORES

void multi_initialize_default_parallelization_setup () {
  
  // !!! IF MODIFYING, also CHECK make.py::MakeMLMC::get_multi_total() for possible necessary modifications
  
  // DEFAULT_MULTI[*][L], DEFAULT_MULTIN[L] are set by MULTI*
  DEFAULT_MULTI [X] [L] = MULTIX;
  DEFAULT_MULTI [Y] [L] = MULTIY;
  DEFAULT_MULTI [Z] [L] = MULTIZ;
  DEFAULT_MULTI [M] [L] = MULTIM;
  DEFAULT_MULTIN    [L] = MULTIX * MULTIY * MULTIZ;
  
  // compute remaining DEFAULT_MULTI[X-Z][*], DEFAULT_MULTIN[*], DEFAULT_MULTI[M][*]
  for (int multi_level=L-1; multi_level>=0; multi_level--) {
    
    // compute DEFAULT_MULTI[*][multi_level] by halfing DEFAULT_MULTI[*][multi_level+1]
    for (int d=0; d<NDIR; d++)
      if ( DEFAULT_MULTI[d][multi_level + 1] != 1 )
        // check if DEFAULT_MULTI[d][multi_level+1] is a multiple of 2
        if ( DEFAULT_MULTI[d][multi_level + 1] % 2 == 0 )
          DEFAULT_MULTI[d][multi_level] = DEFAULT_MULTI[d][multi_level + 1] / 2;
        else {
          if (multi_main) std::cerr << " :: ERROR: MULTI[d] is NOT divisible by 2^" << L - multi_level << " for d=" << d << "." << std::endl;
          if (multi_main) std::exit(1);
        }
      else
        DEFAULT_MULTI[d][multi_level] = 1;
    
    // compute MULTIN[multi_level][L]
    DEFAULT_MULTIN[multi_level] = DEFAULT_MULTI[X][multi_level] * DEFAULT_MULTI[Y][multi_level] * DEFAULT_MULTI[Z][multi_level];
    
    // === compute DEFAULT_MULTI[M][*] using WORK_FACTOR
    
    // factor, by which the processing power of each sampler is decreased due to decreasing MULTI[X-Z][*][L]
    int MULTIN_FACTOR = DEFAULT_MULTIN[multi_level+1] / DEFAULT_MULTIN[multi_level];
    
    // WORK_FACTOR = WORK[multi_level+1] / WORK[multi_level]
    int WORK_FACTOR = 1 << ( ACTIVE_DIRS + 1 + HAVE_VISCOSITY - EFFECTIVE_DOUBLED_ORDER );
    
    // MULTI[M][multi_level] should never be smaller than '1'
    if ( DEFAULT_MULTI[M][multi_level+1] * MULTIN_FACTOR >= WORK_FACTOR )
      DEFAULT_MULTI[M][multi_level] = (DEFAULT_MULTI[M][multi_level+1] * MULTIN_FACTOR) / WORK_FACTOR;
    else
      DEFAULT_MULTI[M][multi_level] = 1;
  }
  
  // compute LOWEST_MULTI_LEVEL, i.e.
  // find the largest multi level from which on parallelization is no longer needed
  LOWEST_MULTI_LEVEL = 0;
  for (int multi_level=L-1; multi_level>=0; multi_level--)
    if (DEFAULT_MULTI[M][multi_level] == 1 && DEFAULT_MULTI[M][multi_level+1] == 1 && DEFAULT_MULTIN[multi_level] == 1 && DEFAULT_MULTIN[multi_level+1] == 1 ) {
      LOWEST_MULTI_LEVEL = multi_level;
      break;
    }
  
  // all cores participate in the computation
  REMAINING_CORES = 0;
}

#else

void multi_initialize_default_parallelization_setup_from_MULTI_CORES () {
  
  REMAINING_CORES = MULTI_CORES;
  int cores_per_sampler, multi_level = L;
  
  while ( REMAINING_CORES > 0 && multi_level >= 0 ) {
    
    // find the largest possible parallelization configuration
    cores_per_sampler = 1 << (int) floor ( log2 ( REMAINING_CORES ) / ACTIVE_DIRS );
    
    // number of cores per sampler can NOT exceed the number of cells on the finest mesh level
    cores_per_sampler = min ( cores_per_sampler, NX*NY*NZ );
    
    // set DEFAULT_MULTI[X-Z][multi_level] accordingly
    for ( int d = 0; d < NDIR; d++ )
      DEFAULT_MULTI [d] [multi_level] = HAVE_DIR(d) ? cores_per_sampler : 1;
    
    // set DEFAULT_MULTIN[multi_level] accordingly
    DEFAULT_MULTIN [multi_level] = DEFAULT_MULTI [X] [multi_level] * DEFAULT_MULTI [Y] [multi_level] * DEFAULT_MULTI [Z] [multi_level];
    
    // set DEFAULT_MULTI[M][multi_level] to the largest possible value
    DEFAULT_MULTI [M] [multi_level] = REMAINING_CORES / DEFAULT_MULTIN [multi_level];
    
    // update REMAINING_CORES
    REMAINING_CORES -= DEFAULT_MULTIN [multi_level] * DEFAULT_MULTI [M] [multi_level];
    
    // update LOWEST_MULTI_LEVEL
    LOWEST_MULTI_LEVEL = multi_level;
    
    // update multi_level
    multi_level--;
  }
}

#endif

void multi_check_core_consistency () {
  
  // compute MULTI_TOTAL_COMPUTED
  int MULTI_TOTAL_COMPUTED = 0;
  for (int multi_level=L; multi_level>=LOWEST_MULTI_LEVEL; multi_level--)
    MULTI_TOTAL_COMPUTED += DEFAULT_MULTI[M][multi_level] * DEFAULT_MULTIN[multi_level];
  
  // check for consistency
  if ( MULTI_TOTAL_COMPUTED + REMAINING_CORES != MULTI_TOTAL ) {
    
    // if fewer cores are needed, adjust REMAINING_CORES and display a warning
    if ( MULTI_TOTAL_COMPUTED < MULTI_TOTAL ) {
      if (multi_main) {
        std::cout << std::endl;
        std::cout << " :: WARNING: inconsistent number of cores: " << MULTI_TOTAL_COMPUTED << " + " << REMAINING_CORES << " != " << MULTI_TOTAL << std::endl;
        std::cout << "  : -> value for REMAINING_CORES was adjusted: " << MULTI_TOTAL_COMPUTED << " + " << MULTI_TOTAL - MULTI_TOTAL_COMPUTED << " == " << MULTI_TOTAL << std::endl;
        std::cout << std::endl;
      }
      REMAINING_CORES = MULTI_TOTAL - MULTI_TOTAL_COMPUTED;
    }
    
    // if more cores are needed, display an error
    else if (multi_main) {
      std::cout << std::endl;
      std::cerr << " ERROR: inconsistent number of cores: " << MULTI_TOTAL_COMPUTED << " + " << REMAINING_CORES << " != " << MULTI_TOTAL << std::endl;
      std::cout << std::endl;
      std::exit(1);
    }
  }
}

void multi_report_parallelization_setup () {
  
  if ( ! multi_main ) return;
  std::cout << std::endl << " :: INFO: Parallelization setup:";
  for (int multi_level = L; multi_level >= LOWEST_MULTI_LEVEL; multi_level-- ) {
    std::cout << " " << DEFAULT_MULTI [M] [multi_level] << "x(" << DEFAULT_MULTI [X] [multi_level];
    if ( HAVE_DIR(Y) ) std::cout << "x" << DEFAULT_MULTI [Y] [multi_level];
    if ( HAVE_DIR(Z) ) std::cout << "x" << DEFAULT_MULTI [Z] [multi_level];
    std::cout << ")";
  }
  if ( REMAINING_CORES > 0 )
    std::cout << " [" << REMAINING_CORES << " core(s) unused]";
  std::cout << std::endl;
}

void multi_compute_parallelization_setup () {
  
  // if modifications for the parallelization setup are NOT allowed
  // then check for possible parameter inconsistencies and report warnings
  if ( ENFORCE_DDM_SETUP ) {
    
    // report a warning if MIN_CELLS_PER_CORE is inconsistent with DEFAULT_MULTI[X-Z]
    for (int d = 0; d < NDIR; d++)
      if ( HAVE_DIR(d) && DEFAULT_MULTI[d][L] > 1 && N [d] [L] / DEFAULT_MULTI[d][L] < MIN_CELLS_PER_CORE ) {
        if (multi_main) std::cerr << std::endl;
        if (multi_main) std::cerr << " :: WARNING: Uniform enforcement of MIN_CELLS_PER_CORE=" << MIN_CELLS_PER_CORE << " is NOT possible." << std::endl;
        if (multi_main) std::cerr << "  : -> Reduce MIN_CELLS_PER_CORE." << std::endl;
      }
    
    // report a warning if ENABLE_SLICED_PARALLELIZATION == 1 is inconsistent with DEFAULT_MULTIN[L][L]
    if ( ENABLE_SLICED_PARALLELIZATION && NX < DEFAULT_MULTIN[L] ) {
      if (multi_main) std::cerr << std::endl;
      if (multi_main) std::cout << " :: WARNING: Uniform enforcement of ENABLE_SLICED_PARALLELIZATION=1 is NOT possible." << std::endl;
      if (multi_main) std::cout << "  : -> Only partial sliced parallelization will be used (e.g. for FFT)." << std::endl;
    }
  }
  
  // compute MULTI[X-Z][*][*], MULTIN[*][*], MULTI[M][*][*]
  int nothing_modified_yet = 1;
  for (int multi_level = L; multi_level >= LOWEST_MULTI_LEVEL; multi_level--) {
    for (int mesh_level = L; mesh_level >= 0; mesh_level--) {
      
      // initialize MULTI [*] [multi_level] [mesh_level] with default values from DEFAULT_MULTI [*] [multi_level]
      MULTI [X] [multi_level] [mesh_level] = DEFAULT_MULTI [X] [multi_level];
      MULTI [Y] [multi_level] [mesh_level] = DEFAULT_MULTI [Y] [multi_level];
      MULTI [Z] [multi_level] [mesh_level] = DEFAULT_MULTI [Z] [multi_level];
      MULTI [M] [multi_level] [mesh_level] = DEFAULT_MULTI [M] [multi_level];
      MULTIN    [multi_level] [mesh_level] = DEFAULT_MULTIN    [multi_level];
      
      // record that NO modification was performed
      MULTI_MODIFIED [multi_level] [mesh_level] = 0;
      
      // as long as modification is needed and is possible
      while ( multi_modification_needed (multi_level, mesh_level) && multi_modification_possible (multi_level, mesh_level) ) {
        
        // iteratively reduce MULTI [X-Z] [multi_level] [MESH_LEVEL]
        // and increase MULTI [M] [multi_level] [MESH_LEVEL]
        multi_modify_ddm (multi_level, mesh_level);
        
        // record that modification was performed
        MULTI_MODIFIED [multi_level] [mesh_level] = 1;
      }
      
      // report the modification
      if (MULTI_MODIFIED [multi_level] [mesh_level] && multi_main) {
        if (nothing_modified_yet) {
          std::cout << std::endl << " :: INFO: Parallelization configuration was modified for:" << std::endl;
          nothing_modified_yet = 0;
        }
        std::cout << "  : -> MULTI_LEVEL=" << multi_level;
        std::cout << ", MESH_LEVEL=" << mesh_level;
        std::cout << " (" << N[X] [mesh_level] << "x" << N[Y] [mesh_level] << "x" << N[Z] [mesh_level] << ")";
        std::cout << ": from " << DEFAULT_MULTI [X] [multi_level] << "x" << DEFAULT_MULTI [Y] [multi_level] << "x" << DEFAULT_MULTI [Z] [multi_level];
        std::cout << " to " << MULTI [X] [multi_level] [mesh_level] << "x" << MULTI [Y] [multi_level] [mesh_level] << "x" << MULTI [Z] [multi_level] [mesh_level];
        std::cout << std::endl;
      }
    }
  }
  
  // if no modifications are present, also report
  if (nothing_modified_yet && multi_main)
    std::cout << std::endl << " :: INFO: Parallelization configuration was NOT modified on any MULTI_LEVEL." << std::endl;
  
  // check if the number of samplers does not exceed the total available number of samples
  for (int multi_level = L; multi_level >= LOWEST_MULTI_LEVEL; multi_level--)
    for (int mesh_level = L; mesh_level >= 0; mesh_level--)
      if ( MULTI [M] [multi_level] [mesh_level] > NM [mesh_level] )
        if ( multi_main && mesh_level < multi_level ) {
          std::cout << " :: WARNING: MULTI [M] ["<< multi_level << "] [" << mesh_level << "] > NM [" << mesh_level << "]: ";
          std::cout << MULTI [M] [multi_level] [mesh_level] << " > " << NM [mesh_level] << std::endl;
        }
}

bool multi_modification_needed (const int multi_level, const int mesh_level) {
  
  // check if modifications in parallelization setup are allowed
  if ( ENFORCE_DDM_SETUP ) return false;
  
  // check if the specified MIN_CELLS_PER_CORE is satisfied
  for (int d = 0; d < NDIR; d++ )
    if ( HAVE_DIR(d) )
      if ( N [d] [mesh_level] / MULTI [d] [multi_level] [mesh_level] < MIN_CELLS_PER_CORE )
        return true;
  
  // check if ENABLE_SLICED_PARALLELIZATION is satisfied
  if ( ENABLE_SLICED_PARALLELIZATION )
    if ( N [X] [mesh_level] < MULTIN [multi_level] [mesh_level] )
      return true;
  
  // no need for modification was found
  return false;
}

bool multi_modification_possible (const int multi_level, const int mesh_level) {
  return ( MULTIN [multi_level] [mesh_level] > 1 );
}

void multi_modify_ddm (const int multi_level, const int mesh_level) {
  
  int factor = 1;
  
  // reduce MULTI [X-Z] [multi_level] [mesh_level]
  for (int d = 0; d < NDIR; d++ )
    if ( HAVE_DIR(d) && MULTI [d] [multi_level] [mesh_level] > 1 ) {
      MULTI [d] [multi_level] [mesh_level] /= 2;
      factor *= 2;
    }
  
  // increase MULTI [M] [multi_level] [mesh_level]
  MULTI [M] [multi_level] [mesh_level] *= factor;
  
  // reduce MULTIN [multi_level] [mesh_level]
  MULTIN [multi_level] [mesh_level] /= factor;
}

void compute_comm_level_root_ranks () {
  comm_level_root_ranks [L] = 0;
  for (int multi_level = L-1; multi_level >= LOWEST_MULTI_LEVEL; multi_level--)
    comm_level_root_ranks [multi_level] = comm_level_root_ranks [multi_level+1] + DEFAULT_MULTI[M][multi_level+1] * DEFAULT_MULTIN[multi_level+1];
}

void compute_comm_level_last_ranks () {
  comm_level_last_ranks [LOWEST_MULTI_LEVEL] = MULTI_TOTAL - 1 - REMAINING_CORES;
  for (int multi_level = LOWEST_MULTI_LEVEL + 1; multi_level <= L; multi_level++)
    comm_level_last_ranks [multi_level] = comm_level_last_ranks [multi_level-1] - DEFAULT_MULTI[M][multi_level-1] * DEFAULT_MULTIN[multi_level-1];
}

void set_multi_level () {
  
  // special treatment of the remaining cores
  if ( REMAINING_CORES > 0 && rank_main > MULTI_TOTAL - 1 - REMAINING_CORES ) {
    MULTI_LEVEL = L+1;
    multi_level_root = 0;
  }
  
  // otherwise, proceed as normal
  else {
    MULTI_LEVEL = L;
    while ( rank_main > comm_level_last_ranks [MULTI_LEVEL] )
      MULTI_LEVEL--;
    multi_level_root = (L == MULTI_LEVEL);
  }
}

int compute_ddm_rank_send (const int &multix, const int &multiy, const int &multiz, const int &multi_sampler) {
  return multix * MULTI[Y][L][L] * MULTI[Z][L][L] + multiy * MULTI[Z][L][L] + multiz + multi_sampler * MULTIN[L][L];
}

int compute_ddm_rank_recv (const int &multix, const int &multiy, const int &multiz, const int &multi_level, const int &multi_sampler) {
  
  int needed_multix = multix / (MULTI[X][L][L] / MULTI[X][multi_level][MESH_LEVEL]);
  int needed_multiy = multiy / (MULTI[Y][L][L] / MULTI[Y][multi_level][MESH_LEVEL]);
  int needed_multiz = multiz / (MULTI[Z][L][L] / MULTI[Z][multi_level][MESH_LEVEL]);
  
  int rank = comm_level_root_ranks [multi_level];
  
  rank += needed_multix * MULTI[Y][multi_level][MESH_LEVEL] * MULTI[Z][multi_level][MESH_LEVEL];
  rank += needed_multiy * MULTI[Z][multi_level][MESH_LEVEL];
  rank += needed_multiz;
  
  rank += multi_sampler * MULTIN[multi_level][MESH_LEVEL];
  
  return rank;
}
