
#include "multi.h"

#include "multi_mpi.h"
#include "multi-mpi-helper.h"
#include "multi-mpi-timers.h"
#include "mlmc.h"
#include "stats.h"
#include "balancer.h"

#include <cstdlib>
#include <iostream>
#include <vector>

///::title:: MPI parallelisation
///// Number of zones in each direction (each zone gets its own core)
///::vars:: MULTIX=1 MULTIY=1 MULTIZ=1
///// Number of separate groups of cores for sampling
///::vars:: MULTIM=1
///// If MULTI_CORES > 0, the above is ignored and all MULTI* are computed from MULTI_CORES
///::vars:: MULTI_CORES=0
///// Indicator for overlapping computations and communication in DDM
///::vars:: HIDE_DDM_COMMUNICATION=0
///// DDM configuration is enforced to have at least MIN_CELLS_PER_CORE^NDIR cells/core
///::vars:: MIN_CELLS_PER_CORE=16
///// DDM configuration is enforced to have NXl >= MULTINl
///::consts:: ENABLE_SLICED_PARALLELIZATION=0
///// Other parameters
///::consts:: BC_REPEAT_X=0 BC_REPEAT_Y=0 BC_REPEAT_Z=0
///::depends:: multi-mpi-helper multi-mpi-timers

// === for declarations and documentation, refer to 'multi_cpp.h' === //

int multi_bc_rank[NDIR][2];

// MPI data types for the communication in DDM
MPI_Datatype mpi_SUB_VOLUME_BC_CV [NDIR];
MPI_Datatype mpi_SUB_VOLUME_BC_XV [NDIR];

// MPI requests for asynchronous DDM communication
MPI_Request ddm_send_req    [NDIR] [2];
MPI_Request ddm_recv_req    [NDIR] [2];

// vector of MPI requests for asynchronous SENDS in MLMC communication
std::vector<MPI_Request> mlmc_send_req;

// default MPI requests for asynchronous SENDS in MLMC communication
MPI_Request req_CV, req_XV;

// MPI communicators

MPI_Comm comm_level    = MPI_COMM_NULL;
MPI_Comm comm_samplers = MPI_COMM_NULL;
MPI_Comm comm_domain   = MPI_COMM_NULL;

// MPI ranks of current core within various MPI communicators

int rank_main, rank_level, rank_samplers, rank_domain;

// MPI message tags

int TAG_DDM (const int d, const int p) { return d + p*NDIR; }

int TAG_FUNCTIONAL () { return STAT; }
int TAG_CV         () { return STAT; }
int TAG_XV         () { return STATS_COUNT + STAT; }

// parallelization configuration flags

const int INHERIT_RANK_ORDER = 0;

// === routines

void multi_init (int argc, char **argv) {
  
  // MPI initialization
  
  MPI_Init(&argc, &argv);
  
  // set 'rank_main'
  // must be before 'set_multi_level()'
  MPI_Comm_rank (MPI_COMM_WORLD, &rank_main);
  multi_main = (rank_main == 0);
  
  // obtain MULTI_TOTAL
  MPI_Comm_size (MPI_COMM_WORLD, &MULTI_TOTAL);
  
  // output number of cores
  if (multi_main) std::cout << " :: Simulation on " << MULTI_TOTAL << " core(s)" << std::endl << std::endl;
}

void multi_mlmc_init () {
  
  // compute default parallelization parameters MULTI[*][*][L]
  #if ! MULTI_CORES
  multi_initialize_default_parallelization_setup ();
  #else
  multi_initialize_default_parallelization_setup_from_MULTI_CORES ();
  #endif
  
  // check consistency in number of processes
  multi_check_core_consistency ();
  
  // report parallelization setup
  multi_report_parallelization_setup ();
  
  // compute actual parallelization parameters MULTI[*][*][*]
  multi_compute_parallelization_setup ();
  
  // compute comm_level_root_ranks[*]
  compute_comm_level_root_ranks ();
  
  // compute comm_level_last_ranks[*]
  compute_comm_level_last_ranks ();
  
  // set MULTI_LEVEL
  set_multi_level ();
  
  // create level communicators
  multi_create_level_communicators ();
}

void multi_sync () {
  MPI_Barrier (MPI_COMM_WORLD);
}

void multi_sync_domain () {
  MPI_Barrier (comm_domain);
}

void multi_create_level_communicators () {
  
  // create level communicators, which are used
  // for creation of 'comm_domain' and 'comm_samplers'
  
  MPI_Comm_split ( MPI_COMM_WORLD, MULTI_LEVEL, INHERIT_RANK_ORDER, &comm_level );
  MPI_Comm_rank (comm_level, &rank_level);
}

void multi_free_level_communicators () {
  MPI_Comm_free (&comm_level);
}

void multi_level_init () {
  
  // check if configuration is exactly the same as before
  int same_config = (MULTINl == MULTIN [MULTI_LEVEL][MESH_LEVEL] && MULTIMl == MULTI [M] [MULTI_LEVEL][MESH_LEVEL]);
  
  // set MULTI*l
  MULTIXl = MULTI[X][MULTI_LEVEL][MESH_LEVEL];
  MULTIYl = MULTI[Y][MULTI_LEVEL][MESH_LEVEL];
  MULTIZl = MULTI[Z][MULTI_LEVEL][MESH_LEVEL];
  MULTIMl = MULTI[M][MULTI_LEVEL][MESH_LEVEL];
  MULTINl = MULTIN  [MULTI_LEVEL][MESH_LEVEL];
  
  // if communicators have not been initialized yet
  // or if parallelization configuration is different
  // create domain and samplers communicators
  if ( !same_config || (comm_domain == MPI_COMM_NULL && comm_samplers == MPI_COMM_NULL) ) {
    
    // initialize parameters MULTI*l, etc.
    multi_level_init_parameters (MESH_LEVEL);
    
    // free existing communicators
    if (comm_domain != MPI_COMM_NULL && comm_samplers != MPI_COMM_NULL)
      multi_level_exit ();
    
    // create new communicators
    multi_create_domain_communicators ();
    multi_create_samplers_communicators ();
    
    // initialize multi_bc[NDIR*2], multi_bc_rank[NDIR*2]
    multi_init_bc_parameters ();
  }
}

void multi_level_init_parameters (const int mesh_level) {
  
  // set MULTI*l
  MULTIXl = MULTI[X][MULTI_LEVEL][mesh_level];
  MULTIYl = MULTI[Y][MULTI_LEVEL][mesh_level];
  MULTIZl = MULTI[Z][MULTI_LEVEL][mesh_level];
  MULTIMl = MULTI[M][MULTI_LEVEL][mesh_level];
  MULTINl = MULTIN  [MULTI_LEVEL][mesh_level];
  
  // setup domain
  rank_domain = rank_level % MULTINl;
  multi_domain_root = (rank_domain == 0);
  
  // set MULTI_SUBDOMAIN
  MULTI_SUBDOMAIN = rank_domain;
  
  // compute multix, multiy, multiz
  // must be after 'rank_domain' is computed
  // multix is the slowest index, multiz is the fastest index
  multix = rank_domain / (MULTIYl * MULTIZl);
  multiy = (rank_domain / MULTIZl) % MULTIYl;
  multiz = rank_domain % MULTIZl;
  
  // setup samplers
  rank_samplers = rank_level / MULTINl;
  multi_samplers_root = (rank_samplers == 0);
  MULTI_SAMPLER = rank_samplers;
}

void multi_level_exit () {
  // free existing domain and samplers communicators
  multi_free_domain_communicators ();
  multi_free_samplers_communicators ();
}

void multi_create_domain_communicators () {
  
  // create domain communicators
  // used in 'multi_bc()'
  
  int color = rank_level / MULTINl;
  MPI_Comm_split ( comm_level, color, INHERIT_RANK_ORDER, &comm_domain );
}

void multi_free_domain_communicators () {
  MPI_Comm_free (&comm_domain);
}

void multi_create_samplers_communicators () {
  
  // create samplers communicators
  
  MPI_Comm_split ( comm_level, rank_domain, INHERIT_RANK_ORDER, &comm_samplers );
}

void multi_free_samplers_communicators () {
  MPI_Comm_free (&comm_samplers);
}

void multi_init_bc_parameters () {
  
  // keep in mind, that multix is the slowest index and multiz is the fastest index
  
  if (BC_REPEAT_Z && MULTIZl > 1) {
    multi_bc[Z][HEAD] = 1;
    multi_bc_rank[Z][HEAD] = rank_domain - rank_domain % MULTIZl + (rank_domain + MULTIZl - 1) % MULTIZl;
    multi_bc[Z][TAIL] = 1;
    multi_bc_rank[Z][TAIL] = rank_domain - rank_domain % MULTIZl + (rank_domain + 1) % MULTIZl;
  }
  
  else {
    multi_bc[Z][HEAD] = (multiz > 0);
    multi_bc_rank[Z][HEAD] = rank_domain - 1;
    multi_bc[Z][TAIL] = (multiz < MULTIZl - 1);
    multi_bc_rank[Z][TAIL] = rank_domain + 1;
  }
  
  const int MYZ = MULTIYl * MULTIZl;
  
  if (BC_REPEAT_Y && MULTIYl > 1) {
    multi_bc[Y][HEAD] = 1;
    multi_bc_rank[Y][HEAD] = rank_domain - rank_domain % MYZ + (rank_domain + MYZ - MULTIZl) % MYZ;
    multi_bc[Y][TAIL] = 1;
    multi_bc_rank[Y][TAIL] = rank_domain - rank_domain % MYZ + (rank_domain + MULTIZl) % MYZ;
  }
  
  else {
    multi_bc[Y][HEAD] = (multiy > 0);
    multi_bc_rank[Y][HEAD] = rank_domain - MULTIZl;
    multi_bc[Y][TAIL] = (multiy < MULTIYl - 1);
    multi_bc_rank[Y][TAIL] = rank_domain + MULTIZl;
  }
  
  const int MXYZ = MULTIXl * MYZ;
  
  if (BC_REPEAT_X && MULTIXl > 1) {
    multi_bc[X][HEAD] = 1;
    multi_bc_rank[X][HEAD] = (rank_domain + MXYZ - MYZ) % MXYZ;
    multi_bc[X][TAIL] = 1;
    multi_bc_rank[X][TAIL] = (rank_domain + MYZ) % MXYZ;
  }
  
  else {
    multi_bc[X][HEAD] = (multix > 0);
    multi_bc_rank[X][HEAD] = rank_domain - MYZ;
    multi_bc[X][TAIL] = (multix < MULTIXl-1);
    multi_bc_rank[X][TAIL] = rank_domain + MYZ;
  }
}

// create indexed MPI datatype to map a generic 'sub-volume' of size (SUB_NX, SUB_NY, SUB_NZ)
// from a given volume with full resolution (including ghost cells) of (VNX, VNY, VNZ)
// 'real_count' specifies the number of 'reals' per each cell
// when using this datatype, buffer should point to the origin (0,0,0) of the desired 'sub-volume'
MPI_Datatype multi_create_indexed_MPI_datatype (const int SUB_NX, const int SUB_NY, const int SUB_NZ,
                                                const int VNX,    const int VNY,    const int VNZ,
                                                const int real_count) {
  
  // compute the number of contiguous blocks
  const int block_count = HAVE_DIR(Z) ? SUB_NX * SUB_NY: HAVE_DIR(Y) ? SUB_NX : 1;
  
  // compute block lengths
  int *block_lengths = (int*) malloc ( block_count * sizeof(int) );
  for (int b = 0; b < block_count; b++)
    block_lengths[b] = real_count * ( HAVE_DIR(Z) ? SUB_NZ: HAVE_DIR(Y) ? SUB_NY : SUB_NX );
  
  // compute displacements
  int *displacements = (int*) malloc ( block_count * sizeof(int) );
  displacements[0] = 0;
  int next_displacement;
  for (int b = 1; b < block_count; b++) {
    next_displacement = HAVE_DIR(Z) ? VNZ : HAVE_DIR(Y) ? VNY : 0;
    if ( HAVE_DIR(Z) && b % SUB_NY == 0 )
      next_displacement += ( VNY - SUB_NY ) * VNZ;
    displacements[b] = displacements[b-1] + real_count * next_displacement;
  }
  
  // create MPI datatype
  MPI_Datatype mpi_SUB_VOLUME;
  MPI_Type_indexed (block_count, block_lengths, displacements, mpi_REAL, &mpi_SUB_VOLUME);
  MPI_Type_commit (&mpi_SUB_VOLUME);
  
  // cleanup memory
  free (block_lengths);
  free (displacements);
  
  return mpi_SUB_VOLUME;
}

void multi_mesh_level_init () {
  
  timer_mpi_ddm.resume();
  
  // create mpi_SUB_VOLUME_BC_CV and mpi_SUB_VOLUME_BC_XV
  
  if ( multi_bc[X][HEAD] || multi_bc[X][TAIL] ) {
    mpi_SUB_VOLUME_BC_CV [X] = multi_create_indexed_MPI_datatype (GHOST_SIZE, NYl, NZl,  NXl+2*GX, NYl+2*GY, NZl+2*GZ,  SIZE_CV);
    mpi_SUB_VOLUME_BC_XV [X] = multi_create_indexed_MPI_datatype (GHOST_SIZE, NYl, NZl,  NXl+2*GX, NYl+2*GY, NZl+2*GZ,  SIZE_XV);
  }
  
  if ( multi_bc[Y][HEAD] || multi_bc[Y][TAIL] ) {
    mpi_SUB_VOLUME_BC_CV [Y] = multi_create_indexed_MPI_datatype (NXl, GHOST_SIZE, NZl,  NXl+2*GX, NYl+2*GY, NZl+2*GZ,  SIZE_CV);
    mpi_SUB_VOLUME_BC_XV [Y] = multi_create_indexed_MPI_datatype (NXl, GHOST_SIZE, NZl,  NXl+2*GX, NYl+2*GY, NZl+2*GZ,  SIZE_XV);
  }
  
  if ( multi_bc[Z][HEAD] || multi_bc[Z][TAIL] ) {
    mpi_SUB_VOLUME_BC_CV [Z] = multi_create_indexed_MPI_datatype (NXl, NYl, GHOST_SIZE,  NXl+2*GX, NYl+2*GY, NZl+2*GZ,  SIZE_CV);
    mpi_SUB_VOLUME_BC_XV [Z] = multi_create_indexed_MPI_datatype (NXl, NYl, GHOST_SIZE,  NXl+2*GX, NYl+2*GY, NZl+2*GZ,  SIZE_XV);
  }
  
  timer_mpi_ddm.pause();
}

void multi_mesh_level_exit () {
  
  timer_mpi_ddm.resume();
  
  // free mpi_SUB_VOLUME_BC_CV and mpi_SUB_VOLUME_BC_XV
  
  if ( multi_bc[X][HEAD] || multi_bc[X][TAIL] ) {
    MPI_Type_free (& mpi_SUB_VOLUME_BC_CV [X]);
    MPI_Type_free (& mpi_SUB_VOLUME_BC_XV [X]);
  }
  
  if ( multi_bc[Y][HEAD] || multi_bc[Y][TAIL] ) {
    MPI_Type_free (& mpi_SUB_VOLUME_BC_CV [Y]);
    MPI_Type_free (& mpi_SUB_VOLUME_BC_XV [Y]);
  }
  
  if ( multi_bc[Z][HEAD] || multi_bc[Z][TAIL] ) {
    MPI_Type_free (& mpi_SUB_VOLUME_BC_CV [Z]);
    MPI_Type_free (& mpi_SUB_VOLUME_BC_XV [Z]);
  }
  
  timer_mpi_ddm.pause();
}

void multi_ddm_start_send_recv_xtra () {
  
  timer_mpi_ddm.resume();
  
  // start sending (asynchronously) all multi boundary data, HEAD and TAIL
  
  for (int d = 0; d < NDIR; d++)
    if (multi_bc[d][HEAD])
      MPI_Isend ( & Xtra [0][0][0], 1, mpi_SUB_VOLUME_BC_XV [d], multi_bc_rank[d][HEAD], TAG_DDM(d,TAIL), comm_domain, &ddm_send_req[d][HEAD]);
  
  for (int d = 0; d < NDIR; d++)
    if (multi_bc[d][TAIL])
      MPI_Isend ( & Xtra [(NXl-GX)*DX[d]] [(NYl-GY)*DY[d]] [(NZl-GZ)*DZ[d]], 1, mpi_SUB_VOLUME_BC_XV [d], multi_bc_rank[d][TAIL], TAG_DDM(d,HEAD), comm_domain, &ddm_send_req[d][TAIL]);
  
  // start receiving (asynchronously) all multi boundary data, HEAD and TAIL
  
  for (int d = 0; d < NDIR; d++)
    if (multi_bc[d][HEAD])
      MPI_Irecv ( & Xtra [-GX*DX[d]] [-GY*DY[d]] [-GZ*DZ[d]], 1, mpi_SUB_VOLUME_BC_XV [d], multi_bc_rank[d][HEAD], TAG_DDM(d,HEAD), comm_domain, &ddm_recv_req[d][HEAD]);
  
  for (int d = 0; d < NDIR; d++)
    if (multi_bc[d][TAIL])
      MPI_Irecv ( & Xtra [NXl*DX[d]] [NYl*DY[d]] [NZl*DZ[d]], 1, mpi_SUB_VOLUME_BC_XV [d], multi_bc_rank[d][TAIL], TAG_DDM(d,TAIL), comm_domain, &ddm_recv_req[d][TAIL]);
  
  timer_mpi_ddm.pause();
}

real multi_dt (real dt)
{
  timer_mpi_ddm.resume();
  
  real dt_recv;
  MPI_Allreduce(&dt, &dt_recv, 1, mpi_REAL, MPI_MIN, comm_domain);
  
  timer_mpi_ddm.pause();
  
  return dt_recv;
}

void multi_ddm_start_send_recv () {
  
  timer_mpi_ddm.resume();
  
  // start sending (asynchronously) all multi boundary data, HEAD and TAIL
  
  for (int d = 0; d < NDIR; d++)
    if (multi_bc[d][HEAD])
      MPI_Isend ( & U [0][0][0], 1, mpi_SUB_VOLUME_BC_CV [d], multi_bc_rank[d][HEAD], TAG_DDM(d,TAIL), comm_domain, &ddm_send_req[d][HEAD]);
  
  for (int d = 0; d < NDIR; d++)
    if (multi_bc[d][TAIL])
      MPI_Isend ( & U [(NXl-GX)*DX[d]] [(NYl-GY)*DY[d]] [(NZl-GZ)*DZ[d]], 1, mpi_SUB_VOLUME_BC_CV [d], multi_bc_rank[d][TAIL], TAG_DDM(d,HEAD), comm_domain, &ddm_send_req[d][TAIL]);
  
  // start receiving (asynchronously) all multi boundary data, HEAD and TAIL
  
  for (int d = 0; d < NDIR; d++)
    if (multi_bc[d][HEAD])
      MPI_Irecv ( & U [-GX*DX[d]] [-GY*DY[d]] [-GZ*DZ[d]], 1, mpi_SUB_VOLUME_BC_CV [d], multi_bc_rank[d][HEAD], TAG_DDM(d,HEAD), comm_domain, &ddm_recv_req[d][HEAD]);
  
  for (int d = 0; d < NDIR; d++)
    if (multi_bc[d][TAIL])
      MPI_Irecv ( & U [NXl*DX[d]] [NYl*DY[d]] [NZl*DZ[d]], 1, mpi_SUB_VOLUME_BC_CV [d], multi_bc_rank[d][TAIL], TAG_DDM(d,TAIL), comm_domain, &ddm_recv_req[d][TAIL]);
  
  timer_mpi_ddm.pause();
}

void multi_ddm_confirm_recv () {
  
  timer_mpi_ddm.resume();
  
  // confirm the completion of the receives
  for (int d = 0; d < NDIR; d++)
    for (int p = 0; p < 2; p++)
      if (multi_bc[d][p])
        MPI_Wait (&ddm_recv_req[d][p], MPI_STATUS_IGNORE);
  
  timer_mpi_ddm.pause();
}

void multi_ddm_confirm_send () {
  
  timer_mpi_ddm.resume();
  
  // confirm the completion of the sends
  for (int d = 0; d < NDIR; d++)
    for (int p = 0; p < 2; p++)
      if (multi_bc[d][p])
        MPI_Wait (&ddm_send_req[d][p], MPI_STATUS_IGNORE);
  
  timer_mpi_ddm.pause();
}

#if STATS

real multi_reduce_functional (real value) {
  
  timer_mpi_ddm.resume();
  
  real value_recv;
  MPI_Allreduce (&value, &value_recv, 1, mpi_REAL, MPI_SUM, comm_domain);
  
  timer_mpi_ddm.pause();
  
  return value_recv;
}

int COUNT_CV (const int &multi_level, const int &mesh_level) {
  return STATS_SIZE * SIZE_CV * max ( 1, N[X][MESH_LEVEL] * N[Y][MESH_LEVEL] * N[Z][MESH_LEVEL] / MULTIN[multi_level][mesh_level] );
}

int COUNT_XV (const int &multi_level, const int &mesh_level) {
  return STATS_SIZE * SIZE_XV * max ( 1, N[X][MESH_LEVEL] * N[Y][MESH_LEVEL] * N[Z][MESH_LEVEL] / MULTIN[multi_level][mesh_level] );
}

int COUNT_FUNCTIONAL () {
  return STATS_SIZE;
}

#if L != 0 || MULTIM != 1

void multi_combine_mc_send () {
  
  // check if any samples were computed
  if ( SAMPLES[MULTI_LEVEL][MULTI_SAMPLER][MESH_LEVEL][TYPE].empty() )
    return;
  
  // === SEND partial MC estimates
  
  timer_mpi_mc.resume();
  
  // need to combine all snapshots
  for (SNAPSHOT = 0; SNAPSHOT < NSAVES+SAVE0; SNAPSHOT++)
    
    // need to combine all statistics
    for (STAT = 0; STAT < STATS_COUNT; STAT++)
      
      // check if samples should be sent for statistic STAT
      if (STATS_SEND_RECV [STAT]) {
        
        // send partial MC estimates to 'multi_samplers_root'
        
        #if STATS_FUNCTIONAL
        
        MPI_Send (stats_send_buffer_functional(), COUNT_FUNCTIONAL(), mpi_REAL, 0, TAG_FUNCTIONAL(), comm_samplers);
        
        #else
        
        MPI_Send (stats_send_buffer_CV(), COUNT_CV(MULTI_LEVEL,MESH_LEVEL), mpi_REAL, 0, TAG_CV(), comm_samplers);
        MPI_Send (stats_send_buffer_XV(), COUNT_XV(MULTI_LEVEL,MESH_LEVEL), mpi_REAL, 0, TAG_XV(), comm_samplers);
        
        #endif
      }
  
  timer_mpi_mc.pause();
}

void multi_combine_mc_recv () {
  
  // make sure statistical volumes are initialized for current MESH_LEVEL and TYPE
  if ( SAMPLES[MULTI_LEVEL][MULTI_SAMPLER][MESH_LEVEL][TYPE].empty() )
    stats_volumes_level_init();
  
  // receive partial MC estimtes from EACH non-multi_samplers_root core
  for (int multi_sampler=1; multi_sampler<MULTIMl; multi_sampler++) {
    
    // get the number of samples of the incoming partial MC estimate
    long_int weight = SAMPLES[MULTI_LEVEL][multi_sampler][MESH_LEVEL][TYPE].size();
    
    // check if any samples were used for the incoming MC estimate
    if ( weight == 0 )
      continue;
    
    // initialize receive buffers for statistics
    stats_recv_volumes_level_init();
    
    // === RECEIVE partial MC estimates
    
    // need to combine all snapshots
    for (SNAPSHOT = 0; SNAPSHOT < NSAVES+SAVE0; SNAPSHOT++) {
      
      timer_mpi_mc.resume();
      
      // need to combine all statistics
      for (STAT = 0; STAT < STATS_COUNT; STAT++)
        
        // check if samples should be received for statistic STAT
        if (STATS_SEND_RECV [STAT]) {
          
          // receive partial MC estimates from sampler 'multi_sampler'
          
          #if STATS_FUNCTIONAL
          
          MPI_Send (stats_recv_buffer_functional(), COUNT_FUNCTIONAL(), mpi_REAL, multi_sampler, TAG_FUNCTIONAL(), comm_samplers);
          
          #else
          
          MPI_Recv (stats_recv_buffer_CV(), COUNT_CV(MULTI_LEVEL,MESH_LEVEL), mpi_REAL, multi_sampler, TAG_CV(), comm_samplers, MPI_STATUS_IGNORE);
          MPI_Recv (stats_recv_buffer_XV(), COUNT_XV(MULTI_LEVEL,MESH_LEVEL), mpi_REAL, multi_sampler, TAG_XV(), comm_samplers, MPI_STATUS_IGNORE);
          
          #endif
        }
      
      timer_mpi_mc.pause();
      
      // combine partial MC estimates
      stats_combine (weight);
    }
  }
}

void multi_combine_mc_reduce() {
  
  // === initialize missing statistics volumes
  
  if ( ! multi_samplers_root && SAMPLES[MULTI_LEVEL][MULTI_SAMPLER][MESH_LEVEL][TYPE].empty() )
    for (STAT = 0; STAT < STATS_COUNT; STAT++)
      if (STATS_REDUCE [STAT]) {
        stats_volumes_level_init();
        break;
      }
  
  // === REDUCE partial MC estimates
  
  // custom send buffer for use of MPI_IN_PLACE
  void *mpi_send_buf;
  
  timer_mpi_mc.resume();
  
  // need to combine all snapshots
  for (SNAPSHOT = 0; SNAPSHOT < NSAVES+SAVE0; SNAPSHOT++)
    
    // need to combine all statistics
    for (STAT = 0; STAT < STATS_COUNT; STAT++)
      
      // check if samples should be reduced for statistic STAT
      if (STATS_REDUCE [STAT]) {
        
        // reduce partial MC estimates to 'multi_samplers_root'
        
        #if STATS_FUNCTIONAL
        
        mpi_send_buf = multi_samplers_root ? MPI_IN_PLACE : stats_send_buffer_functional();
        MPI_Reduce ( mpi_send_buf, stats_send_buffer_functional(), COUNT_FUNCTIONAL(), mpi_REAL, MPI_SUM, 0, comm_samplers );
        
        #else
        
        mpi_send_buf = multi_samplers_root ? MPI_IN_PLACE : stats_send_buffer_CV();
        MPI_Reduce ( mpi_send_buf, stats_send_buffer_CV(), COUNT_CV(MULTI_LEVEL,MESH_LEVEL), mpi_REAL, MPI_SUM, 0, comm_samplers );
        
        mpi_send_buf = multi_samplers_root ? MPI_IN_PLACE : stats_send_buffer_XV();
        MPI_Reduce ( mpi_send_buf, stats_send_buffer_XV(), COUNT_XV(MULTI_LEVEL,MESH_LEVEL), mpi_REAL, MPI_SUM, 0, comm_samplers );
        
        #endif
      }
  
  timer_mpi_mc.pause();
}

#endif

// combine all statistics from samplers to the multi_samplers_root cores
void multi_combine_mc () {
  
  #if L != 0 || MULTIM != 1
  
  #ifndef QUIET
  if (multi_main) std::cout << " :: Combining MC estimates from samplers...";
  if (multi_main) std::cout.flush();
  #endif
  
  // need to combine all mesh levels
  for ( MESH_LEVEL=0; MESH_LEVEL<=L; MESH_LEVEL++ ) {
    
    // initialize parallelization setup for current MESH_LEVEL
    multi_level_init ();
    
    // need to combine all types
    for (TYPE=0; TYPE<2; TYPE++)
      
      // check if there are any samples to combine
      if (MULTI[M][MULTI_LEVEL][MESH_LEVEL] > 1) {
        
        // check if any samples were computed (among ALL samplers)
        if ( SAMPLES_COUNT[MULTI_LEVEL][MESH_LEVEL][TYPE] == 0 )
          continue;
        
        // initialize parallelization setup for current MESH_LEVEL
        multi_level_init ();
        
        // non-multi_samplers_root cores SEND partial MC estimates to multi_samplers_root cores
        // within 'comm_samplers' communicator
        if ( ! multi_samplers_root ) multi_combine_mc_send();
        
        // multi_samplers_root cores RECEIVE partial MC estimates from non-multi_samplers_root cores
        // within 'comm_samplers' communicator
        else multi_combine_mc_recv();
        
        // REDUCE partial MC estimates (if needed)
        multi_combine_mc_reduce();
        
      } // END for (TYPE ...
  }
  
  #ifndef QUIET
  if (multi_main) std::cout << " done." << std::endl;
  #endif
  
  #endif
}

#if L != 0

void multi_combine_mlmc_send () {
  
  // for MULTI_LEVEL == L, check if the sending (rearrangement) is needed
  if ( MULTI_LEVEL == L && ( ! MULTI_MODIFIED [MULTI_LEVEL] [MESH_LEVEL] || STATS_FUNCTIONAL ) ) return;
  
  // check if any samples were used in the MC estimate on current multi_level
  if ( SAMPLES_COUNT [MULTI_LEVEL][MESH_LEVEL][TYPE] == 0 )
    return;
  
  #if STATS_FUNCTIONAL
  
  timer_mpi_mlmc.resume();
  
  // need to combine all snapshots
  for (SNAPSHOT=0; SNAPSHOT<NSAVES+SAVE0; SNAPSHOT++)
    
    // need to combine all statistics
    for (STAT = 0; STAT < STATS_COUNT; STAT++)
      
      // send MC estimates to multi_main
      MPI_Send ((real*) stats_send_buffer_functional(), COUNT_FUNCTIONAL(), mpi_REAL, 0, TAG_FUNCTIONAL(), MPI_COMM_WORLD);
  
  timer_mpi_mlmc.pause();
  
  #else
  
  // number of destinations (w.r.t. each direction)
  const int DESTN [NDIR] = {MULTI[X][L][L] / MULTIXl, MULTI[Y][L][L] / MULTIYl, MULTI[Z][L][L] / MULTIZl};
  
  // resolution of 'sub-volume' that needs to be sent
  // if N[*][MESH_LEVEL] < MULTI*, then the SAME single-cell 'sub-volume' is sent to MULTIPLE cores
  const int SUB_NX = max (1, N[X][MESH_LEVEL] / MULTI[X][L][L]);
  const int SUB_NY = max (1, N[Y][MESH_LEVEL] / MULTI[Y][L][L]);
  const int SUB_NZ = max (1, N[Z][MESH_LEVEL] / MULTI[Z][L][L]);
  
  // create indexed MPI datatypes to map a generic 'sub-volume' of a statistical volumes
  MPI_Datatype mpi_SUB_VOLUME_CV = multi_create_indexed_MPI_datatype (SUB_NX, SUB_NY, SUB_NZ, N[X][MESH_LEVEL] / MULTIXl, N[Y][MESH_LEVEL] / MULTIYl, N[Z][MESH_LEVEL] / MULTIZl, STATS_SIZE * SIZE_CV);
  MPI_Datatype mpi_SUB_VOLUME_XV = multi_create_indexed_MPI_datatype (SUB_NX, SUB_NY, SUB_NZ, N[X][MESH_LEVEL] / MULTIXl, N[Y][MESH_LEVEL] / MULTIYl, N[Z][MESH_LEVEL] / MULTIZl, STATS_SIZE * SIZE_XV);
  
  int destx, desty, destz, offset, dest;
  
  for (destx=0; destx < DESTN[X]; destx++)
  for (desty=0; desty < DESTN[Y]; desty++)
  for (destz=0; destz < DESTN[Z]; destz++) {
    
    // compute offset of the sub-volumes in the statistical volumes
    // if N[*][MESH_LEVEL] < MULTI*, then the SAME single-cell 'sub-volume' is sent to MULTIPLE cores
    offset = 0;
    offset += destx * (N[X][MESH_LEVEL] / MULTI[X][L][L]) * (N[Y][MESH_LEVEL] / MULTIYl) * (N[Z][MESH_LEVEL] / MULTIZl);
    offset += desty * (N[Y][MESH_LEVEL] / MULTI[Y][L][L]) * (N[Z][MESH_LEVEL] / MULTIZl);
    offset += destz * (N[Z][MESH_LEVEL] / MULTI[Z][L][L]);
    offset *= STATS_SIZE;
    
    // send all samples to the appropriate core on (the finest) MULTI_LEVEL = L
    dest = compute_ddm_rank_send ( destx + multix*DESTN[X], desty + multiy*DESTN[Y], destz + multiz*DESTN[Z] );
    
    timer_mpi_mlmc.resume();
    
    // need to combine all snapshots
    for (SNAPSHOT = 0; SNAPSHOT < NSAVES+SAVE0; SNAPSHOT++)
      
      // need to combine all statistics
      for (STAT = 0; STAT < STATS_COUNT; STAT++)
        
        // send MC estimates with dest != rank_main
        // using SYNCHRONOUS communication
        if (dest != rank_main) {
          MPI_Send ((real*) stats_send_buffer_CV() + offset * SIZE_CV, 1, mpi_SUB_VOLUME_CV, dest, TAG_CV(), MPI_COMM_WORLD);
          MPI_Send ((real*) stats_send_buffer_XV() + offset * SIZE_XV, 1, mpi_SUB_VOLUME_XV, dest, TAG_XV(), MPI_COMM_WORLD);
        }
        
        // for dest == rank_main, use ASYNCHRONOUS send WITHIN rank_main
        else {
          MPI_Request req_CV, req_XV;
          MPI_Isend ((real*) stats_send_buffer_CV() + offset * SIZE_CV, 1, mpi_SUB_VOLUME_CV, dest, TAG_CV(), MPI_COMM_WORLD, &req_CV);
          MPI_Isend ((real*) stats_send_buffer_XV() + offset * SIZE_XV, 1, mpi_SUB_VOLUME_XV, dest, TAG_XV(), MPI_COMM_WORLD, &req_XV);
        }
    
    timer_mpi_mlmc.pause();
    
  } // END FOR (destx, desty, destz...)
  
  #endif // STATS_FUNCTIONAL
}

void multi_combine_mlmc_recv () {
  
  // initialize receive buffers for statistics
  stats_recv_volumes_level_init();
  
  // receive partial MC estimtes from each multi_level (rearranging might be needed for multi_level == L)
  // IMPORTANT: one MUST start with receiving (rearranging) from multi_level=L to avoid data overriding
  for (int multi_level=L; multi_level>=LOWEST_MULTI_LEVEL; multi_level--) {
    
    // need to combine both types
    for (TYPE=0; TYPE<2; TYPE++) {
      
      // make sure statistical volumes are initialized for current MESH_LEVEL and TYPE
      if ( multi_level == L && SAMPLES_COUNT [MULTI_LEVEL][MESH_LEVEL][TYPE] == 0 )
        stats_volumes_level_init();
      
      // for multi_level == L, check if the receiving (rearrangement) is needed
      if ( multi_level == L && ( ! MULTI_MODIFIED [MULTI_LEVEL] [MESH_LEVEL] || STATS_FUNCTIONAL ) ) continue;
      
      // compute the number of samples in the partial MC estimate on multi_level
      long_int weight = SAMPLES_COUNT [multi_level][MESH_LEVEL][TYPE];
      
      // check if any samples were used in the partial MC estimate on multi_level
      if ( weight == 0 ) continue;
      
      // set source ranks
      const int source = compute_ddm_rank_recv ( multix, multiy, multiz, multi_level );
      
      // need to combine all snapshots
      for (SNAPSHOT=0; SNAPSHOT<NSAVES+SAVE0; SNAPSHOT++) {
        
        timer_mpi_mlmc.resume();
        
        // need to combine all statistics
        for (STAT = 0; STAT < STATS_COUNT; STAT++) {
          
          // receive MC estimates from required 'multi_level'
          
          #if STATS_FUNCTIONAL
          
          if (multi_main)
            MPI_Recv (stats_recv_buffer_functional(), COUNT_FUNCTIONAL(), mpi_REAL, source, TAG_FUNCTIONAL(), MPI_COMM_WORLD, MPI_STATUS_IGNORE);
          
          #else
          
          // receive MC estimates using SYNCHRONOUS communication
          MPI_Recv (stats_recv_buffer_CV(), COUNT_CV(MULTI_LEVEL,L), mpi_REAL, source, TAG_CV(), MPI_COMM_WORLD, MPI_STATUS_IGNORE);
          MPI_Recv (stats_recv_buffer_XV(), COUNT_XV(MULTI_LEVEL,L), mpi_REAL, source, TAG_XV(), MPI_COMM_WORLD, MPI_STATUS_IGNORE);
          
          #endif
        
        } // END FOR (STAT...
        
        timer_mpi_mlmc.pause();
        
        // for stats from multi_level == L
        // reinitialize local statistics and update them from the received statistics
        if (multi_level == L) {
          stats_volumes_snapshot_level_init ();
          stats_copy ();
        }
        
        // for stats from multi_level != L
        // combine local statistics with the received statistics
        else {
          stats_combine (weight);
          for (STAT = 0; STAT < STATS_COUNT; STAT++)
            if (STATS_REDUCE [STAT])
              stats_reduce ();
        }
        
      } // END FOR (SNAPSHOT...
    
    } // END FOR (TYPE...
  
  } // END FOR (multi_level...
}

#endif

void multi_combine_mlmc () {
  
  #if L != 0
  
  #ifndef QUIET
  if (multi_main) std::cout << " :: Combining MC estimates from levels...";
  if (multi_main) std::cout.flush();
  #endif
  
  // === for each MESH_LEVEL, call multi_level_init_setup_only() and SEND/RECV DATA
  
  for (MESH_LEVEL=0; MESH_LEVEL<=L; MESH_LEVEL++) {
    
    // initialize parallelization setup for sending (use current MESH_LEVEL)
    multi_level_init_parameters (MESH_LEVEL);
    
    // all information is in samplers_roots (and domain_roots, if STATS_FUNCTIONAL=1)
    if (multi_samplers_root && (!STATS_FUNCTIONAL || multi_domain_root) )
      
      // need to combine both types
      for (TYPE=0; TYPE<2; TYPE++)
        
        // SEND data to multi_level_root cores
        multi_combine_mlmc_send();
    
    // initialize parallelization setup for receiving (use MESH_LEVEL == L)
    multi_level_init_parameters (L);
    
    // only multi_level_root and multi_samplers_root receive samples (and domain_roots, if STATS_FUNCTIONAL=1)
    if ( multi_level_root && multi_samplers_root && (!STATS_FUNCTIONAL || multi_domain_root) )
        
      // RECEIVE data
      multi_combine_mlmc_recv();
  }
  
  #ifndef QUIET
  if (multi_main) std::cout << " done." << std::endl;
  #endif
  
  #endif
}

#endif

#if ESTIMATE_LOADS
void multi_balancer_gather_loads () {
  
  // sync all cores
  multi_sync();
  
  timer_mpi_bcast_loads.start();
  
  // compute the required buffer size, i.e. total number of loads
  #if DISTRIBUTE_IN_DIFFERENCES
  long_int BUFFER_SIZE = NMT_TYPE0;
  #else
  long_int BUFFER_SIZE = NMT;
  #endif
  
  // reserve space on each core
  LOADS.resize(BUFFER_SIZE);
  
  // === define MPI data type 'mpi_LOAD' as MPI analogue for 'Load'
  
  MPI_Datatype mpi_LOAD;
  
  const int num = 3;
  MPI_Datatype old_types [num] = {MPI_INT, mpi_LONG_INT, mpi_REAL};
  int blocklens          [num] = {2, 1, 1};
  MPI_Aint indices       [num];
  
  // compute 'indices'
  indices[0] = 0;
  MPI_Aint extent;
  for (int i = 1; i < num; i++) {
    MPI_Type_extent (old_types[i-1], &extent);
    indices[i] = indices[i-1] + blocklens[i-1]*extent;
  }
  
  // create MPI data type 'mpi_LOAD'
  MPI_Type_struct (num, blocklens, indices, old_types, &mpi_LOAD);
  
  // allocate MPI data type 'mpi_LOAD'
  MPI_Type_commit (&mpi_LOAD);
  
  // === gather loads from samplers in MULTI_LEVEL == L to multi_main
  
  #ifndef QUIET
  if (multi_main) std::cout << "  : -> Gathering loads from all samplers...";
  if (multi_main) std::cout.flush();
  #endif
  
  // custom send buffer for use of MPI_IN_PLACE
  void *mpi_send_buffer;
  
  if ( MULTI_LEVEL == L ) {
    
    // set parallelization setup for MESH_LEVEL = 0
    MESH_LEVEL == 0;
    multi_level_init ();
    
    if ( multi_domain_root ) {
      
      // set counts and displacements for MPI_Gatherv
      int counts [MAX_SAMPLERS], displacements [MAX_SAMPLERS];
      for ( int multi_sampler = 0; multi_sampler < MULTIMl; multi_sampler++ ) {
        counts [multi_sampler] = LOADS_COUNT [multi_sampler];
        if ( multi_sampler == 0 )
          displacements [multi_sampler] = 0;
        else
          displacements [multi_sampler] = displacements [multi_sampler - 1] + LOADS_COUNT [multi_sampler - 1];
      }
      
      // set send buffer
      mpi_send_buffer = multi_samplers_root ? MPI_IN_PLACE : LOADS.data();
      
      // commence!
      MPI_Gatherv (mpi_send_buffer, counts[MULTI_SAMPLER], mpi_LOAD, LOADS.data(), counts, displacements, mpi_LOAD, 0, comm_samplers);
    }
  }
  
  #ifndef QUIET
  if (multi_main) std::cout << " done." << std::endl;
  #endif
  
  // === broadcast loads from multi_main core to all cores
  
  #ifndef QUIET
  if (multi_main) std::cout << "  : -> Broadcasting loads to all cores...";
  if (multi_main) std::cout.flush();
  #endif
  
  // commence!
  MPI_Bcast ( LOADS.data(), BUFFER_SIZE, mpi_LOAD, 0, MPI_COMM_WORLD );
  
  #ifndef QUIET
  if (multi_main) std::cout << " done." << std::endl;
  #endif
  
  // === deallocate MPI data type 'mpi_LOAD'
  MPI_Type_free (&mpi_LOAD);
  
  timer_mpi_bcast_loads.stop();
}
#endif

void multi_cleanup () {
  
  #ifndef QUIET
  if (multi_main) std::cout << " :: Cleaning up...";
  if (multi_main) std::cout.flush();
  #endif
  
  // free existing communicators
  multi_free_level_communicators ();
  if (comm_domain != MPI_COMM_NULL && comm_samplers != MPI_COMM_NULL)
    multi_level_exit ();
  
  #ifndef QUIET
  if (multi_main) std::cout << " done." << std::endl;
  #endif
}

void multi_finalize() {
  MPI_Finalize();
}

double multi_time () {
  return MPI_Wtime();
}
