
/*
 * 1. Type definitions of 'real' and 'long_int'
 * 2. Definition of main dimension-dependent constants
 * 3. Declarations of various numerical types, e.g. vectors, matrices, etc.
 * 4. Declarations of various numerical functions
 */

#ifndef NUMTYPES_H
#define NUMTYPES_H

#include "config.h"

// === define type 'long_int' so that we can easily compile 'long' arrays for MC samples

#include <climits>

//typedef unsigned int long_int;
//static const unsigned int NM_MAX = UINT_MAX;

typedef unsigned long long_int;
static const unsigned long NM_MAX = ULONG_MAX;

// this data type has no analogue in MPI, hence NOT compatible with balance:adaptive,
// or any other balance that used COMPUTE_LOADS=1.
//typedef unsigned long long long_int;
//static const unsigned long long NM_MAX = ULLONG_MAX;

// === define real so that we can easily compile single or double precision versions of the code.

typedef double real;

// === dimension related constants

static const int X = 0, Y = 1, Z = 2, NDIR = 3;
static const int XU = 3, YU = 4, ZU = 5;
static const int DX[] = { 1, 0, 0 };
static const int DY[] = { 0, 1, 0 };
static const int DZ[] = { 0, 0, 1 };
static const int ACTIVE_DIRS = (NX>1) + (NY>1) + (NZ>1);

static const int MAX[] = {MAXX, MAXY, MAXZ};

static const int GX = GHOST_SIZE*(NX>1?1:0);
static const int GY = GHOST_SIZE*(NY>1?1:0);
static const int GZ = GHOST_SIZE*(NZ>1?1:0);

// === dimension related functions

static bool HAVE_DIR (int d) { return ( ( d==X?NX:(d==Y?NY:NZ) ) > 1 ); }

// === some useful inline numerical functions

#include <cmath>

static inline real sqr  (real n)         {return n*n;}
static inline real sign (real n)         {if (n<=0) return -1; else return 1;}
static inline real min  (real a, real b) {if (a < b) return a; else return b;}
static inline real max  (real a, real b) {if (a > b) return a; else return b;}
static inline real ssqrt(real x)         {return sqrt(max(0, x));}
static inline int  mod  (int a, int b)   {int r = a % b; return r < 0 ? r + b : r;}

template <class T> static inline void Zero        (T &in)       {in = 0;}
template <class T> static inline void Exp         (T &in)       {in = exp(in);}
template <class T> static inline void Log         (T &in)       {in = log(in);}
template <class T> static inline void Sqr         (T &in)       {in = sqr(in);}
template <class T> static inline void NonNegative (T &in)       {if (in < 0) in = 0;}

// === other

// define pi
static const real pi = 3.1415926535897931;

// used for structures to cast to arbitrary pointer to 'real' pointer
template <class T>
real* realCast (T &v) { return (real*)&v; }

// === generic 3d 'vector' type

template <class T>
class raw_vector {
  
  T v[NDIR];

public:
  
  // constructors
  
  raw_vector (const T x, const T y, const T z);
  raw_vector (const raw_vector &a);
  raw_vector ();
  
  // iterators
  
  const T& operator[] (int i) const;
        T& operator[] (int i);
  
  // various numerical functions
  
  T            sqr  ()  const;
  T            norm ()  const;
  raw_vector   abs  ()  const;
  raw_vector & zero ();
  
  // operator overloading
  
  raw_vector & operator += (const raw_vector &a);
  raw_vector & operator -= (const raw_vector &a);
  raw_vector & operator *= (const raw_vector &a);
  raw_vector & operator /= (const raw_vector &a);
  
  raw_vector & operator += (const real r);
  raw_vector & operator -= (const real r);
  raw_vector & operator *= (const real r);
  raw_vector & operator /= (const real r);
  
  raw_vector operator + (const raw_vector &a) const;
  raw_vector operator - (const raw_vector &a) const;
  raw_vector operator * (const real r) const;
  raw_vector operator / (const real r) const;
  
  // overloaded 'dot' product
  real operator * (const raw_vector &a) const;
};

// should stay in the header file
template <class T>
raw_vector<T> operator * (const real r, const raw_vector<T> &a) {
  return raw_vector<T>(a) *= r;
}

// vector of reals
typedef raw_vector<real> vector;

// === various distance functions

// evaluates the distances between two points
// if dir >= 0, only distance in the given direction 'dir' is returned
static real dist (const vector &x1, const vector &x2 = vector(0,0,0), const int dir = -1);

// evaluates the distances between two points in a periodic domain
// if dir >= 0, only distance in the given direction 'dir' is returned
static real dist_periodic (const vector &x1, const vector &x2 = vector(0,0,0), const int dir = -1);

// must be here in order to have arbitrary templated objects and inlining
#include "numtypes.cpp"

#endif
