
#include "solver.h"

#include "equation.h"
#include "solver-util-hll.h"
#include "powell.h"

///::title:: HLL3 solver ("second" 3 wave) | for MHD
///::depends:: solver-mhd-hll-util
///::consts:: HLL_dt=1
///::vars:: POWELL=1

// variables for local wave speeds
real sl, sr, cs;

// vectors for velocities and magnetic fields
vector us, Bs;

void solver_calc_flux (ConservedVars &F,
		                   const AllVariables &lV, const AllVariables &rV,
		                   int i, int j, int k, int d)
{
  ConservedVars lF, rF, Us;
  
  hll_speeds_ext(lV, rV, d, 1e-5, sl, sr, cs);
  calc_point_flux(lF, lV, d);
  calc_point_flux(rF, rV, d);

  real ptl = lV.XU.p + 0.5*lV.U.B.sqr();
  real ptr = rV.XU.p + 0.5*rV.U.B.sqr();

  real udl = lV.XU.u[d]-sl;
  real udr = rV.XU.u[d]-sr;

  real aa = udr*rV.U.rho - udl*lV.U.rho;
  real bb = rV.U.B[d] - lV.U.B[d];
  real cc = sl - sr;

  real sm = (rV.U.m[d]*udr - lV.U.m[d]*udl + ptr - ptl - (rV.U.B[d]*rV.U.B[d] - lV.U.B[d]*lV.U.B[d])/2)/aa;

  vector Gm = rF.m - lF.m - sr*rV.U.m + sl*lV.U.m;
  vector GB = rF.B - lF.B - sr*rV.U.B + sl*lV.U.B;
  real acmbb = aa*cc-bb*bb;
  
  us = (cc*Gm + bb*GB)/acmbb;
  us[d] = sm;
  Bs = (bb*Gm + aa*GB)/acmbb;
  Us.B = Bs;

  if (sl == 0) Zero(F);
  else if (sl > 0) {F = lF; Bs[d] = lV.U.B[d];}
  else if (sr < 0) {F = rF; Bs[d] = rV.U.B[d];}
  else if (sm >= 0) {
    Us.rho = lV.U.rho*udl/(sm-sl);
    Us.m = Us.rho * us;
    Us.B[d] = lV.U.B[d];
    real pts = ptl + lV.U.rho*(lV.XU.u[d]-sm)*udl;
    Us.E = (udl*lV.U.E + ptl*lV.XU.u[d] - pts*sm + lV.U.B[d]*(Us.B*us - lV.U.B*lV.XU.u)) / (sm-sl);
    F = lF + sl*(Us-lV.U);
  } else {
    Us.rho = rV.U.rho*udr/(sm-sr);
    Us.m = Us.rho * us;
    Us.B[d] = rV.U.B[d];
    real pts = ptr + rV.U.rho*(rV.XU.u[d]-sm)*udr;
    Us.E = (udr*rV.U.E + ptr*rV.XU.u[d] - pts*sm + rV.U.B[d]*(Us.B*us - rV.U.B*rV.XU.u)) / (sm-sr);
    F = rF + sr*(Us-rV.U);
  }
  
  if (POWELL) powell_source (us, Bs, lV, rV, i, j, k, d);
}

real solver_calc_max_speed (const AllVariables &lV, const AllVariables &rV,
                            const int i, const int j, const int k, const int d)
{
  hll_speeds (lV, rV, d, sl, sr, cs);
  return max ( fabs(sl), fabs(sr) );
}
