
/*
 * Declarations STATISTICS routines:
 * Definitions of all declared functions are available in corresponding 'stats_*.cpp' file
 */

#ifndef STATS_H
#define STATS_H

#include "structures.h"
#include "volumes.h"

#include <vector>

// indicates active statistic
extern int STAT;

// switches for statistical communication
extern int STATS_SEND_RECV [STATS_COUNT];
extern int STATS_REDUCE    [STATS_COUNT];

// general STATS routines, implemented in 'stats.cpp' file

// assemble MLMC estimate for a given class T
template <class T>
void stats_assemble_mlmc (T (&a) [L+1][2] [NSAVES+SAVE0]);

// initializes statistical volumes (e.g. mean, variance, kurtosis, histograms, etc.) for currently set MESH_LEVEL and TYPE, for ALL snapshots
void stats_volumes_level_init ();

// STATS-specific routines, implemented in corresponding 'stats_*.cpp' file

// initializes arrays for the statistical routines
void stats_init();

// initializes statistical volumes (e.g. mean, variance, kurtosis, histograms, etc.) for currently set MESH_LEVEL, TYPE and SNAPSHOT
void stats_volumes_snapshot_level_init ();

// initializes statistical volumes for receive buffers required in parallel implementations
void stats_recv_volumes_level_init ();

// copies statistical volumes from the receive buffers to statistics volumes for currently set MESH_LEVEL, TYPE and SNAPSHOT
void stats_copy ();

// updates statistics (e.g. mean, variance, threshold)
// i.e. appends an additional (just computed) sample (U, Xtra)
void stats_update();

// finalizes (intermediate) statistical estimates
// i.e. division by number of samples, normalization, etc.
// executed AFTER all 'stats_update()' and BEFORE all 'stats_combine()' routines
void stats_finalize_updates();

// returns pointer to the 'send' buffer for MC estimates
// for ConservedVars with specified stat_index in 0, ..., STATS_COUNT-1
// needs to be defined for STATS_FUNCTIONAL=0
void* stats_send_buffer_CV ();

// returns pointer to the 'send' buffer for MC estimates
// for XtraVars with specified stat_index in 0, ..., STATS_COUNT-1
// needs to be defined for STATS_FUNCTIONAL=0
void* stats_send_buffer_XV ();

// returns pointer to the 'send' buffer for MC estimates
// for functional with specified stat_index in 0, ..., STATS_COUNT-1
// needs to be defined for STATS_FUNCTIONAL=1
void* stats_send_buffer_functional ();

// returns pointer to the 'receive' buffer for MC estimates
// for ConservedVars with specified stat_index in 0, ..., STATS_COUNT-1
// needs to be defined for STATS_FUNCTIONAL=0
void* stats_recv_buffer_CV ();

// returns pointer to the 'receive' buffer for MC estimates
// for XtraVars with specified stat_index in 0, ..., STATS_COUNT-1
// needs to be defined for STATS_FUNCTIONAL=0
void* stats_recv_buffer_XV ();

// returns pointer to the 'receive' buffer for MC estimates
// for functional with specified stat_index in 0, ..., STATS_COUNT-1
// needs to be defined for STATS_FUNCTIONAL=1
void* stats_recv_buffer_functional ();

// combines statistics (e.g. mean, variance, threshold)
// from different samplers to the 'multi_samplers_root'
// 'weight' is the weight (number of sample) of the ensemble to be appended
void stats_combine (const real weight);

// values from 'stats_recv_buffer' are added (+) to values of 'stats_send_buffer'
void stats_reduce ();

// finalizes statistical estimates
// i.e. division by number of samples, normalization, etc.
// executed AFTER all 'stats_combine()' routines
void stats_finalize_combines ();

// applies constraints for mean and variance
void stats_apply_constraints ();

// assemble MLMC estimate from combined MC estimates
void stats_assemble_mlmc ();

// write statistical datafiles for the set MESH_LEVEL and TYPE
void stats_write_datafiles ();

// === must be here in order to have arbitrary templated routines

template <class T>
void stats_assemble_mlmc (T (&a) [L+1][2] [NSAVES+SAVE0]) {
  
  for (int MESH_LEVEL=L-1; MESH_LEVEL>=0; MESH_LEVEL--)
    
    for (int SNAPSHOT=0; SNAPSHOT<NSAVES+SAVE0; SNAPSHOT++) {
      
      // adding TYPE=0 MC estimates to the MLMC estimate
      a [L][0] [SNAPSHOT] += a [MESH_LEVEL][0] [SNAPSHOT];
      
      // subtracting TYPE=1 MC estimates from the MLMC estimate
      a [L][0] [SNAPSHOT] -= a [MESH_LEVEL][1] [SNAPSHOT];
    }
}

#endif
