
///::title:: Histograms at each domain point
///::consts:: STATS_HISTOGRAMS=1 STATS_COUNT=1 STATS_SIZE=NBINS STATS_FUNCTIONAL=0
///::vars:: NBINS=512 HISTOGRAM_MIN=0.5 HISTOGRAM_MAX=3.5

#include "stats.h"
#include "stats-histograms.h"

#if STATS

#include "mlmc.h"
#include "multi.h"
#include "saver.h"

// volumes (no ghost cells) for MC estimates of histograms (for each snapshot)
// iterators [mesh_level][type] indicate estimate of type 'type' at mesh level 'mesh_level'
volume < histogram <ConservedVars> > histogram_U [L+1][2] [NSAVES+SAVE0];
volume < histogram <XtraVars> >      histogram_X [L+1][2] [NSAVES+SAVE0];

void* stats_send_buffer_CV () {
  return histogram_U [MESH_LEVEL][TYPE][SNAPSHOT] .data;
}

void* stats_send_buffer_XV () {
  return histogram_X [MESH_LEVEL][TYPE][SNAPSHOT] .data;
}

// volumes for receive buffers required in parallel implementations

volume < histogram <ConservedVars> > recv_histogram_U;
volume < histogram <XtraVars> >      recv_histogram_X;

void* stats_recv_buffer_CV () {
  return recv_histogram_U.data;
}

void* stats_recv_buffer_XV () {
  return recv_histogram_X.data;
}

void stats_init () {
  histogram_init ();
  STATS_SEND_RECV[0] = 0;
  STATS_REDUCE[0] = 1;
}

void stats_volumes_snapshot_level_init () {
  
  // resizing
  histogram_U [MESH_LEVEL][TYPE] [SNAPSHOT] .resize (max(1, N[X][MESH_LEVEL] / MULTIXl), max(1, N[Y][MESH_LEVEL] / MULTIYl), max(1, N[Z][MESH_LEVEL] / MULTIZl), 0);
  histogram_X [MESH_LEVEL][TYPE] [SNAPSHOT] .resize (max(1, N[X][MESH_LEVEL] / MULTIXl), max(1, N[Y][MESH_LEVEL] / MULTIYl), max(1, N[Z][MESH_LEVEL] / MULTIZl), 0);
  
  // setting to 0
  histogram_U [MESH_LEVEL][TYPE] [SNAPSHOT] .zero();
  histogram_X [MESH_LEVEL][TYPE] [SNAPSHOT] .zero();
}

void stats_recv_volumes_level_init () {
  recv_histogram_U .resize (max(1, N[X][MESH_LEVEL] / MULTIXl), max(1, N[Y][MESH_LEVEL] / MULTIYl), max(1, N[Z][MESH_LEVEL] / MULTIZl), 0);
  recv_histogram_X .resize (max(1, N[X][MESH_LEVEL] / MULTIXl), max(1, N[Y][MESH_LEVEL] / MULTIYl), max(1, N[Z][MESH_LEVEL] / MULTIZl), 0);
}

void stats_copy () {
  histogram_U [MESH_LEVEL][TYPE] [SNAPSHOT] += recv_histogram_U;
  histogram_X [MESH_LEVEL][TYPE] [SNAPSHOT] += recv_histogram_X;
}

// templated histogram update function
template<class T>
void histogram_update (volume < histogram<T> > &hist, const volume <T> &sample) {
  for (int i=0; i<hist.NXl; i++)
    for (int j=0; j<hist.NYl; j++)
      for (int k=0; k<hist.NZl; k++)
        hist [i][j][k] += sample [i][j][k];
}

void stats_update () {
  
  // update histogram for ConservedVars
  histogram_update (histogram_U [MESH_LEVEL][TYPE][SNAPSHOT], U);
  
  // update histogram for XtraVars
  histogram_update (histogram_X [MESH_LEVEL][TYPE][SNAPSHOT], Xtra);
}

void stats_finalize_updates () {}

void stats_combine (const real weight) {  
  // REDUCE_MEAN=1 is sufficient
}

void stats_reduce () {
  histogram_U [MESH_LEVEL][TYPE] [SNAPSHOT] += recv_histogram_U;
  histogram_X [MESH_LEVEL][TYPE] [SNAPSHOT] += recv_histogram_X;
}

void stats_finalize_combines () {
  // computing empirical probability distribution function (integral should be equal to 1)
  histogram_U [MESH_LEVEL][TYPE] [SNAPSHOT] /= NM[SAMPLES_LEVEL()] * WIDTH;
  histogram_X [MESH_LEVEL][TYPE] [SNAPSHOT] /= NM[SAMPLES_LEVEL()] * WIDTH;
}

void stats_assemble_mlmc () {
  stats_assemble_mlmc (histogram_U);
  stats_assemble_mlmc (histogram_X);
}

void stats_apply_constraints () {
  
  // histogram values should be non-negative
  histogram_U [L][0][SNAPSHOT] .nonnegative();
  histogram_X [L][0][SNAPSHOT] .nonnegative();
  
  // probability mass at each bin should not exceed 1
  real MAX_VALUE = 1 / WIDTH;
  for (int i = 0; i < histogram_U[L][0][SNAPSHOT].NXl; i++)
    for (int j = 0; j < histogram_U[L][0][SNAPSHOT].NYl; j++)
      for (int k = 0; k < histogram_U[L][0][SNAPSHOT].NZl; k++)
        for (int b = 0; b < NBINS; b++) {
          for (int m = 0; m < SIZE_CV; m++)
            if ( realCast ( histogram_U [L][0][SNAPSHOT] [i][j][k] [b] ) [m] > MAX_VALUE )
              realCast ( histogram_U [L][0][SNAPSHOT] [i][j][k] [b] ) [m] = MAX_VALUE;
          for (int m = 0; m < SIZE_XV; m++)
            if ( realCast ( histogram_X [L][0][SNAPSHOT] [i][j][k] [b] ) [m] > MAX_VALUE )
              realCast ( histogram_X [L][0][SNAPSHOT] [i][j][k] [b] ) [m] = MAX_VALUE;
        }
}

void stats_write_datafiles (const int &mesh_level, const int &type) {  
  Saver saver ("histograms_");
  saver_mean.snapshots (&histogram_U[mesh_level][type][0], &histogram_X[mesh_level][type][0]);
}

#endif
