
///::title:: Mean and variance of a functional (stable version, Schwab)
///::consts:: STATS_MEAN_VAR_FUNCTIONAL=1 STATS_COUNT=2 STATS_SIZE=1 STATS_FUNCTIONAL=1
///::vars:: VARIANCE_FIX=1 STATS_FUNCTIONAL_INTEGRAND=CV[i][j][k].u*pow(x-MAXX/2,3)

#include "stats.h"

#if STATS

#include "mlmc.h"
#include "multi.h"
#include "saver.h"

// structures for MC estimates of mean and variance (for each snapshot)
// iterators [mesh_level][type] indicate estimate of type 'type' at mesh level 'mesh_level'

real mean     [L+1][2] [NSAVES+SAVE0];
real variance [L+1][2] [NSAVES+SAVE0];

void* stats_send_buffer_functional () {
  switch (STAT) {
    case 0: return &mean     [MESH_LEVEL][TYPE][SNAPSHOT];
    case 1: return &variance [MESH_LEVEL][TYPE][SNAPSHOT];
  }
}

// structures for receive buffers required in parallel implementations

real recv_mean;
real recv_variance;

void* stats_recv_buffer_functional () {
  switch (STAT) {
    case 0: return &recv_mean;
    case 1: return &recv_variance;
  }
}

// WEIGHTS[mesh_level][type]
// number of type ='type' samples computed on the mesh level 'mesh_level' so far
long_int WEIGHTS[L+1][2];

void stats_init () {
  
  WEIGHTS[MESH_LEVEL][TYPE] = 0;
  
  STATS_SEND_RECV[0] = 1;
  STATS_SEND_RECV[1] = 0;
  
  STATS_REDUCE[0] = 0;
  STATS_REDUCE[1] = 1;
}

real stats_apply_functional (volume<ConservedVars> &CV, volume<XtraVars> &XV) {
  
  real value = 0;
  real x, y, z;
  
  for (int i = 0; i < NXl; i++)
    for (int j = 0; j < NYl; j++)
      for (int k = 0; k < NZl; k++) {
        x = (i + NXl*multix + 0.5) * dx[X];
        y = (j + NYl*multiy + 0.5) * dx[Y];
        z = (k + NZl*multiz + 0.5) * dx[Z];
        value += STATS_FUNCTIONAL_INTEGRAND;
      }
  
  return value * MAXX * MAXY * MAXZ / ( NXl_FULL * NYl_FULL * NZl_FULL );
}

void stats_volumes_snapshot_level_init () {}
void stats_recv_volumes_level_init     () {}
void stats_copy                        () {}

// numerically stable variance computation algorithm from NumODE course by Ch. Schwab
template<class T>
void variance_update (const T &mean, T &variance, const T &sample, const long_int &i) {
  
  if ( i==1 ) return;
  
  T delta (sample);
  delta -= mean;
  delta *= delta;
  delta *= (1 - 1.0/i);
  variance += delta;
}

// numerically stable mean computation algorithm by Chan et al.
template<class T>
void mean_update (T &mean, const T &sample, const long_int &i) {
  T delta (sample);
  delta -= mean;
  delta /= i;
  mean += delta;
}

void stats_update () {
  
  // update WEIGHTS (during the FIRST snapshot)
  if ( SNAPSHOT == FIRST_SNAPSHOT )
    WEIGHTS[MESH_LEVEL][TYPE]++;
  
  // compute value of the functional
  real value = multi_reduce_functional ( stats_apply_functional (U, Xtra) );
  
  // update mean and variance
  
  variance_update (mean[MESH_LEVEL][TYPE][SNAPSHOT], variance[MESH_LEVEL][TYPE][SNAPSHOT], value, WEIGHTS[MESH_LEVEL][TYPE]);
  
  mean_update (mean[MESH_LEVEL][TYPE][SNAPSHOT], value, WEIGHTS[MESH_LEVEL][TYPE]);
}

void stats_finalize_updates () {}

// numerically stable variance combination algorithm by Chan et al.
// combined variance is stored in 'var1'
// for the complete combination, second variance 'var2' should be added to 'var1'
// i.e. STATS_REDUCE[1] = 1 is assumed
template<class T>
void variance_combine (const T &mean1, T &var1, const real &w1, const T &mean2, const real &w2) {
  
  if ( w1 == 0 || w2 == 0 ) return;
  
  T delta (mean2);
  delta -= mean1;
  delta *= delta;
  delta *= (w1 * w2) / (w1 + w2);
  var1 += delta;
}

// numerically stable mean computation algorithm by Chan et al.
// combined mean is stored in 'mean1'
template<class T>
void mean_combine (T &mean1, const real &w1, const T &mean2, const real &w2) {
  
  // if w2 is much smaller than w1
  if ( w2 < 0.01 * w1 ) {
    T delta (mean2);
    delta -= mean1;
    delta /= w2 / (w1 + w2);
    mean1 += delta;
  }
  
  // if w1 ~ w2 and both large
  else {
    mean1 *= w1 / (w1 + w2);
    mean1 += mean2 * ( w2 / (w1 + w2) );
  }
}

void stats_combine (const real weight) {
  
  // combine mean and variance
  
  variance_combine (mean[MESH_LEVEL][TYPE][SNAPSHOT], variance[MESH_LEVEL][TYPE][SNAPSHOT], WEIGHTS[MESH_LEVEL][TYPE], recv_mean, weight);
  
  mean_combine (mean[MESH_LEVEL][TYPE][SNAPSHOT], WEIGHTS[MESH_LEVEL][TYPE], recv_mean, weight);
  
  // update the total number of samples (during the LAST snapshot)
  if ( SNAPSHOT == LAST_SNAPSHOT )
    WEIGHTS[MESH_LEVEL][TYPE] += weight;
}

void stats_reduce () {
  switch (STAT) {
    case 0:
      mean [MESH_LEVEL][TYPE] [SNAPSHOT] += recv_mean;
    case 1:
      variance [MESH_LEVEL][TYPE] [SNAPSHOT] += recv_variance;
  }
}

void stats_finalize_combines () {
  
  // computing variance
  if (NM[SAMPLES_LEVEL()] > 2) variance [MESH_LEVEL][TYPE] [SNAPSHOT] /= NM[SAMPLES_LEVEL()] - 1;
}

void stats_assemble_mlmc () {
  
  // assemble MLMC estimate of mean
  stats_assemble_mlmc (mean);
  
  // assemble MLMC estimate of variance
  stats_assemble_mlmc (variance);
}

void stats_apply_constraints () {
  
  // variance should be non-negative
  #if VARIANCE_FIX
  NonNegative (variance [L][0][SNAPSHOT]);
  #endif
}

void stats_write_datafiles () {
  
  if ( multi_domain_root ) {
    // MLMC mean estimates are in mean_U[L][0], mean_X[L][0]
    Saver saver_mean ("");
    saver_mean.snapshots (&mean[L][0][0]);
    
    // MLMC variance estimates are in variance_U[L][0], variance_X[L][0]
    Saver saver_var ("var_");
    saver_var.snapshots (&variance[L][0][0]);
  }
}

#endif
