
///::title:: Mean and variance (stable version, Schwab)
///::consts:: STATS_MEAN_VAR=1 STATS_COUNT=2 STATS_SIZE=1 STATS_FUNCTIONAL=0
///::vars:: VARIANCE_FIX=1

#include "stats.h"

#if STATS

#include "mlmc.h"
#include "multi.h"
#include "saver.h"

// volumes for MC estimates of mean and variance (for each snapshot)
// iterators [mesh_level][type] indicate estimate of type 'type' at mesh level 'mesh_level'

volume <ConservedVars> mean_U [L+1][2] [NSAVES+SAVE0];
volume <XtraVars>      mean_X [L+1][2] [NSAVES+SAVE0];

volume <ConservedVars> variance_U  [L+1][2] [NSAVES+SAVE0];
volume <XtraVars>      variance_X  [L+1][2] [NSAVES+SAVE0];

void* stats_send_buffer_CV () {
  switch (STAT) {
    case 0: return mean_U     [MESH_LEVEL][TYPE][SNAPSHOT] .data;
    case 1: return variance_U [MESH_LEVEL][TYPE][SNAPSHOT] .data;
  }
}

void* stats_send_buffer_XV () {
  switch (STAT) {
    case 0: return mean_X     [MESH_LEVEL][TYPE][SNAPSHOT] .data;
    case 1: return variance_X [MESH_LEVEL][TYPE][SNAPSHOT] .data;
  }
}

// volumes for receive buffers required in parallel implementations

volume <ConservedVars> recv_mean_U;
volume <XtraVars>      recv_mean_X;

volume <ConservedVars> recv_variance_U;
volume <XtraVars>      recv_variance_X;

void* stats_recv_buffer_CV () {
  switch (STAT) {
    case 0: return recv_mean_U     .data;
    case 1: return recv_variance_U .data;
  }
}

void* stats_recv_buffer_XV () {
  switch (STAT) {
    case 0: return recv_mean_X     .data;
    case 1: return recv_variance_X .data;
  }
}

// WEIGHTS[mesh_level][type]
// number of type ='type' samples computed on the mesh level 'mesh_level' so far
long_int WEIGHTS[L+1][2];

void stats_init () {
  
  WEIGHTS[MESH_LEVEL][TYPE] = 0;
  
  STATS_SEND_RECV[0] = 1;
  STATS_SEND_RECV[1] = 0;
  
  STATS_REDUCE[0] = 0;
  STATS_REDUCE[1] = 1;
}

void stats_volumes_snapshot_level_init () {
  
  mean_U [MESH_LEVEL][TYPE] [SNAPSHOT] .reinit (max(1, N[X][MESH_LEVEL] / MULTIXl), max(1, N[Y][MESH_LEVEL] / MULTIYl), max(1, N[Z][MESH_LEVEL] / MULTIZl), 0);
  mean_X [MESH_LEVEL][TYPE] [SNAPSHOT] .reinit (max(1, N[X][MESH_LEVEL] / MULTIXl), max(1, N[Y][MESH_LEVEL] / MULTIYl), max(1, N[Z][MESH_LEVEL] / MULTIZl), 0);
  
  variance_U [MESH_LEVEL][TYPE] [SNAPSHOT] .reinit (max(1, N[X][MESH_LEVEL] / MULTIXl), max(1, N[Y][MESH_LEVEL] / MULTIYl), max(1, N[Z][MESH_LEVEL] / MULTIZl), 0);
  variance_X [MESH_LEVEL][TYPE] [SNAPSHOT] .reinit (max(1, N[X][MESH_LEVEL] / MULTIXl), max(1, N[Y][MESH_LEVEL] / MULTIYl), max(1, N[Z][MESH_LEVEL] / MULTIZl), 0);
}

void stats_recv_volumes_level_init () {
  
  recv_mean_U .resize (max(1, N[X][MESH_LEVEL] / MULTIXl), max(1, N[Y][MESH_LEVEL] / MULTIYl), max(1, N[Z][MESH_LEVEL] / MULTIZl), 0);
  recv_mean_X .resize (max(1, N[X][MESH_LEVEL] / MULTIXl), max(1, N[Y][MESH_LEVEL] / MULTIYl), max(1, N[Z][MESH_LEVEL] / MULTIZl), 0);
  
  recv_variance_U .resize (max(1, N[X][MESH_LEVEL] / MULTIXl), max(1, N[Y][MESH_LEVEL] / MULTIYl), max(1, N[Z][MESH_LEVEL] / MULTIZl), 0);
  recv_variance_X .resize (max(1, N[X][MESH_LEVEL] / MULTIXl), max(1, N[Y][MESH_LEVEL] / MULTIYl), max(1, N[Z][MESH_LEVEL] / MULTIZl), 0);
}

void stats_copy () {
  
  mean_U [MESH_LEVEL][TYPE] [SNAPSHOT] += recv_mean_U;
  mean_X [MESH_LEVEL][TYPE] [SNAPSHOT] += recv_mean_X;
  
  variance_U [MESH_LEVEL][TYPE] [SNAPSHOT] += recv_variance_U;
  variance_X [MESH_LEVEL][TYPE] [SNAPSHOT] += recv_variance_X;
}

// numerically stable variance computation algorithm from NumODE course by Ch. Schwab
template<class T>
void variance_update (const T &mean, T &variance, const T &sample, const long_int &i) {
  
  if ( i==1 ) return;
  
  T delta (sample);
  delta -= mean;
  delta.sqr();
  delta *= (1 - 1.0/i);
  variance += delta;
}

// numerically stable mean computation algorithm by Chan et al.
template<class T>
void mean_update (T &mean, const T &sample, const long_int &i) {
  T delta (sample);
  delta -= mean;
  delta /= i;
  mean += delta;
}

void stats_update () {
  
  // update WEIGHTS (during the FIRST snapshot)
  if ( SNAPSHOT == FIRST_SNAPSHOT )
    WEIGHTS[MESH_LEVEL][TYPE]++;
  
  // update mean and variance for ConservedVars
  
  variance_update (mean_U[MESH_LEVEL][TYPE][SNAPSHOT], variance_U[MESH_LEVEL][TYPE][SNAPSHOT], U, WEIGHTS[MESH_LEVEL][TYPE]);
  
  mean_update (mean_U[MESH_LEVEL][TYPE][SNAPSHOT], U, WEIGHTS[MESH_LEVEL][TYPE]);
  
  // update mean and variance for XtraVars
  
  variance_update (mean_X[MESH_LEVEL][TYPE][SNAPSHOT], variance_X[MESH_LEVEL][TYPE][SNAPSHOT], Xtra, WEIGHTS[MESH_LEVEL][TYPE]);
  
  mean_update (mean_X[MESH_LEVEL][TYPE][SNAPSHOT], Xtra, WEIGHTS[MESH_LEVEL][TYPE]);
}

void stats_finalize_updates () {}

// numerically stable variance combination algorithm by Chan et al.
// combined variance is stored in 'var1'
// for the complete combination, second variance 'var2' should be added to 'var1'
// i.e. STATS_REDUCE[1] = 1 is assumed
template<class T>
void variance_combine (const T &mean1, T &var1, const real &w1, const T &mean2, const real &w2) {
  
  if ( w1 == 0 || w2 == 0 ) return;
  
  T delta (mean2);
  delta -= mean1;
  delta.sqr();
  delta *= (w1 * w2) / (w1 + w2);
  var1 += delta;
}

// numerically stable mean computation algorithm by Chan et al.
// combined mean is stored in 'mean1'
template<class T>
void mean_combine (T &mean1, const real &w1, const T &mean2, const real &w2) {
  
  // if w2 is much smaller than w1
  if ( w2 < 0.01 * w1 ) {
    T delta (mean2);
    delta -= mean1;
    delta *= w2 / (w1 + w2);
    mean1 += delta;
  }
  
  // if w1 ~ w2 and both large
  else {
    mean1 *= w1 / (w1 + w2);
    mean1 += mean2 * ( w2 / (w1 + w2) );
  }
}

void stats_combine (const real weight) {
  
  // combine mean and variance for ConservedVars
  
  variance_combine (mean_U[MESH_LEVEL][TYPE][SNAPSHOT], variance_U[MESH_LEVEL][TYPE][SNAPSHOT], WEIGHTS[MESH_LEVEL][TYPE], recv_mean_U, weight);
  
  mean_combine (mean_U[MESH_LEVEL][TYPE][SNAPSHOT], WEIGHTS[MESH_LEVEL][TYPE], recv_mean_U, weight);
  
  // combine mean and variance for XtraVars
  
  variance_combine (mean_X[MESH_LEVEL][TYPE][SNAPSHOT], variance_X[MESH_LEVEL][TYPE][SNAPSHOT], WEIGHTS[MESH_LEVEL][TYPE], recv_mean_X, weight);
  
  mean_combine (mean_X[MESH_LEVEL][TYPE][SNAPSHOT], WEIGHTS[MESH_LEVEL][TYPE], recv_mean_X, weight);
  
  // update the total number of samples (during the LAST snapshot)
  if ( SNAPSHOT == LAST_SNAPSHOT )
    WEIGHTS[MESH_LEVEL][TYPE] += weight;
}

void stats_reduce () {
  switch (STAT) {
    case 0:
      mean_U [MESH_LEVEL][TYPE] [SNAPSHOT] += recv_mean_U;
      mean_X [MESH_LEVEL][TYPE] [SNAPSHOT] += recv_mean_X;
    case 1:
      variance_U [MESH_LEVEL][TYPE] [SNAPSHOT] += recv_variance_U;
      variance_X [MESH_LEVEL][TYPE] [SNAPSHOT] += recv_variance_X;
  }
}

void stats_finalize_combines () {
  
  // computing variance
  if (NM[SAMPLES_LEVEL()] > 2) variance_U [MESH_LEVEL][TYPE] [SNAPSHOT] /= NM[SAMPLES_LEVEL()] - 1;
  if (NM[SAMPLES_LEVEL()] > 2) variance_X [MESH_LEVEL][TYPE] [SNAPSHOT] /= NM[SAMPLES_LEVEL()] - 1;
}

void stats_assemble_mlmc () {
  
  // assemble MLMC estimate of mean
  stats_assemble_mlmc (mean_U);
  stats_assemble_mlmc (mean_X);
  
  // assemble MLMC estimate of variance
  stats_assemble_mlmc (variance_U);
  stats_assemble_mlmc (variance_X);
}

void stats_apply_constraints () {
  
  // apply generic constraints to mean
  for (int i = 0; i < mean_U[L][0][SNAPSHOT].NXl; i++)
    for (int j = 0; j < mean_U[L][0][SNAPSHOT].NYl; j++)
      for (int k = 0; k < mean_U[L][0][SNAPSHOT].NZl; k++)
        apply_constraints (mean_U[L][0][SNAPSHOT] [i][j][k], mean_X[L][0][SNAPSHOT] [i][j][k]);
  
  // variance should be non-negative
  #if VARIANCE_FIX
  variance_U [L][0][SNAPSHOT] .nonnegative();
  variance_X [L][0][SNAPSHOT] .nonnegative();
  #endif
}

void stats_write_datafiles () {
  
  Saver saver_mean ("");
  saver_mean.snapshots (&mean_U[MESH_LEVEL][TYPE][0], &mean_X[MESH_LEVEL][TYPE][0]);
  
  Saver saver_var ("var_");
  saver_var.snapshots (&variance_U[MESH_LEVEL][TYPE][0], &variance_X[MESH_LEVEL][TYPE][0]);
}

#endif
