
#include "time.h"

#include "equation.h"
#include "multi.h"
#include "space.h"
#include "solver.h"

#include <iostream>

int TIME_SUB_STEP;

void time_step () {
  
  // start parallel communication of multi boundaries for ConservedVars
  multi_ddm_start_send_recv ();
  
  TIME_SUB_STEP = 1;
  
  // run FVM on the 'internal' cells DURING communication
  #if HIDE_DDM_COMMUNICATION
  time_step_internal_cells ();
  #endif
  
  // set boundary conditions for ConservedVars
  time_step_set_bc (U);
  
  // compute XtraVars for boundary conditions
  time_step_calc_xtra_bc ();
  
  // confirm receives of multi boundaries for ConservedVars
  multi_ddm_confirm_recv ();
  
  // compute XtraVars for multi boundaries
  time_step_calc_xtra_ddm ();
  
  // confirm sends of multi boundaries for ConservedVars
  multi_ddm_confirm_send ();
  
  // run FVM on the 'internal' cells once communication is FINISHED
  #if ! HIDE_DDM_COMMUNICATION
  time_step_internal_cells ();
  #endif
  
  // run fvm on the remaining ('external') cells
  time_step_external_cells ();
  
  // increment time t
  t += dt;
}

// vector for saving maximal wave speed in each direction
vector maxspeed;

void time_sub_step () {
  
  // compute sources for all cells
  Source.zero();
  for (int i = 0; i < NXl; i++)
    for (int j = 0; j < NYl; j++)
      for (int k = 0; k < NZl; k++)
        time_compute_source (i, j, k);
  
  PRE_RECONSTRUCTION_HOOK
  
  // initialize vector for local maximal wave speeds
  maxspeed.zero();
  
  // compute fluxes
  for (int d = 0; d < NDIR; d++)
    if (HAVE_DIR(d))
      for (int i = 0; i < FLUX_NX(d); i++)
        for (int j = 0; j < FLUX_NY(d); j++)
          for (int k = 0; k < FLUX_NZ(d); k++)
            time_compute_flux (i, j, k, d);
  
  POST_RECONSTRUCTION_HOOK
  
  // apply fluxes and sources to 'U'
  time_apply_fluxes_and_sources ();
}

void time_compute_source (const int &i, const int &j, const int &k) {
  SOURCE_HOOK
  #if HAVE_VISCOSITY
  VISCOSITY_HOOK
  #endif
}

// left and right fluxes
AllVariables lV, rV;

void time_compute_flux (const int &i, const int &j, const int &k, const int &d) {
  
  // check if flux needs to be set to zero due to boundary conditions
  
  #if BC_ZERO_FLUX_X_HEAD
  if ( d == X && i == 0            ) { Zero ( Flux[d][i][j][k] ); return; }
  #endif
  
  #if BC_ZERO_FLUX_X_TAIL
  if ( d == X && i == FLUX_NX(X)-1 ) { Zero ( Flux[d][i][j][k] ); return; }
  #endif
  
  #if BC_ZERO_FLUX_Y_HEAD
  if ( d == Y && j == 0            ) { Zero ( Flux[d][i][j][k] ); return; }
  #endif
  
  #if BC_ZERO_FLUX_Y_TAIL
  if ( d == Y && j == FLUX_NY(Y)-1 ) { Zero ( Flux[d][i][j][k] ); return; }
  #endif
  
  #if BC_ZERO_FLUX_Z_HEAD
  if ( d == Z && k == 0            ) { Zero ( Flux[d][i][j][k] ); return; }
  #endif
  
  #if BC_ZERO_FLUX_Z_TAIL
  if ( d == Z && k == FLUX_NZ(Z)-1 ) { Zero ( Flux[d][i][j][k] ); return; }
  #endif
  
  // compute right and left traces of the reconstructed solution
  right (lV, i-DX[d], j-DY[d], k-DZ[d], d);
  left  (rV, i, j, k, d);
  
  // compute local maximal wave speed
  if (TIME_SUB_STEP == 1)
    #if HAVE_VISCOSITY
    maxspeed[d] = max ( maxspeed[d], solver_calc_max_speed (lV, rV, i, j, k, d) + 2*(VISCOSITY / dx[d]) );
    #else
    maxspeed[d] = max ( maxspeed[d], solver_calc_max_speed (lV, rV, i, j, k, d) );
    #endif
  
  // compute flux
  solver_calc_flux (Flux[d][i][j][k], lV, rV, i, j, k, d);
}

void time_apply_fluxes_and_sources () {
  
  // compute global time step
  if (TIME_SUB_STEP == 1)
    dt = min ( t_halt - t, multi_dt ( CFL/(maxspeed[X]/dx[X] + maxspeed[Y]/dx[Y] + maxspeed[Z]/dx[Z]) ) );
  
  // check if time step is not zero
  if (multi_main && dt == 0) {
    std::cout << " :: ERROR: time step is 0." << std::endl;
    std::exit(1);
  }
  
  // scale fluxes according to 'dt' and 'dx'
  for (int d = 0; d < NDIR; d++)
    if ( HAVE_DIR(d) )
      Flux[d] *= dt/dx[d];
  
  // apply fluxes to 'U'
  for (int d = 0; d < NDIR; d++)
    if ( HAVE_DIR(d) )
      for (int i = 0; i < NXl; i++)
        for (int j = 0; j < NYl; j++)
          for (int k = 0; k < NZl; k++)
            U[i][j][k] -= Flux[d][i+DX[d]][j+DY[d]][k+DZ[d]] - Flux[d][i][j][k];
  
  // apply sources to 'U'
  U += (Source *= dt);
  
  // increment sub-step counter
  TIME_SUB_STEP++;
}

void time_sub_step_internal_cells () {
  
  // if boundary layer is sufficiently thick, all cells are 'boundary' cells
  if ( NXl < 2*GX || NYl < 2*GY || NZl < 2*GZ )
    return;
  
  // compute sources for internal cells
  Source.zero();
  for (int i = GX; i < NXl-GX; i++)
    for (int j = GY; j < NYl-GY; j++)
      for (int k = GZ; k < NZl-GZ; k++)
        time_compute_source (i, j, k);
  
  PRE_RECONSTRUCTION_HOOK
  
  // initialize vector for local maximal wave speeds
  maxspeed.zero();
  
  // compute fluxes for internal cells
  for (int d = 0; d < NDIR; d++)
    if (HAVE_DIR(d))
      for (int i = GX; i < FLUX_NX(d)-GX; i++)
        for (int j = GY; j < FLUX_NY(d)-GY; j++)
          for (int k = GZ; k < FLUX_NZ(d)-GZ; k++)
            time_compute_flux (i, j, k, d);
  
  POST_RECONSTRUCTION_HOOK
}

void time_sub_step_external_cells () {
  
  // if boundary layer is sufficiently thick, all cells are 'external' cells
  if ( NXl < 2*GX || NYl < 2*GY || NZl < 2*GZ ) {
    time_sub_step ();
    return;
  }
  
  // compute sources for all external cells except 'corner cells'
  if ( ACTIVE_DIRS > 1 && HAVE_DIR(X) )
    for (int p = 0; p < 2; p++)
      for (int i = (p==0?0:NXl-GX); i < (p==0?GX:NXl); i++)
        for (int j = GY; j < NYl-GY; j++)
          for (int k = GZ; k < NZl-GZ; k++)
            time_compute_source (i, j, k);
  if ( ACTIVE_DIRS > 1 && HAVE_DIR(Y) )
    for (int p = 0; p < 2; p++)
      for (int i = GX; i < NXl-GX; i++)
        for (int j = (p==0?0:NYl-GY); j < (p==0?GY:NYl); j++)
          for (int k = GZ; k < NZl-GZ; k++)
            time_compute_source (i, j, k);
  if ( ACTIVE_DIRS > 1 && HAVE_DIR(Z) )
    for (int p = 0; p < 2; p++)
      for (int i = GX; i < NXl-GX; i++)
        for (int j = GY; j < NYl-GY; j++)
          for (int k = (p==0?0:NZl-GZ); k < (p==0?GZ:NZl); k++)
            time_compute_source (i, j, k);
  
  // compute sources for 'corner cells'
  for (int ci = 0; ci <= HAVE_DIR(X); ci++)
    for (int cj = 0; cj <= HAVE_DIR(Y); cj++)
      for (int ck = 0; ck <= HAVE_DIR(Z); ck++)
        for (int i = (ci==0?0:max(0,NXl-GX)); i < (ci==0?min(max(GX,1),NXl):NXl); i++)
          for (int j = (cj==0?0:max(0,NYl-GY)); j < (cj==0?min(max(GY,1),NYl):NYl); j++)
            for (int k = (ck==0?0:max(0,NZl-GZ)); k < (ck==0?min(max(GZ,1),NZl):NZl); k++)
              time_compute_source (i, j, k);
  
  PRE_RECONSTRUCTION_HOOK
  
  // compute fluxes for all external cells except 'corner cells'
  for (int d = 0; d < NDIR; d++)
    if ( ACTIVE_DIRS > 1 && HAVE_DIR(d) ) {
      if ( ACTIVE_DIRS > 1 && HAVE_DIR(X) )
        for (int p = 0; p < 2; p++)
          for (int i = (p==0?0:FLUX_NX(d)-GX); i < (p==0?GX:FLUX_NX(d)); i++)
            for (int j = GY; j < FLUX_NY(d)-GY; j++)
              for (int k = GZ; k < FLUX_NZ(d)-GZ; k++)
                time_compute_flux (i, j, k, d);
      if ( ACTIVE_DIRS > 1 && HAVE_DIR(Y) )
        for (int p = 0; p < 2; p++)
          for (int i = GX; i < FLUX_NX(d)-GX; i++)
            for (int j = (p==0?0:FLUX_NY(d)-GY); j < (p==0?GY:FLUX_NY(d)); j++)
              for (int k = GZ; k < FLUX_NZ(d)-GZ; k++)
                time_compute_flux (i, j, k, d);
      if ( ACTIVE_DIRS > 1 && HAVE_DIR(Z) )
        for (int p = 0; p < 2; p++)
          for (int i = GX; i < FLUX_NX(d)-GX; i++)
            for (int j = GY; j < FLUX_NY(d)-GY; j++)
              for (int k = (p==0?0:FLUX_NZ(d)-GZ); k < (p==0?GZ:FLUX_NZ(d)); k++)
                time_compute_flux (i, j, k, d);
    }
  
  // compute fluxes for 'corner cells'
  for (int d = 0; d < NDIR; d++)
    if (HAVE_DIR(d))
      for (int ci = 0; ci <= HAVE_DIR(X); ci++)
        for (int cj = 0; cj <= HAVE_DIR(Y); cj++)
          for (int ck = 0; ck <= HAVE_DIR(Z); ck++)
            for (int i = (ci==0?0:max(0,FLUX_NX(d)-GX)); i < (ci==0?min(max(GX,1),FLUX_NX(d)):FLUX_NX(d)); i++)
              for (int j = (cj==0?0:max(0,FLUX_NY(d)-GY)); j < (cj==0?min(max(GY,1),FLUX_NY(d)):FLUX_NY(d)); j++)
                for (int k = (ck==0?0:max(0,FLUX_NZ(d)-GZ)); k < (ck==0?min(max(GZ,1),FLUX_NZ(d)):FLUX_NZ(d)); k++)
                  time_compute_flux (i, j, k, d);
  
  POST_RECONSTRUCTION_HOOK
  
  // apply fluxes and sources to 'U'
  time_apply_fluxes_and_sources ();
}
