
/*
 * Definitions of 'volume' object
 */

#include "structures.h"

#include <cstdlib>
#include <iostream>

// required for 'memcpy()'
#include <cstring>

template <class T>
volume<T>::volume () : \
  
  NXl(1), NYl(1), NZl(1), NTl(1), GH(1), VX0(GX), VY0(GY), VZ0(GZ), VNX(1 + 2*GX), VNY(1 + 2*GY), VNZ(1 + 2*GZ) {
  
  NTl = NXl * NYl * NZl;
  VNT = VNX * VNY * VNZ;
  
  data = (T*) malloc ( VNT * sizeof(T) );
  
  if ( data == NULL ) {
    std::cerr << " :: ERROR: out of memory for volume allocation." << std::endl;
    std::exit(1);
  }
  
  zero ();
}

template <class T>
volume<T>::volume (const int NXl_, const int NYl_, const int NZl_, const int GH_) : \
  
  NXl(max(1,NXl_)), NYl(max(1,NYl_)), NZl(max(1,NZl_)), GH(GH_), VX0(-GX*GH_), VY0(-GY*GH_), VZ0(-GZ*GH_), VNX(max(1,NXl_) + 2*GX*GH_), VNY(max(1,NYl_) + 2*GY*GH_), VNZ(max(1,NZl_) + 2*GZ*GH_) {
  
  NTl = NXl * NYl * NZl;
  VNT = VNX * VNY * VNZ;
  
  if ( VNT == 0 ) {
    std::cerr << " :: ERROR: zero size volumes are not allowed." << std::endl;
    std::exit(1);
  }
  
  data = (T*) malloc ( VNT * sizeof(T) );
  
  if ( data == NULL ) {
    std::cerr << " :: ERROR: out of memory for volume allocation." << std::endl;
    std::exit(1);
  }
  
  zero();
}

template <class T>
volume<T>::volume (const volume<T> &a) {
  
  NXl = a.NXl;
  NYl = a.NYl;
  NZl = a.NZl;
  NTl = a.NTl;
  
  GH  = a.GH;
  
  VX0 = a.VX0;
  VY0 = a.VY0;
  VZ0 = a.VZ0;
  
  VNX = a.VNX;
  VNY = a.VNY;
  VNZ = a.VNZ;
  VNT = a.VNT;
  
  if ( VNT == 0 ) {
    std::cerr << " :: ERROR: zero size volumes are not allowed." << std::endl;
    std::exit(1);
  }
  
  data = (T*) malloc ( VNT * sizeof(T) );
  
  if ( data == NULL ) {
    std::cerr << " :: ERROR: out of memory for volume allocation." << std::endl;
    std::exit(1);
  }
  
  memcpy ( data, a.data, VNT * sizeof(T) );
}

template <class T>
volume<T>::~volume () {
  if ( data != NULL ) free (data);
  data = NULL;
}

template <class T>
volume<T> & volume<T>::zero() {
  for (int i=0; i<VNT; i++)
    Zero(data[i]);
  return *this;
}

template <class T>
volume<T> & volume<T>::reset () {
  return zero();
}

#include "multi.h"

template <class T>
volume<T> & volume<T>::resize (const int NXl_, const int NYl_, const int NZl_, const int GH_) {
  
  if ( NXl == NXl_ && NYl == NYl_ && NZl == NZl_ && GH == GH_ )
    return *this;
  
  // zero size volumes are not allowed
  // for consistency, a minimum of 1 cell is required in each direction
  NXl = max (1, NXl_);
  NYl = max (1, NYl_);
  NZl = max (1, NZl_);
  NTl = NXl * NYl * NZl;
  
  GH  = GH_;
  
  VX0 = -GX*GH_;
  VY0 = -GY*GH_;
  VZ0 = -GZ*GH_;
  
  VNX = NXl + 2*GX*GH_;
  VNY = NYl + 2*GY*GH_;
  VNZ = NZl + 2*GZ*GH_;
  VNT = VNX * VNY * VNZ;
  
  if ( VNT == 0 ) {
    std::cerr << " :: ERROR: zero size volumes are not allowed." << std::endl;
    std::exit(1);
  }
  
  T* data_ = (T*) realloc ( data, VNT * sizeof(T) );
  
  if ( data_ == NULL ) {
    // memory leak prevention
    if ( data != NULL ) free (data);
    data = NULL;
    std::cerr << " :: ERROR: out of memory for volume reinitialization." << std::endl;
    std::exit(1);
  }
  
  data = data_;
  
  return *this;
}

template <class T>
volume<T> & volume<T>::reinit (const int NXl_, const int NYl_, const int NZl_, const int GH_) {
  resize (NXl_, NYl_, NZl_, GH_);
  return zero();
}

template <class T>
const slice<T> volume<T>::operator [] (const int i) const {
  #if CHECK_VOLUME_INDICES
  if ( i < VX0 || i >= NXl - VX0 ) {
    std::cerr << " :: ERROR: Volume index 'i' is out of bounds." << std::endl;
    std::exit(1);
  }
  #endif
  return slice<T> (data, i-VX0, VY0, VZ0, VNY, VNZ);
}

template <class T>
slice<T> volume<T>::operator [] (const int i) {
  #if CHECK_VOLUME_INDICES
  if ( i < VX0 || i >= NXl - VX0 ) {
    std::cerr << " :: ERROR: Volume index 'i' is out of bounds." << std::endl;
    std::exit(1);
  }
  #endif
  return slice<T> (data, i-VX0, VY0, VZ0, VNY, VNZ);
}

template <class T> volume<T> & volume<T>::exp()         { for (int i=0; i<VNT; i++) Exp         ( data[i] ); return *this; }
template <class T> volume<T> & volume<T>::log()         { for (int i=0; i<VNT; i++) Log         ( data[i] ); return *this; }
template <class T> volume<T> & volume<T>::sqr ()        { for (int i=0; i<VNT; i++) data[i] *=    data[i]  ; return *this; }
template <class T> volume<T> & volume<T>::nonnegative() { for (int i=0; i<VNT; i++) NonNegative ( data[i] ); return *this; }

template <class T>
volume<T> & volume<T>::multiply_elementwise (const T v) {
  for (int i=0; i<VNT; i++)
    data[i] *= v;
  return *this;
}

template <class T>
const volume<T> & volume<T>::validate (const volume<T> &a) const {
  if ( NXl % a.NXl != 0 || NYl % a.NYl != 0 || NZl % a.NZl != 0 ) {
    std::cerr << " :: ERROR: non-compatible volume sizes: ";
    std::cerr << "(" << NXl << "," << NYl << "," << NZl << ")";
    std::cerr << " and (" << a.NXl << "," << a.NYl << "," << a.NZl << ")";
    std::cerr << std::endl;
    std::exit(1);
  }
  return *this;
}

template <class T>
volume<T> & volume<T>::operator += (const volume<T> &a) {
  
  validate(a);
  
  if ( NXl == a.NXl && NYl == a.NYl && NZl == a.NZl && GH == a.GH )
    for (int i=0; i<VNT; i++)
      data[i] += a.data[i];
  
  else {
    
    int factor_X = NXl / a.NXl;
    int factor_Y = NYl / a.NYl;
    int factor_Z = NZl / a.NZl;
    
    for (int i=0; i<NXl; i++)
      for (int j=0; j<NYl; j++)
        for (int k=0; k<NZl; k++)
          data[getLoc(i,j,k)] += a[i/factor_X][j/factor_Y][k/factor_Z];
  }
  
  return *this;
}

template <class T>
volume<T> & volume<T>::operator -= (const volume<T> &a) {
  
  validate(a);
  
  if ( NXl == a.NXl && NYl == a.NYl && NZl == a.NZl && GH == a.GH )
    for (int i=0; i<VNT; i++)
      data[i] -= a.data[i];
  
  else {
    
    int factor_X = NXl / a.NXl;
    int factor_Y = NYl / a.NYl;
    int factor_Z = NZl / a.NZl;
    
    for (int i=0; i<NXl; i++)
      for (int j=0; j<NYl; j++)
        for (int k=0; k<NZl; k++)
          data[getLoc(i,j,k)] -= a[i/factor_X][j/factor_Y][k/factor_Z];
  }
  
  return *this;
}

template <class T>
volume<T> & volume<T>::operator *= (const volume<T> &a) {
  
  validate(a);
  
  if ( NXl == a.NXl && NYl == a.NYl && NZl == a.NZl && GH == a.GH )
    for (int i=0; i<VNT; i++)
      data[i] *= a.data[i];
  
  else {
    
    int factor_X = NXl / a.NXl;
    int factor_Y = NYl / a.NYl;
    int factor_Z = NZl / a.NZl;
    
    for (int i=0; i<NXl; i++)
      for (int j=0; j<NYl; j++)
        for (int k=0; k<NZl; k++)
          data[getLoc(i,j,k)] *= a[i/factor_X][j/factor_Y][k/factor_Z];
  }
  
  return *this;
}

template <class T>
volume<T> & volume<T>::operator /= (const volume<T> &a) {
  
  validate(a);
  
  if ( NXl == a.NXl && NYl == a.NYl && NZl == a.NZl && GH == a.GH )
    for (int i=0; i<VNT; i++)
      data[i] /= a.data[i];
  
  else {
    
    int factor_X = NXl / a.NXl;
    int factor_Y = NYl / a.NYl;
    int factor_Z = NZl / a.NZl;
    
    for (int i=0; i<NXl; i++)
      for (int j=0; j<NYl; j++)
        for (int k=0; k<NZl; k++)
          data[getLoc(i,j,k)] /= a[i/factor_X][j/factor_Y][k/factor_Z];
  }
  
  return *this;
}

template <class T>
volume<T> & volume<T>::operator += (const real r) {
  
  for (int i=0; i<VNT; i++)
    data[i] += r;
  return *this;
}

template <class T>
volume<T> & volume<T>::operator -= (const real r) {
  
  for (int i=0; i<VNT; i++)
    data[i] -= r;
  return *this;
}

template <class T>
volume<T> & volume<T>::operator *= (const real r) {
  
  for (int i=0; i<VNT; i++)
    data[i] *= r;
  return *this;
}

template <class T>
volume<T> & volume<T>::operator /= (const real r) {
  
  for (int i=0; i<VNT; i++)
    data[i] /= r;
  return *this;
}

template <class T>
volume<T> volume<T>::operator - (const volume &a) const {
  return volume(*this) -= a;
}

template <class T>
volume<T> volume<T>::operator / (const real r) const {
  return volume(*this) /= r;
}

template <class T>
volume<T> volume<T>::operator * (const volume &a) const {
  return volume(*this) *= a;
}

template <class T>
volume<T> volume<T>::operator * (const real r) const {
  return volume(*this) *= r;
}

template <class T>
volume<T> & volume<T>::average (int levels) {
  
  if (levels < 1)
    return *this;
  
  // copy data to temporary volume
  // TODO: this could be avoided
  volume *v = new volume (*this);
  
  // resize current volume
  this -> reinit ( max(NXl/2,1), max(NYl/2,1), max(NZl/2,1), GH );
  
  // compute cell averages
  for (int i=0; i<NXl; i++)
    for (int j=0; j<NYl; j++)
      for (int k=0; k<NZl; k++) {
        (*this)[i][j][k]    = (*v)[2*i][2*j][  2*k]   + (*v)[2*i+1][2*j][  2*k];
        if (HAVE_DIR(Y))
          (*this)[i][j][k] += (*v)[2*i][2*j+1][2*k]   + (*v)[2*i+1][2*j+1][2*k];
        if (HAVE_DIR(Z)) {
          (*this)[i][j][k] += (*v)[2*i][2*j][  2*k+1] + (*v)[2*i+1][2*j][  2*k+1];
          (*this)[i][j][k] += (*v)[2*i][2*j+1][2*k+1] + (*v)[2*i+1][2*j+1][2*k+1];
        }
      }
  
  // scale values according to the number of dimensions
  (*this) /= 1 << ACTIVE_DIRS;
  
  // delete temporary volume
  delete v;
  
  // recursive averaging
  return this -> average (levels - 1);
}
/*
template <class T>
volume<T> & volume<T>::print2d (const char* name) {
  
  std::cout << std::endl << " === VOLUME: " << name << " (" << NXl << "x" << NYl << ")" std::endl;
  for (int i=0; i<NXl; i++) {
    for (int j=0; j<NYl; j++)
      std::cout << (*this)[i][j][0] << " ";
    std::cout << std::endl;
  }
  
  return *this;
}
*/
