
/*
 * Declarations of 'volume' object
 */

#ifndef VOLUMES_H
#define VOLUMES_H

#include "numtypes.h"

// 'slice' struct for subscript operator [][][]

template <class T>
struct slice {
  T* d;
  int VY0, VZ0, VNY, VNZ;
  slice ( T *d_, const int shift, const int VY0_, const int VZ0_, const int VNY_, const int VNZ_ ) : \
  d(d_ + shift*VNY_*VNZ_), VY0(VY0_), VZ0(VZ0_), VNY(VNY_), VNZ(VNZ_) {}
  
  const T* const operator [] (const int j) const {
    #if CHECK_VOLUME_INDICES
    if ( j < VY0 || j >= VNY + VY0 ) {
      std::cerr << " :: ERROR: Volume index 'j' is out of bounds." << std::endl;
      std::exit(1);
    }
    #endif
    return d + (j-VY0)*VNZ - VZ0;
  };
  
  T* operator [] (const int j) {
    #if CHECK_VOLUME_INDICES
    if ( j < VY0 || j >= VNY + VY0 ) {
      std::cerr << " :: ERROR: Volume index 'j' is out of bounds." << std::endl;
      std::exit(1);
    }
    #endif
    return d + (j-VY0)*VNZ - VZ0;
  };
};

// 3d volume array type (dynamic version)
// overloaded operators are supported for hierarchical grids
// ghost cells array indices are either negative or larger than grid size

template <class T>
class volume {

private:
  
  // returns location of the [i][j][k]-th element in the volume
  inline int getLoc (int i, int j, int k) { return (i-VX0)*VNY*VNZ + (j-VY0)*VNZ + (k-VZ0); }
  
public:
  
  // pointer to actual array
  T* data;
  
  // grid sizes, total size
  int NXl, NYl, NZl, NTl;
  
  // whether ghost cells are present or not
  int GH;
  
  // offsets taking into account the ghost cells
  int VX0, VY0, VZ0;
  
  // grid sizes including ghost cells, total size
  int VNX, VNY, VNZ, VNT;
  
  // constructors
  volume ();
  volume (const int NXl_, const int NYl_, const int NZl_, const int GH_ = 1);
  volume (const volume &);
  
  // destructors
  ~volume ();
  
  // initializes volume elements to 0
  volume & zero  ();
  volume & reset ();
  
  // adjusts the size of the volume
  volume & resize (const int NXl_, const int NYl_, const int NZl_, const int GH_ = 1);
  
  // adjusts the size of the volume and initializes elements to 0
  volume & reinit (const int NXl_, const int NYl_, const int NZl_, const int GH_ = 1);
  
  // subscript operator [i][j][k]:
  // slice = volume[i]; slice[j] = data + ...; data[k] = required element
  const slice<T> operator [] (const int i) const;
        slice<T> operator [] (const int i);
  
  // some useful numerical functions
  
  volume & exp();
  volume & log();
  volume & sqr();
  volume & nonnegative();
  volume & multiply_elementwise (const T v);
  
  // validates compatibility of volumes for use with operators
  const volume & validate (const volume &a) const;
  
  // === operator overloading
  
  volume & operator += (const volume &a);
  volume & operator -= (const volume &a);
  volume & operator *= (const volume &a);
  volume & operator /= (const volume &a);
  
  volume & operator += (const real r);
  volume & operator -= (const real r);
  volume & operator *= (const real r);
  volume & operator /= (const real r);
  
  // === operators creating NEW objects should be used with care: memory efficiency is not guarranteed.
  
  // needed for variance computation during parallel runs
  volume operator - (const volume &a) const;
  
  // needed for numerically stable variance computation
  volume operator / (const real r) const;
  
  // needed for stats:mean-kurtosis
  volume operator * (const volume &a) const;
  volume operator * (const real r) const;
  
  // === shape modification routines
  
  // averages volume by the required number of 'levels'
  volume & average (const int levels = 1);
  
  // === output routines
  volume & print2d (const char* name);
};

// must be here in order to have arbitrary templated objects and inlining
#include "volumes.cpp"

#endif
