
# # # # # # # # # # # # # # # # # # # # # # # # # #
# Error convergence and scaling plots and tables, #
# + exporting multiple simulations to MATLAB      #
#                                                 #
# Jonas Sukys                                     #
# SAM, ETH Zurich, Switzerland                    #
# sukys.jonas@gmail.com                           #
# # # # # # # # # # # # # # # # # # # # # # # # # #

# === local imports

from plot import *

# === global imports

import pylab
import numpy

# === configuration

minfloat = 4.9406564584124654e-324
maxfloat = 8.9884656743115795e+307

import matplotlib as mpl
mpl.rcParams['lines.linewidth'] = 2
mpl.rcParams['lines.markeredgewidth'] = 2
mpl.rcParams['lines.markersize'] = 8
mpl.rcParams['font.size'] = 18
mpl.rcParams['legend.fontsize'] = 18

import warnings
warnings.filterwarnings("ignore", message="divide by zero encountered in log10")

# === helpers

# error norms

def L1(data, refd=None, area=1):
  if refd is None:
    n = area*reduce((lambda x,y:x*y), data.shape)
    return abs(data).sum()/n
  else:
    return abs(data - refd).sum()/abs(refd).sum()

def L2(data, refd=None, area=1):
  if refd is None:
    n = area*reduce((lambda x,y:x*y), data.shape)
    return (numpy.sqr(data).sum()/n)**0.5
  else:
    return ((data - refd)**2).sum()**0.5/((refd**2).sum()**0.5)

def Linf(data, refd=None):
  if refd is None:
    return numpy.max(abs(data))
  else:
    return numpy.max(abs(data - refd))/numpy.max(abs(refd))

# loads the reference solution from the required directory
def load_reference (ref_path=None):
  if ref_path == None:
    r = Run()
    print ( 'Reference solution loaded from the current directory.')
  else:
    r = Run(path=ref_path)
    print ( 'Reference solution loaded from: ' + ref_path)
  
  return r

# upscales the reference solution to a required level
def upscale_reference (refd, NX, NY, NZ):
  
  refdd = refd
  
  shape = refd.shape
  zoom = shape[0]/NX
  
  if len(shape) == 1:
    while zoom > 1:
      refdd = (refdd[0:shape[0]-1:2]+refdd[1:shape[0]:2])/2
      zoom /= 2
  
  elif len(shape) == 2:
    while zoom > 1:
      refdd = (  refdd[0:shape[0]-1:2, 0:shape[1]-1:2] \
               + refdd[0:shape[0]-1:2, 1:shape[1]:2  ] \
               + refdd[1:shape[0]:2,   0:shape[1]-1:2] \
               + refdd[1:shape[0]:2,   1:shape[1]:2  ] ) / 4
      zoom /= 2
  
  else:
    while zoom > 1:
      refdd = (  refdd[0:shape[0]-1:2, 0:shape[1]-1:2, 0:shape[2]-1:2] \
               + refdd[0:shape[0]-1:2, 0:shape[1]-1:2, 1:shape[2]:2  ] \
               + refdd[0:shape[0]-1:2, 1:shape[1]:2,   0:shape[2]-1:2] \
               + refdd[0:shape[0]-1:2, 1:shape[1]:2,   1:shape[2]:2  ] \
               + refdd[1:shape[0]:2,   0:shape[1]-1:2, 0:shape[2]-1:2] \
               + refdd[1:shape[0]:2,   0:shape[1]-1:2, 1:shape[2]:2  ] \
               + refdd[1:shape[0]:2,   1:shape[1]:2,   0:shape[2]-1:2] \
               + refdd[1:shape[0]:2,   1:shape[1]:2,   1:shape[2]:2  ] ) / 8
      zoom /= 2
  
  return refdd

# generators for lists of mesh sizes

make1d_N = lambda refs, start=32 : [[start*2**i,1,1] for i in range(refs)]
make2d_N = lambda refs, start=32 : [[start*2**i,start*2**i,1] for i in range(refs)]
make3d_N = lambda refs, start=32 : [[start*2**i,start*2**i,start*2**i] for i in range(refs)]

# === plotting routines

# plots the errors for the refinements in the list N vs number of cells and vs runtime
def plotError ( name, v, N, case='mean', frame=False, color=None, refr=None, K=1, infolines=False, save_suffix='', file_suffix='pdf', path='', label_suffix='', std_in_label=True, order=1, functional=None, marker='x', s=None ):
  
  pylab.subplot(121)
  plotErrorVsCells   ( name, v, N, case=case, frame=True, color=color, refr=refr, K=K, infolines=False, save_suffix=save_suffix, file_suffix=file_suffix, path=path, label_suffix=label_suffix, std_in_label=std_in_label, order=order, functional=functional, marker=marker, plot_slopes=(not frame), s=s )
  pylab.subplot(122)
  plotErrorVsRuntime ( name, v, N, case=case, frame=True, color=color, refr=refr, K=K, infolines=False, save_suffix=save_suffix, file_suffix=file_suffix, path=path, label_suffix=label_suffix, order=order, functional=functional, marker=marker, plot_slopes=(not frame), s=s )
  
  pylab.subplots_adjust(right=0.98)
  pylab.subplots_adjust(left=0.07)
  pylab.subplots_adjust(top=0.92)
  
  if infolines:
    text = refr.generateInfo('Ref. ')
    pylab.figtext(0.01, 0.015, text, fontsize='small')
    pylab.subplots_adjust(bottom=0.28)
  else:
    pylab.subplots_adjust(bottom=0.12)
  
  if not frame:
    save = getFileName(refr, 'conv-' + case + save_suffix, file_suffix=file_suffix)
    pylab.savefig(save)
    pylab.savefig(save[:-3] + 'eps')
    refr.generateTexTable(save)
    pylab.show()

# plots the errors for the refinements in the list N vs number of cells
def plotErrorVsCells ( name, v, N, case='mean', frame=False, color=None, refr=None, K=1, infolines=False, save_suffix='', file_suffix='pdf', path='', label_suffix='', std_in_label=True, order=1, functional=None, marker='x', plot_slopes=True, s=None ):
  
  reft = len(refr.times)-1
  refd = refr.var(v, reft, case)
  
  errors   = []
  errors_K = []
  ncells   = []
  rtimes   = []
  
  if functional:
    functional_ref = evaluate_functional (refd, refr, functional)
  
  for i , (NX, NY, NZ) in enumerate(N):
    
    if not functional:
      refdd = upscale_reference (refd, NX, NY, NZ)
    
    name_base = name + '_%dx%dx%d' % (NX, NY, NZ)
    
    error = 0
    mean_K = 0
    M2_K = 0
    
    # try to load runs and get needed values
    try:
      for k in range(1, K+1):
        r = Run('%s-%d' % (name_base, k), path=path)
        
        t = len(r.times)-1
        d = r.var(v, t, case)
        
        if functional:
          rel_err = ( 100 * abs( evaluate_functional(d,r,functional) - functional_ref ) ) ** 2
        
        else:
          if order == 1:
            rel_err = ( 100 * L1(d, refdd) ) ** 2
          elif order == 2:
            rel_err = ( 100 * L2(d, refdd) ) ** 2
        
        error += rel_err
        delta = rel_err - mean_K
        mean_K += delta / k
        if k > 1:
          M2_K += delta * (rel_err - mean_K)
    
    # if some load failed, move to another
    except:
      #raise
      continue
    
    errors.append ( numpy.sqrt(error/K) )
    ncells.append ( NX )
    variance_K  = M2_K / max(1,K-1)
    rel_error_K = variance_K / max(1,K-1) / (error/K)
    errors_K.append ( rel_error_K )
  
  # plot error vs ncells
  
  log_ncells = numpy.log10(ncells)
  log_errors = numpy.log10(errors)
  
  if K == 1 or std_in_label == False:
	  pylab.plot(log_ncells, log_errors, '-' + marker, label=name+label_suffix, color=color, markeredgecolor=color)
  elif errors_K[-1] < 0.01:
	  pylab.plot(log_ncells, log_errors, '-' + marker, label=name+label_suffix+r' ($\sigma^2_K$<0.01)', color=color, markeredgecolor=color)
  else:
    pylab.plot(log_ncells, log_errors, '-' + marker, label=name+label_suffix+r' ($\sigma^2_K$'+'=%.2f)' % errors_K[-1], color=color, markeredgecolor=color)
  
  if functional:
    pylab.title(r'Rel. $L^2(\Omega)$-error of the ' + case + ' of ' + r'$\Phi$' + ' (K=%d)' % K)
    pylab.ylabel(r'log10( $L^2(\Omega)$-error )')
  else:
    pylab.title(r'Rel. $L^2(L^%d)$-error of the ' % order + case + ' of ' + v.name + ' (K=%d)' % K)
    pylab.ylabel(r'log10( $L^2(\Omega,\ L^%d(D))$-error )' % order)
  
  pylab.xlabel('log10(cells) in x-direction')
  pylab.legend(loc='lower left')
  pylab.grid(True)
  
  if plot_slopes or not frame:
    slopes = [[1,2], [1,1]]
    for slope in slopes:
      plotSlope(slope, pylab.gca())
  
  if s:
    p = numpy.polyfit(log_ncells[1:], log_errors[1:], 1)
    #log_errors -= s * ( numpy.log10(refr.f.MAXX) - log_ncells )
    print ' :: Error vs. mesh size: ' + name + label_suffix + ': ' + 'constant %1.2e' % numpy.exp(p[1]) + ', ' + 'rate %1.2f' % -p[0] + ' (expected %1.2f' % s + ')'
  
  pylab.subplots_adjust(left=0.16)
  
  if infolines:
    text = refr.generateInfo('Ref. ')
    pylab.figtext(0.01, 0.015, text, fontsize='small')
    pylab.subplots_adjust(bottom=0.28)
  else:
    pylab.subplots_adjust(bottom=0.12)
  
  if not frame:
    save = getFileName(refr, 'conv-vs-cells-' + case + save_suffix, file_suffix=file_suffix)
    pylab.savefig(save)
    pylab.savefig(save[:-3] + 'eps')
    refr.generateTexTable(save)
    pylab.show()

# plots the errors for the refinements in the list N vs runtime
def plotErrorVsRuntime ( name, v, N, case='mean', frame=False, color=None, refr=None, K=1, infolines=False, save_suffix='', file_suffix='pdf', path='', label_suffix='', order=1, functional=None, marker='x', plot_slopes=True, s=None ):
  
  reft = len(refr.times)-1
  refd = refr.var(v, reft, case)
  
  errors = []
  rtimes = []
  
  if functional:
    functional_ref = evaluate_functional (refd, refr, functional)
  
  for i , (NX, NY, NZ) in enumerate(N):
    
    if not functional:
      refdd = upscale_reference (refd, NX, NY, NZ)
    
    name_base = name + '_%dx%dx%d' % (NX, NY, NZ)
    
    error = 0
    rtime = 0
    
    # try to load runs and get needed values
    try:
      for k in range(1, K+1):
        r = Run('%s-%d' % (name_base, k), path=path)
        
        t = len(r.times)-1
        d = r.var(v, t, case)
        
        if functional:
          rel_err = ( 100 * abs( evaluate_functional(d,r,functional) - functional_ref ) ) ** 2
        
        else:
          if order == 1:
            rel_err = (100*L1(d, refdd))**2
          elif order == 2:
            rel_err = (100*L2(d, refdd))**2
        
        error += rel_err
        
        if 'wallclock_runtime_global' in r.runtime:
          rtime += r.runtime['wallclock_runtime_global']
        elif 'wall_runtime_total' in r.runtime:
          rtime += r.runtime['wall_runtime_total'] # legacy
        else:
          rtime += r.runtime['clock_runtime']
    
    # if some load failed, move to another
    except:
      continue
    
    errors.append ( numpy.sqrt(error/K) )
    rtimes.append ( rtime/K )
  
  # plot error vs runtime
  
  pylab.plot(numpy.log10(rtimes), numpy.log10(errors), '-' + marker, label=name+label_suffix, color=color, markeredgecolor=color)
  
  if functional:
    pylab.title(r'Rel. $L^2(\Omega)$-error of the ' + case + ' of ' + r'$\Phi$' + ' (K=%d)' % K)
    pylab.ylabel(r'log10( $L^2(\Omega)$-error )')
  else:
    pylab.title(r'Rel. $L^2(L^%d)$-error of the ' % order + case + ' of ' + v.name + ' (K=%d)' % K)
    pylab.ylabel(r'log10( $L^2(\Omega,\ L^%d(D))$-error )' % order)
  
  pylab.xlabel('log10(seconds)')
  pylab.legend(loc='lower left')
  pylab.grid(True)
  
  dim = int(refr.i.NX>1) + int(refr.i.NY>1) + int(refr.i.NZ>1)
  
  if s:
    while rtimes[0] == 0:
      rtimes = numpy.delete(rtimes,0)
      errors = numpy.delete(errors,0)
    
    log_rtimes = numpy.log10(rtimes)
    log_errors = numpy.log10(errors)
    
    p = numpy.polyfit(log_rtimes[1:], log_errors[1:], 1)
    #log_errors -= - s / float(dim + 1) * log_rtimes
    print ' :: Error vs. runtime: ' + name + label_suffix + ': ' + 'constant %1.2e' % numpy.exp(p[1]) + ', ' + 'rate %1.2f' % -p[0] + ' (expected %1.2f' % (s / float(dim + 1)) + ')'
  
  if plot_slopes or not frame:
    if 'HAVE_VISCOSITY' in refr.vars:
      slopes = [[1,2+2*dim+2*refr.i.HAVE_VISCOSITY], [1,1+dim+refr.i.HAVE_VISCOSITY]]
    else:
      slopes = [[1,2+2*dim], [1,1+dim]]
    for slope in slopes:
      plotSlope(slope, pylab.gca())
  
  pylab.subplots_adjust(left=0.16)
  
  if infolines:
    text = refr.generateInfo('Ref. ')
    pylab.figtext(0.01, 0.015, text, fontsize='small')
    pylab.subplots_adjust(bottom=0.28)
  else:
    pylab.subplots_adjust(bottom=0.12)
  
  if not frame:
    save = getFileName(refr, 'conv-vs-runtime-' + case + save_suffix, file_suffix=file_suffix)
    pylab.savefig(save)
    pylab.savefig(save[:-3] + 'eps')
    refr.generateTexTable(save)
    pylab.show()

# plots the slope together with a numerical value in the plt axis
def plotSlope(slope, plt, type='linear'):
  
  from pylab import xlim, ylim
  from numpy import log10, power
  
  xmin, xmax = xlim()
  ymin, ymax = ylim()
  
  if type == 'linear':
    x1 = xmax - (xmax-xmin)/3
    x2 = xmax - (xmax-xmin)/10
    
    y1 = ymax - (ymax-ymin)/20
    y2 = y1 - float(slope[0])/slope[1] * (x2 - x1)
  
  else:
    x1 = log10(xmax) - (log10(xmax)-log10(xmin))/3
    x2 = log10(xmax) - (log10(xmax)-log10(xmin))/10
    
    y1 = log10(ymax) - (log10(ymax)-log10(ymin))/20
    y2 = y1 - float(slope[0])/slope[1] * (x2 - x1)
    
    x1 = power(10, x1)
    x2 = power(10, x2)
    y1 = power(10, y1)
    y2 = power(10, y2)
  
  plt.plot([x1, x2], [y1, y2], '--k', label='_nolegend_')
  plt.annotate(' %d/%d' % (slope[0], slope[1]), xy=(x2, y2),  xycoords='data')

# plots the specified variable for the refinements in the list N vs number of cells
def plotEnergy ( name, v, N, case='mean', frame=False, color=None, K=1, infolines=False, save_suffix='', file_suffix='pdf', path='', label_suffix='', order=1, marker='x'):
  
  errors   = []
  errors_K = []
  ncells   = []
  
  for i , (NX, NY, NZ) in enumerate(N):
    
    name_base = name + '_%dx%dx%d' % (NX, NY, NZ)
    
    error = 0
    mean_K = 0
    M2_K = 0
    
    # try to load runs and get needed values
    try:
      for k in range(1, K+1):
        r = Run('%s-%d' % (name_base, k), path=path)
        
        if r.vars['SAVE0'] == 0:
          raise Exception ("Snapshot at t=0 is not available! Use SAVE0=1 when compiling.")
        
        ts = len(r.times)-1
        dT = r.var(v, ts, case)
        d0 = r.var(v, 0,  case)
        
        area = r.f.MAXX * r.f.MAXY * r.f.MAXZ
        
        if order == 1:
          error_sample = numpy.fabs( L1(dT, area=area) - L1(d0, area=area) ) ** 2
        elif order == 2:
          error_sample = numpy.fabs( L2(dT, area=area) - L2(d0, area=area) ) ** 2
        
        error += error_sample
        delta = error_sample - mean_K
        mean_K += delta / k
        if k > 1:
          M2_K += delta * (error_sample - mean_K)
    
    # if some load failed, move to another
    except:
      continue
    
    errors.append ( numpy.sqrt(error/K) )
    ncells.append ( NX )
    variance_K  = M2_K / max(1,K-1)
    error_K = numpy.sqrt(variance_K / K) / (error/K)
    errors_K.append ( error_K )
  
  # plot error vs ncells
  
  if K == 1:
	  pylab.plot(numpy.log10(ncells), errors, '-' + marker, label=name+label_suffix, color=color, markeredgecolor=color)
  elif values_K[-1] < 0.01:
	  pylab.plot(numpy.log10(ncells), errors, '-' + marker, label=name+label_suffix+r' ($\sigma_K$<0.01)', color=color, markeredgecolor=color)
  else:
    pylab.plot(numpy.log10(ncells), errors, '-' + marker, label=name+label_suffix+r' ($\sigma_K$'+'=%.2f)' % values_K[-1], color=color, markeredgecolor=color)
  
  pylab.title(r'Discrepancy of $L^%d$-norm of the ' % order + case + ' of ' + v.name + ' (K=%d)' % K)
  pylab.ylabel(r'$L^2(\Omega,\ L^%d(D))$ discrepancy' % order)
  pylab.xlabel('log10(cells) in x-direction')
  pylab.legend(loc='lower left')
  pylab.grid(True)
  
  pylab.subplots_adjust(left=0.16)
  
  if infolines:
    text = r.generateInfo('Ref. ')
    pylab.figtext(0.01, 0.015, text, fontsize='small')
    pylab.subplots_adjust(bottom=0.28)
  else:
    pylab.subplots_adjust(bottom=0.12)
  
  if not frame:
    save = getFileName(r, 'variable-vs-cells-' + case + save_suffix, file_suffix=file_suffix)
    pylab.savefig(save)
    pylab.savefig(save[:-3] + 'eps')
    r.generateTexTable(save)
    pylab.show()

# plots the parallelisation efficiency for the refinements in the list N
def plotEfficiency ( name, N, frame=False, color=None, K=1, save_suffix='', file_suffix='pdf', path='', label_suffix='', marker='x' ):
  
  ncells = []
  
  wallclock_runtimes_global     = []
  wallclock_runtimes_global_mpi = []
  mpi_efficiencies              = []
  
  for i, (NX, NY, NZ) in enumerate(N):
    
    name_base = name + '_%dx%dx%d' % (NX, NY, NZ)
    
    wallclock_runtime_global     = 0
    wallclock_runtime_global_mpi = 0
    mpi_efficiency               = 0
    
    # try to load runs and get needed values
    try:
      for k in range(1, K+1):
        container = {}
        execfile(path + '%s_%s-%d.py' % ('runtime', name_base, k), globals(), container)
        
        if 'wallclock_runtime_global' in container:
          wallclock_runtime_global += container['wallclock_runtime_global']
        else:
          wallclock_runtime_global += container['wall_runtime_total'] # legacy
        
        if 'wallclock_runtime_global_mpi' in container:
          wallclock_runtime_global_mpi += container['wallclock_runtime_global_mpi']
        else:
          wallclock_runtime_global_mpi += container['mpi_wall_runtime_total'] # legacy
        
        if 'mpi_efficiency' in container:
          mpi_efficiency += container['mpi_efficiency']
        else:
          mpi_efficiency += container['mpi_scaling'] # legacy
    
    # if some load failed, move to another
    except:
      continue
    
    ncells.append ( NX * NY * NZ )
    
    wallclock_runtimes_global     .append ( wallclock_runtime_global/K )
    wallclock_runtimes_global_mpi .append ( wallclock_runtime_global_mpi/K )
    mpi_efficiencies              .append ( mpi_efficiency/K )
  
  # runtimes vs ncells
  
  pylab.subplot(121)
  pylab.plot(numpy.log10(ncells), numpy.log10(wallclock_runtimes_global), '-' + marker, label=name+label_suffix, color=color, markeredgecolor=color)
  pylab.plot(numpy.log10(ncells), numpy.log10(wallclock_runtimes_global_mpi), color+'--', markeredgecolor=color)
  pylab.title('Wall runtime vs. MPI calls | avg. of %d runs' % K)
  pylab.ylabel('log10(seconds)')
  pylab.xlabel('log10(cells)')
  pylab.legend(loc='best')
  pylab.grid(True)
  
  # Parallel efficiency vs ncells
  
  pylab.subplot(122)
  pylab.plot(numpy.log10(ncells), mpi_efficiencies, '-' + marker, label=name+label_suffix, color=color, markeredgecolor=color)
  pylab.ylim(0,1.02)
  pylab.title('Parallel efficiency | average of %d runs' % K)
  pylab.ylabel('efficiency')
  pylab.xlabel('log10(cells)')
  pylab.legend(loc='lower right')
  pylab.grid(True)
  
  pylab.subplots_adjust(right=0.98)
  pylab.subplots_adjust(left=0.06)
  pylab.subplots_adjust(top=0.92)
  pylab.subplots_adjust(bottom=0.12)
  
  if not frame:
    pylab.savefig(getFileName(Run(name_base + '-1'), 'efficiency' + save_suffix, file_suffix=file_suffix))
    pylab.savefig(getFileName(Run(name_base + '-1'), 'efficiency' + save_suffix, file_suffix='eps'))
    pylab.show()

# plots the strong scaling for the refinements 'REF_START', ..., 'REF_END'
def plotStrongScaling ( name, REF_START, REF_END, frame=False, color=None, K=1, save_suffix='', file_suffix='pdf', path='', label_suffix='', marker='x', detailed=False ):
  
  # === strong scaling
  
  wallclock_runtimes = []
  mpi_efficiencies   = []
  cores              = []
  
  # needed for 'detailed == True'
  loads               = []
  mpi_ddm_loads       = []
  mpi_bcast_loads     = []
  balancers           = []
  mpi_ddms            = []
  mpi_mcs             = []
  mpi_mlmcs           = []
  mpi_idles           = []
  
  balancer_info_available = False
  
  for i in range(REF_START, REF_END+1):
    
    name_base = name + '_%d' % i
    
    wallclock_runtime = 0
    mpi_efficiency    = 0
    
    load           = 0
    mpi_ddm_load   = 0
    mpi_bcast_load = 0
    balancer       = 0
    mpi_ddm        = 0
    mpi_mc         = 0
    mpi_mlmc       = 0
    mpi_idle       = 0
    
    # try to load runs and get needed values
    try:
      for k in range(1, K+1):
        container = {}
        execfile(path + '%s_%s-%d.py' % ('runtime', name_base, k), globals(), container)
        
        if 'wallclock_runtime' in container:
          wallclock_runtime += container['wallclock_runtime']
        else:
          wallclock_runtime += container['wall_runtime'] # legacy
        
        if 'mpi_efficiency' in container:
          mpi_efficiency += container['mpi_efficiency']
        else:
          mpi_efficiency += container['mpi_scaling'] # legacy
        
        balancer_info_available = 'wallclock_runtime_global_loads' in container
        
        if balancer_info_available:
          load += container['wallclock_runtime_global_loads']
          balancer += container['wallclock_runtime_global_balancer']
        
        if detailed:
          if balancer_info_available:
            mpi_ddm_load   += container['wallclock_runtime_global_mpi_ddm_loads']
            mpi_bcast_load += container['wallclock_runtime_global_mpi_bcast_loads']
          mpi_ddm  += container['wallclock_runtime_global_mpi_ddm']
          mpi_mc   += container['wallclock_runtime_global_mpi_mc']
          mpi_mlmc += container['wallclock_runtime_global_mpi_mlmc']
          mpi_idle += container['wallclock_runtime_global_mpi_idle']
    
    # if some load failed, move to another
    except:
      continue
    
    if 'cores' in container:
      cores .append( container['cores'] )
    else:
      cores .append( container['nodes'] ) # legacy
    
    wallclock_runtimes .append ( wallclock_runtime/K )
    mpi_efficiencies   .append ( mpi_efficiency/K )
    
    if balancer_info_available:
      loads     .append ( load/K / (wallclock_runtime * cores[-1]) )
      balancers .append ( balancer/K / (wallclock_runtime * cores[-1]) )
    
    if detailed:
      if balancer_info_available:
        mpi_ddm_loads   .append ( mpi_ddm_load/K / (wallclock_runtime * cores[-1]) )
        mpi_bcast_loads .append ( mpi_bcast_load/K / (wallclock_runtime * cores[-1]) )
      mpi_ddms  .append ( mpi_ddm/K / (wallclock_runtime * cores[-1]) )
      mpi_mcs   .append ( mpi_mc/K / (wallclock_runtime * cores[-1]) )
      mpi_mlmcs .append ( mpi_mlmc/K / (wallclock_runtime * cores[-1]) )
      mpi_idles .append ( mpi_idle/K / (wallclock_runtime * cores[-1]) )
  
  pylab.subplot(121)
  pylab.loglog(cores, wallclock_runtimes, '-' + marker, label=name+label_suffix, color=color, markeredgecolor=color)
  if K != 1:
    pylab.title('Strong scaling | average of %d runs' % K)
  else:
    pylab.title('Strong scaling')
  pylab.ylabel('wallclock runtime')
  pylab.xlabel('cores')
  pylab.legend(loc='lower left')
  pylab.grid(True)
  
  if not frame:
    plotSlope([1,1], pylab.gca(), type='loglog')
  
  # === efficiency
  
  # reduce legend font size for large legends
  if detailed:
    legend_font_size_backup = mpl.rcParams['legend.fontsize']
    mpl.rcParams['legend.fontsize'] = 14
  
  pylab.subplot(122)
  
  if detailed:
    pylab.semilogx(cores, mpi_efficiencies, '-' + marker, label='efficiency'+label_suffix, color=color, markeredgecolor=color)
  else:
    pylab.semilogx(cores, mpi_efficiencies, '-' + marker, label=name+label_suffix, color=color, markeredgecolor=color)
  
  if balancer_info_available:
    if detailed:
      pylab.semilogx(cores, loads,           '--', label='loads est.'+label_suffix)
      pylab.semilogx(cores, balancers,       '--', label='loads dist.'+label_suffix)
      pylab.semilogx(cores, mpi_ddm_loads,   '--', label='loads DDM'+label_suffix)
      pylab.semilogx(cores, mpi_bcast_loads, '--', label='loads Bcast'+label_suffix)
    else:
      pylab.semilogx(cores, [loads[i]+balancers[i] for i in range(len(loads))], '--' + marker, label='balancing'+label_suffix, color=color, markeredgecolor=color)
  
  if detailed:
    pylab.semilogx(cores, mpi_ddms,        label='MPI DDM'+label_suffix)
    pylab.semilogx(cores, mpi_mcs,         label='MPI MC'+label_suffix)
    pylab.semilogx(cores, mpi_mlmcs,       label='MPI MLMC'+label_suffix)
    pylab.semilogx(cores, mpi_idles,       label='MPI idle'+label_suffix)
  
  if K != 1:
    if detailed:
      pylab.title('Parallel efficiency and overhead runtimes')
    else:
      pylab.title('Parallel efficiency | average of %d runs' % K)
  else:
    if detailed:
      pylab.title('Parallel efficiency and overhead runtimes')
    else:
      pylab.title('Parallel efficiency')
  
  pylab.ylim(0,1.02)
  
  if detailed:
    pylab.ylabel('relative runtime')
  else:
    pylab.ylabel('efficiency')
  
  pylab.xlabel('cores')
  
  if detailed or balancer_info_available:
    pylab.legend(loc='center left')
  else:  
    pylab.legend(loc='lower center')
  
  pylab.grid(True)
  
  # restore legend font size
  if detailed:
    mpl.rcParams['legend.fontsize'] = legend_font_size_backup
  
  # === formatting
  
  pylab.subplots_adjust(right=0.97)
  pylab.subplots_adjust(left=0.07)
  pylab.subplots_adjust(top=0.93)
  pylab.subplots_adjust(bottom=0.15)
  
  if not frame:
    pylab.savefig(getFileName(Run(name_base + '-1', load_times=False), 'scaling-strong' + save_suffix, file_suffix=file_suffix))
    pylab.savefig(getFileName(Run(name_base + '-1', load_times=False), 'scaling-strong' + save_suffix, file_suffix='eps'))
    pylab.show()

# plots the weak scaling for the refinements 1, ..., 'REF_END'
def plotWeakScaling ( name, REF_END, frame=False, color=None, K=1, save_suffix='', file_suffix='pdf', path='', label_suffix='', marker='x' ):
  
  # === weak scaling
  
  wallclock_runtimes_global = []
  mpi_efficiencies          = []
  cores                     = []
  
  for i in range(1, REF_END+1):
    
    name_base = name + '_%d' % i
    
    wallclock_runtime_global = 0
    mpi_efficiency           = 0
    
    # try to load runs and get needed values
    try:
      for k in range(1, K+1):
        container = {}
        execfile(path + '%s_%s-%d.py' % ('runtime', name_base, k), globals(), container)
        
        if 'wallclock_runtime_global' in container:
          wallclock_runtime_global += container['wallclock_runtime_global']
        else:
          wallclock_runtime_global += container['wall_runtime_total'] # legacy
        
        if 'mpi_efficiency' in container:
          mpi_efficiency += container['mpi_efficiency']
        else:
          mpi_efficiency += container['mpi_scaling'] # legacy
    
    # if some load failed, move to another
    except:
      continue
    
    if 'cores' in container:
      cores .append( container['cores'] )
    else:
      cores .append( container['nodes'] ) # legacy
    
    wallclock_runtimes_global .append ( wallclock_runtime_global/K )
    mpi_efficiencies          .append ( mpi_efficiency/K )
  
  pylab.subplot(121)
  pylab.loglog(cores, wallclock_runtimes_global, '-' + marker, label=name+label_suffix, color=color, markeredgecolor=color)
  if K != 1:
    pylab.title('Weak scaling | average of %d runs' % K)
  else:
    pylab.title('Weak scaling')
  pylab.ylabel('global wallclock-time')
  pylab.xlabel('cores (~ problem size)')
  pylab.legend(loc='best')
  pylab.grid(True)
  
  # === efficiency
  
  pylab.subplot(122)
  pylab.semilogx(cores, mpi_efficiencies, '-' + marker, label=name+label_suffix, color=color, markeredgecolor=color)
  pylab.ylim(0,1.02)
  if K != 1:
    pylab.title('Parallel efficiency | average of %d runs' % K)
  else:
    pylab.title('Parallel efficiency')
  pylab.ylabel('efficiency')
  pylab.xlabel('cores (~ problem size)')
  pylab.legend(loc='lower center')
  pylab.grid(True)
  
  # === formatting
  
  pylab.subplots_adjust(right=0.98)
  pylab.subplots_adjust(left=0.06)
  pylab.subplots_adjust(top=0.93)
  pylab.subplots_adjust(bottom=0.15)
  
  if not frame:
    pylab.savefig(getFileName(Run(name_base + '-1', load_times=False), 'scaling-weak' + save_suffix, file_suffix=file_suffix))
    pylab.savefig(getFileName(Run(name_base + '-1', load_times=False), 'scaling-weak' + save_suffix, file_suffix='eps'))
    pylab.show()

# plots the stochastic scaling for the number of stochastic dimensions in the list KL
def plotStochasticScaling ( name, KL, frame=False, color=None, K=1, save_suffix='', file_suffix='pdf', path='', label_suffix='', marker='x' ):
  
  rtimes = []
  
  for m in KL:
    
    if name == '':
      name_base = 'KL=%d' % m
    else:
      name_base = name + '_KL=%d' % m
    
    rtime = 0
    
    # try to load runs and get needed values
    try:
      for k in range(1, K+1):
        container = {}
        execfile(path + '%s_%s-%d.py' % ('runtime', name_base, k), globals(), container)
        if 'wallclock_runtime_global' in r.runtime:
          rtime += r.runtime['wallclock_runtime_global']
        elif 'wallclock_runtime_total' in r.runtime:
          rtime += r.runtime['wallclock_runtime_total'] # legacy
        else:
          rtime += r.runtime['clock_runtime']
    
    # if some load failed, move to another
    except:
      continue
      
    rtimes.append ( rtime/K )
  
  pylab.plot(numpy.log10(KL), numpy.log10(rtimes), '-' + marker, label=name+label_suffix, color='k', edgecolor='k')
  pylab.title('Stochastic scaling | average of %d runs' % K)
  pylab.ylabel('log10(seconds)')
  pylab.xlabel('log10(KL-coefficients)')
  pylab.legend(loc='upper left')
  pylab.grid(True)
  
  if not frame:
    plotSlope([-1,1], pylab.gca())
  
  Run('%s-1' % name_base).showInfo()
  
  # === formatting
  
  pylab.subplots_adjust(right=0.96)
  pylab.subplots_adjust(left=0.10)
  pylab.subplots_adjust(top=0.94)
  
  if not frame:
    pylab.savefig(getFileName(Run(name_base + '-1', load_times=False), 'scaling-stochastic' + save_suffix, file_suffix=file_suffix))
    pylab.savefig(getFileName(Run(name_base + '-1', load_times=False), 'scaling-stochastic' + save_suffix, file_suffix='eps'))
    pylab.show()

# === exporters

# saves all refinement data to Matlab files
def save_matlab ( name, v, N, K=1, matlab_path='', path='' ):
  
  for i , (NX, NY, NZ) in enumerate(N):
    
    name_base = name + '_%dx%dx%d' % (NX, NY, NZ)
    
    for k in range(1, K+1):
      r = Run('%s-%d' % (name_base, k), path=path)
      r.matlab(v, path=matlab_path)

# creates error tables
def tableref(r, runref, v):
  fmt="%10.1e"
  scale=1
  t1 = len(r.times)-1
  t2 = len(runref.times)-1
  d = r.var(v, t1)
  refd = runref.var(v, t2)
  zoom = runref.i.NX/r.i.NX
  if runref.i.NY == 1:
    while zoom>1:
      refd = (refd[0:runref.i.NX-1:2,:,:]+refd[1:runref.i.NX:2,:,:])/2
      zoom=zoom/2
  else:
    while zoom>1:
      refd = (refd[0:runref.i.NX-1:2,0:runref.i.NY-1:2,:]+refd[1:runref.i.NX:2,0:runref.i.NY-1:2,:]+refd[1:runref.i.NX:2,1:runref.i.NY:2,:]+refd[0:runref.i.NX-1:2,1:runref.i.NY:2,:])/4
      zoom=zoom/2
  td = (L1(d, refd)*scale, Linf(d, refd)*scale)
  print ((r"%s & %s & %s \\"% (r.name,fmt,fmt)) % td)

# helper for 'tableref'
def rebin(d, f):
  nx,ny,nz = d.shape
  if ny == 1:
    assert nz == 1
    return sum(d[i:nx-f+1+i:f,:,:] for i in range(f))/f
  elif nz == 1:
    return sum(d[i:nx-f+1+i:f,i:ny-f+1+i:f,:] for i in range(f))/(f*f)
  else:
    return sum(d[i:nx-f+1+i:f,i:ny-f+1+i:f,i:nz-f+1+i:f] for i in range(f))/(f*f*f)

# helper for 'tableref'
def table(runs, v, t=None, fmt="%12.1e", refd=None, reft=None, refr=None, scale=1, dt=0.01):
  if t is None:
    t = max(r.times[-1] for r in runs)
  print (r'       Run &        $L_1$ &   $L_\infty$ \\')
  for r in runs:
    ts = None
    for n,tm in enumerate(r.times):
      if abs(tm - t) < dt:
        ts = n
        break
    if ts is None:
      print (r.name + r' & crashed \\')
      continue
    d = r.var(v, ts)
    if reft is not None:
      refts = None
      for n,tm in enumerate(r.times):
        if tm == reft:
          refts = n
          break
      refd = r.var(v, refts)
    elif refr is not None:
      refd = refr.var(v, ts)
      if refr.i.NX != r.i.NX:
        assert (refr.i.NX//r.i.NX) * r.i.NX == refr.i.NX
        refd = rebin(refd, refr.i.NX // r.i.NX)
    td = (L1(d, refd)*scale, Linf(d, refd)*scale)
    print ((r"%10s & %s & %s \\"% (r.name,fmt,fmt)) % td)
